/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007-2015 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@posteo.de

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************

   Fotoxx image editor - image metadata functions.
   View and edit metadata: EXIF, IPTC, etc.

   m_meta_view_short          metadata short report
   m_meta_view_long           report all metadata
   meta_view                  callable function
   m_captions                 write captions and comments at top of current image
   m_edit_metadata            primary edit metadata dialog
   pdate_metadate             convert yyyy-mm-dd to yyyymmdd
   ptime_metatime             convert hh:mm:ss to hhmmss
   metadate_pdate             convert yyyymmddhhmmss to yyyy-mm-dd and hh:mm:ss
   manage_tags                maintain list of defined tags
   add_tag                    add tag to a tag list
   del_tag                    remove tag from a tag list
   add_recentag               add tag to recent tags list, remove oldest if needed
   load_deftags               load defined tags list from tags file and image index
   save_deftags               save defined tags list to tags file
   find_deftag                check if given tag is in defined tags list
   add_deftag                 add new tag to defined tags list or change category
   del_deftag                 remove tag from defined tags list
   deftags_stuff              stuff defined tags into dialog text widget
   tag_orphans                report tags defined and not used in any image file
   m_meta_edit_any            dialog to fetch and save any image file metadata by name
   m_meta_delete              dialog to delete any image file metadata by name
   m_batchTags                batch add and delete tags - selected image files
   m_batch_renametags         convert tag names for all image files
   load_filemeta              load image file metadata into memory (indexed data only)
   save_filemeta              save metadata to image file EXIF and to image index
   update_image_index         update index data for current image file
   delete_image_index         delete index record for deleted image file

   Geotag editing, image mapping, image search functions

   validate_latlong           validate latitude/longitude data
   init_geolocs               load geolocations table from image index file
   get_geolocs                get latitude/longitude for a city or location
   put_geotags                save geotags in image file EXIF and image index file
   m_load_geomap              load a geographic map chosen by user
   geomap_coordinates         convert map pixel position to latitude/longitude
   geomap_position            convert latitude/longitude to map pixel position
   geomap_paint_dots          paint red dots on map where images are located
   geomap_mousefunc           respond to mouse movement and left clicks on geomap
   free_geomap                free huge memory for geomap image
   m_mapsearch_range          set the map search range from a clicked position, kilometers
   m_edit_geotags             edit image city/location, country, latitude, longitude
   m_batch_add_geotags        add given geotag to selected set of images
   geotags_choosecity         get possible location matches for partial input, choose one
   web_geocode                use web geocoding service to map location to latitude/longitude
   m_geotag_groups            list locations/image counts/time groups, choose one for gallery
   m_search_images            general image search using file names and any metadata

   EXIF etc. data storage and retrieval

   exif_get                   get image metadata from list of keys
   exif_put                   update image metadata from list of keys and data
   exif_copy                  copy metadata from one image to another, with revisions
   exif_server                start exiftool server process, send data requests
   init_image_index           initialization for image file index read/write

   Image index file functions

   get_sxrec                  get image index record for image file
   get_sxrec_min              get index data used for gallery view
   put_sxrec                  add or update index record for an image file
   read_sxrec_seq             read all index records sequentially, one per call
   write_sxrec_seq            write all index records sequentially


***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"                                                        //  (variables in fotoxx.h are refs)

/**************************************************************************/

char  *pdate_metadate(cchar *pdate);                                       //  "yyyy-mm-dd" to "yyyymmdd"
char  *ptime_metatime(cchar *ptime);                                       //  "hh:mm" to "hhmm"
void  metadate_pdate(cchar *metadate, char *pdate, char *ptime);           //  "yyyymmddhhmm" to "yyyy-mm-dd" and "hh:mm"
int   get_mouse_tag(GtkTextView *, int px, int py, cchar *);               //  get tag selected by mouse
int   add_tag(char *tag, char *taglist, int maxcc);                        //  add tag if unique and enough space
int   del_tag(char *tag, char *taglist);                                   //  remove tag from tag list
int   add_recentag(char *tag);                                             //  add tag to recent tags, keep recent
void  load_deftags();                                                      //  tags_defined file >> tags_deftags[]
void  save_deftags();                                                      //  tags_deftags[] >> defined_tags file
int   find_deftag(char *tag);                                              //  find tag in tags_deftags[]
int   add_deftag(char *catg, char *tag);                                   //  add tag to tags_deftags[]
int   del_deftag(char *tag);                                               //  remove tag from tags_deftags[]
void  deftags_stuff(zdialog *zd);                                          //  tags_deftags[] >> dialog widget

char  *tags_deftags[maxtagcats];                                           //  defined tags: catg: tag1, ... tagN,
char  tags_recentags[tagRcc] = "";                                         //  recently added tags list

/**************************************************************************/

//  menu function and popup dialog to show EXIF/IPTC data
//  window is updated when navigating to another image

/**************************************************************************/

//  menu function and popup dialog to show EXIF/IPTC data
//  window is updated when navigating to another image

void m_meta_view_short(GtkWidget *, cchar *menu)
{
   if (menu) F1_help_topic = "view_metadata";
   meta_view(1);
   return;
}

void m_meta_view_long(GtkWidget *, cchar *menu)
{
   if (menu) F1_help_topic = "view_metadata";
   meta_view(2);
   return;
}


void meta_view(int arg)
{
   int   meta_view_dialog_event(zdialog *zd, cchar *event);

   cchar       *keyname[19] = { "ImageSize", "FileSize", exif_date_key, "FileModifyDate", "Make", "Model",
                              exif_focal_length_key, "ExposureTime", "FNumber", "ISO",
                              exif_city_key, exif_country_key, exif_latitude_key, exif_longitude_key,
                              iptc_keywords_key, iptc_rating_key, iptc_caption_key, exif_comment_key,
                              exif_editlog_key };
   char        *keyval[19];

   static int     reportype = 1;
   char           *buff, chsec[12], **text;
   static char    *file = 0, *filen, *chsize;
   float          fsize, fsec;
   int            err, ii, nn, contx = 0;
   cchar          *textdelims = "!-,.:;?/)}]";
   cchar          *editdelims = ":|,";
   GtkWidget      *widget;

   if (file) zfree(file);
   file = 0;

   if (clicked_file) {                                                     //  use clicked file if present
      file = clicked_file;
      clicked_file = 0;
   }
   else if (curr_file)                                                     //  else current file
      file = zstrdup(curr_file);
   else return;

   if (arg > 0) reportype = arg;                                           //  change report type

   if (! zdexifview)                                                       //  popup dialog if not already
   {
      zdexifview = zdialog_new(ZTX("View Metadata"),Mwin,Bcancel,null);
      zdialog_add_widget(zdexifview,"scrwin","scroll","dialog",0,"expand");
      zdialog_add_widget(zdexifview,"text","exifdata","scroll",0,"expand|wrap");
      if (reportype == 1) zdialog_resize(zdexifview,650,300);
      else zdialog_resize(zdexifview,800,800);
      zdialog_run(zdexifview,meta_view_dialog_event);
   }

   widget = zdialog_widget(zdexifview,"exifdata");
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing
   gtk_widget_override_font(widget,zfuncs::monofont);
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_NONE);       //  disable text wrap
   wclear(widget);

   if (reportype == 1)                                                     //  short report        better format  v.15.02
   {
      err = exif_get(file,keyname,keyval,19);
      if (err) {
         zmessageACK(Mwin,"exif failure");
         return;
      }

      filen = strrchr(file,'/');                                           //  get file name without directory
      if (filen) filen++;
      else filen = file;

      fsize = 0;
      if (keyval[1]) fsize = atof(keyval[1]);                              //  get file size in B/KB/MB units
      chsize = formatKBMB(fsize,3);

      if (keyval[2] && strlen(keyval[2]) > 16) keyval[2][16] = 0;          //  truncate dates to yyyy-mm-dd hh:mm
      if (keyval[3] && strlen(keyval[3]) > 16) keyval[3][16] = 0;

      wprintf(widget,"File       %s \n",filen);
      wprintf(widget,"Size       %s  %s \n",keyval[0],chsize);
      wprintf(widget,"Dates      photo: %s  file: %s \n",keyval[2],keyval[3]);

      if (keyval[4] || keyval[5])
         wprintf(widget,"Camera     %s  %s \n",keyval[4],keyval[5]);

      if (keyval[6] || keyval[7] || keyval[8] || keyval[9])
      {
         strcpy(chsec,"null");
         if (keyval[7]) {
            fsec = atof(keyval[7]);                                        //  convert 0.008 seconds to 1/125 etc.
            if (fsec > 0 && fsec <= 0.5) {
               fsec = 1/fsec;
               snprintf(chsec,12,"1/%.0f",fsec);
            }
            else if (fsec > 0)
               snprintf(chsec,12,"%.0f",fsec);
         }
         wprintf(widget,"Exposure   %s mm  %s sec  F%s  ISO %s \n",keyval[6],chsec,keyval[8],keyval[9]);
      }

      if (keyval[10] || keyval[11] || keyval[12] || keyval[13])
         wprintf(widget,"Location   %s %s  %s %s \n",keyval[10],keyval[11],keyval[12],keyval[13]);

      if (keyval[14])
         wprintf(widget,"Keywords   %s \n",keyval[14]);

      if (keyval[15])
         wprintf(widget,"Rating     %s \n",keyval[15]);

      if (keyval[16]) {
         nn = breakup_text(keyval[16],text,textdelims,40,60);
         wprintf(widget,"Caption    %s \n",text[0]);
         for (ii = 1; ii < nn; ii++)
            wprintf(widget,"           %s \n",text[ii]);
         for (ii = 0; ii < nn; ii++)
            zfree(text[ii]);
         zfree(text);
      }

      if (keyval[17]) {
         nn = breakup_text(keyval[17],text,textdelims,40,60);
         wprintf(widget,"Comments   %s \n",text[0]);
         for (ii = 1; ii < nn; ii++)
            wprintf(widget,"           %s \n",text[ii]);
         for (ii = 0; ii < nn; ii++)
            zfree(text[ii]);
         zfree(text);
      }

      if (keyval[18]) {
         nn = breakup_text(keyval[18],text,editdelims,40,60);
         wprintf(widget,"Edits      %s \n",text[0]);
         for (ii = 1; ii < nn; ii++)
            wprintf(widget,"           %s \n",text[ii]);
         for (ii = 0; ii < nn; ii++)
            zfree(text[ii]);
         zfree(text);
      }

      wscroll(widget,1);

      for (ii = 0; ii < 19; ii++)
         if (keyval[ii]) zfree(keyval[ii]);
   }

   if (reportype == 2)                                                     //  long report
   {
      snprintf(command,CCC,"exiftool -s -e \"%s\" ", file);

      while ((buff = command_output(contx,command))) {                     //  run command, output into window
         wprintf(widget,"%s\n",buff);
         zfree(buff);
      }
      wscroll(widget,1);
      command_status(contx);
   }

   gtk_window_present(MWIN);                                               //  keep focus on main window          v.15.03
   return;
}


//  dialog event and completion callback function

int meta_view_dialog_event(zdialog *zd, cchar *event)                      //  kill dialog
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   if (! zd->zstat) return 0;
   zdialog_free(zdexifview);
   return 0;
}


/**************************************************************************/

//  Show captions and comments on top of current image in main window.
//  Menu call (menu arg not null): toggle switch on/off.
//  Non-menu call: write caption/comment on image if switch is ON.

void m_captions(GtkWidget *, cchar *menu)
{
   cchar        *keynames[2] = { iptc_caption_key, exif_comment_key };
   char         *keyvals[2];
   char         caption[200], comment[200];
   static char  text[402];

   F1_help_topic = "show_captions";

   if (menu) {                                                             //  if menu call, flip toggle
      Fcaptions = 1 - Fcaptions;
      if (curr_file) f_open(curr_file);
      return;
   }

   if (! Fcaptions) return;

   if (! curr_file) return;
   *caption = *comment = 0;

   exif_get(curr_file,keynames,keyvals,2);                                 //  get captions and comments metadata

   if (keyvals[0]) {
      strncpy0(caption,keyvals[0],200);
      zfree(keyvals[0]);
   }

   if (keyvals[1]) {
      strncpy0(comment,keyvals[1],200);
      zfree(keyvals[1]);
   }

   *text = 0;

   if (*caption) strcpy(text,caption);
   if (*caption && *comment) strcat(text,"\n");
   if (*comment) strcat(text,comment);

   for (int ii = 0; text[ii]; ii++)                                        //  replace "\n" with newline chars.   v.15.01.1
      if (text[ii] == '\\' && text[ii+1] == 'n')
         memmove(text+ii,"\n ",2);

   if (*text) add_toptext(1,0,0,text,"Sans 10");

   return;
}


/**************************************************************************/

//  edit metadata menu function

void m_edit_metadata(GtkWidget *, cchar *menu)                             //  overhauled
{
   void  edit_imagetags_clickfunc(GtkWidget *widget, int line, int pos);
   void  edit_recentags_clickfunc(GtkWidget *widget, int line, int pos);
   void  edit_matchtags_clickfunc(GtkWidget *widget, int line, int pos);
   void  edit_deftags_clickfunc(GtkWidget *widget, int line, int pos);
   int   editmeta_dialog_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   zdialog     *zd;
   char        *ppv, pdate[12], ptime[12];
   char        cctext[exif_maxcc+50];
   char        RN[4] = "R0";
   int         ii;

   if (menu) F1_help_topic = "edit_metadata";

   if (! curr_file) {
      if (zdeditmeta) zdialog_free(zdeditmeta);
      zdeditmeta = 0;
      return;
   }
   
/**
          ________________________________________________________
         |                Edit Metadata                           |
         |                                                        |
         |  Image File: filename.jpg                              |
         |  Image Date: [_________]  Time: [______]  [prev]       |
         |  Rating (stars): ʘ 1  ʘ 2  ʘ 3  ʘ 4  ʘ 5  ʘ 6       |
         |  Caption [___________________________________________] |
         |  Comments [__________________________________________] |
         |                                                        |
         |  Image Tags [________________________________________] |
         |  Recent Tags [_______________________________________] |
         |  Type New Tag [______________________________________] |
         |  Matching Tags [_____________________________________] |
         |                                                        |
         |  Defined Tags: ______________________________________  |
         |  |                                                   | |
         |  |                                                   | |
         |  |                                                   | |
         |  |                                                   | |
         |  |___________________________________________________| |
         |                                                        |
         |                        [Manage Tags] [Apply] [Cancel]  |
         |________________________________________________________|

**/

   if (! zdeditmeta)                                                       //  (re) start edit dialog
   {
      zd = zdialog_new(ZTX("Edit Metadata"),Mwin,Bmanagetags,Bapply,Bcancel,null);
      zdeditmeta = zd;

      //  File: xxxxxxxxx.jpg
      zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labf","hbf",ZTX("Image File:"),"space=3");
      zdialog_add_widget(zd,"label","file","hbf","filename.jpg","space=5");

      //  Image Date yyyy-mm-dd [__________]  Time hh:mm [_____]  [prev]
      zdialog_add_widget(zd,"hbox","hbdt","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labdate","hbdt",ZTX("Image Date"),"space=3");
      zdialog_add_widget(zd,"entry","date","hbdt",0,"size=12");
      zdialog_add_widget(zd,"label","space","hbdt",0,"space=5");
      zdialog_add_widget(zd,"label","labtime","hbdt",ZTX("Time"),"space=3");
      zdialog_add_widget(zd,"entry","time","hbdt",0,"size=8");
      zdialog_add_widget(zd,"button","ppdate","hbdt",Bprev,"space=8");

      //  Rating (stars): ʘ 1  ʘ 2  ʘ 3  ʘ 4  ʘ 5  ʘ 6                  //  v.14.12
      zdialog_add_widget(zd,"hbox","hbrate","dialog");
      zdialog_add_widget(zd,"label","labrate","hbrate",ZTX("Rating (stars):"),"space=3");
      zdialog_add_widget(zd,"radio","R0","hbrate","0","space=6");
      zdialog_add_widget(zd,"radio","R1","hbrate","1","space=6");
      zdialog_add_widget(zd,"radio","R2","hbrate","2","space=6");
      zdialog_add_widget(zd,"radio","R3","hbrate","3","space=6");
      zdialog_add_widget(zd,"radio","R4","hbrate","4","space=6");
      zdialog_add_widget(zd,"radio","R5","hbrate","5","space=6");

      zdialog_add_widget(zd,"hbox","space","dialog",0,"space=3");

      //  Caption [___________________________________________]
      zdialog_add_widget(zd,"hbox","hbcap","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labcap","hbcap",ZTX("Caption"),"space=3");
      zdialog_add_widget(zd,"frame","frame4","hbcap",0,"space=3|expand");
      zdialog_add_widget(zd,"edit","caption","frame4",0,"wrap");

      //  Comments [__________________________________________]
      zdialog_add_widget(zd,"hbox","hbcom","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labcom","hbcom",ZTX("Comments"),"space=3");
      zdialog_add_widget(zd,"frame","frcom","hbcom",0,"space=3|expand");
      zdialog_add_widget(zd,"edit","comments","frcom",0,"wrap");

      zdialog_add_widget(zd,"hsep","sep","dialog",0,"space=3");

      //  Enter New Tag [_____________________________________]
      zdialog_add_widget(zd,"hbox","hbnt","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labnt","hbnt",ZTX("Enter New Tag"),"space=3");
      //zdialog_add_widget(zd,"frame","frnt","hbnt",0,"space=3|expand");   //  works OK, but GTK warnings galore
      //zdialog_add_widget(zd,"edit","newtag","frnt");
      zdialog_add_widget(zd,"entry","newtag","hbnt");

      //  Matching Tags [_____________________________________]
      zdialog_add_widget(zd,"hbox","hbmt","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labmt","hbmt",ZTX("Matching Tags"),"space=3");
      zdialog_add_widget(zd,"frame","frmt","hbmt",0,"space=3|expand");
      zdialog_add_widget(zd,"text","matchtags","frmt",0,"wrap");

      zdialog_add_widget(zd,"hsep","sep","dialog",0,"space=3");

      //  Image Tags [________________________________________]
      zdialog_add_widget(zd,"hbox","hbit","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labit","hbit",ZTX("Image Tags"),"space=3");
      zdialog_add_widget(zd,"frame","frit","hbit",0,"space=3|expand");
      zdialog_add_widget(zd,"text","imagetags","frit",0,"wrap");

      //  Recent Tags [_______________________________________]
      zdialog_add_widget(zd,"hbox","hbrt","dialog",0,"space=1");
      zdialog_add_widget(zd,"label","labrt","hbrt",ZTX("Recent Tags"),"space=3");
      zdialog_add_widget(zd,"frame","frrt","hbrt",0,"space=3|expand");
      zdialog_add_widget(zd,"text","recentags","frrt",0,"wrap");
      
      //  Defined Tags 
      zdialog_add_widget(zd,"hbox","space","dialog");
      zdialog_add_widget(zd,"hbox","hbdt1","dialog");
      zdialog_add_widget(zd,"label","labdt","hbdt1",ZTX("Defined Tags:"),"space=3");
      zdialog_add_widget(zd,"hbox","hbdt2","dialog",0,"expand");
      zdialog_add_widget(zd,"frame","frdt2","hbdt2",0,"expand|space=3");
      zdialog_add_widget(zd,"scrwin","swdt2","frdt2",0,"expand");
      zdialog_add_widget(zd,"text","deftags","swdt2",0,"wrap");

      load_deftags();                                                      //  stuff defined tags into dialog
      deftags_stuff(zd);

      widget = zdialog_widget(zd,"imagetags");                             //  tag widget mouse functions
      textwidget_set_clickfunc(widget,edit_imagetags_clickfunc);

      widget = zdialog_widget(zd,"recentags");
      textwidget_set_clickfunc(widget,edit_recentags_clickfunc);

      widget = zdialog_widget(zd,"matchtags");
      textwidget_set_clickfunc(widget,edit_matchtags_clickfunc);

      widget = zdialog_widget(zd,"deftags");
      textwidget_set_clickfunc(widget,edit_deftags_clickfunc);

      zdialog_resize(zd,400,500);                                          //  run dialog
      zdialog_run(zd,editmeta_dialog_event);
   }

   zd = zdeditmeta;

   ppv = (char *) strrchr(curr_file,'/');
   zdialog_stuff(zd,"file",ppv+1);                                         //  stuff dialog fields from curr. image file

   metadate_pdate(meta_pdate,pdate,ptime);
   zdialog_stuff(zd,"date",pdate);
   zdialog_stuff(zd,"time",ptime);

   for (ii = 0; ii <= 5; ii++) {                                           //  set all rating radio buttons OFF   v.14.12
      RN[1] = '0' + ii;
      zdialog_stuff(zd,RN,0);
   }
   RN[1] = meta_rating[0];                                                 //  set radio button ON for current rating
   zdialog_stuff(zd,RN,1);

   repl_1str(meta_caption,cctext,"\\n","\n");
   zdialog_stuff(zd,"caption",cctext);
   repl_1str(meta_comments,cctext,"\\n","\n");
   zdialog_stuff(zd,"comments",cctext);

   zdialog_stuff(zd,"imagetags",meta_tags);
   zdialog_stuff(zd,"recentags",tags_recentags);

   gtk_window_present(MWIN);                                               //  keep focus on main window          v.15.03
   return;
}


//  mouse click functions for various text widgets for tags

void edit_imagetags_clickfunc(GtkWidget *widget, int line, int pos)        //  existing image tag was clicked
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   del_tag(txtag,meta_tags);                                               //  remove tag from image
   zdialog_stuff(zdeditmeta,"imagetags",meta_tags);
   Fmetamod++;                                                             //  note change                        v.15.03

   zfree(txline);
   zfree(txtag);
   return;
}


void edit_recentags_clickfunc(GtkWidget *widget, int line, int pos)        //  recent tag was clicked
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   add_tag(txtag,meta_tags,tagFcc);                                        //  add recent tag to image
   zdialog_stuff(zdeditmeta,"imagetags",meta_tags);
   Fmetamod++;                                                             //  note change                        v.15.03

   zfree(txline);
   zfree(txtag);
   return;
}


void edit_matchtags_clickfunc(GtkWidget *widget, int line, int pos)        //  matching tag was clicked           v.15.07
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   add_tag(txtag,meta_tags,tagFcc);                                        //  add matching tag to image
   Fmetamod++;                                                             //  note change                        v.15.03
   add_recentag(txtag);                                                    //  and add to recent tags

   zdialog_stuff(zdeditmeta,"imagetags",meta_tags);                        //  update dialog widgets
   zdialog_stuff(zdeditmeta,"recentags",tags_recentags);
   zdialog_stuff(zdeditmeta,"newtag","");
   zdialog_stuff(zdeditmeta,"matchtags","");

   zdialog_goto(zdeditmeta,"newtag");                                      //  put focus back on newtag widget

   zfree(txline);
   zfree(txtag);
   return;
}


void edit_deftags_clickfunc(GtkWidget *widget, int line, int pos)          //  defined tag was clicked
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;:",end);
   if (! txtag || end == ':') { zfree(txline); return; }                   //  tag category clicked, ignore

   add_tag(txtag,meta_tags,tagFcc);                                        //  add new tag to image
   zdialog_stuff(zdeditmeta,"imagetags",meta_tags);                        //    from defined tags list
   Fmetamod++;                                                             //  note change                        v.15.03

   add_recentag(txtag);                                                    //  and add to recent tags
   zdialog_stuff(zdeditmeta,"recentags",tags_recentags);

   zfree(txline);
   zfree(txtag);
   return;
}


//  dialog event and completion callback function

int editmeta_dialog_event(zdialog *zd, cchar *event)                       //  overhauled
{
   char     pdate[12], ptime[12];                                          //  yyyy-mm-dd  and  hh:mm:ss
   char     *metadate, *metatime;                                          //  yyyymmdd  and  hhmmss
   int      ii, jj, nt, cc1, cc2, ff;
   char     *pp1, *pp2;
   char     newtag[tagcc], matchtags[20][tagcc];
   char     matchtagstext[(tagcc+2)*20];
   char     cctext[exif_maxcc+50];
   char     RN[4] = "R0";

   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (! curr_file) return 1;

   if (strstr("date time caption comments",event))                         //  note change but process later
      Fmetamod++;

   if (strmatch(event,"ppdate")) {                                         //  repeat last date used
      if (*meta_ppdate) {
         metadate_pdate(meta_ppdate,pdate,ptime);
         zdialog_stuff(zd,"date",pdate);
         zdialog_stuff(zd,"time",ptime);
         Fmetamod++;
         return 1;
      }
   }

   if (strstr("R0 R1 R2 R3 R4 R5",event)) {                                //  note if rating changed
      Fmetamod++;
      return 1;
   }

   if (strmatch(event,"newtag"))                                           //  new tag is being typed in          v.15.07
   {
      zdialog_stuff(zd,"matchtags","");                                    //  clear matchtags in dialog

      zdialog_fetch(zd,"newtag",newtag,tagcc);                             //  get chars. typed so far
      cc1 = strlen(newtag);
      
      for (ii = jj = 0; ii <= cc1; ii++) {                                 //  remove foul characters
         if (strchr(",:;",newtag[ii])) continue;
         newtag[jj++] = newtag[ii];
      }
      
      if (jj < cc1) {                                                      //  something was removed
         newtag[jj] = 0;
         cc1 = jj;
         zdialog_stuff(zd,"newtag",newtag);
      }

      if (cc1 < 2) return 1;                                               //  wait for at least 2 chars.

      for (ii = nt = 0; ii < maxtagcats; ii++)                             //  loop all categories
      {
         pp2 = tags_deftags[ii];                                           //  category: aaaaaa, bbbbb, ... tagN,
         if (! pp2) continue;                                              //            |     |
         pp2 = strchr(pp2,':');                                            //            pp1   pp2
         
         while (true)                                                      //  loop all deftags in category
         {
            pp1 = pp2 + 2;
            if (! *pp1) break;
            pp2 = strchr(pp1,',');
            if (! pp2) break;
            if (strmatchcaseN(newtag,pp1,cc1)) {                           //  deftag matches chars. typed so far
               cc2 = pp2 - pp1;
               strncpy(matchtags[nt],pp1,cc2);                             //  save deftags that match
               matchtags[nt][cc2] = 0;
               if (++nt == 20) return 1;                                   //  quit if 20 matches or more
            }
         }
      }
      
      if (nt == 0) return 1;                                               //  no matches

      pp1 = matchtagstext;

      for (ii = 0; ii < nt; ii++)                                          //  make deftag list: aaaaa, bbb, cccc ...
      {
         strcpy(pp1,matchtags[ii]);
         pp1 += strlen(pp1);
         strcpy(pp1,", ");
         pp1 += 2;
      }
      
      zdialog_stuff(zd,"matchtags",matchtagstext);                         //  stuff matchtags in dialog
      return 1;
   }

   if (strmatch(event,"enter"))                                            //  KB Enter, new tag finished         v.15.07
   {
      zdialog_fetch(zd,"newtag",newtag,tagcc);                             //  get finished tag
      cc1 = strlen(newtag);
      if (! cc1) return 1;
      if (newtag[cc1-1] == '\n') {                                         //  remove newline character
         cc1--;
         newtag[cc1] = 0;
      }

      for (ii = ff = 0; ii < maxtagcats; ii++)                             //  loop all categories
      {
         pp2 = tags_deftags[ii];                                           //  category: aaaaaa, bbbbb, ... tagN,
         if (! pp2) continue;                                              //            |     |
         pp2 = strchr(pp2,':');                                            //            pp1   pp2
         
         while (true)                                                      //  loop all deftags in category
         {
            pp1 = pp2 + 2;
            if (! *pp1) break;
            pp2 = strchr(pp1,',');
            if (! pp2) break;
            cc2 = pp2 - pp1;
            if (cc2 != cc1) continue;
            if (strmatchcaseN(newtag,pp1,cc1)) {                           //  entered tag matches deftag
               strncpy(newtag,pp1,cc1);                                    //  use deftag upper/lower case
               ff = 1;
               break;
            }
         }

         if (ff) break;
      }

      add_tag(newtag,meta_tags,tagFcc);                                    //  add to image tag list
      Fmetamod++;                                                          //  note change
      add_recentag(newtag);                                                //  and add to recent tags

      if (! ff) {                                                          //  if new tag, add to defined tags
         add_deftag((char *) "nocatg",newtag);
         deftags_stuff(zd);
      }

      zdialog_stuff(zd,"newtag","");                                       //  update dialog widgets
      zdialog_stuff(zd,"imagetags",meta_tags);
      zdialog_stuff(zd,"recentags",tags_recentags);
      zdialog_stuff(zd,"matchtags","");

      zdialog_goto(zd,"newtag");                                           //  put focus back on newtag widget
      return 1;
   }

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat == 1) {                                                   //  manage tags
      zd->zstat = 0;                                                       //  keep dialog active
      zdialog_show(zd,0);                                                  //  hide parent dialog
      manage_tags();
      zdialog_show(zd,1);
      return 1;
   }

   if (zd->zstat != 2) {                                                   //  cancel or [x]
      zdialog_free(zdeditmeta);                                            //  kill dialog
      zdeditmeta = 0;
      Fmetamod = 0;
      return 1;
   }

   zd->zstat = 0;                                                          //  apply - keep dialog active
   gtk_window_present(MWIN);                                               //  keep focus on main window

   if (! Fmetamod) return 1;                                               //  nothing changed

   zdialog_fetch(zd,"date",pdate,12);                                      //  get photo date and time
   zdialog_fetch(zd,"time",ptime,12);
   if (*pdate) {                                                           //  date available
      metadate = pdate_metadate(pdate);                                    //  validate
      if (! metadate) return 1;                                            //  bad, re-input
      strcpy(meta_pdate,metadate);                                         //  convert to yyyymmdd
      if (*ptime) {                                                        //  time available
         metatime = ptime_metatime(ptime);                                 //  validate
         if (! metatime) return 1;                                         //  bad, re-input
         strcat(meta_pdate,metatime);                                      //  append hhmmss
      }
   }
   else *meta_pdate = 0;                                                   //  leave empty

   strcpy(meta_rating,"0");                                                //  v.14.12
   for (ii = 0; ii <= 5; ii++) {                                           //  get which rating radio button ON
      RN[1] = '0' + ii;
      zdialog_fetch(zd,RN,jj);
      if (jj) meta_rating[0] = '0' + ii;                                   //  set corresponding rating
   }

   zdialog_fetch(zd,"caption",cctext,exif_maxcc);                          //  get new caption
   repl_1str(cctext,meta_caption,"\n","\\n");                              //  replace newlines with "\n"
   zdialog_fetch(zd,"comments",cctext,exif_maxcc);                         //  get new comments
   repl_1str(cctext,meta_comments,"\n","\\n");                             //  replace newlines with "\n"

   save_filemeta(curr_file);                                               //  save metadata changes to image file
   return 1;
}


//  Convert pretty date/time format from "yyyy-mm-dd" and "hh:mm:ss" to "yyyymmdd" and "hhmmss".
//  Missing month or day ("yyyy" or "yyyy-mm") is replaced with "-01".
//  Missing seconds ("hh:mm") are replaced with zero ("hh:mm:00").
//  Output user message and return null if not valid.

char * pdate_metadate(cchar *pdate)                                        //  "yyyy-mm-dd" >> "yyyymmdd"
{
   int         monlim[12] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
   int         cc, year, mon, day;
   char        pdate2[12];
   static char mdate[12];

   cc = strlen(pdate);
   if (cc > 10) goto badformat;

   strcpy(pdate2,pdate);

   if (cc == 4)                                                            //  conv. "yyyy" to "yyyy-01-01"
      strcat(pdate2,"-01-01");
   else if (cc == 7)                                                       //  conv. "yyyy-mm" to "yyyy-mm-01"
      strcat(pdate2,"-01");

   if (strlen(pdate2) != 10) goto badformat;
   if (pdate2[4] != '-' || pdate2[7] != '-') goto badformat;

   year = atoi(pdate2);
   mon = atoi(pdate2+5);
   day = atoi(pdate2+8);

   if (year < 0 || year > 2999) goto baddate;
   if (mon < 1 || mon > 12) goto baddate;
   if (day < 1 || day > monlim[mon-1]) goto baddate;
   if (mon == 2 && day == 29 && (year % 4)) goto baddate;

   memcpy(mdate,pdate2,4);                                                 //  return "yyyymmdd"
   memcpy(mdate+4,pdate2+5,2);
   memcpy(mdate+6,pdate2+8,3);
   return mdate;

badformat:
   zmessageACK(Mwin,ZTX("date format is YYYY-MM-DD"));
   return 0;

baddate:
   zmessageACK(Mwin,ZTX("date is invalid"));
   return 0;
}


char * ptime_metatime(cchar *ptime)                                        //  "hh:mm:ss" >> "hhmmss"
{
   int         cc, hour, min, sec;
   char        ptime2[12];
   static char mtime[8];

   cc = strlen(ptime);
   if (cc > 8) goto badformat;

   strcpy(ptime2,ptime);

   if (cc == 5) strcat(ptime2,":00");                                      //  conv. "hh:mm" to "hh:mm:00"

   if (strlen(ptime2) != 8) goto badformat;
   if (ptime2[2] != ':' || ptime2[5] != ':') goto badformat;

   hour = atoi(ptime2);
   min = atoi(ptime2+3);
   sec = atoi(ptime2+6);
   if (hour < 0 || hour > 23) goto badtime;
   if (min < 0 || min > 59) goto badtime;
   if (sec < 0 || sec > 59) goto badtime;

   memcpy(mtime,ptime2,2);                                                 //  return "hhmmss"
   memcpy(mtime+2,ptime2+3,2);
   memcpy(mtime+4,ptime2+6,2);
   return mtime;

badformat:
   zmessageACK(Mwin,ZTX("time format is HH:MM [:SS]"));
   return 0;

badtime:
   zmessageACK(Mwin,ZTX("time is invalid"));
   return 0;
}


//  Convert metadata date/time "yyyymmddhhmmss" to pretty format "yyyy-mm-dd" and "hh:mm:ss"

void metadate_pdate(cchar *metadate, char *pdate, char *ptime)
{
   if (*metadate) {
      memcpy(pdate,metadate,4);                                            //  yyyymmdd to yyyy-mm-dd
      memcpy(pdate+5,metadate+4,2);
      memcpy(pdate+8,metadate+6,2);
      pdate[4] = pdate[7] = '-';
      pdate[10] = 0;

      memcpy(ptime,metadate+8,2);                                          //  hhmmss to hh:mm:ss
      memcpy(ptime+3,metadate+10,2);
      ptime[2] = ':';
      ptime[5] = 0;
      if (metadate[12] > '0' || metadate[13] > '0') {                      //  append :ss only if not :00
         memcpy(ptime+6,metadate+12,2);
         ptime[5] = ':';
         ptime[8] = 0;
      }
   }
   else *pdate = *ptime = 0;                                               //  missing
   return;
}


/**************************************************************************/

//  manage tags function - auxilliary dialog

zdialog  *zdmanagetags = 0;

void manage_tags()
{
   void  manage_deftags_clickfunc(GtkWidget *widget, int line, int pos);
   int   managetags_dialog_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   zdialog     *zd;

   if (zdmanagetags) return;
   zd = zdialog_new(ZTX("Manage Tags"),Mwin,ZTX("orphan tags"),Bdone,null);
   zdmanagetags = zd;

   zdialog_add_widget(zd,"hbox","hb7","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcatg","hb7",ZTX("category"),"space=5");
   zdialog_add_widget(zd,"entry","catg","hb7",0,"size=12");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"label","labtag","hb7",ZTX("tag"),"space=5");
   zdialog_add_widget(zd,"entry","tag","hb7",0,"size=20|expand");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"button","create","hb7",Bcreate);
   zdialog_add_widget(zd,"button","delete","hb7",Bdelete);

   zdialog_add_widget(zd,"hbox","hb8","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hb8",ZTX("Defined Tags:"),"space=5");
   zdialog_add_widget(zd,"hbox","hb9","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frame8","hb9",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwin8","frame8",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrwin8",0,"expand|wrap");

   widget = zdialog_widget(zd,"deftags");                                  //  tag widget mouse function
   textwidget_set_clickfunc(widget,manage_deftags_clickfunc);

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   zdialog_resize(zd,0,400);
   zdialog_run(zd,managetags_dialog_event);                                //  run dialog
   zdialog_wait(zd);
   zdialog_free(zd);                                                       //  v.14.06

   return;
}


//  mouse click functions for widget having tags

void manage_deftags_clickfunc(GtkWidget *widget, int line, int pos)        //  tag or tag category was clicked
{
   char     *txline, *txword, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txword = textwidget_get_word(txline,pos,",;:",end);
   if (! txword) { zfree(txline); return; }
   
   if (end == ':') zdialog_stuff(zdmanagetags,"catg",txword);              //  selected category >> dialog widget
   else zdialog_stuff(zdmanagetags,"tag",txword);                          //  selected tag >> dialog widget

   zfree(txline);
   zfree(txword);
   return;
}


//  dialog event and completion callback function

int managetags_dialog_event(zdialog *zd, cchar *event)
{
   void tag_orphans();

   char        tag[tagcc], catg[tagcc];
   int         err, changed = 0;

   if (strmatch(event,"enter")) zd->zstat = 2;                             //  [done]  v.14.03
   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  report orphan tags
         zd->zstat = 0;                                                    //  keep dialog active
         tag_orphans();
      }

      else {                                                               //  done or [x]
         zdmanagetags = 0;
         return 0;
      }
   }

   if (strmatch(event,"create")) {                                         //  add new tag to defined tags
      zdialog_fetch(zd,"catg",catg,tagcc);
      zdialog_fetch(zd,"tag",tag,tagcc);
      err = add_deftag(catg,tag);
      if (! err) changed++;
   }

   if (strmatch(event,"delete")) {                                         //  remove tag from defined tags
      zdialog_fetch(zd,"tag",tag,tagcc);
      del_deftag(tag);
      changed++;
   }

   if (changed) {
      save_deftags();                                                      //  save tag updates to file
      deftags_stuff(zd);                                                   //  update dialog "deftags" window
      if (zdeditmeta)                                                      //  and edit metadata dialog if active
         deftags_stuff(zdeditmeta);
      if (zdbatchtags)                                                     //  and batch tags dialog if active
         deftags_stuff(zdbatchtags);
   }

   return 0;
}


/**************************************************************************/

//  add input tag to output tag list if not already there and enough room
//  returns:   0 = added OK     1 = already there (case ignored)
//             2 = overflow     3 = bad tag name     4 = null tag

int add_tag(char *tag, char *taglist, int maxcc)
{
   char     *pp1, *pp2, tag1[tagcc];
   int      cc, cc1, cc2;

   if (! tag) return 4;
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag1);
   if (! cc) return 4;
   if (utf8_check(tag1)) return 3;                                         //  look for bad characters
   if (strpbrk(tag1,",;:")) return 3;
   strcpy(tag,tag1);

   pp1 = taglist;
   cc1 = strlen(tag);

   while (true)                                                            //  check if already in tag list
   {
      while (*pp1 == ' ' || *pp1 == ',') pp1++;
      if (! *pp1) break;
      pp2 = pp1 + 1;
      while (*pp2 && *pp2 != ',') pp2++;
      cc2 = pp2 - pp1;
      if (cc2 == cc1 && strmatchcaseN(tag,pp1,cc1)) return 1;
      pp1 = pp2;
   }

   cc2 = strlen(taglist);                                                  //  append to tag list if space enough
   if (cc1 + cc2 + 3 > maxcc) return 2;
   strcpy(taglist + cc2,tag);
   strcpy(taglist + cc2 + cc1, ", ");                                      //  add delimiter + space

   if (taglist == meta_tags)                                               //  image tags were changed
      Fmetamod++;

   return 0;
}


//  remove tag from taglist, if present
//  returns: 0 if found and deleted, otherwise 1

int del_tag(char *tag, char *taglist)
{
   int         ii, ftcc, atcc, found;
   char        *temptags;
   cchar       *pp;

   temptags = zstrdup(taglist);

   *taglist = 0;
   ftcc = found = 0;

   for (ii = 1; ; ii++)
   {
      pp = strField(temptags,",;",ii);                                     //  next tag
      if (! pp) {
         zfree(temptags);
         if (found && taglist == meta_tags)                                //  image tags were changed
            Fmetamod++;
         return 1-found;
      }
      if (*pp == ' ') continue;

      if (strmatchcase(pp,tag)) {                                          //  skip matching tag
         found = 1;
         continue;
      }

      atcc = strlen(pp);                                                   //  copy non-matching tag
      strcpy(taglist + ftcc, pp);
      ftcc += atcc;
      strcpy(taglist + ftcc, ", ");                                        //  + delim + blank
      ftcc += 2;
   }
}


//  add new tag to recent tags, if not already.
//  remove oldest to make space if needed.

int add_recentag(char *tag)
{
   int         err;
   char        *pp, temptags[tagRcc];

   err = add_tag(tag,tags_recentags,tagRcc);                               //  add tag to recent tags

   while (err == 2)                                                        //  overflow
   {
      strncpy0(temptags,tags_recentags,tagRcc);                            //  remove oldest to make room
      pp = strpbrk(temptags,",;");
      if (! pp) return 0;
      strcpy(tags_recentags,pp+2);                                         //  delimiter + blank before tag
      err = add_tag(tag,tags_recentags,tagRcc);
   }

   return 0;
}


/**************************************************************************/

//  Load tags_defined file into tags_deftags[ii] => category: tag1, tag2, ...
//  Read image_index recs. and add unmatched tags: => nocatg: tag1, tag2, ...

void load_deftags()
{
   int tags_Ucomp(cchar *tag1, cchar *tag2);

   static int  Floaded = 0;
   FILE *      fid;
   sxrec_t     sxrec;
   int         ii, jj, ntags, err, cc, tcc, ftf;
   int         ncats, catoverflow;
   int         nocat, nocatcc;
   char        tag[tagcc], catg[tagcc];
   char        tagsbuff[tagGcc];
   char        *pp1, *pp2;
   char        ptags[maxtags][tagcc];                                      //  10000 * 50 = 0.5 MB

   if (Floaded) return;                                                    //  use memory tags if already there
   Floaded++;

   for (ii = 0; ii < maxtagcats; ii++)                                     //  clean memory
      tags_deftags[ii] = 0;

   ncats = catoverflow = 0;

   fid = fopen(tags_defined_file,"r");                                     //  read tags_defined file
   if (fid) {
      while (true) {
         pp1 = fgets_trim(tagsbuff,tagGcc,fid);
         if (! pp1) break;
         pp2 = strchr(pp1,':');                                            //  isolate "category:"
         if (! pp2) continue;                                              //  no colon
         cc = pp2 - pp1 + 1;
         if (cc > tagcc-1) continue;                                       //  category name too long
         strncpy0(catg,pp1,cc);                                            //  (for error message)
         if (strlen(pp1) > tagGcc-2) goto cattoobig;                       //  all category tags too long
         pp2++;
         while (*pp2 == ' ') pp2++;
         if (strlen(pp2) < 3) continue;                                    //  category with no tags
         while (*pp2) { 
            if (*pp2 == ';') *pp2 = ',';                                   //  replace ';' with ',' for Fotoxx    v.15.07
            pp2++; 
         }
         tags_deftags[ncats] = zstrdup(pp1);                               //  tags_deftags[ii]
         ncats++;                                                          //   = category: tag1, tag2, ... tagN,
         if (ncats == maxtagcats-1) goto toomanycats;                      //  leave 1 for "nocatg"
      }
      err = fclose(fid);
      fid = 0;                                                             //  v.15.07
      if (err) goto deftagsfilerr;
   }

   nocat = ncats;                                                          //  make last category "nocatg" for
   ncats++;                                                                //   unmatched tags in image_index recs.
   tags_deftags[nocat] = (char *) zmalloc(tagGcc);
   strcpy(tags_deftags[nocat],"nocatg: ");
   nocatcc = 8;

   ftf = 1;                                                                //  read all image index recs.

   while (true)
   {
      err = read_sxrec_seq(sxrec,ftf);
      if (err) break;

      pp1 = sxrec.tags;                                                    //  may be "null,"

      while (true)
      {
         while (*pp1 && strchr(",; ",*pp1)) pp1++;                         //  next image tag start
         if (! *pp1) break;
         pp2 = strpbrk(pp1,",;");                                          //  end
         if (! pp2) pp2 = pp1 + strlen(pp1);
         cc = pp2 - pp1;
         if (cc > tagcc-1) {
            pp1 = pp2;
            continue;                                                      //  ignore huge tag
         }

         strncpy0(tag,pp1,cc+1);                                           //  look for tag in defined tags
         if (find_deftag(tag)) {
            pp1 = pp2;                                                     //  found
            continue;
         }

         if (nocatcc + cc + 2 > tagGcc-2) {
            catoverflow = 1;                                               //  nocatg: length limit reached
            break;
         }
         else {
            strcpy(tags_deftags[nocat] + nocatcc, tag);                    //  append tag to list
            nocatcc += cc;
            strcpy(tags_deftags[nocat] + nocatcc, ", ");                   //  + delim + blank
            nocatcc += 2;
         }

         pp1 = pp2;
      }

      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.comms);
      zfree(sxrec.capt);
      zfree(sxrec.gtags);
   }

   if (catoverflow) goto cattoobig;

//  parse all the tags in each category and sort in ascending order

   for (ii = 0; ii < ncats; ii++)
   {
      pp1 = tags_deftags[ii];
      pp2 = strchr(pp1,':');
      cc = pp2 - pp1 + 1;
      strncpy0(catg,pp1,cc);
      pp1 = pp2 + 1;
      while (*pp1 == ' ') pp1++;
      tcc = 0;

      for (jj = 0; jj < maxtags; jj++)
      {
         if (! *pp1) break;
         pp2 = strchr(pp1,',');
         if (pp2) cc = pp2 - pp1;
         else cc = strlen(pp1);
         if (cc > tagcc-1) cc = tagcc-1;
         strncpy0(ptags[jj],pp1,cc+1);
         pp1 += cc + 1;
         tcc += cc;
         while (*pp1 == ' ') pp1++;
      }

      ntags = jj;
      if (ntags == maxtags) goto toomanytags;                              //  v.15.07
      HeapSort((char *) ptags,tagcc,ntags,tags_Ucomp);

      pp1 = tags_deftags[ii];
      tcc += strlen(catg) + 2 + 2 * ntags + 2;                             //  category, all tags, delimiters
      pp2 = (char *) zmalloc(tcc);

      tags_deftags[ii] = pp2;                                              //  swap memory
      zfree(pp1);

      strcpy(pp2,catg);
      pp2 += strlen(catg);
      strcpy(pp2,": ");                                                    //  pp2 = "category: "
      pp2 += 2;

      for (jj = 0; jj < ntags; jj++)                                       //  add the sorted tags
      {
         strcpy(pp2,ptags[jj]);                                            //  append tag + delim + blank
         pp2 += strlen(pp2);
         strcpy(pp2,", ");
         pp2 += 2;
      }

      *pp2 = 0;
   }

//  sort the categories in ascending order
//  leave "nocatg" at the end

   for (ii = 0; ii < ncats-1; ii++)
   for (jj = ii+1; jj < ncats-1; jj++)
   {
      pp1 = tags_deftags[ii];
      pp2 = tags_deftags[jj];
      if (strcasecmp(pp1,pp2) > 0) {
         tags_deftags[ii] = pp2;
         tags_deftags[jj] = pp1;
      }
   }

   return;

toomanycats:
   zmessLogACK(Mwin,"more than %d categories",maxtagcats);
   if (fid) fclose(fid);
   return;

cattoobig:
   zmessLogACK(Mwin,"category %s is too big",catg);
   if (fid) fclose(fid);
   return;

toomanytags:                                                               //  v.15.07
   zmessLogACK(Mwin,"category %s has too many tags",catg);
   if (fid) fclose(fid);
   return;

deftagsfilerr:
   zmessLogACK(Mwin,"tags_defined file error: %s",strerror(errno));
   return;
}


//  compare function for tag sorting

int tags_Ucomp(cchar *tag1, cchar *tag2)
{
   return strcasecmp(tag1,tag2);
}


//  write tags_deftags[] memory data to the defined tags file if any changes were made

void save_deftags()
{
   int         ii, err;
   FILE        *fid;

   fid = fopen(tags_defined_file,"w");                                     //  write tags_defined file
   if (! fid) goto deftagserr;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      if (! tags_deftags[ii+1]) break;                                     //  omit last category, "nocatg"
      err = fprintf(fid,"%s\n",tags_deftags[ii]);                          //  each record:
      if (err < 0) goto deftagserr;                                        //    category: tag1, tag2, ... tagN,
   }

   err = fclose(fid);
   if (err) goto deftagserr;
   return;

deftagserr:
   zmessLogACK(Mwin,"tags_defined file error: %s",strerror(errno));
   return;
}


//  find a given tag in tags_deftags[]
//  return: 1 = found, 0 = not found

int find_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp;

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,", ");
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];                                               //  category: tag1, tag2, ... tagN,
      if (! pp) return 0;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for delim + blank + tag + delim
         if (! pp) break;
         if (strchr(",;:", pp[-2])) return 1;                              //  cat: tag,  or  priortag, tag,
         pp += cc;                                                         //       |                   |
      }                                                                    //       pp                  pp
   }

   return 1;
}


//  add new tag to tags_deftags[] >> category: tag1, tag2, ... newtag,
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad name     4 = null/blank tag
//  if tag present under another category, it is moved to new category

int add_deftag(char *catg, char *tag)
{
   int         ii, cc, cc1, cc2;
   char        catg1[tagcc], tag1[tagcc];
   char        *pp1, *pp2;

   if (! catg) strcpy(catg1,"nocatg");                                     //  revised v.15.05
   else strncpy0(catg1,catg,tagcc);
   cc = strTrim2(catg1);                                                   //  remove leading and trailing blanks
   if (! cc) strcpy(catg1,"nocatg");
   if (utf8_check(catg1)) goto badcatname;                                 //  look for bad characters
   if (strpbrk(catg1,",;:\"")) goto badcatname;

   if (! tag) return 4;
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag1);
   if (! cc) return 4;
   if (utf8_check(tag1)) goto badtagname;                                  //  look for bad characters
   if (strpbrk(tag1,",;:\"")) goto badtagname;

   del_deftag(tag1);                                                       //  delete tag if already there

   cc1 = strlen(catg1);

   for (ii = 0; ii < maxtagcats; ii++)                                     //  look for given category
   {
      pp1 = tags_deftags[ii];
      if (! pp1) goto newcatg;
      if (! strmatchN(catg1,pp1,cc1)) continue;                            //  match on "catname:"
      if (pp1[cc1] == ':') goto oldcatg;
   }

newcatg:
   if (ii == maxtagcats) goto toomanycats;
   cc1 = strlen(catg1) + strlen(tag1) + 6;
   pp1 = (char *) zmalloc(cc1);
   *pp1 = 0;
   strncatv(pp1,cc1,catg1,": ",tag1,", ",null);                            //  category: + tag + delim + blank
   tags_deftags[ii] = tags_deftags[ii-1];                                  //  move "nocatg" record to next slot
   tags_deftags[ii-1] = pp1;                                               //  insert new record before
   return 0;

oldcatg:                                                                   //  logic simplified                   v.15.05
   pp2 = pp1 + cc1 + 2;                                                    //  char following "catname: "
   cc1 = strlen(tag1);
   while (true) {
      pp2 = strcasestr(pp2,tag1);                                          //  look for "tagname, "
      if (! pp2) break;                                                    //  not found
      if (strchr(",;",pp2[cc1])) return 1;                                 //  found, tag not unique
      pp2 += cc1;                                                          //  keep searching
   }
   cc2 = strlen(pp1);                                                      //  add new tag to old record
   if (cc1 + cc2 + 4 > tagGcc) goto cattoobig;
   pp2 = zstrdup(pp1,cc1+cc2+4);                                           //  expand string
   zfree(pp1);
   tags_deftags[ii] = pp2;
   strcpy(pp2+cc2,tag1);                                                   //  old record + tag + delim + blank
   strcpy(pp2+cc2+cc1,", ");
   return 0;

badcatname:
   zmessageACK(Mwin,"bad category name");
   return 3;

badtagname:
   zmessageACK(Mwin,"bad tag name");
   return 3;

toomanycats:
   zmessageACK(Mwin,"too many categories");
   return 2;

cattoobig:
   zmessageACK(Mwin,"too many tags");
   return 2;
}


//  delete tag from defined tags list, tags_deftags[]
//  return: 0 = found and deleted, 1 = not found

int del_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp, *pp1, *pp2;

   if (! tag || *tag <= ' ') return 1;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,", ");
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for prior delim or colon
         if (! pp) break;
         if (strchr(",;:", pp[-2])) goto found;                            //  cat: tag,  or  priortag, tag,
         pp += cc;                                                         //       |                   |
      }                                                                    //       pp                  pp
   }

found:
   for (pp1 = pp, pp2 = pp+cc; *pp2; pp1++, pp2++)                         //  eliminate tag, delim, blank
      *pp1 = *pp2;
   *pp1 = 0;

   return 0;
}


//  stuff tags_deftags[] into text widget and format by category

void deftags_stuff(zdialog *zd)
{
   GtkWidget      *widget;
   int            ii, cc;
   char           catgname[tagcc+3];
   char           *pp1, *pp2;

   widget = zdialog_widget(zd,"deftags");
   wclear(widget);

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp1 = tags_deftags[ii];
      if (! pp1) break;
      pp2 = strchr(pp1,':');
      if (! pp2) continue;
      if (pp2 > pp1 + tagcc-3) continue;
      pp2 += 2;
      cc = pp2 - pp1 + 1;
      strncpy0(catgname,pp1,cc);
      wprintx(widget,0,catgname,"monospace bold 9");                       //  "category: " in bold text
      if (*pp2) wprintx(widget,0,pp2,"monospace 9");                       //  "cat1, cat2, ... catN," 
      wprintx(widget,0,"\n");
   }

   return;
}


//  report tags defined and not used in any image file

void tag_orphans()
{
   FILE        *fid;
   sxrec_t     sxrec;
   int         ii, cc, err, ftf;
   int         Ndeftags;
   char        **deftags;
   char        usedtag[tagcc], tagsbuff[tagGcc];
   char        *pp1, *pp2, text[20];

   deftags = (char **) zmalloc(maxtags * sizeof(char *));                  //  allocate memory
   Ndeftags = 0;

   fid = fopen(tags_defined_file,"r");                                     //  read tags_defined file
   if (fid) {
      while (true) {
         pp1 = fgets_trim(tagsbuff,tagGcc,fid);
         if (! pp1) break;
         pp1 = strchr(pp1,':');                                            //  skip over "category:"
         if (! pp1) continue;
         cc = pp1 - tagsbuff;
         if (cc > tagcc) continue;                                         //  reject bad data (manual edit?)
         pp1++;
         for (ii = 1; ; ii++) {                                            //  get tags: tag1, tag2, ...
            pp2 = (char *) strField(pp1,",;",ii);
            if (! pp2) break;
            if (strlen(pp2) < 3) continue;                                 //  reject bad data
            if (strlen(pp2) > tagcc) continue;
            deftags[Ndeftags] = zstrdup(pp2);
            Ndeftags++;
         }
      }
      fclose(fid);
   }

   ftf = 1;                                                                //  read all image index recs.

   while (true)
   {
      err = read_sxrec_seq(sxrec,ftf);
      if (err) break;

      pp1 = sxrec.tags;                                                    //  image tags
      if (! pp1) continue;
      if (strmatchN(pp1,"null",4)) continue;

      while (true)
      {
         while (*pp1 && strchr(",; ",*pp1)) pp1++;                         //  next image tag start
         if (! *pp1) break;
         pp2 = strpbrk(pp1,",;");                                          //  end
         if (! pp2) pp2 = pp1 + strlen(pp1);
         cc = pp2 - pp1;
         if (cc > tagcc-1) {
            pp1 = pp2;
            continue;                                                      //  ignore huge tag
         }

         strncpy0(usedtag,pp1,cc+1);                                       //  used tag, without delimiter

         for (ii = 0; ii < Ndeftags; ii++)                                 //  find in defined tags
            if (strmatch(usedtag,deftags[ii])) break;

         if (ii < Ndeftags) {                                              //  found
            zfree(deftags[ii]);
            Ndeftags--;
            while (ii < Ndeftags) {                                        //  defined tag is in use
               deftags[ii] = deftags[ii+1];                                //  remove from list and pack down
               ii++;
            }
         }

         pp1 = pp2;
      }
   }

   write_popup_text("open","unused tags",200,200,Mwin);
   for (ii = 0; ii < Ndeftags; ii++)
      write_popup_text("write",deftags[ii]);
   snprintf(text,20,"%d unused tags",Ndeftags);
   write_popup_text("write",text);

   for (ii = 0; ii < Ndeftags; ii++)
      zfree(deftags[ii]);
   zfree(deftags);

   return;
}


/**************************************************************************/

//  edit EXIF/IPTC data - add or change specified EXIF/IPTC/etc. key

void m_meta_edit_any(GtkWidget *, cchar *menu)
{
   int  meta_edit_any_dialog_event(zdialog *zd, cchar *event);

   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1];
   char        *pp2[1];

   if (menu) F1_help_topic = "edit_any_metadata";
   if (! curr_file) return;

   if (! zdexifedit)                                                       //  popup dialog if not already
   {
      zdexifedit = zdialog_new(ZTX("Edit Metadata"),Mwin,Bfetch,Bsave,Bdone,null);
      zdialog_add_widget(zdexifedit,"vbox","hb1","dialog");
      zdialog_add_widget(zdexifedit,"hbox","hbkey","dialog",0,"space=2");
      zdialog_add_widget(zdexifedit,"hbox","hbdata","dialog",0,"space=2");
      zdialog_add_widget(zdexifedit,"label","labkey","hbkey",ZTX("key name"));
      zdialog_add_widget(zdexifedit,"entry","keyname","hbkey",0,"size=20");
      zdialog_add_widget(zdexifedit,"label","labdata","hbdata",ZTX("key value"));
      zdialog_add_widget(zdexifedit,"entry","keydata","hbdata",0,"expand");
      zdialog_run(zdexifedit,meta_edit_any_dialog_event);
   }

   zdialog_fetch(zdexifedit,"keyname",keyname,40);                         //  get key name from dialog
   strCompress(keyname);

   if (*keyname)                                                           //  update live dialog
   {
      pp1[0] = keyname;                                                    //  look for key data
      exif_get(curr_file,pp1,pp2,1);
      if (pp2[0]) {
         strncpy0(keydata,pp2[0],exif_maxcc);
         zfree(pp2[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zdexifedit,"keydata",keydata);                         //  stuff into dialog
   }

   return;
}


//  dialog event and completion callback function

int  meta_edit_any_dialog_event(zdialog *zd, cchar *event)
{
   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1], *pp2[1];
   char        *pp3[1];
   int         err;

   if (strmatch(event,"escape")) zd->zstat = 4;                            //  escape = cancel                    v.15.07

   if (! zd->zstat) return 1;
   if (! curr_file) return 1;

   zdialog_fetch(zd,"keyname",keyname,40);
   zdialog_fetch(zd,"keydata",keydata,exif_maxcc);
   strCompress(keyname);                                                   //  remove blanks

   if (zd->zstat == 1)                                                     //  fetch
   {
      zd->zstat = 0;
      if (! *keyname) return 0;
      pp1[0] = keyname;
      exif_get(curr_file,pp1,pp3,1);
      if (pp3[0]) {
         strncpy0(keydata,pp3[0],exif_maxcc);
         zfree(pp3[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zd,"keydata",keydata);
   }

   else if (zd->zstat == 2)                                                //  save
   {
      zd->zstat = 0;                                                       //  keep dialog active
      if (! *keyname) return 1;
      pp1[0] = keyname;
      pp2[0] = keydata;
      err = exif_put(curr_file,pp1,pp2,1);                                 //  change metadata in image file
      if (err) zmessageACK(Mwin,"error: %s",strerror(err));
      load_filemeta(curr_file);                                            //  update image index in case         v.15.03
      update_image_index(curr_file);                                       //    searchable metadata item updated
      if (zdexifview) meta_view(0);                                        //  update exif view if active
   }

   else zdialog_free(zdexifedit);                                          //  other
   return 1;
}


/**************************************************************************/

//  delete EXIF/IPTC data, specific key or all data

void m_meta_delete(GtkWidget *, cchar *menu)
{
   int   meta_delete_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;

   F1_help_topic = "delete_metadata";
   if (! curr_file) return;

   zd = zdialog_new(ZTX("Delete Metadata"),Mwin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","kall","hb1",ZTX("All"),"space=5");
   zdialog_add_widget(zd,"radio","key1","hb1",ZTX("One Key:"));
   zdialog_add_widget(zd,"entry","keyname","hb1",0,"size=20");
   zdialog_stuff(zd,"key1",1);

   zdialog_run(zd,meta_delete_dialog_event);
   return;
}


//  dialog event and completion callback function

int meta_delete_dialog_event(zdialog *zd, cchar *event)
{
   int         kall, key1;
   char        keyname[40];

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat != 1 || ! curr_file) {                                    //  canceled
      zdialog_free(zd);
      return 1;
   }

   zd->zstat = 0;                                                          //  dialog remains active
   gtk_window_present(MWIN);                                               //  keep focus on main window          v.14.09

   zdialog_fetch(zd,"kall",kall);
   zdialog_fetch(zd,"key1",key1);
   zdialog_fetch(zd,"keyname",keyname,40);
   strCompress(keyname);

   if (kall)                                                               //  update file metadata
      shell_ack("exiftool -m -q -overwrite_original -all=  \"%s\"",curr_file);
   else if (key1)
      shell_ack("exiftool -m -q -overwrite_original -%s=  \"%s\"",keyname,curr_file);
   else return 1;

   load_filemeta(curr_file);                                               //  update image index in case a       v.15.03
   update_image_index(curr_file);                                          //    searchable metadata deleted

   if (zdexifview) meta_view(0);                                           //  update exif view if active
   return 1;
}


/**************************************************************************/

//  menu function - add and remove tags for many files at once

namespace batchtags
{
   char        **filelist = 0;                                             //  files to process
   int         filecount = 0;                                              //  file count
   char        addtags[tagMcc];                                            //  tags to add, list
   char        deltags[tagMcc];                                            //  tags to remove, list
}


void m_batchTags(GtkWidget *, cchar *)                                     //  combine batch add/del tags
{
   using namespace batchtags;

   void  batch_addtags_clickfunc(GtkWidget *widget, int line, int pos);
   void  batch_deltags_clickfunc(GtkWidget *widget, int line, int pos);
   void  batch_deftags_clickfunc(GtkWidget *widget, int line, int pos);
   int   batchTags_dialog_event(zdialog *zd, cchar *event);

   char        *ptag, *file;
   int         zstat, ii, jj, err;
   zdialog     *zd;
   GtkWidget   *widget;

   F1_help_topic = "batch_tags";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
               Batch Add/Remove Tags

         [Select Files]  NN files selected

         (o) tags to add    [_________________________]
         (o) tags to remove [_________________________]

          __ defined tags ____________________________
         |                                            |
         |                                            |
         |                                            |
         |                                            |
         |____________________________________________|

                      [manage tags] [proceed] [cancel]
***/

   zd = zdialog_new(ZTX("Batch Add/Remove Tags"),Mwin,Bmanagetags,Bproceed,Bcancel,null);
   zdbatchtags = zd;

   zdialog_add_widget(zd,"hbox","hbfiles","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","files","hbfiles",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","labcount","hbfiles",Bnofileselected,"space=10");

   zdialog_add_widget(zd,"hbox","hbtags","dialog",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb1","hbtags",0,"space=3|homog");
   zdialog_add_widget(zd,"vbox","vb2","hbtags",0,"space=3|homog|expand");

   zdialog_add_widget(zd,"radio","radadd","vb1",ZTX("tags to add"));
   zdialog_add_widget(zd,"radio","raddel","vb1",ZTX("tags to remove"));

   zdialog_add_widget(zd,"frame","fradd","vb2",0,"expand");
   zdialog_add_widget(zd,"text","addtags","fradd",0,"expand|wrap");
   zdialog_add_widget(zd,"frame","frdel","vb2",0,"expand");
   zdialog_add_widget(zd,"text","deltags","frdel",0,"expand|wrap");

   zdialog_add_widget(zd,"hbox","hbspace","dialog",0,"space=3");
   zdialog_add_widget(zd,"hbox","hbdef1","dialog");
   zdialog_add_widget(zd,"label","labdef1","hbdef1",ZTX("Defined Tags:"),"space=5");
   zdialog_add_widget(zd,"hbox","hbdef2","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frdef2","hbdef2",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrdef2","frdef2",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrdef2",0,"expand|wrap");

   zdialog_stuff(zd,"radadd",1);
   zdialog_stuff(zd,"raddel",0);

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   filelist = 0;
   filecount = 0;
   *addtags = *deltags = 0;

   widget = zdialog_widget(zd,"addtags");                                  //  tag widget mouse functions
   textwidget_set_clickfunc(widget,batch_addtags_clickfunc);

   widget = zdialog_widget(zd,"deltags");
   textwidget_set_clickfunc(widget,batch_deltags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,batch_deftags_clickfunc);

   zdialog_resize(zd,500,400);                                             //  run dialog
   zdialog_run(zd,batchTags_dialog_event);
   zstat = zdialog_wait(zd);                                               //  wait for dialog completion
   zdialog_free(zd);                                                       //  kill dialog

   zdbatchtags = 0;

   if (zstat != 2)                                                         //  cancel
   {
      if (filecount) {
         for (ii = 0; filelist[ii]; ii++)
            zfree(filelist[ii]);
         zfree(filelist);
      }

      Fblock = 0;
      return;
   }

   write_popup_text("open","Batch Tags",500,200,Mwin);                     //  status monitor popup window

   for (ii = 0; filelist[ii]; ii++)                                        //  loop all selected files
   {
      file = filelist[ii];                                                 //  display image
      err = f_open(file,0,0,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      for (jj = 1; ; jj++)                                                 //  remove tags if present
      {
         ptag = (char *) strField(deltags,",;",jj);
         if (! ptag) break;
         if (*ptag == ' ') continue;
         err = del_tag(ptag,meta_tags);
         if (err) continue;
      }

      for (jj = 1; ; jj++)                                                 //  add new tags unless already
      {
         ptag = (char *) strField(addtags,",;",jj);
         if (! ptag) break;
         if (*ptag == ' ') continue;
         err = add_tag(ptag,meta_tags,tagFcc);
         if (err == 2) {
            zmessageACK(Mwin,ZTX("%s \n too many tags"),file);
            break;
         }
      }

      save_filemeta(file);                                                 //  save tag changes
   }

   write_popup_text("write","COMPLETED");

   for (ii = 0; filelist[ii]; ii++)
      zfree(filelist[ii]);
   zfree(filelist);

   Fblock = 0;
   return;
}


//  mouse click functions for widgets holding tags

void batch_addtags_clickfunc(GtkWidget *widget, int line, int pos)         //  a tag in the add list was clicked
{
   using namespace batchtags;

   char     *txline, *txtag, end;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   del_tag(txtag,addtags);                                                 //  remove tag from list
   zdialog_stuff(zdbatchtags,"addtags",addtags);

   zfree(txline);
   zfree(txtag);
   return;
}


void batch_deltags_clickfunc(GtkWidget *widget, int line, int pos)         //  a tag in the remove list was clicked
{
   using namespace batchtags;

   char     *txline, *txtag, end;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   del_tag(txtag,deltags);                                                 //  remove tag from list
   zdialog_stuff(zdbatchtags,"deltags",deltags);

   zfree(txline);
   zfree(txtag);
   return;
}


void batch_deftags_clickfunc(GtkWidget *widget, int line, int pos)         //  a defined tag was clicked
{
   using namespace batchtags;

   char     *txline, *txtag, end;
   int      radadd;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;:",end);
   if (! txtag || end == ':') { zfree(txline); return; }                   //  tag category clicked, ignore       v.15.5

   zdialog_fetch(zdbatchtags,"radadd",radadd);                             //  which radio button?

   if (radadd) {
      add_tag(txtag,addtags,tagMcc);                                       //  add defined tag to tag add list
      zdialog_stuff(zdbatchtags,"addtags",addtags);
   }
   else {
      add_tag(txtag,deltags,tagMcc);                                       //  add defined tag to tag remove list
      zdialog_stuff(zdbatchtags,"deltags",deltags);
   }

   zfree(txline);
   zfree(txtag);
   return;
}


//  batchTags dialog event function

int batchTags_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batchtags;

   int      ii;
   char     countmess[50];

   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  manage tags
         zd->zstat = 0;                                                    //  keep dialog active
         zdialog_show(zd,0);                                               //  hide parent dialog
         manage_tags();
         zdialog_show(zd,1);
      }

      if (zd->zstat == 2) {                                                //  proceed
         if (! filecount || (*addtags <= ' ' && *deltags <= ' ')) {
            zmessageACK(Mwin,ZTX("specify files and tags"));
            zd->zstat = 0;                                                 //  keep dialog active
         }
      }

      return 1;                                                            //  cancel
   }

   if (strmatch(event,"files"))                                            //  select images to process
   {
      if (filelist) {                                                      //  free prior list
         for (ii = 0; filelist[ii]; ii++)
            zfree(filelist[ii]);
         zfree(filelist);
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      filelist = gallery_getfiles();                                       //  get file list from user
      zdialog_show(zd,1);

      filelist = filelist;

      if (filelist)                                                        //  count files in list
         for (ii = 0; filelist[ii]; ii++);
      else ii = 0;
      filecount = ii;

      snprintf(countmess,50,Bfileselected,filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }

   return 1;
}


/**************************************************************************/

//  Convert tag names for all image files, based on a list 
//    of old tag names and corresponding new tag names.

void m_batch_renametags(GtkWidget *, cchar *menu)                          //  v.15.07
{
   char     *tagfile, buff[200], logrec[200];
   char     *taglist[maxtags][2], *filelist[maximages];
   char     *oldtaglist[100], *newtaglist[100];
   char     *pp, *file;
   char     *oldtag, *newtag, *filetag;
   int      ntags = 0, nfiles = 0, nlist;
   int      err, ii, jj, kk, ff, yn, ftf = 1;
   FILE     *fid = 0;
   sxrec_t  sxrec;

   F1_help_topic = "batch_rename_tags";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

   tagfile = zgetfile(ZTX("tag names file"),MWIN,"file",getenv("HOME"));   //  get tag names file from user
   if (! tagfile) {
      Fblock = 0;
      return;
   }

   fid = fopen(tagfile,"r");                                               //  open file
   if (! fid) {
      zmessageACK(Mwin,strerror(errno));
      Fblock = 0;
      return;
   }

   write_popup_text("open","batch rename tags",600,400,Mwin);

   while (true)                                                            //  read old and new tag names
   {
      pp = fgets_trim(buff,200,fid,1);
      if (! pp) break;
      if (blank_null(pp)) continue;
      oldtag = (char *) strField(buff,',',1);
      if (! oldtag) continue;
      taglist[ntags][0] = zstrdup(oldtag);                                 //  tag to replace
      newtag = (char *) strField(buff,',',2);
      if (newtag) taglist[ntags][1] = zstrdup(newtag);                     //  replacement tag
      else taglist[ntags][1] = 0;                                          //  none - tag will be deleted
      snprintf(logrec,200,"old tag: %s  new tag: %s",oldtag,newtag);
      write_popup_text("write",logrec);
      ntags++;
      if (ntags == maxtags) {
         zmessageACK(Mwin,"max tags exceeded: %d",maxtags);
         goto cleanup;
      }
   }
   
   fclose(fid);
   fid = 0;
   
   while (true)                                                            //  read image index file
   {
      err = read_sxrec_seq(sxrec,ftf);                                     //  loop each image file
      if (err) break;
      ff = 0;
      if (sxrec.tags) {                                                    //  search for tags to rename
         for (ii = 1; ; ii++) {
            pp = (char *) strField(sxrec.tags,',',ii);
            if (! pp) break;
            if (strmatch(pp,"null")) continue;
            for (jj = 0; jj < ntags; jj++) {
               if (strmatchcase(pp,taglist[jj][0])) {                      //  this file has one or more tags
                  ff = 1;                                                  //    that will be renamed
                  break;
               }
            }
            if (ff) break;
         }
      }
      
      if (ff) {
         filelist[nfiles] = zstrdup(sxrec.file);                           //  add to list of files to process
         nfiles++;
         snprintf(logrec,200,"file included: %s",sxrec.file);
         write_popup_text("write",logrec);
      }

      zfree(sxrec.file);                                                   //  free index rec. memory
      if (sxrec.tags) zfree(sxrec.tags);
      if (sxrec.capt) zfree(sxrec.capt);
      if (sxrec.comms) zfree(sxrec.comms);
      if (sxrec.gtags) zfree(sxrec.gtags);
   }
   
   yn = zmessageYN(Mwin,ZTX("%d tags to rename \n"
                            "in %d image files. \n"
                            "Proceed?"),ntags,nfiles);
   if (! yn) goto cleanup;

   for (ii = 0; ii < nfiles; ii++)                                         //  loop all selected files
   {
      file = filelist[ii];                                                 //  open image file
      err = f_open(file,0,0,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress

      nlist = 0;

      for (jj = 1; ; jj++) {                                               //  loop file tags
         filetag = (char *) strField(meta_tags,',',jj);
         if (! filetag) break;
         for (kk = 0; kk < ntags; kk++) {                                  //  loop tag replacement list
            oldtag = taglist[kk][0];
            newtag = taglist[kk][1];
            if (strmatchcase(filetag,oldtag)) {                            //  file tag matches tag to replace
               oldtaglist[nlist] = oldtag;                                 //  save old and new tags 
               newtaglist[nlist] = newtag;
               nlist++;
               break;                                                      //  next file tag
            }
         }
      }

      for (jj = 0; jj < nlist; jj++)                                       //  loop tags to remove
         err = del_tag(oldtaglist[jj],meta_tags);

      for (jj = 0; jj < nlist; jj++) {                                     //  loop tags to add
         if (! newtaglist[jj]) continue;                                   //  must be after removals
         write_popup_text("write",newtaglist[jj]);
         err = add_tag(newtaglist[jj],meta_tags,tagFcc);
         if (err && err != 1) write_popup_text("write","ERROR");           //  ignore if already there
      }

      save_filemeta(file);                                                 //  save tag changes
   }

   write_popup_text("write","COMPLETED");

cleanup:                                                                   //  free resources

   Fblock = 0;

   if (fid) fclose(fid);
   
   for (ii = 0; ii < ntags; ii++) {
      zfree(taglist[ii][0]);
      if (taglist[ii][1]) zfree(taglist[ii][1]);
   }
   
   for (ii = 0; ii < nfiles; ii++)
      zfree(filelist[ii]);
      
   return;
}


/**************************************************************************/

//  image file EXIF/IPTC data >> memory data:
//    meta_pdate, meta_rating, meta_tags, meta_comments, meta_caption,
//    meta_city, meta_country, meta_latitude, meta_longitude

void load_filemeta(cchar *file)
{
   int      ii, jj, cc;
   char     *pp;
   cchar    *exifkeys[10] = { exif_date_key, iptc_keywords_key,
                              iptc_rating_key, exif_size_key,
                              exif_comment_key, iptc_caption_key,
                              exif_city_key, exif_country_key,
                              exif_latitude_key, exif_longitude_key };

   char     *ppv[10], *imagedate, *imagekeywords, *imagestars, *imagesize;
   char     *imagecomms, *imagecapt;
   char     *imagecity, *imagecountry, *imagelatitude, *imagelongitude;

   *meta_tags = *meta_pdate = *meta_comments = *meta_caption = 0;
   strcpy(meta_rating,"0");
   *meta_city = *meta_country = *meta_latitude = *meta_longitude = 0;

   exif_get(file,exifkeys,ppv,10);                                         //  get metadata from image file
   imagedate = ppv[0];
   imagekeywords = ppv[1];
   imagestars = ppv[2];
   imagesize = ppv[3];
   imagecomms = ppv[4];
   imagecapt = ppv[5];
   imagecity = ppv[6];
   imagecountry = ppv[7];
   imagelatitude = ppv[8];
   imagelongitude = ppv[9];

   if (imagedate) {
      exif_tagdate(imagedate,meta_pdate);                                  //  EXIF date/time >> yyyymmddhhmmss
      strcpy(meta_ppdate,meta_pdate);
      zfree(imagedate);
   }

   if (imagekeywords)
   {
      for (ii = 1; ; ii++)
      {
         pp = (char *) strField(imagekeywords,",;",ii);
         if (! pp) break;
         if (*pp == ' ') continue;
         cc = strlen(pp);
         if (cc >= tagcc) continue;                                        //  reject tags too big
         for (jj = 0; jj < cc; jj++)
            if (pp[jj] > 0 && pp[jj] < ' ') break;                         //  reject tags with control characters
         if (jj < cc) continue;
         add_tag(pp,meta_tags,tagFcc);                                     //  add to file tags if unique
      }

      zfree(imagekeywords);
   }

   if (imagestars) {
      meta_rating[0] = *imagestars;
      if (meta_rating[0] < '0' || meta_rating[0] > '5') meta_rating[0] = '0';
      meta_rating[1] = 0;
      zfree(imagestars);
   }

   if (imagesize) {
      strncpy0(meta_size,imagesize,15);
      zfree(imagesize);
   }

   if (imagecomms) {
      strncpy0(meta_comments,imagecomms,exif_maxcc);
      zfree(imagecomms);
   }

   if (imagecapt) {
      strncpy0(meta_caption,imagecapt,exif_maxcc);
      zfree(imagecapt);
   }

   if (imagecity) {                                                        //  geotags
      strncpy0(meta_city,imagecity,99);
      zfree(imagecity);
   }
   else strcpy(meta_city,"null");                                          //  replace missing data with "null"

   if (imagecountry) {
      strncpy0(meta_country,imagecountry,99);
      zfree(imagecountry);
   }
   else strcpy(meta_country,"null");

   if (imagelatitude) {
      strncpy0(meta_latitude,imagelatitude,12);
      zfree(imagelatitude);
   }
   else strcpy(meta_latitude,"null");

   if (imagelongitude) {
      strncpy0(meta_longitude,imagelongitude,12);
      zfree(imagelongitude);
   }
   else strcpy(meta_longitude,"null");

   Fmetamod = Fgtagmod = 0;                                                //  no pending changes                 v.15.03
   return;
}


//  add metadata in memory to image file EXIF/IPTC data and image_index recs.

void save_filemeta(cchar *file)
{
   cchar    *exifkeys[10] = { exif_date_key, iptc_keywords_key,
                              iptc_rating_key, exif_size_key,
                              exif_comment_key, iptc_caption_key,
                              exif_city_key, exif_country_key,
                              exif_latitude_key, exif_longitude_key };
   cchar       *exifdata[10];
   char        imagedate[24];

   *imagedate = 0;

   if (*meta_pdate) {
      tag_exifdate(meta_pdate,imagedate);                                  //  yyyymmddhhmmss >> EXIF date/time
      strcpy(meta_ppdate,meta_pdate);
   }

   exifdata[0] = imagedate;                                                //  update file EXIF/IPTC data
   exifdata[1] = meta_tags;
   exifdata[2] = meta_rating;
   exifdata[3] = meta_size;
   exifdata[4] = meta_comments;
   exifdata[5] = meta_caption;

   if (strmatch(meta_city,"null")) exifdata[6] = "";                       //  geotags
   else exifdata[6] = meta_city;                                           //  if "null" erase EXIF
   if (strmatch(meta_country,"null")) exifdata[7] = "";
   else exifdata[7] = meta_country;
   if (strmatch(meta_latitude,"null") || strmatch(meta_longitude,"null"))
      exifdata[8] = exifdata[9] = "";
   else {
      exifdata[8] = meta_latitude;
      exifdata[9] = meta_longitude;
   }

   exif_put(file,exifkeys,exifdata,10);                                    //  write EXIF

   update_image_index(file);                                               //  update image index file

   if (zdexifview) meta_view(0);                                           //  live EXIF/IPTC update

   Fmetamod = Fgtagmod = 0;                                                //  no pending changes                 v.15.03
   return;
}


//  update image index record (replace updated file data)

void update_image_index(cchar *file)                                       //  overhauled
{
   char     gtags[200];
   int      err;
   sxrec_t  sxrec;
   STATB    statb;

   err = stat(file,&statb);                                                //  build new metadata record to insert
   if (err) {                                                              //    or replace
      zmessageACK(Mwin,ZTX("file not found"));
      return;
   }

   memset(&sxrec,0,sizeof(sxrec_t));

   sxrec.file = (char *) file;                                             //  image filespec

   compact_time(statb.st_mtime,sxrec.fdate);                               //  convert file date to "yyyymmddhhmmss"

   strncpy0(sxrec.pdate,meta_pdate,15);                                    //  photo date, "yyyymmddhhmmss"

   sxrec.rating[0] = meta_rating[0];                                       //  rating '0' to '5' stars
   sxrec.rating[1] = 0;

   strncpy0(sxrec.size,meta_size,15);                                      //  2345x12345

   if (*meta_tags)                                                         //  tags
      sxrec.tags = meta_tags;

   if (*meta_caption)                                                      //  user caption
      sxrec.capt = meta_caption;

   if (*meta_comments)                                                     //  user comments
      sxrec.comms = meta_comments;

   if (! *meta_city) strcpy(meta_city,"null");                             //  geotags
   if (! *meta_country) strcpy(meta_country,"null");
   if (! *meta_latitude) strcpy(meta_latitude,"null");                     //  "null" for city/country is searchable
   if (! *meta_longitude) strcpy(meta_longitude,"null");

   snprintf(gtags,200,"%s^ %s^ %s^ %s",meta_city, meta_country,
                                  meta_latitude, meta_longitude);
   sxrec.gtags = gtags;

   put_sxrec(&sxrec,file);
   return;
}


//  delete given image file from image index recs.

void delete_image_index(cchar *file)
{
   put_sxrec(null,file);
   return;
}


/**************************************************************************
            GEOTAG functions
***************************************************************************/

struct geolocs_t {                                                         //  geotag location table, memory DB
   char     *city, *country;
   char     *lati, *longi;
   float    flati, flongi;
};
geolocs_t   *geolocs;                                                      //  maps location names <-> latitude/longitude
int         Ngeolocs = 0;                                                  //  size of geolocations table

zdialog  *zd_geotags;                                                      //  active zdialog to get geotags data
float    geomap_range = 10;                                                //  default geotag search range, km
char     prevcity[100] = "", prevcountry[100] = "";                        //  last geotag data used
char     prevlatitude[20] = "", prevlongitude[20] = "";                    //  (recalled with [prev] button)

int  geomap_position(float flat, float flong, int &mx, int &my);           //  latitude/longitude > map position
int  geomap_coordinates(int mx, int my, float &flat, float &flong);        //  map position > latitude/longitude
void geomap_paint_dots();                                                  //  paint red dots on map where images located
int  geotags_choosecity(char *location[2], char *coord[2]);                //  choose one city from multiple options
cchar * web_geocode(char *location[2], char *coord[2]);                    //  find latitude/longitude via web service


/**************************************************************************/

//  validate and convert earth coordinates, latitude and longitude
//  return: 0  OK
//          1  both are missing ("null" or "")
//          2  invalid data
//  if status is > 0, 0.0 is returned for both values

int validate_latlong(char *latitude, char *longitude, float &flati, float &flongi)
{
   if (! latitude || *latitude == 0 || strmatch(latitude,"null"))
      if (! longitude || *longitude == 0 || strmatch(longitude,"null"))
         goto status1;                                                     //  both missing

   if (! latitude || *latitude == 0 || strmatch(latitude,"null"))
      goto status2;                                                        //  one missing
   if (! longitude || *longitude == 0 || strmatch(longitude,"null"))
      goto status2;

   for (int ii = 0; latitude[ii]; ii++)                                    //  replace comma decimal point        v.14.12
      if (latitude[ii] == ',') latitude[ii] = '.';                         //    with period
   for (int ii = 0; longitude[ii]; ii++)
      if (longitude[ii] == ',') longitude[ii] = '.';

   flati = atof(latitude);                                                 //  convert strings to float
   flongi = atof(longitude);
   if (flati < -90.0 || flati > +90.0) goto status2;                       //  check range
   if (flongi < -180.0 || flongi > +180.0) goto status2;
   if (flati == 0.0 && flongi == 0.0) goto status2;                        //  reject both = 0.0
   return 0;

status1:
   flati = flongi = 0.0;                                                   //  both missing
   return 1;

status2:                                                                   //  either missing or invalid
   flati = flongi = 0.0;
   return 2;
}


/**************************************************************************/

//  Initialize for geotag functions.
//  Load geolocations data into memory from image index files.
//  Returns no. geolocations loaded.

int init_geolocs()
{
   int  init_glocs_comp(cchar *rec1, cchar *rec2);                         //  code rationalized                  v.14.11

   char     city[100], country[100];
   char     latitude[20], longitude[20];
   char     *gtags, *pp;
   float    flati, flongi;
   int      err, ftf, cc, ii, jj, keep, Nimages;
   sxrec_t  sxrec;

   if (Ngeolocs) return Ngeolocs;                                          //  already done

   Ffuncbusy = 1;

   cc = maxgeolocs * sizeof(geolocs_t);                                    //  get memory for geotag locations DB
   geolocs = (geolocs_t *) zmalloc(cc);

   //  populate geolocs[] from search-index file (= image EXIF data)

   Nimages = 0;
   ftf = 1;

   while (true)
   {
      zmainloop(100);

      if (Ngeolocs == maxgeolocs) {                                        //  table full
         zmessLogACK(Mwin,"max. geotags %d exceeded",maxgeolocs);
         break;
      }

      err = read_sxrec_seq(sxrec,ftf);                                     //  read image index recs.
      if (err) break;

      gtags = sxrec.gtags;
      strcpy(city,"null");
      strcpy(country,"null");
      strcpy(latitude,"null");
      strcpy(longitude,"null");

      pp = (char *) strField(gtags,'^',1);                                 //  city name or "null"
      if (pp) strncpy0(city,pp,99);
      pp = (char *) strField(gtags,'^',2);                                 //  country
      if (pp) strncpy0(country,pp,99);
      pp = (char *) strField(gtags,'^',3);                                 //  latitude
      if (pp) strncpy0(latitude,pp,12);
      pp = (char *) strField(gtags,'^',4);                                 //  longitude
      if (pp) strncpy0(longitude,pp,12);

      zfree(sxrec.file);                                                   //  free sxrec allocations
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);

      if (! strmatch(latitude,"null") || ! strmatch(longitude,"null")) {
         err = validate_latlong(latitude,longitude,flati,flongi);          //  validate and replace bad data
         if (err) {
            strcpy(latitude,"null");                                       //  1/2 = missing/bad data
            strcpy(longitude,"null");
            continue;
         }
      }

      Nimages++;                                                           //  images with valid geotags

      if (Ngeolocs) {
         ii = Ngeolocs - 1;
         if (strmatch(city,geolocs[ii].city))                              //  skip same city in sequence (frequent)
            if (strmatch(country,geolocs[ii].country)) continue;           //    to reduce subsequent sort
      }

      ii = Ngeolocs++;                                                     //  fill next entry in table
      geolocs[ii].city = zstrdup(city);
      geolocs[ii].country = zstrdup(country);
      geolocs[ii].lati = zstrdup(latitude);
      geolocs[ii].longi = zstrdup(longitude);
      geolocs[ii].flati = atof(latitude);
      geolocs[ii].flongi = atof(longitude);
   }

   cc = sizeof(geolocs_t);
   HeapSort((char *) geolocs, cc, Ngeolocs, init_glocs_comp);              //  sort by country, city

   for (ii = 0, jj = 1; jj < Ngeolocs; jj++)                               //  eliminate duplicate cities
   {
      keep = 0;
      if (! strmatch(geolocs[jj].country,geolocs[ii].country)) keep = 1;
      if (! strmatch(geolocs[jj].city,geolocs[ii].city)) keep = 1;
      if (keep) {
         ii++;
         geolocs[ii] = geolocs[jj];
      }
      else {
         zfree(geolocs[jj].country);
         zfree(geolocs[jj].city);
         zfree(geolocs[jj].lati);
         zfree(geolocs[jj].longi);
      }
   }

   Ngeolocs = ii + 1;
   printz("%d images have %d unique geotags \n",Nimages,Ngeolocs);

   Ffuncbusy = 0;
   return Ngeolocs;
}


//  Compare 2 geolocs records by country and city
//  return  <0  0  >0   for   rec1  <  ==  >  rec2.

int  init_glocs_comp(cchar *rec1, cchar *rec2)
{
   int      ii;

   char * country1 = ((geolocs_t *) rec1)->country;                        //  compare countries
   char * country2 = ((geolocs_t *) rec2)->country;
   ii = strcmp(country1,country2);
   if (ii) return ii;

   char * city1 = ((geolocs_t *) rec1)->city;                              //  compare cities
   char * city2 = ((geolocs_t *) rec2)->city;
   ii = strcmp(city1,city2);
   return ii;
}


/**************************************************************************/

//  get latitude/longitude data for a city [ country ]
//  inputs:        location[0] = city
//                 location[1] = country (opt)
//  outputs:       coord[2] = latitude -90 to +90, longitude -180 to +180
//                 matches[20][2] up to 20 matching city/country locations
//  returns:       no. matches for input city [ country ]
//                 (max. 20)
//
//  use null or "" for missing city or country input (no NULL pointer)
//  coordinates are returned for the first match only, if any
//

int get_geolocs(char *location[2], char *coord[2], char *matches[20][2])
{
   int      cc, ii, jj, Nmatch;
   int      fcity = 0, fcountry = 0;

   if (location[0])                                                        //  v.14.12
      if (*location[0]) fcity = 1;
   if (location[1])
      if (*location[1]) fcountry = 1;

   if (! fcity && ! fcountry) return 0;                                    //  one of these must be present

   for (ii = Nmatch = 0; ii < Ngeolocs; ii++)                              //  search for exact city [ country ]
   {
      if (fcity && ! strmatchcase(location[0],geolocs[ii].city)) continue;
      if (fcountry && ! strmatchcase(location[1],geolocs[ii].country)) continue;

      for (jj = 0; jj < Nmatch; jj++) {                                    //  look for duplicate match
         if (strmatch(geolocs[ii].city,matches[jj][0]))                    //  (EXIF and cities-geotag file)
            if (strmatch(geolocs[ii].country,matches[jj][1])) break;
      }
      if (jj < Nmatch) continue;                                           //  discard duplicates

      matches[Nmatch][0] = geolocs[ii].city;
      matches[Nmatch][1] = geolocs[ii].country;
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolocs[ii].lati;                                      //  return lat/long of 1st match
         coord[1] = geolocs[ii].longi;
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }

   if (Nmatch) return Nmatch;                                              //  exact match found

   for (ii = Nmatch  = 0; ii < Ngeolocs; ii++)                             //  search for partial city [ country ]
   {
      if (strmatch("null",geolocs[ii].city)) continue;

      if (fcity) {
         cc = strlen(location[0]);
         if (! strmatchcaseN(location[0],geolocs[ii].city,cc)) continue;
      }
      if (fcountry) {
         cc = strlen(location[1]);
         if (! strmatchcaseN(location[1],geolocs[ii].country,cc)) continue;
      }
      matches[Nmatch][0] = geolocs[ii].city;
      matches[Nmatch][1] = geolocs[ii].country;
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolocs[ii].lati;                                      //  return geolat/long of 1st match
         coord[1] = geolocs[ii].longi;
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }

   return Nmatch;
}


/**************************************************************************/

//  Save geotags in image file EXIF and image index file.
//  Update geolocs[*] memory table
//
//  location[2] = city and country
//  coord[2] = latitude and longitude
//  return value:  0    OK, no geotag revision (incomplete data)
//                 1    OK, no geotag revision (matches existing data)
//                 2    OK, geotag lat/long updated
//                 3    OK, geotag new location added
//                -1    error, lat/long bad

int put_geotags(char *location[2], char *coord[2])
{
   char        acoord[2][20];
   float       flati, flongi;
   int         ii, err, Fnew = 0, Fchange = 0;
   int         retval;

   if (! curr_file) return 0;

   err = validate_latlong(coord[0],coord[1],flati,flongi);
   if (err) {
      if (err == 2) goto badcoord;                                         //  reject bad data
      strcpy(acoord[0],"null");                                            //  replace missing data with "null"
      strcpy(acoord[1],"null");
      flati = flongi = 0;                                                  //  lati/longi missing value
   }
   else {
      snprintf(acoord[0],20,"%.4f",flati);                                 //  reformat with std. precision
      snprintf(acoord[1],20,"%.4f",flongi);
      StripZeros(acoord[0]);
      StripZeros(acoord[1]);
   }

   if (! strmatch(location[0],"null"))                                     //  unless null,
      *location[0] = toupper(*location[0]);                                //  force capitalization
   if (! strmatch(location[1],"null"))
      *location[1] = toupper(*location[1]);

   strncpy0(meta_city,location[0],99);                                     //  save geotags in image file EXIF
   strncpy0(meta_country,location[1],99);                                  //    and in search-index file
   strncpy0(meta_latitude,acoord[0],12);
   strncpy0(meta_longitude,acoord[1],12);

   save_filemeta(curr_file);

   if (! *location[0] || strmatch(location[0],"null")) return 0;           //  quit here if city data not complete
   if (! *location[1] || strmatch(location[1],"null")) return 0;

   for (ii = 0; ii < Ngeolocs; ii++) {                                     //  search geotags for city, country
      if (! strmatchcase(location[0],geolocs[ii].city)) continue;          //  (case-insensitive compare)
      if (strmatchcase(location[1],geolocs[ii].country)) break;
   }

   if (ii < Ngeolocs) {                                                    //  found, check for revised lat/long
      if (! strmatch(geolocs[ii].lati,acoord[0])) Fchange = 1;
      if (! strmatch(geolocs[ii].longi,acoord[1])) Fchange = 1;
      if (! strmatch(geolocs[ii].city,location[0])) Fchange = 1;           //  or revised capitalization
      if (! strmatch(geolocs[ii].country,location[1])) Fchange = 1;
   }
   else Fnew = 1;                                                          //  a new city, country

   if (Fnew + Fchange == 0) return 1;                                      //  no change

   if (Fchange)
   {
      zfree(geolocs[ii].city);                                             //  change geotag data in memory
      geolocs[ii].city = zstrdup(location[0]);                             //  (to be used subsequently)
      zfree(geolocs[ii].country);
      geolocs[ii].country = zstrdup(location[1]);                          //  presence in image EXIF will make
      zfree(geolocs[ii].lati);                                             //    this the preferred version
      geolocs[ii].lati = zstrdup(acoord[0]);
      zfree(geolocs[ii].longi);
      geolocs[ii].longi = zstrdup(acoord[1]);
      geolocs[ii].flati = flati;
      geolocs[ii].flongi = flongi;
      retval = 2;
   }

   else if (Fnew)
   {
      if (Ngeolocs == maxgeolocs) {
         zmessLogACK(Mwin,"max. geotags %d exceeded",maxgeolocs);
         return -1;
      }
      for (ii = Ngeolocs; ii > 0; ii--)                                    //  shift all geotag data up
         geolocs[ii] = geolocs[ii-1];
      Ngeolocs++;

      ii = 0;
      geolocs[ii].city = zstrdup(location[0]);                             //  new geotag is now first
      geolocs[ii].country = zstrdup(location[1]);                          //  (find again faster)
      geolocs[ii].lati = zstrdup(acoord[0]);
      geolocs[ii].longi = zstrdup(acoord[1]);
      geolocs[ii].flati = flati;
      geolocs[ii].flongi = flongi;
      retval = 3;
   }

   else return -1;                                                         //  should not happen
   return retval;

badcoord:
   zmessageACK(Mwin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return -1;
}


/**************************************************************************/

//  Geographic Map Functions
//  Maps of any scale can be user-installed.
//  Mercator projection is assumed (but unimportant for maps < 100 km).

namespace geomap
{
   char     mapname[100];
   int      mapww, maphh;                                                  //  map width, height
   float    mlat[2];                                                       //  latitude range, low - high
   float    mlong[2];                                                      //  longitude range, low - high
}


//  load the default world map or a map chosen by the user

void m_load_geomap(GtkWidget *, cchar *menu)                               //  v.14.11
{
   using namespace geomap;

   int  load_geomap_dialog_event(zdialog *zd, cchar *event);

   char     mapindex[200], mapfile[200];
   char     buff[200];
   cchar    *pp;
   zdialog  *zd, *zdbusy;
   int      err, zstat, ii;
   FILE     *fid;
   float    lat1, lat2, long1, long2;
   STATB    statb;

   F1_help_topic = "images_by_map";

   if (checkpend("all")) return;
   if (! init_geolocs() ) return;                                          //  insure geolocations are loaded

   if (Fnoindex) {
      zmessageACK(Mwin,ZTX("-noindex in use, disabled"));                  //  v.14.12
      return;
   }

   snprintf(mapindex,200,"%s/maps_index",maps_dirk);                       //  v.14.11
   err = stat(mapindex,&statb);
   if (err) goto nomapsinstalled;

   if (menu && strmatch(menu,"default")) {
      strcpy(mapname,"World.jpg");                                         //  load default world map
      goto load_map;
   }

   fid = fopen(mapindex,"r");                                              //  open file with list of avail. maps
   if (! fid) goto nomapsinstalled;

   zd = zdialog_new(ZTX("choose map file"),Mwin,Bcancel,null);
   zdialog_add_widget(zd,"combo","mapname","dialog",0,"space=5");

   while (true)
   {
      pp = fgets_trim(buff,200,fid,1);                                     //  get map file names
      if (! pp) break;
      pp = strField(pp,",",1);
      if (! pp) continue;
      zdialog_cb_app(zd,"mapname",pp);                                     //  add to dialog popup list
   }

   fclose(fid);

   if (*mapname && Wstate.fpxb)                                            //  show current map if any
      zdialog_stuff(zd,"mapname",mapname);

   zdialog_resize(zd,300,100);
   zdialog_run(zd,load_geomap_dialog_event);                               //  run dialog, get user choice

   zstat = zdialog_wait(zd);

   if (zstat != 1) {
      zdialog_free(zd);                                                    //  cancel
      return;
   }

   zdialog_fetch(zd,"mapname",mapname,100);                                //  user choice
   zdialog_free(zd);

load_map:

   snprintf(mapfile,200,"%s/%s",maps_dirk,mapname);                        //  check map file exists
   err = stat(mapfile,&statb);
   if (err) goto mapfilemissing;

   fid = fopen(mapindex,"r");                                              //  read map file list again
   if (! fid) goto nomapsinstalled;

   while (true)
   {
      pp = fgets_trim(buff,200,fid,1);                                     //  find chosen map file
      if (! pp) break;
      pp = strField(buff,",",1);
      if (! pp) continue;
      if (strmatch(pp,mapname)) break;
   }

   fclose(fid);
   if (! pp) goto mapfilemissing;

   lat1 = lat2 = long1 = long2 = 0;

   pp = strField(buff,",",2);                                              //  get map latitude/longitude range
   if (! pp) goto latlongerr;                                              //    and verify data OK
   err = convSF(pp,lat1,-80,+80);
   if (err) goto latlongerr;

   pp = strField(buff,",",3);
   if (! pp) goto latlongerr;
   err = convSF(pp,lat2,-80,+80);
   if (err) goto latlongerr;

   pp = strField(buff,",",4);
   if (! pp) goto latlongerr;
   err = convSF(pp,long1,-200,+200);
   if (err) goto latlongerr;

   pp = strField(buff,",",5);
   if (! pp) goto latlongerr;
   err = convSF(pp,long2,-200,+200);
   if (err) goto latlongerr;

   if (lat2 < lat1 + 0.1) goto latlongerr;                                 //  require min. 0.1 degree map range
   if (long2 < long1 + 0.1) goto latlongerr;                               //    (11 km)

   printz("load geomap: %s \n",mapname);                                   //  no errors, commit to load map

   Ffuncbusy = 1;

   zdbusy = zmessage_post(Mwin,0,"Loading geomap ...",null);
   for (ii = 0; ii < 5; ii++) {                                            //  required to get message out
      zmainloop();                                                         //  X11 or GDK problem
      zsleep(0.01);
   }

   free_geomap();                                                          //  free prior map

   Wstate.fpxb = PXB_load(mapfile,1);                                      //  load map file (with diagnostic)
   if (! Wstate.fpxb) {
      Ffuncbusy = 0;
      return;
   }

   mapww = Wstate.fpxb->ww;                                                //  save map pixel dimensions
   maphh = Wstate.fpxb->hh;

   mlat[0] = lat1;                                                         //  save map latitude/longitude range
   mlat[1] = lat2;
   mlong[0] = long1;
   mlong[1] = long2;

   zdialog_free(zdbusy);                                                   //  kill the message                   v.14.08
   Ffuncbusy = 0;

   m_zoom(null,"fit");
   return;

nomapsinstalled:
   zmessageACK(Mwin,ZTX("fotoxx-maps package not installed \n"
                        "(see http://kornelix.com/packages and /tarballs)"));
   return;

mapfilemissing:
   zmessageACK(Mwin,ZTX("map file %s is missing"),mapname);
   return;

latlongerr:
   zmessageACK(Mwin,ZTX("map latitude/longitude data unreasonable \n"
                        " %.3f %.3f %.3f %.3f"),lat1,lat2,long1,long2);
   return;
}


//  dialog event and completion function

int  load_geomap_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   if (strmatch(event,"mapname")) zd->zstat = 1;
   return 1;
}


//  Convert map position px/py into latitude and longitude.
//  Return 0 if OK, +N if error (off the map).

int geomap_coordinates(int px, int py, float &flat, float &flong)          //  v.14.11
{
   using namespace geomap;

   float    lat1, lat2, long1, long2;
   float    zww, qy, qy2;

   flat = flong = 0;
   if (px < 0 || px > mapww) return 1;                                     //  px/py outside map size
   if (py < 0 || py > maphh) return 1;

   lat1 = mlat[0];                                                         //  map latitude low - high range
   lat2 = mlat[1];
   long1 = mlong[0];                                                       //  map longitude low - high range
   long2 = mlong[1];

   flong = long1 + (1.0 * px / mapww) * (long2 - long1);                   //  return longitude

   zww = mapww * 360.0 / (long2 - long1);                                  //  width for -180 to +180 longitude

   lat1 = lat1 / RAD;                                                      //  convert to radians
   lat2 = lat2 / RAD;

   qy2 = (zww/2/PI) * (log(tan(lat2/2 + PI/4)));                           //  lat2 distance from equator
   qy2 = qy2 - py;                                                         //  py distance from equator
   qy = fabsf(qy2);

   flat = 2 * atan(exp(2*PI*qy/zww)) - PI/2;
   if (qy2 < 0) flat = -flat;

   flat = flat * RAD;                                                      //  return latitude
   return 0;
}


//  Convert latitude and longitude into map position px/py.
//  Return 0 if OK, +N if error (off the map).

int geomap_position(float flat, float flong, int &px, int &py)             //  v.14.11
{
   using namespace geomap;

   float    lat1, lat2, long1, long2;
   float    zww, qy, qy2;

   lat1 = mlat[0];                                                         //  map latitude low - high range
   lat2 = mlat[1];
   long1 = mlong[0];                                                       //  map longitude low - high range
   long2 = mlong[1];

   px = py = 0;

   if (flat < lat1 || flat >= lat2) return 1;                              //  flat/flong outside map limits
   if (flong < long1 || flong >= long2) return 1;

   px = (flong - long1) / (long2 - long1) * mapww;                         //  return px position

   zww = mapww * 360.0 / (long2 - long1);                                  //  width for -180 to +180 longitude

   lat1 = lat1 / RAD;                                                      //  convert to radians
   lat2 = lat2 / RAD;
   flat = flat / RAD;

   qy2 = (zww/2/PI) * (log(tan(lat2/2 + PI/4)));                           //  lat2 distance from equator
   qy =  (zww/2/PI) * (log(tan(flat/2 + PI/4)));                           //  flat distance from equator
   py = qy2 - qy;                                                          //  return py position

   if (px < 2 || px > mapww-3) return 1;                                   //  out of bounds
   if (py < 2 || py > maphh-3) return 1;                                   //  includes margins for red dot

   return 0;
}


//  paint red dots corresponding to image locations on map

void geomap_paint_dots()                                                   //  v.15.01
{
   int      ii, err;
   int      mx, my, dx, dy;
   float    flati, flongi, radius;

   if (! Wstate.fpxb) return;                                              //  no map loaded

   radius = 3;
   if (Cstate->mscale >= 1) radius = 4;

   cairo_set_source_rgb(mwcr,1,0,0);

   for (ii = 0; ii < Ngeolocs; ii++)                                       //  paint red dots on map
   {                                                                       //    where images are present
      flati = geolocs[ii].flati;
      flongi = geolocs[ii].flongi;
      err = geomap_position(flati,flongi,mx,my);
      if (err) continue;
      dx = Cstate->mscale * mx - Cstate->morgx + Cstate->dorgx;
      dy = Cstate->mscale * my - Cstate->morgy + Cstate->dorgy;
      if (dx < 0 || dx > Dww-1) continue;
      if (dy < 0 || dy > Dhh-1) continue;
      cairo_arc(mwcr,dx,dy,radius,0,2*PI);
      cairo_fill(mwcr);
   }

   return;
}


/**************************************************************************/

//  Respond to mouse movement and left clicks on geomap image.
//  Set longitude and latitude, and city and country.

void geomap_mousefunc()
{
   void  find_geoloc_images(float flati, float flongi);                    //  private function

   int         err, mx, my, ii, minii;
   char        *city, *country;
   float       flati, flongi, glati, glongi;
   float       dist, mindist, distrange;
   float       mscale = Cstate->mscale;
   static      char  *pcity = 0;

   if (checkpend("all")) return;                                           //  check nothing pending
   if (Cstate != &Wstate) return;                                          //  view mode not world maps
   if ((Mxdrag || Mydrag)) return;                                         //  pan/scroll - handle normally
   if (RMclick) return;                                                    //  zoom - fit window, handle normally
   if (LMclick && mscale < 1) return;                                      //  handle normally if not full size
   if (! Wstate.fpxb) return;

   mx = Mxposn;                                                            //  mouse position, image space
   my = Myposn;

   err = geomap_coordinates(mx,my,flati,flongi);                           //  v.14.11
   if (err) return;

   dist = mindist = 999999;
   minii = 0;

   for (ii = 0; ii < Ngeolocs; ii++)                                       //  find nearest city/country
   {
      glati = geolocs[ii].flati;
      dist = (flati - glati) * (flati - glati);
      if (dist > mindist) continue;
      glongi = geolocs[ii].flongi;
      dist += (flongi - glongi) * (flongi - glongi);                       //  degrees**2
      if (dist > mindist) continue;
      mindist = dist;
      minii = ii;
   }

   mindist = 111 * sqrtf(mindist);                                         //  degrees**2 to km at earth surface
   distrange = geomap_range / mscale;                                      //  increase if not full scale         v.15.01

   if (mindist <= distrange) {                                             //  within range of a known place
      ii = minii;
      city = geolocs[ii].city;
      country = geolocs[ii].country;
      flati = geolocs[ii].flati;
      flongi = geolocs[ii].flongi;
   }
   else  city = country = 0;

   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;
      poptext_window(0,0,0,0,0,0);                                         //  remove popup

      if (zd_geotags) {
         zdialog_stuff(zd_geotags,"city",city);                            //  stuff calling dialog
         zdialog_stuff(zd_geotags,"country",country);
         zdialog_stuff(zd_geotags,"latitude",flati);
         zdialog_stuff(zd_geotags,"longitude",flongi);
         zdialog_send_event(zd_geotags,"geomap");                          //  activate calling dialog
      }

      else if (city)
         find_geoloc_images(flati,flongi);                                 //  show images for location           v.14.11
   }

   else if (city) {                                                        //  mouse movement, no click
      if (! pcity || ! strmatch(city,pcity)) {
         poptext_window(city,MWIN,Mwxposn,Mwyposn,0.1,1);                  //  popup the city name at mouse
         pcity = city;
      }
   }

   else if (pcity) {
      poptext_window(0,0,0,0,0,0);                                         //  remove popup
      pcity = 0;
   }

   return;
}


/**************************************************************************/

//  find images in range of given geolocation, show gallery of images
//  privat function for geomap_mousefunc(), called when a location is clicked

void find_geoloc_images(float flati, float flongi)
{
   int            ftf, err, nn = 0;
   char           imagefile[XFCC];
   float          glati, glongi, grange;
   cchar          *pp;
   FILE           *fid;
   sxrec_t        sxrec;

   if (Fnoindex) {
      zmessageACK(Mwin,ZTX("-noindex in use, disabled"));                  //  v.14.12
      return;
   }

   fid = fopen(searchresults_file,"w");                                    //  open output file
   if (! fid) {
      zmessLogACK(Mwin,"output file error: %s",strerror(errno));
      return;
   }

   ftf = 1;

   while (true)                                                            //  read image index recs.
   {
      err = read_sxrec_seq(sxrec,ftf);
      if (err) break;

      strncpy0(imagefile,sxrec.file,XFCC);                                 //  save filespec

      pp = strField(sxrec.gtags,'^',3);                                    //  latitude
      if (! pp) goto freemem;
      if (strmatch(pp,"null")) goto freemem;
      glati = atof(pp);

      pp = strField(sxrec.gtags,'^',4);                                    //  longitude
      if (! pp) goto freemem;
      if (strmatch(pp,"null")) goto freemem;
      glongi = atof(pp);

      grange = (flati - glati) * (flati - glati);
      grange += (flongi - glongi) * (flongi - glongi);
      grange = 111.0 * sqrt(grange);

      if (grange <= geomap_range) {                                        //  within distance limit, select
         fprintf(fid,"%s\n",imagefile);                                    //  output matching file
         nn++;
      }

   freemem:
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   fclose(fid);

   if (! nn) {
      zmessageACK(Mwin,ZTX("No matching images found"));
      return;
   }

   free_resources();
   navi::gallerytype = 2;                                                  //  search results
   gallery(searchresults_file,"initF");                                    //  generate gallery of matching files
   m_viewmode(0,"G");
   gallery(0,"paint",0);                                                   //  v.15.05

   return;
}


/**************************************************************************/

//  free memory lorge memory used for geomap image
//  used by edit_setup() to maximize available memory

void free_geomap()                                                         //  v.14.11
{
   if (Wstate.fpxb) PXB_free(Wstate.fpxb);
   Wstate.fpxb = 0;
   return;
}


/**************************************************************************/

//  set the map search range from a clicked position, kilometers

void m_mapsearch_range(GtkWidget *, cchar *)
{
   char     trange[12];
   char     *pp;
   float    range = 0;

   F1_help_topic = "images_by_map";

   snprintf(trange,12,"%.1f",geomap_range);                                //  v.14.11  float
   pp = zdialog_text(Mwin,ZTX("search range (km)"),trange);
   if (pp) range = atof(pp);
   if (pp) zfree(pp);

   if (range > 0 && range < 1000)
      geomap_range = range;

   return;
}


/**************************************************************************/

//  add or edit image geotags - city, country, latitude, longitude

void m_edit_geotags(GtkWidget *, cchar *menu)
{
   int  edit_geotags_dialog_event(zdialog *zd, cchar *event);

/**
                  Edit Geotags

      city [______________]  country [______________]
      latitude [_______] longitude [_______]

      [find] [web] [prev] [map] [apply] [clear] [cancel]

**/

   cchar    *title = ZTX("Edit Geotags");
   cchar    *mapquest1 = ZTX("Geocoding web service courtesy of");
   cchar    *mapquest2 = "http://www.mapquest.com";

   zdialog  *zd;

   F1_help_topic = "edit_geotags";

   if (! init_geolocs()) return;                                           //  initialize geotags
   if (! curr_file) return;

   if (! zdeditgeotags)                                                    //  start dialog if not already
   {
      zdeditgeotags = zdialog_new(title,Mwin,Bfind,Bweb,Bprev,Bmap,Bapply,Bclear,Bcancel,null);
      zd = zdeditgeotags;
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labcity","hb1",ZTX("city"),"space=5");
      zdialog_add_widget(zd,"entry","city","hb1",0,"expand");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labcountry","hb1",ZTX("country"),"space=5");
      zdialog_add_widget(zd,"entry","country","hb1",0,"expand");
      zdialog_add_widget(zd,"hbox","hb4","dialog");
      zdialog_add_widget(zd,"label","lablat","hb4","Latitude","space=3");
      zdialog_add_widget(zd,"entry","latitude","hb4",0,"size=10");
      zdialog_add_widget(zd,"label","space","hb4",0,"space=5");
      zdialog_add_widget(zd,"label","lablong","hb4","Longitude","space=3");
      zdialog_add_widget(zd,"entry","longitude","hb4",0,"size=10");
      zdialog_add_widget(zd,"hbox","hbmq","dialog");
      zdialog_add_widget(zd,"label","labmq","hbmq",mapquest1,"space=3");
      zdialog_add_widget(zd,"link","MapQuest","hbmq",mapquest2);

      zdialog_run(zd,edit_geotags_dialog_event);
   }

   zd = zdeditgeotags;
   zd_geotags = zd;                                                        //  activate geomap clicks             v.14.12

   if (strmatch(meta_city,"null")) zdialog_stuff(zd,"city","");
   else zdialog_stuff(zd,"city",meta_city);

   if (strmatch(meta_country,"null")) zdialog_stuff(zd,"country","");
   else zdialog_stuff(zd,"country",meta_country);

   if (strmatch(meta_latitude,"null")) zdialog_stuff(zd,"latitude","");
   else zdialog_stuff(zd,"latitude",meta_latitude);

   if (strmatch(meta_longitude,"null")) zdialog_stuff(zd,"longitude","");
   else zdialog_stuff(zd,"longitude",meta_longitude);

   return;
}


//  dialog event and completion callback function

int edit_geotags_dialog_event(zdialog *zd, cchar *event)
{
   int          zstat, Nmatch;
   char         *location[2], *coord[2], *matches[20][2];
   char         city[100], country[100];
   char         latitude[20], longitude[20], temp[20];
   float        fcoord[2];
   cchar        *errmess;

   if (strmatch(event,"escape")) zd->zstat = 7;                            //  escape = cancel                    v.15.07

   if (! curr_file) return 1;

   if (strmatch(event,"enter")) zd->zstat = 5;                             //  [apply]                            v.14.03

   if (strmatch(event,"geomap")) {                                         //  have geotags data from geomap
      Fgtagmod++;                                                          //  dialog inputs changed
      return 1;
   }

   if (strstr("city country latitude longitude",event)) {                  //  dialog inputs changed              v.14.09
      Fgtagmod++;
      return 1;
   }

   if (! zd->zstat) return 1;                                              //  wait for action button

   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active
   m_viewmode(0,"F");                                                      //  back to F view

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",temp,20);                                   //  and latitude, longitude
   repl_1str(temp,latitude,",",".");                                       //  (replace comma decimal)
   zdialog_fetch(zd,"longitude",temp,20);
   repl_1str(temp,longitude,",",".");
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geolocs(location,coord,matches);                        //  find in geolocs[*] table
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(Mwin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
         Fgtagmod++;                                                       //  dialog inputs changed
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
            Fgtagmod++;                                                    //  dialog inputs changed
         }
      }

      return 1;
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess)
         zmessageACK(Mwin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
         Fgtagmod++;                                                       //  dialog inputs changed
      }

      return 1;
   }

   else if (zstat == 3)                                                    //  [prev]
   {
      zdialog_stuff(zd,"city",prevcity);                                   //  get last-used geotags
      zdialog_stuff(zd,"country",prevcountry);
      zdialog_stuff(zd,"latitude",prevlatitude);
      zdialog_stuff(zd,"longitude",prevlongitude);
      Fgtagmod++;                                                          //  dialog inputs changed
      return 1;
   }

   else if (zstat == 4)                                                    //  [map]
   {
      m_viewmode(0,"W");
      return 1;
   }

   else if (zstat == 5)                                                    //  [apply]
   {
      if (! Fgtagmod) return 1;                                            //  nothing changed                    v.14.12

      if (strmatch(coord[0],"null")) *coord[0] = 0;                        //  replace "null" with ""
      if (strmatch(coord[1],"null")) *coord[1] = 0;

      if (*coord[0] || *coord[1]) {                                        //  if coordinates present, validate
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      put_geotags(location,coord);                                         //  update EXIF, image index, geolocs[*] table

      strcpy(prevcity,city);                                               //  save data for later use as [prev]
      strcpy(prevcountry,country);
      strcpy(prevlatitude,latitude);
      strcpy(prevlongitude,longitude);

      gtk_window_present(MWIN);                                            //  keep focus on main window          v.14.09
      return 1;
   }

   else if (zstat == 6)                                                    //  [clear]
   {
      zdialog_stuff(zd,"city","");                                         //  erase dialog fields
      zdialog_stuff(zd,"country","");
      zdialog_stuff(zd,"latitude","");
      zdialog_stuff(zd,"longitude","");
      Fgtagmod++;                                                          //  bugfix  v.14.11
      return 1;
   }

   else {                                                                  //  cancel
      Fgtagmod = 0;                                                        //  no changes made                    v.15.03
      zdialog_free(zd);
      zdeditgeotags = 0;
      zd_geotags = 0;                                                      //  deactivate geomap clicks
      m_viewmode(0,"F");
      return 1;
   }

   return 1;

badcoord:
   zmessageACK(Mwin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;
}


/**************************************************************************/

//  batch add geotags - set geotags for multiple image files

char     **batch_add_geotags_filelist = 0;
int      batch_add_geotags_filecount = 0;

void m_batch_add_geotags(GtkWidget *, cchar *menu)
{
   int   batch_add_geotags_dialog_event(zdialog *zd, cchar *event);

   cchar       *title = ZTX("Batch Add Geotags");
   char        *file, **flist;
   zdialog     *zd;
   int         ii, err;
   char        *location[2], *coord[2];
   char        city[100], country[100];
   char        latitude[20], longitude[20];
   cchar       *mapquest1 = "Geocoding web service courtesy of";
   cchar       *mapquest2 = "http://www.mapquest.com";

   F1_help_topic = "batch_add_geotags";

   if (! init_geolocs()) return;                                           //  initialize geotags

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/**
                Batch Add Geotags

      [select files]  NN files selected
      city [______________]  country [______________]
      latitude [_______] longitude [_______]

      [find] [web] [prev] [map] [proceed] [cancel]

**/

   zd = zdialog_new(title,Mwin,Bfind,Bweb,Bprev,Bmap,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb1",Bnofileselected,"space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcity","hb2",ZTX("city"),"space=5");
   zdialog_add_widget(zd,"entry","city","hb2",0,"expand");
   zdialog_add_widget(zd,"label","space","hb2",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hb2",ZTX("country"),"space=5");
   zdialog_add_widget(zd,"entry","country","hb2",0,"expand");
   zdialog_add_widget(zd,"hbox","hb3","dialog");
   zdialog_add_widget(zd,"label","lablat","hb3","Latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hb3",0,"size=10");
   zdialog_add_widget(zd,"label","space","hb3",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hb3","Longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hb3",0,"size=10");
   zdialog_add_widget(zd,"hbox","hbmq","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmq","hbmq",mapquest1,"space=3");
   zdialog_add_widget(zd,"link","MapQuest","hbmq",mapquest2);

   batch_add_geotags_filelist = 0;
   batch_add_geotags_filecount = 0;

   flist = 0;
   zd_geotags = zd;                                                        //  activate geomap clicks             v.14.12

   zdialog_run(zd,batch_add_geotags_dialog_event);                         //  run dialog
   zdialog_wait(zd);                                                       //  wait for dialog completion

   if (zd->zstat != 5) goto cleanup;                                       //  status not [proceed]
   if (! batch_add_geotags_filecount) goto cleanup;                        //  no files selected

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,20);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,20);
   coord[0] = latitude;
   coord[1] = longitude;

   zdialog_free(zd);                                                       //  kill dialog

   flist = batch_add_geotags_filelist;                                     //  selected files
   if (! flist) goto cleanup;

   write_popup_text("open","Adding Geotags",500,200,Mwin);                 //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image
      err = f_open(file,0,0,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      put_geotags(location,coord);                                         //  update file geotags
   }

   write_popup_text("write","COMPLETED");

cleanup:

   if (zd) zdialog_free(zd);
   zd_geotags = 0;                                                         //  deactivate geomap clicks

   if (flist) {
      for (ii = 0; flist[ii]; ii++)
         zfree(flist[ii]);
      zfree(flist);
   }

   Fblock = 0;
   return;
}


//  batch_add_geotags dialog event function

int batch_add_geotags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii, yn, zstat, Nmatch;
   char     **flist = batch_add_geotags_filelist;
   char     countmess[50];
   char     *location[2], *coord[2], *matches[20][2];
   char     city[100], country[100];
   char     latitude[20], longitude[20], temp[20];
   cchar    *errmess;
   float    fcoord[2];

   if (strmatch(event,"escape")) zd->zstat = 6;                            //  escape = cancel                    v.15.07

   if (strmatch(event,"files"))                                            //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++)
            zfree(flist[ii]);
         zfree(flist);
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      flist = gallery_getfiles();                                          //  get file list from user
      zdialog_show(zd,1);

      batch_add_geotags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batch_add_geotags_filecount = ii;

      snprintf(countmess,50,Bfileselected,batch_add_geotags_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }

   if (! zd->zstat) return 1;                                              //  wait for action button

   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",temp,20);                                   //  and latitude, longitude
   repl_1str(temp,latitude,",",".");                                       //  (replace comma decimal)
   zdialog_fetch(zd,"longitude",temp,20);
   repl_1str(temp,longitude,",",".");
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geolocs(location,coord,matches);                        //  find in geolocs[*] table
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(Mwin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
         }
      }
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess)
         zmessageACK(Mwin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   else if (zstat == 3)                                                    //  [prev]
   {
      zdialog_stuff(zd,"city",prevcity);                                   //  get last-used geotags
      zdialog_stuff(zd,"country",prevcountry);
      zdialog_stuff(zd,"latitude",prevlatitude);
      zdialog_stuff(zd,"longitude",prevlongitude);
   }

   else if (zstat == 4)                                                    //  [map]
      m_viewmode(0,"W");

   else if (zstat == 5)                                                    //  [proceed]
   {
      if (strmatch(coord[0],"null")) *coord[0] = 0;                        //  replace "null" with ""
      if (strmatch(coord[1],"null")) *coord[1] = 0;

      if (*coord[0] || *coord[1])                                          //  if coordinates present, validate
      {
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      if (! batch_add_geotags_filecount) goto nofiles;

      if (! *city || ! *country || ! *latitude || ! *longitude) {          //  check data is complete
         yn = zmessageYN(Mwin,ZTX("data is incomplete \n proceed?"));
         if (! yn) return 1;
      }

      strcpy(prevcity,city);                                               //  save data for later use as [prev]
      strcpy(prevcountry,country);
      strcpy(prevlatitude,latitude);
      strcpy(prevlongitude,longitude);

      zd->zstat = 5;                                                       //  OK to proceed
      zdialog_destroy(zd);
   }

   else {                                                                  //  cancel
      zdialog_destroy(zd);
      zd_geotags = 0;                                                      //  deactivate geomap clicks
   }

   return 1;

badcoord:
   zmessageACK(Mwin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;

nofiles:
   zmessageACK(Mwin,Bnofileselected);
   return 1;
}


//  dialog to choose one city from multiple options
//  location[2] is input city and optional country
//  (may be substrings, may have multiple matches in city geotags data)
//  location[2] is output unique city and country after user choice
//  coord[2] is output latitude, longitude

char     *geotags_chosenlocation[2];
char     *geotags_chosencoord[2];

int geotags_choosecity(char *location[2], char *coord[2])
{
   int  geotags_choosecity_event(zdialog *zd, cchar *event);

   char     *matches[20][2], text[200];
   int      Nmatch, ii, zstat;
   zdialog  *zd;

   Nmatch = get_geolocs(location,coord,matches);                           //  get matching city geotags data

   if (Nmatch == 0) return 0;                                              //  no match

   if (Nmatch == 1) {                                                      //  one match, done
      location[0] = matches[0][0];
      location[1] = matches[0][1];
      return 1;
   }

   zd = zdialog_new(ZTX("choose city"),Mwin,BOK,Bcancel,null);             //  multiple matches, start dialog
   zdialog_add_widget(zd,"comboE","cities","dialog",0,"space=5");
   for (ii = 0; ii < Nmatch; ii++) {                                       //  list matching cities to choose from
      snprintf(text,200,"%s | %s",matches[ii][0],matches[ii][1]);
      zdialog_cb_app(zd,"cities",text);                                    //  duplicates are removed
   }

   zdialog_resize(zd,300,100);
   zdialog_run(zd,geotags_choosecity_event);                               //  run dialog, wait for completion
   zstat = zdialog_wait(zd);
   zdialog_free(zd);

   if (zstat == 1) {                                                       //  valid response available
      location[0] = geotags_chosenlocation[0];
      location[1] = geotags_chosenlocation[1];
      coord[0] = geotags_chosencoord[0];
      coord[1] = geotags_chosencoord[1];
      return 1;
   }

   return 0;
}


//  dialog event function - get chosen city/country from multiple choices

int geotags_choosecity_event(zdialog *zd, cchar *event)
{
   char           text[200];
   static char    city[100], country[100];
   char           *location[2], *coord[2];
   char           *matches[20][2];
   cchar          *pp;
   int            nn;
   static int     ftf = 1;

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [OK]  
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (ftf) {
      zdialog_cb_popup(zd,"cities");                                       //  first time, open combo box list
      ftf = 0;
   }

   if (strmatch(event,"cities")) {                                         //  OK
      zdialog_fetch(zd,"cities",text,200);

      pp = strField(text,'|',1);
      if (pp) strncpy0(city,pp,99);
      pp = strField(text,'|',2);
      if (pp) strncpy0(country,pp,99);
      strTrim2(city);
      strTrim2(country);
      location[0] = city;
      location[1] = country;

      nn = get_geolocs(location,coord,matches);                            //  find in city geotags data
      if (nn) {
         geotags_chosenlocation[0] = location[0];                          //  use 1st match if > 1
         geotags_chosenlocation[1] = location[1];
         geotags_chosencoord[0] = coord[0];
         geotags_chosencoord[1] = coord[1];
         zd->zstat = 1;
      }
      else zd->zstat = 2;                                                  //  bad status
   }

   if (zd->zstat) ftf = 1;

   return 1;
}


/**************************************************************************/

//  Convert a city [country] to latitude/longitude using the
//    MapQuest geocoding service.
//  (incomplete names may be completed with a bad guess)

cchar * web_geocode(char *location[2], char *coord[2])                     //  overhaul 13.05, 13.05.1
{
   int         err;
   static char latitude[20], longitude[20];
   char        outfile[200], URI[300];
   char        *pp1, *pp2, buffer[200];
   float       flati, flongi;
   FILE        *fid;
   cchar       *notfound = ZTX("not found");
   cchar       *badinputs = ZTX("city and country required");
   cchar       *query = "http://open.mapquestapi.com/geocoding/v1/address?"
                        "&key=Fmjtd%7Cluub2qa72d%2C20%3Do5-9u700a"
                        "&maxResults=1"
                        "&outFormat=csv";

   *coord[0] = *coord[1] = 0;                                              //  null outputs
   *latitude = *longitude = 0;

   if (*location[0] < ' ' || *location[1] < ' ')
      return badinputs;

   snprintf(outfile,199,"%s/web-data",tempdir);                            //  v.14.11
   snprintf(URI,299,"\"%s&location=%s,%s\"",query,location[0],location[1]);

   err = shell_quiet("wget -T 10 -o /dev/null -O %s %s",outfile,URI);
   if (err == 4) err = ECOMM;                                              //  replace "interrupted system call"  v.14.11
   if (err) return strerror(err);

   fid = fopen(outfile,"r");                                               //  get response
   if (! fid) return notfound;
   pp1 = fgets(buffer,200,fid);
   pp1 = fgets(buffer,200,fid);
   fclose(fid);
   if (! pp1) return notfound;

   pp2 = (char *) strField(pp1,",",7);
   if (! pp2) return notfound;
   strncpy0(latitude,pp2,12);

   pp2 = (char *) strField(pp1,",",8);
   if (! pp2) return notfound;
   strncpy0(longitude,pp2,12);

   err = validate_latlong(latitude,longitude,flati,flongi);
   if (err) return notfound;

   pp1 = strchr(latitude,'.');                                             //  keep max. 4 decimal digits
   if (pp1) *(pp1+5) = 0;
   pp1 = strchr(longitude,'.');
   if (pp1) *(pp1+5) = 0;

   coord[0] = latitude;
   coord[1] = longitude;
   return 0;
}


/**************************************************************************/

//  Group images by location and date, with a count of images in each group.
//  Click on a group to get a thumbnail gallery of all images in the group.

int   ggroups_comp(cchar *rec1, cchar *rec2);
void  ggroups_click(GtkWidget *widget, int line, int pos);
int   ggroups_getdays(cchar *date);

struct grec_t  {                                                           //  image geotags data
   char        *city, *country;                                            //  group location
   char        pdate[12];                                                  //  nominal group date, yyyymmdd
   int         lodate, hidate;                                             //  range, days since 0 CE
   int         count;                                                      //  images in group
};

grec_t   *grec = 0;
int      Ngrec = 0;
int      ggroups_groupby, ggroups_daterange;


void m_geotag_groups(GtkWidget *, cchar *)
{
   int  geotag_groups_dialog_event(zdialog *zd, cchar *event);

   zdialog        *zd;
   int            zstat, ftf, err, cc, cc1, cc2;
   int            iix, iig, newgroup;
   char           country[100], city[100], buff[300], pdate[12];
   cchar          *pp;
   sxrec_t        sxrec;
   GtkWidget      *textwin;

   F1_help_topic = "images_by_location";
   if (checkpend("all")) return;                                           //  check nothing pending

   if (Fnoindex) {
      zmessageACK(Mwin,ZTX("-noindex in use, disabled"));                  //  v.14.12
      return;
   }

/***
            Report Geotag Groups

         (o) Group by country
         (o) Group by country/city
         (o) Group by country/city/date
             Combine within [ xx |-|+] days

                        [proceed]  [cancel]
***/

   zd = zdialog_new(ZTX("Report Geotag Groups"),Mwin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"radio","country","dialog",ZTX("Group by country"));
   zdialog_add_widget(zd,"radio","city","dialog",ZTX("Group by country/city"));
   zdialog_add_widget(zd,"radio","date","dialog",ZTX("Group by country/city/date"));
   zdialog_add_widget(zd,"hbox","hbr","dialog");
   zdialog_add_widget(zd,"label","space","hbr",0,"space=10");
   zdialog_add_widget(zd,"label","labr1","hbr",ZTX("Combine within"),"space=10");
   zdialog_add_widget(zd,"spin","range","hbr","0|999|1|1");
   zdialog_add_widget(zd,"label","labr2","hbr",ZTX("days"),"space=10");

   zdialog_stuff(zd,"country",0);
   zdialog_stuff(zd,"city",1);
   zdialog_stuff(zd,"date",0);

   zdialog_resize(zd,300,0);
   zdialog_run(zd,geotag_groups_dialog_event);
   zstat = zdialog_wait(zd);
   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }

   zdialog_fetch(zd,"country",iix);
   if (iix) ggroups_groupby = 1;                                           //  group by country
   zdialog_fetch(zd,"city",iix);
   if (iix) ggroups_groupby = 2;                                           //  group by country/city
   zdialog_fetch(zd,"date",iix);
   if (iix) ggroups_groupby = 3;                                           //  group by country/city/date (range)
   zdialog_fetch(zd,"range",ggroups_daterange);

   zdialog_free(zd);

   if (Ngrec) {                                                            //  free prior memory
      for (iix = 0; iix < Ngrec; iix++) {
         if (grec[iix].city) zfree(grec[iix].city);
         if (grec[iix].country) zfree(grec[iix].country);
      }
      zfree(grec);
   }

   cc = maximages * sizeof(grec_t);                                        //  allocate memory
   grec = (grec_t *) zmalloc(cc);
   memset(grec,0,cc);

   Ngrec = 0;
   ftf = 1;

   while (true)
   {
      err = read_sxrec_seq(sxrec,ftf);                                     //  read image index recs.
      if (err) break;

      iix = Ngrec;

      pp = strField(sxrec.gtags,'^',1);                                    //  get city
      if (pp) grec[iix].city = zstrdup(pp);
      else grec[iix].city = zstrdup("null");

      pp = strField(sxrec.gtags,'^',2);                                    //  country
      if (pp) grec[iix].country = zstrdup(pp);
      else grec[iix].country = zstrdup("null");

      strncpy0(grec[iix].pdate,sxrec.pdate,9);                             //  photo date, truncate to yyyymmdd
      grec[iix].lodate = ggroups_getdays(sxrec.pdate);                     //  days since 0 CE
      grec[iix].hidate = grec[iix].lodate;

      if (++Ngrec == maximages) {
         zmessageACK(Mwin,"too many image files");
         return;
      }

      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   if (! Ngrec) {
      zmessageACK(Mwin,"no geotags data found");
      return;
   }

   if (Ngrec > 1)                                                          //  sort index by country/city/date
      HeapSort((char *) grec, sizeof(grec_t), Ngrec, ggroups_comp);

   iig = 0;                                                                //  1st group from grec[0]
   grec[iig].count = 1;                                                    //  group count = 1

   for (iix = 1; iix < Ngrec; iix++)                                       //  scan following grecs
   {
      newgroup = 0;

      if (! strmatch(grec[iix].country,grec[iig].country))
         newgroup = 1;                                                     //  new country >> new group

      if (ggroups_groupby >= 2)
         if (! strmatch(grec[iix].city,grec[iig].city)) newgroup = 1;      //  new city >> new group if group by city

      if (ggroups_groupby >= 3)
         if (grec[iix].lodate - grec[iig].hidate > ggroups_daterange)      //  new date >> new group if group by date
            newgroup = 1;                                                  //    and date out of range

      if (newgroup)
      {
         iig++;                                                            //  new group
         if (iix > iig) {
            grec[iig] = grec[iix];                                         //  copy and pack down
            grec[iix].city = grec[iix].country = 0;                        //  no zfree()
         }
         grec[iig].count = 1;                                              //  group count = 1
      }
      else
      {
         zfree(grec[iix].city);                                            //  same group
         zfree(grec[iix].country);                                         //  free memory
         grec[iix].city = grec[iix].country = 0;
         grec[iig].hidate = grec[iix].lodate;                              //  expand group date range
         grec[iig].count++;                                                //  increment group count
      }
   }

   Ngrec = iig + 1;                                                        //  unique groups count

   textwin = write_popup_text("open",ZTX("geotag groups"),620,400,Mwin);   //  write groups to popup window

   if (ggroups_groupby == 1)                                               //  group by country
   {
      snprintf(buff,300,"%-30s  %5s ","Country","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);
         cc1 = 30 + strlen(country) - utf8len(country);
         snprintf(buff,300,"%-*s  %5d ",cc1,country,grec[iig].count);
         write_popup_text("write",buff);
      }
   }

   if (ggroups_groupby == 2)                                               //  group by country/city
   {
      snprintf(buff,300,"%-30s  %-30s  %5s ","Country","City","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);
         cc1 = 30 + strlen(country) - utf8len(country);
         utf8substring(city,grec[iig].city,0,30);
         cc2 = 30 + strlen(city) - utf8len(city);
         snprintf(buff,300,"%-*s  %-*s  %5d ",
                  cc1,country,cc2,city,grec[iig].count);
         write_popup_text("write",buff);
      }
   }

   if (ggroups_groupby == 3)                                               //  group by country/city/date (range)
   {
      snprintf(buff,300,"%-30s  %-30s  %-10s  %5s ","Country","City","Date","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);                    //  get graphic cc for UTF-8 names
         cc1 = 30 + strlen(country) - utf8len(country);
         utf8substring(city,grec[iig].city,0,30);
         cc2 = 30 + strlen(city) - utf8len(city);

         strncpy(pdate,grec[iig].pdate,8);                                 //  date, yyyymmdd
         if (! strmatch(pdate,"null")) {
            memcpy(pdate+8,pdate+6,2);                                     //  convert to yyyy-mm-dd
            memcpy(pdate+5,pdate+4,2);
            pdate[4] = pdate[7] = '-';
            pdate[10] = 0;
         }

         snprintf(buff,300,"%-*s  %-*s  %-10s  %5d ",
                  cc1,country,cc2,city,pdate,grec[iig].count);
         write_popup_text("write",buff);
      }
   }

   write_popup_text("top");
   textwidget_set_clickfunc(textwin,ggroups_click);                        //  response function for mouse click
   return;
}


//  dialog event and completion function

int geotag_groups_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   return 1;
}


//  Compare 2 grec records by geotags and date,
//  return < 0  = 0  > 0   for   rec1  <  =  >  rec2.

int ggroups_comp(cchar *rec1, cchar *rec2)
{
   int      ii;

   char * country1 = ((grec_t *) rec1)->country;                           //  compare countries
   char * country2 = ((grec_t *) rec2)->country;
   ii = strcmp(country1,country2);
   if (ii) return ii;

   char * city1 = ((grec_t *) rec1)->city;                                 //  compare cities
   char * city2 = ((grec_t *) rec2)->city;
   ii = strcmp(city1,city2);
   if (ii) return ii;

   int date1 = ((grec_t *) rec1)->lodate;                                  //  compare dates
   int date2 = ((grec_t *) rec2)->lodate;
   ii = date1 - date2;
   return ii;
}


//  convert yyyymmdd date into days from 0 C.E.
//  "null" date returns 999999 (year 2737)

int ggroups_getdays(cchar *date)
{
   int   CEdays(int year, int mon, int day);

   int      year, month, day;
   char     temp[8];

   year = month = day = 0;

   strncpy0(temp,date,5);
   year = atoi(temp);

   strncpy0(temp,date+4,3);
   month = atoi(temp);

   strncpy0(temp,date+6,3);
   day = atoi(temp);

   return CEdays(year,month,day);
}


//   convert year/month/day into days since Jan 1, 0001 (day 0 CE)
//   year is 0001 to 9999, month is 1-12, day is 1-31

int CEdays(int year, int month, int day)
{
   int    montab[12] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
   int    elaps;

   elaps = 365 * (year-1) + (year-1) / 4;                                  //  elapsed days in prior years
   elaps += montab[month-1];                                               //  + elapsed days in prior months
   if (year % 4 == 0 && month > 2) elaps += 1;                             //  + 1 for Feb. 29
   elaps += day-1;                                                         //  + elapsed days in month
   return elaps;
}


//  Receive clicks on report window and generate gallery of images
//  matching the selected country/city/date

void ggroups_click(GtkWidget *widget, int line, int pos)
{
   int      iix, ftf, err, lodate, hidate, datex;
   cchar    *pp;
   char     city[100], country[100];
   FILE     *fid;
   sxrec_t  sxrec;

   if (checkpend("all")) return;                                           //  check nothing pending

   textwidget_get_line(widget,line,1);                                     //  hilite clicked line

   iix = line - 1;                                                         //  clicked grec[iix]
   if (iix < 0 || iix > Ngrec-1) return;

   strncpy0(country,grec[iix].country,99);                                 //  selected country/city/date range
   strncpy0(city,grec[iix].city,99);
   lodate = grec[iix].lodate;
   hidate = grec[iix].hidate;

   fid = fopen(searchresults_file,"w");                                    //  open output file
   if (! fid) goto filerror;

   ftf = 1;

   while (true)                                                            //  read image index recs.
   {
      err = read_sxrec_seq(sxrec,ftf);
      if (err) break;

      pp = strField(sxrec.gtags,'^',2);
      if (! pp) pp = "null";                                               //  enable search for "null"

      if (! strmatch(pp,country)) goto freemem;                            //  no country match

      if (ggroups_groupby >= 2) {
         pp = strField(sxrec.gtags,'^',1);
         if (! pp) pp = "null";
         if (! strmatch(pp,city)) goto freemem;                            //  no city match
      }

      if (ggroups_groupby == 3) {
         datex = ggroups_getdays(sxrec.pdate);
         if (datex < lodate || datex > hidate) goto freemem;               //  no date match
      }

      fprintf(fid,"%s\n",sxrec.file);                                      //  output matching file

   freemem:
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   fclose(fid);

   free_resources();
   navi::gallerytype = 2;                                                  //  search results
   gallery(searchresults_file,"initF");                                    //  generate gallery of matching files
   m_viewmode(0,"G");
   gallery(0,"paint",0);
   return;

filerror:
   zmessLogACK(Mwin,"file error: %s",strerror(errno));
   return;
}


/**************************************************************************/

//  Search image tags, geotags, dates, stars, comments, captions
//  to find matching images. This is fast using the image index.
//  Search also any other metadata, but relatively slow.

namespace search_images
{
   zdialog  *zdsearchimages = 0;                                           //  search images dialog

   char     searchDateFrom[16] = "";                                       //  search images
   char     searchDateTo[16] = "";
   char     searchStarsFrom[4] = "";
   char     searchStarsTo[4] = "";

   char     searchtags[tagScc] = "";                                       //  search tags list
   char     searchtext[tagScc] = "";                                       //  search comments & captions word list
   char     searchfiles[tagScc] = "";                                      //  search files list

   char     searchCity[100] = "";                                          //  search geotags
   char     searchCountry[100] = "";
   char     searchLatitude[20] = "";
   char     searchLongitude[20] = "";
   float    searchRange = 100;
   float    flatitude = 0, flongitude = 0;

   int      Fscanall, Fscancurr, Fnewset, Faddset, Fremset;
   int      Fdates, Ftext, Ffiles, Ftags, Fstars, Fgtags;
   int      Flastver, Fsearchmeta;
   int      Falltags, Falltext, Fallfiles;
   int      Frepgallery, Frepmeta;

   int      Nsearchkeys = 0;
   char     *searchkeys[5];                                                //  search metadata keys and match data
   char     *searchkeydata[5];
   char     searchkeyx[8], searchkeydatax[8];
}

using namespace search_images;


void m_search_images(GtkWidget *, cchar *)                                 //  overhauled
{
   void  search_searchtags_clickfunc(GtkWidget *widget, int line, int pos);
   void  search_matchtags_clickfunc(GtkWidget *widget, int line, int pos);
   void  search_deftags_clickfunc(GtkWidget *widget, int line, int pos);
   int   searchimages_dialog_event(zdialog*, cchar *event);

   zdialog     *zd;
   GtkWidget   *widget;

   F1_help_topic = "search_images";

   if (checkpend("all")) return;                                           //  check nothing pending

   if (Fnoindex) {
      zmessageACK(Mwin,ZTX("-noindex in use, disabled"));                  //  v.14.12
      return;
   }

/***
                  Search Image Metadata

         images to search: (o) all  (o) current set only
         matching images: (o) new set  (o) add to set  (o) remove
         report type: (o) gallery  (o) metadata
         date range   [___________] [___________] (yyyymmdd)
         stars range  [__] [__]   (o) last version         all/any
         search tags  [__________________________________] (o) (o)
         search text  [__________________________________] (o) (o)
         search files [__________________________________] (o) (o)
         other criteria: [geotags] (*)  [other] (*)
         
         Enter Search Tag [________________________]
         Matching Tags [_________________________________________]

         defined tags 
          ------------------------------------------------------
         |                                                      |
         |                                                      |
         |                                                      |
         |                                                      |
         |                                                      |
          ------------------------------------------------------
                                      [clear] [proceed] [cancel]
***/

   zd = zdialog_new(ZTX("Search Image Metadata"),Mwin,Bclear,Bproceed,Bcancel,null);
   zdsearchimages = zd;

   zdialog_add_widget(zd,"hbox","hbs1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labs1","hbs1",ZTX("images to search:"),"space=5");
   zdialog_add_widget(zd,"radio","allimages","hbs1",ZTX("all"),"space=3");
   zdialog_add_widget(zd,"radio","currset","hbs1",ZTX("current set only"),"space=5");

   zdialog_add_widget(zd,"hbox","hbm1","dialog");
   zdialog_add_widget(zd,"label","labs1","hbm1",ZTX("matching images:"),"space=5");
   zdialog_add_widget(zd,"radio","newset","hbm1",ZTX("new set"),"space=5");
   zdialog_add_widget(zd,"radio","addset","hbm1",ZTX("add to set"),"space=5");
   zdialog_add_widget(zd,"radio","remset","hbm1",ZTX("remove"),"space=5");

   zdialog_add_widget(zd,"hbox","hbrt","dialog");
   zdialog_add_widget(zd,"label","labrt","hbrt",ZTX("report type:"),"space=5");
   zdialog_add_widget(zd,"radio","repgallery","hbrt",ZTX("gallery"),"space=5");
   zdialog_add_widget(zd,"radio","repmeta","hbrt",ZTX("metadata"),"space=5");

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|expand");

   zdialog_add_widget(zd,"label","labD","vb1",ZTX("date range"));
   zdialog_add_widget(zd,"label","labS","vb1",ZTX("stars range"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search tags"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search text"));
   zdialog_add_widget(zd,"label","labF","vb1",ZTX("search files"));

   zdialog_add_widget(zd,"hbox","hbD","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","datefrom","hbD",0,"size=12");
   zdialog_add_widget(zd,"entry","dateto","hbD",0,"size=12");
   zdialog_add_widget(zd,"label","labD","hbD",ZTX("(yyyymmdd)"),"space=5");

   zdialog_add_widget(zd,"hbox","hbS","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","starsfrom","hbS",0,"size=2");
   zdialog_add_widget(zd,"entry","starsto","hbS",0,"size=2");
   zdialog_add_widget(zd,"check","lastver","hbS",ZTX("last version only"),"space=10");
   zdialog_add_widget(zd,"label","space","hbS",0,"expand");
   zdialog_add_widget(zd,"label","all-any","hbS",ZTX("all/any"),"space=2");

   zdialog_add_widget(zd,"hbox","hbT","vb2",0,"space=1");
   zdialog_add_widget(zd,"frame","frameT","hbT",0,"expand");
   zdialog_add_widget(zd,"text","searchtags","frameT",0,"expand|wrap");
   zdialog_add_widget(zd,"radio","alltags","hbT",0);
   zdialog_add_widget(zd,"radio","anytags","hbT",0);

   zdialog_add_widget(zd,"hbox","hbC","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchtext","hbC",0,"expand");
   zdialog_add_widget(zd,"radio","alltext","hbC",0);
   zdialog_add_widget(zd,"radio","anytext","hbC",0);

   zdialog_add_widget(zd,"hbox","hbF","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchfiles","hbF",0,"expand");
   zdialog_add_widget(zd,"radio","allfiles","hbF",0);
   zdialog_add_widget(zd,"radio","anyfiles","hbF",0);

   zdialog_add_widget(zd,"hbox","hbO","dialog","space=3");
   zdialog_add_widget(zd,"label","labO1","hbO",ZTX("other criteria"),"space=5");
   zdialog_add_widget(zd,"button","geotags","hbO",Bgeotags,"space=8");
   zdialog_add_widget(zd,"label","geotags#","hbO","( )");
   zdialog_add_widget(zd,"label","space","hbO",0,"space=5");
   zdialog_add_widget(zd,"button","other","hbO",ZTX("other"),"space=8");
   zdialog_add_widget(zd,"label","other#","hbO","( )");
   
   zdialog_add_widget(zd,"hsep","sep","dialog",0,"space=3");

   zdialog_add_widget(zd,"hbox","hbnt","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labnt","hbnt",ZTX("Enter Search Tag"),"space=3");
   zdialog_add_widget(zd,"entry","entertag","hbnt");

   zdialog_add_widget(zd,"hbox","hbmt","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labmt","hbmt",ZTX("Matching Tags"),"space=3");
   zdialog_add_widget(zd,"frame","frmt","hbmt",0,"space=3|expand");
   zdialog_add_widget(zd,"text","matchtags","frmt",0,"wrap");

   zdialog_add_widget(zd,"hbox","space","dialog",0,"space=1");
   zdialog_add_widget(zd,"hbox","hbA1","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hbA1",ZTX("Defined Tags:"),"space=5");
   zdialog_add_widget(zd,"hbox","hbA2","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frameA","hbA2",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwinA","frameA",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrwinA",0,"expand|wrap");

   widget = zdialog_widget(zd,"searchtags");                               //  tag widget mouse functions
   textwidget_set_clickfunc(widget,search_searchtags_clickfunc);

   widget = zdialog_widget(zd,"matchtags");
   textwidget_set_clickfunc(widget,search_matchtags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,search_deftags_clickfunc);

   zdialog_stuff(zd,"allimages",1);                                        //  defaults
   zdialog_stuff(zd,"currset",0);
   zdialog_stuff(zd,"newset",1);
   zdialog_stuff(zd,"addset",0);
   zdialog_stuff(zd,"remset",0);
   zdialog_stuff(zd,"repgallery",1);
   zdialog_stuff(zd,"repmeta",0);
   zdialog_stuff(zd,"lastver",0);
   zdialog_stuff(zd,"alltags",0);
   zdialog_stuff(zd,"anytags",1);
   zdialog_stuff(zd,"alltext",0);
   zdialog_stuff(zd,"anytext",1);
   zdialog_stuff(zd,"allfiles",0);
   zdialog_stuff(zd,"anyfiles",1);

   zdialog_restore_inputs(zd);                                             //  preload prior user inputs

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   zdialog_resize(zd,0,600);                                               //  start dialog
   zdialog_run(zd,searchimages_dialog_event,"save");
   zdialog_wait(zd);                                                       //  wait for dialog completion
   zdialog_free(zd);

   return;
}

//  mouse click functions for search tags and defined tags widgets

void search_searchtags_clickfunc(GtkWidget *widget, int line, int pos)     //  search tag clicked
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;:",end);
   if (! txtag) { zfree(txline); return; }

   del_tag(txtag,searchtags);                                              //  remove from search list
   zdialog_stuff(zdsearchimages,"searchtags",searchtags);

   zfree(txline);
   zfree(txtag);
   return;
}

void search_matchtags_clickfunc(GtkWidget *widget, int line, int pos)      //  matching tag was clicked           v.15.07
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;",end);
   if (! txtag) { zfree(txline); return; }

   add_tag(txtag,searchtags,tagScc);                                       //  add to search tag list

   zdialog_stuff(zdsearchimages,"entertag","");                            //  update dialog widgets
   zdialog_stuff(zdsearchimages,"matchtags","");
   zdialog_stuff(zdsearchimages,"searchtags",searchtags);

   zdialog_goto(zdsearchimages,"entertag");                                //  focus back to entertag widget

   zfree(txline);
   zfree(txtag);
   return;
}

void search_deftags_clickfunc(GtkWidget *widget, int line, int pos)        //  defined tag clicked
{
   char     *txline, *txtag, end = 0;

   txline = textwidget_get_line(widget,line,0);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,",;:",end);
   if (! txtag || end == ':') { zfree(txline); return; }                   //  tag category clicked, ignore       v.15.05

   add_tag(txtag,searchtags,tagScc);                                       //  add to search tag list
   zdialog_stuff(zdsearchimages,"searchtags",searchtags);

   zfree(txline);
   zfree(txtag);
   return;
}


//  search images dialog event and completion callback function

int searchimages_dialog_event(zdialog *zd, cchar *event)                   //  overhauled
{
   using namespace navi;

   int  searchimages_select(sxrec_t &sxrec);
   int  searchimages_geotags_dialog(zdialog *zd);
   int  searchimages_metadata_dialog(zdialog *zd);
   int  searchimages_metadata_report();

   cchar    dateLoDefault[16] = "000001010000";                            //  date: 0000/01/01  time: 00:00
   cchar    dateHiDefault[16] = "999912312359";                            //  date: 9999/12/31  time: 23:59

   char     *file;
   char     **flist, *pp, buffer[XFCC];
   int      ftf, match, ii, jj, cc, err;
   int      nt, cc1, cc2, ff;
   int      Nadded, Nremoved, Nleft, Npver;
   sxrec_t  sxrec;
   FILE     *fid;
   char     *pp1, *pp2;
   char     entertag[tagcc], matchtags[20][tagcc];
   char     matchtagstext[(tagcc+2)*20];

   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"geotags"))                                          //  get geotags search criteria
      searchimages_geotags_dialog(zd);

   if (strmatch(event,"other"))                                            //  get other metadata criteria
      searchimages_metadata_dialog(zd);

   Fgtags = 0;
   if (*searchCity || *searchCountry ||                                    //  search geotags was given
       *searchLatitude || *searchLongitude) Fgtags = 1;

   Fsearchmeta = 0;
   if (Nsearchkeys) Fsearchmeta = 1;                                       //  search other metadata was given

   if (Fgtags) zdialog_stuff(zd,"geotags#","(*)");                         //  add (*) visual flags to main dialog if
   else zdialog_stuff(zd,"geotags#","( )");                                //    geotags or metadata selection active
   if (Fsearchmeta) zdialog_stuff(zd,"other#","(*)");
   else zdialog_stuff(zd,"other#","( )");

   if (zd->zstat == 1)                                                     //  clear selection criteria           v.14.12
   {
      zdialog_stuff(zd,"allimages",1);
      zdialog_stuff(zd,"currset",0);
      zdialog_stuff(zd,"newset",1);
      zdialog_stuff(zd,"addset",0);
      zdialog_stuff(zd,"remset",0);
      zdialog_stuff(zd,"repgallery",1);
      zdialog_stuff(zd,"repmeta",0);
      zdialog_stuff(zd,"lastver",0);
      zdialog_stuff(zd,"alltags",0);
      zdialog_stuff(zd,"anytags",1);
      zdialog_stuff(zd,"alltext",0);
      zdialog_stuff(zd,"anytext",1);
      zdialog_stuff(zd,"allfiles",0);
      zdialog_stuff(zd,"anyfiles",1);
      zdialog_stuff(zd,"datefrom","");
      zdialog_stuff(zd,"dateto","");
      zdialog_stuff(zd,"starsfrom","");
      zdialog_stuff(zd,"starsto","");
      zdialog_stuff(zd,"searchtags","");
      zdialog_stuff(zd,"searchtext","");
      zdialog_stuff(zd,"searchfiles","");

      zdialog_stuff(zd,"geotags#","( )");
      Fgtags = 0;
      *searchCity = *searchCountry = 0;
      *searchLatitude = *searchLongitude = 0;
      flatitude = flongitude = 0;

      zdialog_stuff(zd,"other#","( )");
      Fsearchmeta = 0;
      Nsearchkeys = 0;

      zd->zstat = 0;                                                       //  keep dialog active
      return 1;
   }

   if (strmatch(event,"entertag"))                                         //  new tag is being typed in          v.15.07
   {
      zdialog_stuff(zd,"matchtags","");                                    //  clear matchtags in dialog

      zdialog_fetch(zd,"entertag",entertag,tagcc);                         //  get chars. typed so far
      cc1 = strlen(entertag);
      
      for (ii = jj = 0; ii <= cc1; ii++) {                                 //  remove foul characters
         if (strchr(",:;",entertag[ii])) continue;
         entertag[jj++] = entertag[ii];
      }
      
      if (jj < cc1) {                                                      //  something was removed
         entertag[jj] = 0;
         cc1 = jj;
         zdialog_stuff(zd,"entertag",entertag);
      }

      if (cc1 < 2) return 1;                                               //  wait for at least 2 chars.

      for (ii = nt = 0; ii < maxtagcats; ii++)                             //  loop all categories
      {
         pp2 = tags_deftags[ii];                                           //  category: aaaaaa, bbbbb, ... tagN,
         if (! pp2) continue;                                              //            |     |
         pp2 = strchr(pp2,':');                                            //            pp1   pp2
         
         while (true)                                                      //  loop all deftags in category
         {
            pp1 = pp2 + 2;
            if (! *pp1) break;
            pp2 = strchr(pp1,',');
            if (! pp2) break;
            if (strmatchcaseN(entertag,pp1,cc1)) {                         //  deftag matches chars. typed so far
               cc2 = pp2 - pp1;
               strncpy(matchtags[nt],pp1,cc2);                             //  save deftags that match
               matchtags[nt][cc2] = 0;
               if (++nt == 20) return 1;                                   //  quit if 20 matches or more
            }
         }
      }
      
      if (nt == 0) return 1;                                               //  no matches

      pp1 = matchtagstext;

      for (ii = 0; ii < nt; ii++)                                          //  make deftag list: aaaaa, bbb, cccc ...
      {
         strcpy(pp1,matchtags[ii]);
         pp1 += strlen(pp1);
         strcpy(pp1,", ");
         pp1 += 2;
      }
      
      zdialog_stuff(zd,"matchtags",matchtagstext);                         //  stuff matchtags in dialog
      return 1;
   }

   if (strmatch(event,"enter"))                                            //  KB Enter, tag finished             v.15.07
   {
      zdialog_fetch(zd,"entertag",entertag,tagcc);                         //  get finished tag
      cc1 = strlen(entertag);
      if (! cc1) return 1;
      if (entertag[cc1-1] == '\n') {                                       //  remove newline character
         cc1--;
         entertag[cc1] = 0;
      }

      for (ii = ff = 0; ii < maxtagcats; ii++)                             //  loop all categories
      {
         pp2 = tags_deftags[ii];                                           //  category: aaaaaa, bbbbb, ... tagN,
         if (! pp2) continue;                                              //            |     |
         pp2 = strchr(pp2,':');                                            //            pp1   pp2
         
         while (true)                                                      //  loop all deftags in category
         {
            pp1 = pp2 + 2;
            if (! *pp1) break;
            pp2 = strchr(pp1,',');
            if (! pp2) break;
            cc2 = pp2 - pp1;
            if (cc2 != cc1) continue;
            if (strmatchcaseN(entertag,pp1,cc1)) {                         //  entered tag matches deftag
               strncpy(entertag,pp1,cc1);                                  //  use deftag upper/lower case
               ff = 1;
               break;
            }
         }

         if (ff) break;
      }
      
      if (! ff) {
         zmessageACK(Mwin,ZTX("not a defined tag: %s"),entertag);
         return 1;
      }

      add_tag(entertag,searchtags,tagScc);                                 //  add to search tag list

      zdialog_stuff(zd,"entertag","");                                     //  update dialog widgets
      zdialog_stuff(zd,"searchtags",searchtags);
      zdialog_stuff(zd,"matchtags","");

      zdialog_goto(zd,"entertag");                                         //  put focus back on entertag widget
      return 1;
   }

   if (! zd->zstat) return 1;                                              //  wait for dialog completion
   if (zd->zstat != 2) return 1;                                           //  cancel if not [proceed]

//  inputs are complete. perform the search.

   zdialog_fetch(zd,"allimages",Fscanall);                                 //  search all images
   zdialog_fetch(zd,"currset",Fscancurr);                                  //  search current set (gallery)
   zdialog_fetch(zd,"newset",Fnewset);                                     //  matching images --> new set
   zdialog_fetch(zd,"addset",Faddset);                                     //  add matching image to set
   zdialog_fetch(zd,"remset",Fremset);                                     //  remove matching images from set

   if (Fremset && Fscanall) {                                              //  illogical search
      zmessageACK(Mwin,ZTX("to remove images from current set, \n"
                           "search current set"));
      zd->zstat = 0;                                                       //  keep dialog active
      return 1;
   }

   if (Faddset && Fscancurr) {
      zmessageACK(Mwin,ZTX("to add images to current set, \n"
                           "search all images"));
      zd->zstat = 0;                                                       //  keep dialog active
      return 1;
   }

   zdialog_fetch(zd,"repgallery",Frepgallery);                             //  gallery report                     v.14.02
   zdialog_fetch(zd,"repmeta",Frepmeta);                                   //  metadata report
   zdialog_fetch(zd,"lastver",Flastver);                                   //  get last versions only             v.14.09

   zdialog_fetch(zd,"datefrom",searchDateFrom,15);                         //  get search date range
   zdialog_fetch(zd,"dateto",searchDateTo,15);
   zdialog_fetch(zd,"starsfrom",searchStarsFrom,2);                        //  get search stars range
   zdialog_fetch(zd,"starsto",searchStarsTo,2);
   zdialog_fetch(zd,"searchtags",searchtags,tagScc);                       //  get search tags
   zdialog_fetch(zd,"searchtext",searchtext,tagScc);                       //  get search text*
   zdialog_fetch(zd,"searchfiles",searchfiles,tagScc);                     //  get search /path*/file*

   zdialog_fetch(zd,"alltags",Falltags);                                   //  get match all/any options
   zdialog_fetch(zd,"alltext",Falltext);
   zdialog_fetch(zd,"allfiles",Fallfiles);

   Fdates = 0;
   if (*searchDateFrom) Fdates++;                                          //  search date from was given
   else strcpy(searchDateFrom,"000001010000");                             //  else search from begining of time

   if (*searchDateTo) Fdates++;                                            //  search date to was given
   else strcpy(searchDateTo,"999912312359");                               //  else search to end of time

   if (Fdates) {                                                           //  complete partial date/time data
      cc = strlen(searchDateFrom);
      for (ii = cc; ii < 12; ii++)                                         //  default date from:
         searchDateFrom[ii] = dateLoDefault[ii];                           //    date: 0000/01/01 time: 00:00
      cc = strlen(searchDateTo);
      for (ii = cc; ii < 12; ii++)                                         //  default date to:
         searchDateTo[ii] = dateHiDefault[ii];                             //    date: 9999/12/31 time: 23:59
   }

   Fstars = 0;
   if (*searchStarsFrom || *searchStarsTo) {
      Fstars = 1;                                                          //  stars was given
      ii = *searchStarsFrom;
      if (! ii) ii = '0';
      if (ii < '0' || ii > '5') Fstars = 0;                                //  validate inputs                    v.14.12
      jj = *searchStarsTo;
      if (! jj) jj = '5';
      if (jj < '0' || jj > '5') Fstars = 0;
      if (jj < ii) Fstars = 0;
      if (! Fstars) {
         zmessageACK(Mwin,ZTX("stars range not reasonable"));
         zd->zstat = 0;
         return 1;
      }
   }

   Ffiles = 0;
   if (! blank_null(searchfiles)) Ffiles = 1;                              //  search path / file (fragment) was given

   Ftext = 0;
   if (! blank_null(searchtext)) Ftext = 1;                                //  search text was given

   Ftags = 0;
   if (! blank_null(searchtags)) Ftags = 1;                                //  search tags was given

   if (Ffiles) strToLower(searchfiles);                                    //  all comparisons in lower case
   if (Ftags) strToLower(searchtags);
   if (Ftext) strToLower(searchtext);

   if (checkpend("all")) return 1;                                         //  check nothing pending
   Ffuncbusy = 1;
   Fblock = 1;

   Nadded = Nremoved = Npver = Nleft = 0;                                  //  image counts

   //  search all images and keep those meeting search criteria
   //  result is gallery of images meeting criteria

   if (Fscanall && Fnewset)
   {
      fid = fopen(searchresults_file,"w");                                 //  open output file
      if (! fid) goto filerror;

      ftf = 1;

      while (true)
      {
         zmainloop(20);

         err = read_sxrec_seq(sxrec,ftf);                                  //  scan all index recs.
         if (err) break;

         match = searchimages_select(sxrec);                               //  test against select criteria
         if (match) {                                                      //  all criteria passed
            Nadded++;                                                      //  count matches
            fprintf(fid,"%s\n",sxrec.file);                                //  save matching filename
         }

         zfree(sxrec.file);                                                //  free allocated strings
         zfree(sxrec.tags);
         zfree(sxrec.capt);
         zfree(sxrec.comms);
         zfree(sxrec.gtags);
      }

      fclose(fid);
      Nleft = Nadded;
   }

   //  search all images and add those meeting search criteria
   //    to current image set (gallery)

   if (Fscanall && Faddset)
   {
      fid = fopen(searchresults_file,"w");                                 //  open output file
      if (! fid) goto filerror;

      for (ii = 0; ii < navi::nfiles; ii++)                                //  scan current gallery
      {
         file = gallery(0,"find",ii);
         if (! file) break;
         if (*file != '!') {                                               //  skip directories
            fprintf(fid,"%s\n",file);                                      //  add image files to output
            Nleft++;
         }
         zfree(file);                                                      //  free memory
      }

      ftf = 1;

      while (true)
      {
         zmainloop(20);

         err = read_sxrec_seq(sxrec,ftf);                                  //  scan all index recs.
         if (err) break;

         match = searchimages_select(sxrec);                               //  test against select criteria
         if (match) {                                                      //  all criteria passed
            Nadded++;                                                      //  count matches
            fprintf(fid,"%s\n",sxrec.file);                                //  save matching filename
         }

         zfree(sxrec.file);                                                //  free memory
         zfree(sxrec.tags);
         zfree(sxrec.capt);
         zfree(sxrec.comms);
         zfree(sxrec.gtags);
      }

      fclose(fid);
      Nleft += Nadded;
   }

   //  search current image set and keep only those meeting search criteria

   if (Fscancurr && Fnewset)
   {
      fid = fopen(searchresults_file,"w");                                 //  open output file
      if (! fid) goto filerror;

      for (ii = 0; ii < navi::nfiles; ii++)                                //  scan current gallery
      {
         zmainloop(20);

         file = gallery(0,"find",ii);
         if (! file) break;
         if (*file == '!') {                                               //  skip directories
            zfree(file);
            continue;
         }

         err = get_sxrec(sxrec,file);
         if (err) {                                                        //  no metadata rec?
            zfree(file);
            continue;
         }

         match = searchimages_select(sxrec);                               //  test against select criteria
         if (match) {                                                      //  passed
            Nleft++;                                                       //  count retained images
            fprintf(fid,"%s\n",file);                                      //  save retained filename
         }
         else Nremoved++;

         zfree(file);                                                      //  free memory

         zfree(sxrec.file);
         zfree(sxrec.tags);
         zfree(sxrec.capt);
         zfree(sxrec.comms);
         zfree(sxrec.gtags);
      }

      fclose(fid);
   }

   //  search current image set and remove those meeting search criteria

   if (Fscancurr && Fremset)
   {
      fid = fopen(searchresults_file,"w");                                 //  open output file
      if (! fid) goto filerror;

      for (ii = 0; ii < navi::nfiles; ii++)                                //  scan current gallery
      {
         zmainloop(20);

         file = gallery(0,"find",ii);
         if (! file) break;
         if (*file == '!') {                                               //  skip directories
            zfree(file);
            continue;
         }

         err = get_sxrec(sxrec,file);
         if (err) {                                                        //  no metadata rec?
            zfree(file);
            continue;
         }

         match = searchimages_select(sxrec);                               //  test against select criteria
         if (! match) {                                                    //  failed
            Nleft++;
            fprintf(fid,"%s\n",file);                                      //  save retained filename
         }
         else Nremoved++;

         zfree(file);                                                      //  free memory

         zfree(sxrec.file);
         zfree(sxrec.tags);
         zfree(sxrec.capt);
         zfree(sxrec.comms);
         zfree(sxrec.gtags);
      }

      fclose(fid);
   }

   if (Flastver)                                                           //  remove all but latest versions     v.14.09
   {
      cc = Nleft * sizeof(char *);
      flist = (char **) zmalloc(cc);

      fid = fopen(searchresults_file,"r");                                 //  read file of selected image files
      if (! fid) goto filerror;

      for (ii = 0; ii < Nleft; ii++)                                       //  build file list in memory
      {
         file = fgets_trim(buffer,XFCC,fid);
         if (! file) break;
         flist[ii] = zstrdup(file);
      }

      fclose(fid);

      for (ii = 1; ii < Nleft; ii++)                                       //  scan file list in memory
      {
         pp = strrchr(flist[ii],'/');                                      //  /directory.../filename.v...
         if (! pp) continue;                                               //  |                     |
         pp = strstr(pp,".v");                                             //  flist[ii]             pp
         if (! pp) continue;
         cc = pp - flist[ii] + 1;
         if (strmatchN(flist[ii],flist[ii-1],cc)) {                        //  compare each filespec with prior
            zfree(flist[ii-1]);                                            //  match: remove prior from list
            flist[ii-1] = 0;
         }
      }

      fid = fopen(searchresults_file,"w");                                 //  write remaining file list
      if (! fid) goto filerror;                                            //    to results file

      Npver = 0;
      for (ii = 0; ii < Nleft; ii++)
      {
         file = flist[ii];
         if (file) {
            fprintf(fid,"%s\n",file);
            zfree(file);
         }
         else Npver++;
      }

      fclose(fid);
      zfree(flist);

      if (Nadded) Nadded -= Npver;                                         //  update counts
      Nremoved += Npver;
      Nleft -= Npver;
   }

   Ffuncbusy = 0;
   Fblock = 0;

   zmessageACK(Mwin,ZTX("images added: %d  removed: %d  new count: %d"),
                           Nadded, Nremoved, Nleft);
   if (Nleft == 0) {
      zmessageACK(Mwin,ZTX("no changes made"));
      zd->zstat = 0;                                                       //  stay in search dialog              v.15.07
      return 1;
   }

   free_resources();
   navi::gallerytype = 2;                                                  //  search results
   gallery(searchresults_file,"initF");                                    //  generate gallery of matching files

   if (Frepmeta)                                                           //  metadata report format
      searchimages_metadata_report();

   m_viewmode(0,"G");
   gallery(0,"paint",0);                                                   //  position at top
   return 1;

filerror:
   zmessLogACK(Mwin,"file error: %s",strerror(errno));
   Ffuncbusy = 0;
   Fblock = 0;
   return 1;
}


//  test a given image against selection criteria, return match status

int searchimages_select(sxrec_t &sxrec)
{
   int  searchimages_geotags_select(char *gtagsrec);
   int  searchimages_metadata_select(char *imagefile);

   cchar    *pps, *ppf;
   int      iis, iif;
   int      Nmatch, Nnomatch;

   if (Ffiles)                                                             //  file name match is wanted
   {
      Nmatch = Nnomatch = 0;

      for (iis = 1; ; iis++)
      {
         pps = strField(searchfiles,' ',iis);                              //  step thru search file names
         if (! pps) break;
         if (strcasestr(sxrec.file,pps)) Nmatch++;
         else Nnomatch++;
      }

      if (Nmatch == 0) return 0;                                           //  no match any file
      if (Fallfiles && Nnomatch) return 0;                                 //  no match all files (dir & file names)
   }

   if (Fdates)                                                             //  date match is wanted
   {
      if (strcmp(sxrec.pdate,searchDateFrom) < 0) return 0;
      if (strcmp(sxrec.pdate,searchDateTo) > 0) return 0;
   }

   if (Ftags)                                                              //  tags match is wanted
   {
      Nmatch = Nnomatch = 0;
      strToLower(sxrec.tags);

      for (iis = 1; ; iis++)                                               //  step thru search tags
      {
         pps = strField(searchtags,",;",iis);                              //  (delimited, wildcards)
         if (! pps) break;
         if (*pps == ' ') continue;

         for (iif = 1; ; iif++)                                            //  step thru file tags (delimited)
         {
            ppf = strField(sxrec.tags,",;",iif);
            if (! ppf) { Nnomatch++; break; }                              //  count matches and fails
            if (*ppf == ' ') continue;
            if (MatchWild(pps,ppf) == 0) { Nmatch++; break; }              //  wildcard match
         }
      }

      if (Nmatch == 0) return 0;                                           //  no match to any tag
      if (Falltags && Nnomatch) return 0;                                  //  no match to all tags
   }

   if (Fstars)                                                             //  rating (stars) match is wanted
   {
      if (*searchStarsFrom && sxrec.rating[0] < *searchStarsFrom) return 0;
      if (*searchStarsTo && sxrec.rating[0] > *searchStarsTo) return 0;
   }

   if (Ftext)                                                              //  text match is wanted
   {
      Nmatch = Nnomatch = 0;

      for (iis = 1; ; iis++)                                               //  step through search words
      {
         pps = strField(searchtext,' ',iis);
         if (! pps) break;
         if (*pps == ' ') continue;
         if (strcasestr(sxrec.capt,pps)) Nmatch++;                         //  search captions for word
         else if (strcasestr(sxrec.comms,pps)) Nmatch++;                   //  search comments for word
         else Nnomatch++;
      }

      if (Nmatch == 0) return 0;                                           //  no match to any word
      if (Falltext && Nnomatch) return 0;                                  //  no match to all words
   }

   if (Fgtags)                                                             //  geotags match is wanted
      if (! searchimages_geotags_select(sxrec.gtags)) return 0;

   if (Fsearchmeta)                                                        //  other metadata match
      if (! searchimages_metadata_select(sxrec.file)) return 0;

   return 1;
}


/**************************************************************************/

//  dialog to get geotags search criteria

int searchimages_geotags_dialog(zdialog *zdp)
{
   int searchimages_geotags_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         zstat;

/***
            Add Geotags Search Criteria

         city [__________]  country [__________]
         latitude [_______]  longitude [_______]
         [find]  [map]   range (km) [___]

                      [clear] [apply] [cancel]
***/

   if (! init_geolocs()) return 0;

   zd = zdialog_new(ZTX("Add Geotags Search Criteria"),Mwin,Bclear,Bapply,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbg1","dialog",0);
   zdialog_add_widget(zd,"label","labcity","hbg1",ZTX("city"),"space=3");
   zdialog_add_widget(zd,"entry","city","hbg1",0,"expand");
   zdialog_add_widget(zd,"label","space","hbg1",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hbg1",ZTX("country"),"space=3");
   zdialog_add_widget(zd,"entry","country","hbg1",0,"expand");

   zdialog_add_widget(zd,"hbox","hbg2","dialog",0);
   zdialog_add_widget(zd,"label","lablat","hbg2","latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hbg2",0,"size=10");
   zdialog_add_widget(zd,"label","space","hbg2",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hbg2","longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hbg2",0,"size=10");

   zdialog_add_widget(zd,"hbox","hbg3","dialog",0);
   zdialog_add_widget(zd,"button","find","hbg3",Bfind,"space=10");
   zdialog_add_widget(zd,"button","map","hbg3",Bmap,"space=10");
   zdialog_add_widget(zd,"label","labkm","hbg3",ZTX("range (km)"),"space=10");
   zdialog_add_widget(zd,"entry","range","hbg3","100","size=5");

   zdialog_show(zdp,0);                                                    //  hide parent dialog

   zdialog_restore_inputs(zd);                                             //  preload prior user inputs
   zd_geotags = zd;                                                        //  activate geomap clicks             v.14.12
   zdialog_run(zd,searchimages_geotags_dialog_event);                      //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion

   zdialog_free(zd);
   zd_geotags = 0;                                                         //  deactivate geomap clicks
   zdialog_show(zdp,1);                                                    //  restore parent dialog

   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion function

int searchimages_geotags_dialog_event(zdialog *zd, cchar *event)
{
   int         fgtags, err, nn;
   char        city[100], country[100];
   char        *location[2], *coord[2];

   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (strmatch(event,"find")) {                                           //  get geotags data from
      zdialog_fetch(zd,"city",city,99);                                    //    (partial) city [country] names
      zdialog_fetch(zd,"country",country,99);
      location[0] = city;
      location[1] = country;
      nn = geotags_choosecity(location,coord);                             //  find matching city geotags data
      if (nn == 1) {                                                       //    and ask user to choose
         zdialog_stuff(zd,"city",location[0]);
         zdialog_stuff(zd,"country",location[1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   if (strmatch(event,"map")) {                                            //  [map]
      m_viewmode(0,"W");
      return 0;
   }

   if (strmatch(event,"enter")) zd->zstat = 2;                             //  [apply]  v.14.03

   if (! zd->zstat) return 0;                                              //  wait for completion

   if (zd->zstat == 1) {
      zdialog_stuff(zd,"city","");                                         //  clear
      zdialog_stuff(zd,"country","");
      zdialog_stuff(zd,"latitude","");
      zdialog_stuff(zd,"longitude","");
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   if (zd->zstat != 2) {                                                   //  cancel
      *searchCity = *searchCountry = 0;                                    //  clear criteria
      *searchLatitude = *searchLongitude = 0;
      zd_geotags = 0;                                                      //  deactivate geomap clicks
      return 0;
   }

   zdialog_fetch(zd,"city",searchCity,99);                                 //  apply - get search geotags
   zdialog_fetch(zd,"country",searchCountry,99);
   zdialog_fetch(zd,"latitude",searchLatitude,20);
   zdialog_fetch(zd,"longitude",searchLongitude,20);
   zdialog_fetch(zd,"range",searchRange);

   if (*searchCity || *searchCountry)                                      //  if city or country supplied,
      if (strmatch(searchLatitude,"null"))                                 //    remove "null" lati/longi search
         *searchLatitude = *searchLongitude = 0;

   strToLower(searchCity);
   strToLower(searchCountry);

   err = 0;
   if (*searchLatitude || *searchLongitude) {                              //  sanity check for latitude/longitude
      if (! *searchLatitude || ! *searchLongitude) err++;
      flatitude = atof(searchLatitude);
      flongitude = atof(searchLongitude);
      if (flatitude < -90 || flatitude > 90) err++;
      if (flongitude < -180 || flongitude > 180) err++;
      if (searchRange < 1 || searchRange > 1000) err++;
      if (strmatch(searchLatitude,"null")) err = 0;                        //  search for null lat/long
   }

   if (err) {
      zmessageACK(Mwin,ZTX("error in latitude/longitude/range"));
      zd->zstat = 0;
      return 1;
   }

   fgtags = 0;
   if (! blank_null(searchCity)) fgtags = 1;                               //  look for inputs
   if (! blank_null(searchCountry)) fgtags = 1;
   if (! blank_null(searchLatitude)) fgtags = 1;
   if (! blank_null(searchLongitude)) fgtags = 1;

   if (fgtags) zd->zstat = 1;                                              //  search criteria was given
   else zd->zstat = 2;                                                     //  no inputs were made

   return 1;
}


//  test the image geotags data against the search criteria

int searchimages_geotags_select(char *gtags)
{
   float       glatitude, glongitude, distance;
   cchar       *pps;

   if (*searchLatitude) {                                                  //  if latitude/longitude present,
      pps = strField(gtags,'^',3);                                         //    ignore city/country data
      if (! pps) return 0;
      if (strmatch(pps,"null"))                                            //  no latitude/longitude data
         if (flatitude == 0 && flongitude == 0) return 1;                  //  "no data" was wanted
      glatitude = atof(pps);
      pps = strField(gtags,'^',4);
      if (! pps) return 0;
      if (strmatch(pps,"null")) return 0;
      glongitude = atof(pps);
      distance = (flatitude - glatitude) * (flatitude - glatitude);
      distance += (flongitude - glongitude) * (flongitude - glongitude);
      distance = 111.0 * sqrt(distance);
      if (distance <= searchRange) return 1;                               //  within distance limit, match
      return 0;                                                            //  outside, no match
   }

   if (*searchCity) {                                                      //  check city and country criteria
      pps = strField(gtags,'^',1);
      if (pps && strcasestr(pps,searchCity) == 0) return 0;
   }

   if (*searchCountry) {
      pps = strField(gtags,'^',2);
      if (pps && strcasestr(pps,searchCountry) == 0) return 0;
   }

   return 1;
}


/**************************************************************************/

//  dialog to get metadata search criteria

int searchimages_metadata_dialog(zdialog *zdp)
{
   int searchimages_metadata_dialog_event(zdialog *zd, cchar *event);

   cchar *metamess = ZTX("These items are always reported: \n"
                         "date, stars, tags, caption, comment");
   zdialog  *zd;
   int      zstat;

/***
         Search and Report Metadata

         These items are always reported:
         date, stars, tags, caption, comment

            Additional Items for Report
            Keyword      Match Criteria
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]

                   [clear] [apply] [cancel]
***/

   zd = zdialog_new("Search and Report Metadata",Mwin,Bclear,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"label","labmeta","dialog",metamess,"space=3");
   zdialog_add_widget(zd,"label","labopts","dialog",ZTX("Additional Items for Report"));

   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|expand");

   zdialog_add_widget(zd,"label","lab1","vb1",ZTX("Keyword"));
   zdialog_add_widget(zd,"entry","key0","vb1");
   zdialog_add_widget(zd,"entry","key1","vb1");
   zdialog_add_widget(zd,"entry","key2","vb1");
   zdialog_add_widget(zd,"entry","key3","vb1");
   zdialog_add_widget(zd,"entry","key4","vb1");

   zdialog_add_widget(zd,"label","lab2","vb2",ZTX("Match Criteria"));
   zdialog_add_widget(zd,"entry","match0","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match1","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match2","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match3","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match4","vb2",0,"expand");

   strcpy(searchkeyx,"keyx");
   strcpy(searchkeydatax,"matchx");

   if (! searchkeys[0])                                                    //  first call initialization
   {
      for (int ii = 0; ii < 5; ii++) {
         searchkeys[ii] = (char *) zmalloc(40);
         searchkeydata[ii] = (char *) zmalloc(100);
         *searchkeys[ii] = *searchkeydata[ii] = 0;
      }
   }

   zdialog_show(zdp,0);                                                    //  hide parent dialog
   zdialog_resize(zd,400,300);
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs
   zdialog_run(zd,searchimages_metadata_dialog_event);                     //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   zdialog_show(zdp,1);                                                    //  restore parent dialog

   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion callback function

int searchimages_metadata_dialog_event(zdialog *zd, cchar *event)
{
   int      ii, jj;
   char     keyx[8] = "keyx", matchx[8] = "matchx";

   if (strmatch(event,"enter")) zd->zstat = 2;                             //  [apply]  v.14.03
   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat == 1) {
      for (ii = 0; ii < 5; ii++) {                                         //  clear
         keyx[3] = '0' + ii;
         matchx[5] = '0' + ii;
         zdialog_stuff(zd,keyx,"");
         zdialog_stuff(zd,matchx,"");
         zd->zstat = 0;                                                    //  keep dialog active
      }
      return 0;
   }

   if (zd->zstat != 2) {
      Nsearchkeys = 0;                                                     //  no search keys
      return 1;
   }

   Nsearchkeys = 0;                                                        //  apply

   for (ii = jj = 0; ii < 5; ii++)                                         //  get metadata keys
   {
      searchkeyx[3] = '0' + ii;
      zdialog_fetch(zd,searchkeyx,searchkeys[ii],40);
      strCompress(searchkeys[ii]);                                         //  remove all blanks from key names
      if (*searchkeys[ii] <= ' ') continue;
      memmove(searchkeys[jj],searchkeys[ii],40);                           //  repack blank keys
      searchkeydatax[5] = '0' + ii;
      zdialog_fetch(zd,searchkeydatax,searchkeydata[ii],100);              //  get corresp. match value if any
      strTrim2(searchkeydata[jj],searchkeydata[ii]);                       //  trim leading and trailing blanks
      if (ii > jj) *searchkeys[ii] = *searchkeydata[ii] = 0;
      jj++;
   }

   Nsearchkeys = jj;                                                       //  keys found, no blanks

   if (Nsearchkeys) zd->zstat = 1;                                         //  keys were entered
   else zd->zstat = 2;                                                     //  no keys were entered

   return 1;
}


//  test image metadata against metadata select criteria

int searchimages_metadata_select(char *file)
{
   char           *kvals[5];
   int            ii, nth;
   cchar          *pp;

   exif_get(file,(cchar **) searchkeys,kvals,Nsearchkeys);                 //  get the image metadata

   for (ii = 0; ii < Nsearchkeys; ii++) {                                  //  loop all metadata search keys
      if (*searchkeydata[ii] > ' ') {                                      //  key match value(s) are present
         if (kvals[ii]) {                                                  //  key values present in metadata
            for (nth = 1; ; nth++) {                                       //  loop all match values
               pp = strField(searchkeydata[ii],' ',nth);                   //  get each match value
               if (! pp) return 0;                                         //  no more, no match found
               if (strcasestr(kvals[ii],pp)) break;                        //  found match (substring also)
            }
         }
         else return 0;                                                    //  empty metadata, no match
      }
   }

   return 1;
}


//  Report the selected metadata using a gallery window layout
//  with image thumbnails and selected metadata text.

int searchimages_metadata_report()
{
   using namespace navi;

   cchar    *keys1[7] = { exif_date_key, iptc_rating_key,
                          iptc_keywords_key,                               //  v.14.02
                          exif_city_key, exif_country_key,
                          iptc_caption_key, exif_comment_key };
   cchar    *keys[20];
   char     *file, *kvals[20];
   char     text1[2*indexrecl], text2[200];
   int      Nkeys, ii, jj, cc;

   if (! mdlist) {
      cc = nfiles * sizeof(char *);                                        //  allocate metadata list
      mdlist = (char **) zmalloc(cc);                                      //  nfiles = curr. gallery files
      memset(mdlist,0,cc);                                                 //  mdlist = corresp. metadata list
   }                                                                       //  (zfree() in image_navigate)

   for (ii = 0; ii < 7; ii++)                                              //  set first 6 key names, fixed
      keys[ii] = keys1[ii];

   for (ii = 0; ii < Nsearchkeys; ii++)                                    //  remaining key names from user
      keys[ii+7] = searchkeys[ii];

   Nkeys = 7 + Nsearchkeys;                                                //  total keys to extract              v.14.02
   mdrows = Nkeys - 1;                                                     //  report rows (city/country 1 row)

   for (ii = 0; ii < nfiles; ii++)                                         //  loop image gallery files
   {
      file = gallery(0,"find",ii);
      if (! file) continue;

      exif_get(file,keys,kvals,Nkeys);                                     //  get the metadata

      snprintf(text2,200,"%s  %s",kvals[3],kvals[4]);                      //  combine city and country
      if (kvals[3]) zfree(kvals[3]);                                       //  bugfix                             v.14.06
      if (kvals[4]) zfree(kvals[4]);
      kvals[3] = zstrdup(text2);
      kvals[4] = 0;

      for (cc = jj = 0; jj < Nkeys; jj++)                                  //  add metadata to report
      {
         if (jj == 4) continue;                                            //  skip country
         if (jj == 0 && kvals[0]) kvals[0][4] = kvals[0][7] = '-';         //  conv. yyyy:mm:dd to yyyy-mm-dd
         snprintf(text2,200,"key: %s  value: %s \n",keys[jj], kvals[jj]);
         strcpy(text1+cc,text2);
         cc += strlen(text2);
      }

      if (mdlist[ii]) zfree(mdlist[ii]);                                   //  attach metadata for gallery paint
      mdlist[ii] = zstrdup(text1);

      for (jj = 0; jj < Nkeys; jj++)                                       //  free memory
         if (kvals[jj]) zfree(kvals[jj]);

      zfree(file);
   }

   gallerytype = 3;                                                        //  gallery type = search results/metadata

   return 0;
}


/**************************************************************************
   Functions to read and write exif/iptc or other metadata
***************************************************************************/

//  get EXIF/IPTC metadata for given image file and EXIF/IPTC key(s)
//  returns array of pointers to corresponding key values
//  if a key is missing, corresponding pointer is null
//  returned strings belong to caller, are subject for zfree()
//  up to 20 keynames may be requested per call
//  returns: 0 = OK, +N = error

int exif_get(cchar *file, cchar **keys, char **kdata, int nkeys)           //  revised                            v.14.12
{
   char        *pp;
   char        *inputs[30], *outputs[99];
   int         cc, ii, jj, err;

   if (nkeys < 1 || nkeys > 20) zappcrash("exif_get nkeys: %d",nkeys);

   cc = nkeys * sizeof(char *);                                            //  clear outputs
   memset(kdata,0,cc);

   inputs[0] = (char *) "-m";                                              //  options for exiftool
   inputs[1] = (char *) "-s2";
   inputs[2] = (char *) "-n";
   inputs[3] = (char *) "-fast2";                                          //  -fast2 reinstated                  v.14.09
   jj = 4;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]);                                               //  -keyname
      inputs[jj] = (char *) zmalloc(cc+2);
      inputs[jj][0] = '-';
      strcpy(inputs[jj]+1,keys[ii]);
      jj++;
   }

   inputs[jj] = zstrdup(file);                                             //  filename last
   jj++;

   err = exif_server(jj,inputs,outputs);                                   //  get exif outputs
   if (err) return 1;                                                      //  exif_server() failure

   for (ii = 4; ii < jj; ii++)                                             //  free memory
      zfree(inputs[ii]);

   for (ii = 0; ii < nkeys; ii++)                                          //  search outputs
   {
      pp = outputs[ii];                                                    //  keyname: keyvalue
      if (! pp) break;

      for (jj = 0; jj < nkeys; jj++)
      {
         uint cc = strlen(keys[jj]);                                       //  look for matching input keyname
         if (! strmatchcaseN(pp,keys[jj],cc)) continue;
         if (strlen(pp) > cc+2)                                            //  if not empty,
            kdata[jj] = zstrdup(pp+cc+2);                                  //    return keyvalue alone
      }

      zfree(pp);
   }

   return 0;
}


/**************************************************************************/

//  create or change EXIF/IPTC metadata for given image file and key(s)
//  up to 20 keys may be processed
//  command:
//    exiftool -m -overwrite_original -keyname="keyvalue" ... "file"
//
//  NOTE: exiftool replaces \n (newline) in keyvalue with . (period).

int exif_put(cchar *file, cchar **keys, cchar **kdata, int nkeys)
{
   int      ii, jj, cc;
   char     *inputs[30];

   if (nkeys < 1 || nkeys > 20) zappcrash("exif_put nkeys: %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  exiftool options
   inputs[1] = (char *) "-overwrite_original";                             //  -P preserve date removed
   jj = 2;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]) + strlen(kdata[ii]) + 3;
      inputs[jj] = (char *) zmalloc(cc);
      inputs[jj][0] = '-';                                                 //  -keyname=value
      strcpy(inputs[jj]+1,keys[ii]);
      cc = strlen(keys[ii]);
      inputs[jj][cc+1] = '=';
      strcpy(inputs[jj]+cc+2,kdata[ii]);
      jj++;

      if (strmatchcase(keys[ii],"GPSLatitude")) {                          //  take care of latitude N/S
         if (*kdata[ii] == '-')
            inputs[jj] = zstrdup("-GPSLatitudeRef=S");
         else
            inputs[jj] = zstrdup("-GPSLatitudeRef=N");
         jj++;
      }

      if (strmatchcase(keys[ii],"GPSLongitude")) {                         //  and longitude E/W
         if (*kdata[ii] == '-')
            inputs[jj] = zstrdup("-GPSLongitudeRef=W");
         else
            inputs[jj] = zstrdup("-GPSLongitudeRef=E");
         jj++;
      }
   }

   inputs[jj] = zstrdup(file);                                             //  last input is filename
   jj++;

   exif_server(jj,inputs,0);                                               //  outputs discarded

   for (ii = 2; ii < jj; ii++)                                             //  free memory
       zfree(inputs[ii]);

   return 0;
}


/**************************************************************************/

//  copy EXIF/IPTC data from one image file to new (edited) image file
//  if nkeys > 0, up to 20 keys may be replaced with new values
//  exiftool -m -tagsfromfile file1 -all -xmp -icc_profile [-keyname=newvalue ...]
//                file2 -overwrite_original

int exif_copy(cchar *file1, cchar *file2, cchar **keys, cchar **kdata, int nkeys)
{
   char     *inputs[30];                                                   //  revised
   int      cc, ii, jj;

   if (nkeys > 20) zappcrash("exif_copy() nkeys %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  -m               (suppress warnings)
   inputs[1] = (char *) "-tagsfromfile";                                   //  -tagsfromfile
   inputs[2] = zstrdup(file1);                                             //  file1
   inputs[3] = (char *) "-all";                                            //  -all
   inputs[4] = (char *) "-xmp";                                            //  -xmp             added v.14.01
   inputs[5] = (char *) "-icc_profile";                                    //  -icc_profile

   jj = 6;                                                                 //  count of inputs so far

   for (int ii = 0; ii < nkeys; ii++)                                      //  -keyname=keyvalue
   {
      cc = strlen(keys[ii]) + strlen(kdata[ii]) + 3;
      inputs[jj] = (char *) zmalloc(cc);
      *inputs[jj] = 0;                                                     //  v.14.03
      strncatv(inputs[jj],cc,"-",keys[ii],"=",kdata[ii],null);
      jj++;
   }

   inputs[jj++] = zstrdup(file2);                                          //  file2
   inputs[jj++] = (char *) "-overwrite_original";                          //  -overwrite_original

   exif_server(jj,inputs,0);

   zfree(inputs[2]);                                                       //  free memory
   for (ii = 6; ii < jj-1; ii++)
   zfree(inputs[ii]);

   return 0;
}


/**************************************************************************

   int exif_server(int Nrecs, char **inputs, char **outputs)

   Server wrapper for exiftool for put/get exif/iptc data.
   This saves perl startup overhead for each call (0.1 >> 0.01 secs).

   Input records: -opt1
                  -opt2
                    ...
                  -keyword1=value1
                  -keyword2=value2
                    ...
                  filename.jpg
                  (null pointer)

   Returned: list of pointers to resulting output records.
   There are <= Nrecs outputs corresponding to keyword inputs.
   These are formatted keyword: text-string-value.
   If < Nrecs output, unused outputs are null pointers.
   These are subject to zfree().
   If outputs = null, outputs are discarded.

   Returns 0 if OK, +N if error.

   First call:
      Starts exiftool with pipe input and output files.
   Subsequent calls:
      The existing exiftool process is re-used so that the
      substantial startup overhead is avoided.
   To kill the exiftool process: exif_server(0,0,0).

***************************************************************************/

int exif_server(int Nrecs, char **inputs, char **outputs)                  //  revised API                        v.14.12
{
   static int   fcf = 1;                                                   //  first call flag
   static FILE  *fid1 = 0, *fid2 = 0;                                      //  exiftool input, output files
   static char  input_file[100] = "";
   int          ii, cc, err;
   char         *pp, command[100];
   char         outrec[exif_maxcc];                                        //  single output record

   if (! *input_file)                                                      //  exif_server input file
      snprintf(input_file,99,"%s/exiftool_input",tempdir);

   if (! Nrecs)                                                            //  kill exiftool process
   {                                                                       //  (also orphaned process)
      fid1 = fopen(input_file,"a");
      if (fid1) {
         fprintf(fid1,"-stay_open\nFalse\n");                              //  tell it to exit
         fclose(fid1);                                                     //  bugfix: fflush after fclose
      }
      remove(input_file);
      if (fid2) pclose(fid2);
      fid2 = 0;
      fcf = 1;                                                             //  start exiftool process if called again
      return 0;
   }

   if (fcf)                                                                //  first call only
   {
      fid1 = fopen(input_file,"w");                                        //  start exiftool input file
      if (! fid1) {
         zmessageACK(Mwin,"exif_server: %s \n",strerror(errno));
         return 1;
      }

      snprintf(command,100,"exiftool -stay_open True -@ %s",input_file);

      fid2 = popen(command,"r");                                           //  start exiftool and output file
      if (! fid2) {
         zmessageACK(Mwin,"exif_server: %s \n",strerror(errno));
         fclose(fid1);
         return 2;
      }

      fcf = 0;
   }

   gallery_monitor("stop");                                                //  stop excess gallery inits

   for (ii = 0; ii < Nrecs; ii++)
      fprintf(fid1,"%s\n",inputs[ii]);                                     //  write to exiftool, 1 per record

   err = fprintf(fid1,"-execute\n");                                       //  tell exiftool to process
   if (err < 0) {
      zmessageACK(Mwin,"exif_server: %s \n",strerror(errno));
      return 3;
   }

   fflush(fid1);                                                           //  flush buffer

   if (outputs) {
      cc = Nrecs * sizeof(char *);                                         //  clear outputs
      memset(outputs,0,cc);
   }

   for (ii = 0; ii < 99; ii++)                                             //  get ALL exiftool outputs
   {
      pp = fgets_trim(outrec,exif_maxcc,fid2,1);
      if (! pp) break;
      if (strncmp(outrec,"{ready}",7) == 0) break;                         //  look for output end
      if (outputs && ii < Nrecs)
         outputs[ii] = zstrdup(outrec);                                    //  add to returned records
   }

   gallery_monitor("start");
   return 0;
}


/**************************************************************************/

//  convert between EXIF and fotoxx tag date formats
//  EXIF date: yyyy:mm:dd hh:mm:ss        20 chars.
//  tag date: yyyymmddhhmmss              16 chars.
//

void exif_tagdate(cchar *exifdate, char *tagdate)
{
   int      cc;

   memset(tagdate,0,15);
   cc = strlen(exifdate);

   if (cc > 3) strncpy(tagdate+0,exifdate+0,4);
   if (cc > 6) strncpy(tagdate+4,exifdate+5,2);
   if (cc > 9) strncpy(tagdate+6,exifdate+8,2);
   if (cc > 12) strncpy(tagdate+8,exifdate+11,2);
   if (cc > 15) strncpy(tagdate+10,exifdate+14,2);
   if (cc > 18) strncpy(tagdate+12,exifdate+17,2);
   tagdate[14] = 0;
   return;
}

void tag_exifdate(cchar *tagdate, char *exifdate)
{
   int      cc;

   memset(exifdate,0,20);
   cc = strlen(tagdate);

   strcpy(exifdate,"1900:01:01 00:00:00");
   if (cc > 3) strncpy(exifdate+0,tagdate+0,4);
   if (cc > 5) strncpy(exifdate+5,tagdate+4,2);
   if (cc > 7) strncpy(exifdate+8,tagdate+6,2);
   if (cc > 9) strncpy(exifdate+11,tagdate+8,2);
   if (cc > 11) strncpy(exifdate+14,tagdate+10,2);
   if (cc > 13) strncpy(exifdate+17,tagdate+12,2);
   exifdate[19] = 0;
   return;
}


/**************************************************************************
   Functions to read and write image index records
***************************************************************************/

//  Initialize for reading/writing to the image index.
//  Build a memory map of the first image file in each image index subfile.
//  Set refresh = 1 to initialize after an index file has been split
//    or the first image file in an image index file has changed.
//  Returns 0 if OK, +N otherwise.
//
//    /.../.fotoxx/image_index/index_001
//    /.../.fotoxx/image_index/index_002
//    /.../.fotoxx/image_index/index_...         up to 999
//
//  Each of these files contains up to 'image_index_max' image filespecs.

char     **image_index_map = 0;                                            //  1st image file in each index file

int init_image_index(int refresh)
{
   int      ii, cc;
   char     *pp;
   char     indexfile[200];
   char     buff[indexrecl];
   FILE     *fid;

   if (image_index_map && ! refresh) return 0;                             //  already initialized

   if (image_index_map) {                                                  //  free memory
      for (ii = 0; image_index_map[ii]; ii++)
         zfree(image_index_map[ii]);
      zfree(image_index_map);
   }

   cc = 1001 * sizeof(char *);                                             //  allocate memory
   image_index_map = (char **) zmalloc(cc);
   memset(image_index_map,0,cc);

   for (ii = 0; ii < 999; ii++)                                            //  read image index files 001 to 999
   {
      snprintf(indexfile,200,"%s/index_%03d",index_dirk,ii+1);             //  image_index_map[ii] is for
      fid = fopen(indexfile,"r");                                          //    image_index_file_(ii+1)
      if (! fid) break;
      pp = fgets_trim(buff,indexrecl,fid);
      if (pp && strmatchN(pp,"file: ",6))
         pp = zstrdup(pp+6);                                               //  map first image file
      else pp = zstrdup("empty");                                          //  if none, map "empty"
      image_index_map[ii] = pp;
      fclose(fid);
   }

   image_index_map[ii] = 0;                                                //  mark EOL

   if (ii == 0) {
      zmessLogACK(Mwin,ZTX("image index is missing"));                     //  v.14.02
      m_quit(0,0);
   }

   if (ii == 999) {
      zmessLogACK(Mwin,"too many image index files");
      m_quit(0,0);
   }

   return 0;
}


//  Return which image index file a given image file belongs in.

int find_image_index_file(cchar *file)
{
   char     *pp;
   int      ii, nn;

   init_image_index(0);

   for (ii = 0; (pp = image_index_map[ii]); ii++)
   {
      if (strmatch(pp,"empty")) continue;                                  //  same sort as image_index_compare()
      nn = strcasecmp(file,pp);
      if (nn == 0) nn = strcmp(file,pp);
      if (nn < 0) break;                                                   //  file < 1st index file, go back 1
   }

   if (ii == 0) ii = 1;                                                    //  file < 1st file in 1st index file

   while (ii > 1 && strmatch(image_index_map[ii-1],"empty")) ii--;
   return ii;                                                              //  (map ii is for index file ii+1)
}


/**************************************************************************/

//  Split an index file when the number of entries > image_index_max
//  All following index files are renumbered +1.
//  Index_index file is refreshed.
//  Returns 0 if OK, else +N.

int split_image_index_file(int Nth)
{
   int      ii, err, nn, last, filecount;
   char     indexfile1[200], indexfile2[200], indexfile3[200];
   char     buff[indexrecl], *pp;
   FILE     *fid1 = 0, *fid2 = 0, *fid3 = 0;
   STATB    statb;

   if (sxrec_fid) {                                                        //  stop get_sxrec()
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   for (last = Nth; last < 999; last++)                                    //  find last index file
   {
      snprintf(indexfile1,200,"%s/index_%03d",index_dirk,last);            //  /.../.fotoxx/image_index/index_NNN
      err = stat(indexfile1,&statb);
      if (err) break;
   }

   last--;

   if (last == 999) {
      zmessLogACK(Mwin,"too many image index files");
      return 1;
   }

   for (ii = last; ii > Nth; ii--)                                         //  rename files following Nth
   {                                                                       //    index_012 >> index_013 etc.
      snprintf(indexfile1,200,"%s/index_%03d",index_dirk,ii);
      snprintf(indexfile2,200,"%s/index_%03d",index_dirk,ii+1);
      err = rename(indexfile1,indexfile2);
      if (err) goto file_err;
   }

   snprintf(indexfile1,200,"%s/index_%03d",index_dirk,Nth);                //  read Nth file
   snprintf(indexfile2,200,"%s/index_%03d_temp2",index_dirk,Nth);          //  write 2 temp. files
   snprintf(indexfile3,200,"%s/index_%03d_temp3",index_dirk,Nth);

   fid1 = fopen(indexfile1,"r");
   if (! fid1) goto file_err;
   fid2 = fopen(indexfile2,"w");
   if (! fid2) goto file_err;
   fid3 = fopen(indexfile3,"w");
   if (! fid3) goto file_err;

   filecount = 0;

   while (true)                                                            //  copy half the entries
   {                                                                       //    to temp2 file
      pp = fgets_trim(buff,indexrecl,fid1);
      if (! pp) break;
      if (strmatchN(pp,"file: ",6)) filecount++;
      if (filecount > image_index_max/2) break;
      nn = fprintf(fid2,"%s\n",pp);
      if (! nn) goto file_err;
   }

   if (pp) {
      nn = fprintf(fid3,"%s\n",pp);                                        //  copy remaining record sets
      if (! nn) goto file_err;                                             //    to temp3 file
   }

   while (true)
   {
      pp = fgets_trim(buff,indexrecl,fid1);
      if (! pp) break;
      nn = fprintf(fid3,"%s\n",pp);
      if (! nn) goto file_err;
   }

   err = fclose(fid1);
   err = fclose(fid2);
   err = fclose(fid3);
   fid1 = fid2 = fid3 = 0;
   if (err) goto file_err;

   err = rename(indexfile2,indexfile1);                                    //  temp2 file replaces Nth index file
   if (err) goto file_err;

   snprintf(indexfile1,200,"%s/index_%03d",index_dirk,Nth+1);              //  temp3 file >> Nth+1 index file
   err = rename(indexfile3,indexfile1);
   if (err) goto file_err;

   err = init_image_index(1);                                              //  update image index map
   return err;

file_err:
   zmessLogACK(Mwin,"split_image_index error \n %s",strerror(errno));
   if (fid1) fclose(fid1);
   if (fid2) fclose(fid2);
   if (fid3) fclose(fid3);
   return 3;
}


/**************************************************************************/

//  Get the image index record for given image file.
//  Returns 0 if OK, 1 if not found, >1 if error (diagnosed).
//  Returned sxrec_t data has allocated fields subject to zfree().
//
//  Index file is kept open across calls to reduce overhead for the
//  normal case of records being accessed in image file sequence
//  (gallery paint). Random access works, but is slower.

int get_sxrec(sxrec_t &sxrec, cchar *file)
{
   FILE           *fid = 0;
   char           indexfile[200];
   static char    buff[indexrecl];
   static char    pfile[XFCC];
   int            err, Nth, nn = 0, Fcontinue;
   cchar          *pp, *pp2;
   static int     pNth, logerr = 0;
   STATB          statb;

   memset(&sxrec,0,sizeof(sxrec));                                         //  clear output record
   err = stat(file,&statb);                                                //  check image file exists
   if (err) return 1;

   Fcontinue = 1;                                                          //  test if prior search can continue
   if (! sxrec_fid) Fcontinue = 0;                                         //  no prior search still open
   Nth = find_image_index_file(file);                                      //  get index file for this image file
   if (Nth != pNth) Fcontinue = 0;                                         //  index file not the same
   if (image_fcomp(file,pfile) <= 0) Fcontinue = 0;                        //  gallery file sequence

   if (Fcontinue) fid = sxrec_fid;
   else {
      if (sxrec_fid) fclose(sxrec_fid);
      snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);              //  /.../.fotoxx/image_index/index_NNN
      fid = fopen(indexfile,"r");                                          //  open index_NNN file
      sxrec_fid = fid;
      if (! fid) goto file_err;
      *buff = 0;                                                           //  no prior data in buffer
   }

   if (! strmatchN(buff,"file: ",6) || ! strmatch(buff+6,file))            //  file in buffer not my file
   {
      while (true)                                                         //  loop until my file found
      {
         pp = fgets_trim(buff,indexrecl,fid);                              //  read existing records
         if (! pp) break;                                                  //  EOF
         if (! strmatchN(pp,"file: ",6)) continue;                         //  to start of next file set
         nn = image_fcomp(pp+6,file);
         if (nn >= 0) break;                                               //  same or after my file
      }

      if (! pp || nn > 0) {                                                //  EOF or after, file not found
         fclose(fid);
         sxrec_fid = 0;
         return 1;
      }
   }

   pNth = Nth;                                                             //  set up for poss. continuation
   strcpy(pfile,file);

   sxrec.file = zstrdup(file);                                             //  build sxrec

   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;
      if (strmatchN(pp,"file: ",6)) break;                                 //  ran into next file rec.

      if (strmatchN(pp,"date: ",6)) {                                      //  EXIF (photo) date
         pp += 6;
         pp2 = strField(pp,' ',1);                                         //  EXIF (photo) date, yyyymmddhhmmss
         if (pp2) strncpy0(sxrec.pdate,pp2,15);
         pp2 = strField(pp,' ',2);
         if (pp2) strncpy0(sxrec.fdate,pp2,15);                            //  file date, yyyymmddhhmmss          v.14.02
      }

      else if (strmatchN(pp,"stars: ",7)) {                                //  rating, '0' to '5' stars
         sxrec.rating[0] = *(pp+7);
         sxrec.rating[1] = 0;
      }

      else if (strmatchN(pp,"size: ",6))                                   //  size, "NNNNxNNNN"
         strncpy0(sxrec.size,pp+6,15);

      else if (strmatchN(pp,"tags: ",6))                                   //  tags
         sxrec.tags = zstrdup(pp+6);

      else if (strmatchN(pp,"capt: ",6))                                   //  caption
         sxrec.capt = zstrdup(pp+6);

      else if (strmatchN(pp,"comms: ",7))                                  //  comments
         sxrec.comms = zstrdup(pp+7);

      else if (strmatchN(pp,"gtags: ",7))                                  //  geotags
         sxrec.gtags = zstrdup(pp+7);
   }

   if (! sxrec.pdate[0])                                                   //  supply defaults for missing items
      strcpy(sxrec.pdate,"null");

   if (! sxrec.rating[0])
      strcpy(sxrec.rating,"0");

   if (! sxrec.size[0])
      strcpy(sxrec.size,"null");

   if (! sxrec.tags)
      sxrec.tags = zstrdup("null, ");

   if (! sxrec.capt)
      sxrec.capt = zstrdup("null");

   if (! sxrec.comms)
      sxrec.comms = zstrdup("null");

   if (! sxrec.gtags)
      sxrec.gtags = zstrdup("null^ null^ null^ null");

   return 0;

file_err:
   if (! logerr) printz("image index read error: %s \n",strerror(errno));
   logerr++;
   if (fid) fclose(fid);
   sxrec_fid = 0;
   return 3;
}


/**************************************************************************/

//  minimized version of get_sxrec() for use by gallery window.
//  fdate[16], pdate[16] and size[16] are provided by caller for returned data.
//  returns 0 if found, 1 if not, 2+ if error (diagnosed)

int get_sxrec_min(cchar *file, char *fdate, char *pdate, char *size)
{
   FILE           *fid = 0;
   char           indexfile[200];
   static char    buff[indexrecl];
   static char    pfile[XFCC];
   int            Nth, nn = 0, Fcontinue;
   cchar          *pp;
   static int     pNth, logerr = 0;

   strcpy(fdate,"");                                                       //  outputs = missing
   strcpy(pdate,"undated");
   strcpy(size,"");

   Fcontinue = 1;                                                          //  test if prior search can continue
   if (! sxrec_fid) Fcontinue = 0;                                         //  no prior search still open
   Nth = find_image_index_file(file);                                      //  get index file for this image file
   if (Nth != pNth) Fcontinue = 0;                                         //  index file not the same
   if (image_fcomp(file,pfile) <= 0) Fcontinue = 0;                        //  req. file not after prior file

   if (Fcontinue) fid = sxrec_fid;
   else {
      if (sxrec_fid) fclose(sxrec_fid);
      snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);              //  /.../.fotoxx/image_index/index_NNN
      fid = fopen(indexfile,"r");                                          //  open index_NNN file
      if (! fid) goto file_err;
      sxrec_fid = fid;
      *buff = 0;                                                           //  no prior data in buffer
   }

   if (! strmatchN(buff,"file: ",6) || ! strmatch(buff+6,file))            //  file in buffer not my file
   {
      while (true)                                                         //  loop until my file found
      {
         pp = fgets_trim(buff,indexrecl,fid);                              //  read existing records
         if (! pp) break;                                                  //  EOF
         if (! strmatchN(pp,"file: ",6)) continue;                         //  to start of next file set
         nn = image_fcomp(pp+6,file);
         if (nn >= 0) break;                                               //  same or after my file
      }

      if (! pp || nn > 0) {                                                //  EOF or after, file not found
         fclose(fid);
         sxrec_fid = 0;
         return 1;
      }
   }

   pNth = Nth;                                                             //  set up for poss. continuation
   strcpy(pfile,file);

   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;
      if (strmatchN(pp,"file: ",6)) break;                                 //  ran into next file rec.

      if (strmatchN(pp,"date: ",6))                                        //  date record
      {
         pp = strField(buff,' ',2);
         if (pp) {
            if (*pp != 'n') strncpy0(pdate,pp,15);                         //  photo date unless "null"
            pp = strField(buff,' ',3);
            if (pp) strncpy0(fdate,pp,15);                                 //  file date
         }
      }

      else if (strmatchN(pp,"size: ",6))                                   //  size, "NNNNxNNNN"
         strncpy0(size,pp+6,15);
   }

   return 0;

file_err:
   if (! logerr) printz("image index read error: %s \n",strerror(errno));
   logerr++;
   if (fid) fclose(fid);
   sxrec_fid = 0;
   return 3;
}


/**************************************************************************/

//  Add or update image index record for given image file.
//  If sxrec is null, delete index record.
//  Return 0 if success, +N if error.

int put_sxrec(sxrec_t *sxrec, cchar *file)
{
   FILE     *fid1 = 0, *fid2 = 0;
   char     indexfile[200], tempfile[200];
   char     buff[indexrecl];
   char     *pp;
   int      err, nn, Nth;
   int      filecount, Fcopy, Finsert, Finserted, Fnewfirst;
   STATB    statb;

   if (sxrec_fid) {                                                        //  stop get_sxrec()
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   err = stat(file,&statb);                                                //  check image file exists
   if (err && sxrec) return 1;

   Nth = find_image_index_file(file);                                      //  construct index file
   snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);                 //  /.../.fotoxx/image_index/index_NNN

   err = stat(indexfile,&statb);                                           //  check if missing
   if (err) {
      nn = creat(indexfile,0640);                                          //  create if needed
      if (nn < 0) goto file_err;
      close(nn);
   }

   fid1 = fopen(indexfile,"r");                                            //  open index_NNN file
   if (! fid1) goto file_err;

   strcpy(tempfile,indexfile);                                             //  temp image index file
   strcat(tempfile,"_temp");

   fid2 = fopen(tempfile,"w");                                             //  open for output
   if (! fid2) goto file_err;

   Finsert = Finserted = Fcopy = Fnewfirst = 0;
   filecount = 0;

   while (true)                                                            //  copy input to output
   {
      pp = fgets_trim(buff,indexrecl,fid1);                                //  read existing records

      if (pp && strmatchN(pp,"file: ",6))                                  //  start of a file record set
      {
         Finsert = Fcopy = Fnewfirst = 0;

         nn = strcasecmp(file,pp+6);                                       //  compare input file to index file
         if (nn == 0) nn = strcmp(file,pp+6);
         if (nn <= 0 && sxrec && ! Finserted) Finsert = 1;                 //  input <= index file, insert sxrec here
         if (nn != 0) Fcopy = 1;                                           //  input != index file, copy to output
         if (filecount == 0) {
            if (Finsert && nn < 0) Fnewfirst = 1;                          //  detect if first image file in index
            if (! Fcopy) Fnewfirst = 1;                                    //    file will be changed or deleted
         }
         filecount += Fcopy + Finsert;
      }

      if (! pp && sxrec && ! Finserted) {                                  //  input EOF, insert at the end
         Finsert = 1;
         if (filecount == 0) Fnewfirst = 1;
      }

      if (! pp && (Finserted || ! sxrec)) break;                           //  done

      if (Finsert && ! Finserted)
      {
         Finserted = 1;                                                    //  new index file recs. go here

         nn = fprintf(fid2,"file: %s\n",file);                             //  write new index file recs.
         if (! nn) goto file_err;

         compact_time(statb.st_mtime,sxrec->fdate);                        //  convert to "yyyymmddhhmmss"

         if (! sxrec->pdate[0]) strcpy(sxrec->pdate,"null");               //  EXIF (photo) date, yyyy:mm:dd

         nn = fprintf(fid2,"date: %s  %s\n",sxrec->pdate,sxrec->fdate);    //  photo and file date rec.
         if (! nn) goto file_err;

         if (sxrec->rating[0])
            nn = fprintf(fid2,"stars: %s\n",sxrec->rating);                //  rating rec.
         else nn = fprintf(fid2,"stars: 0\n");
         if (! nn) goto file_err;

         if (sxrec->size[0])
            nn = fprintf(fid2,"size: %s\n",sxrec->size);                   //  size rec.
         else nn = fprintf(fid2,"size: null\n");

         if (sxrec->tags)
            nn = fprintf(fid2,"tags: %s\n",sxrec->tags);                   //  tags rec.
         else nn = fprintf(fid2,"tags: null, ""\n");
         if (! nn) goto file_err;

         if (sxrec->capt)
            nn = fprintf(fid2,"capt: %s\n",sxrec->capt);                   //  caption rec.
         else nn = fprintf(fid2,"capt: null\n");
         if (! nn) goto file_err;

         if (sxrec->comms)
            nn = fprintf(fid2,"comms: %s\n",sxrec->comms);                 //  comments rec.
         else nn = fprintf(fid2,"comms: null\n");
         if (! nn) goto file_err;

         if (sxrec->gtags)
            nn = fprintf(fid2,"gtags: %s\n",sxrec->gtags);                 //  geotags rec.
         else nn = fprintf(fid2,"gtags: null^ null^ null^ null\n");
         if (! nn) goto file_err;

         nn = fprintf(fid2,"\n");                                          //  EOL blank rec.
         if (! nn) goto file_err;
      }

      if (pp && Fcopy) {                                                   //  copy input to output
         nn = fprintf(fid2,"%s\n",pp);                                     //    unless replaced by sxrec
         if (! nn) goto file_err;
      }
   }

   err = fclose(fid1);                                                     //  close input file
   err = fclose(fid2);                                                     //  close output file
   fid1 = fid2 = 0;
   if (err) goto file_err;

   err = rename(tempfile,indexfile);                                       //  replace index file with temp file
   if (err) goto file_err;

   if (filecount > image_index_max)                                        //  if index file too big, split
      split_image_index_file(Nth);
   else if (Fnewfirst)
      init_image_index(1);                                                 //  update image index map

   return 0;

file_err:
   zmessLogACK(Mwin,"image index write error 3\n %s",strerror(errno));
   if (fid1) fclose(fid1);
   if (fid2) fclose(fid2);
   return 3;
}


/**************************************************************************/

//  Read image index files sequentially, return one index rec. per call.
//  Set ftf = 1 for first read, will be reset to 0.
//  Returns 0 if OK, 1 if EOF, 2 if error (diagnosed).
//  Returned sxrec_t data has allocated fields subject to zfree().

int read_sxrec_seq(sxrec_t &sxrec, int &ftf)
{
   char           indexfile[200];
   static FILE    *fid = 0;
   static char    buff[indexrecl];
   static int     Nth;
   cchar          *pp, *pp2;
   int            err;
   STATB          statb;

   if (ftf)                                                                //  initial call
   {
      ftf = 0;
      snprintf(indexfile,200,"%s/index_001",index_dirk);                   //  first index file
      fid = fopen(indexfile,"r");
      if (! fid) return 2;                                                 //  no index file ?
      Nth = 1;
      *buff = 0;
   }

   while (true)
   {
      if (! strmatchN(buff,"file: ",6))                                    //  next file rec. may be already there
      {
         while (true)                                                      //  get start of next record set
         {
            pp = fgets_trim(buff,indexrecl,fid);
            if (pp && strmatchN(buff,"file: ",6)) break;
            if (pp) continue;
            fclose(fid);                                                   //  EOF, start next index file
            Nth++;
            snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);
            fid = fopen(indexfile,"r");
            if (! fid) return 1;                                           //  no more, final EOF
         }
      }

      *buff = 0;                                                           //  no longer match "file: "
      err = stat(buff+6,&statb);                                           //  check image file exists
      if (err) continue;
      if (S_ISREG(statb.st_mode)) break;
   }

   memset(&sxrec,0,sizeof(sxrec));                                         //  clear record

   sxrec.file = zstrdup(buff+6);                                           //  image file name

   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;

      if (strmatchN(pp,"file: ",6)) break;                                 //  ran into next file rec.

      else if (strmatchN(pp,"date: ",6))
      {
         pp += 6;
         pp2 = strField(pp,' ',1);                                         //  EXIF (photo) date, yyyymmddhhmmss
         if (pp2) strncpy0(sxrec.pdate,pp2,15);
         pp2 = strField(pp,' ',2);
         if (pp2) strncpy0(sxrec.fdate,pp2,15);                            //  file mod date, yyyymmddhhmmss      v.14.02
      }

      else if (strmatchN(pp,"stars: ",7)) {                                //  rating, '0' to '5' stars
         sxrec.rating[0] = *(pp+7);
         sxrec.rating[1] = 0;
      }

      else if (strmatchN(pp,"size: ",6))                                   //  size, "NNNNxNNNN"
         strncpy0(sxrec.size,pp+6,15);

      else if (strmatchN(pp,"tags: ",6))                                   //  tags
         sxrec.tags = zstrdup(pp+6);

      else if (strmatchN(pp,"capt: ",6))                                   //  caption
         sxrec.capt = zstrdup(pp+6);

      else if (strmatchN(pp,"comms: ",7))                                  //  comments
         sxrec.comms = zstrdup(pp+7);

      else if (strmatchN(pp,"gtags: ",7))                                  //  geotags
         sxrec.gtags = zstrdup(pp+7);
   }

   if (! sxrec.pdate[0])                                                   //  supply defaults for missing items
      strcpy(sxrec.pdate,"null");

   if (! sxrec.rating[0])
      strcpy(sxrec.rating,"0");

   if (! sxrec.size[0])
      strcpy(sxrec.size,"null");

   if (! sxrec.tags)
      sxrec.tags = zstrdup("null, ");

   if (! sxrec.capt)
      sxrec.capt = zstrdup("null");

   if (! sxrec.comms)
      sxrec.comms = zstrdup("null");

   if (! sxrec.gtags)
      sxrec.gtags = zstrdup("null^ null^ null^ null");

   return 0;
}


/**************************************************************************/

//  Write the image index files sequentially, 1 rec. per call
//  Set ftf = 1 for first call, will be reset to 0.
//  Set sxrec = 0 to close file after last write.
//  Returns 0 if OK, otherwise +N.
//  Used by index image files function.

int write_sxrec_seq(sxrec_t *sxrec, int &ftf)
{
   static int     Nth, filecount;
   static FILE    *fid = 0;
   char           indexfile[200], oldirk[200];
   int            nn, err;
   STATB          statb;

   if (sxrec_fid) {                                                        //  stop get_sxrec()
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   if (ftf)                                                                //  first call
   {
      ftf = 0;
      err = stat(index_dirk,&statb);
      if (! err && S_ISREG(statb.st_mode)) {                               //  rename old image_index
         strcpy(oldirk,index_dirk);
         strcat(oldirk,"_old");
         rename(index_dirk,oldirk);
      }

      err = stat(index_dirk,&statb);                                       //  create new image_index directory
      if (err) err = mkdir(index_dirk,0750);                               //    if not already there
      if (err) goto file_err;

      err = shell_quiet("rm -f -v %s/index_* > /dev/null",index_dirk);     //  delete all image index files
      Nth = 1;
      snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);              //  create initial index file
      fid = fopen(indexfile,"w");
      if (! fid) goto file_err;
      filecount = 0;
   }

   if (! sxrec) {                                                          //  EOF call
      if (fid) {
         err = fclose(fid);
         fid = 0;
         if (err) goto file_err;
      }
      err = init_image_index(1);                                           //  initz. image index map
      return err;
   }

   err = stat(sxrec->file,&statb);                                         //  check image file exists
   if (err || ! S_ISREG(statb.st_mode)) {
      printz("file %s not found \n",sxrec->file);
      return 0;
   }

   compact_time(statb.st_mtime,sxrec->fdate);                              //  convert to "yyyymmddhhmmss"

   if (filecount > 0.8 * image_index_max) {                                //  if 80% max reached,
      err = fclose(fid);                                                   //    start new index file
      fid = 0;
      if (err) goto file_err;
      if (++Nth > 999) {
         zmessLogACK(Mwin,"too many image index files");
         return 2;
      }
      snprintf(indexfile,200,"%s/index_%03d",index_dirk,Nth);              //  open/write next index file
      fid = fopen(indexfile,"w");
      if (! fid) goto file_err;
      filecount = 0;                                                       //  is empty
   }

   filecount++;                                                            //  new image file record set

   nn = fprintf(fid,"file: %s\n",sxrec->file);                             //  output: filename rec.
   if (! nn) goto file_err;

   if (! sxrec->pdate[0]) strcpy(sxrec->pdate,"null");                     //  EXIF (photo) date, yyyymmddhhmmss

   nn = fprintf(fid,"date: %s  %s\n",sxrec->pdate,sxrec->fdate);           //  photo and file date rec.
   if (! nn) goto file_err;

   if (sxrec->rating[0])
      nn = fprintf(fid,"stars: %c\n",sxrec->rating[0]);                    //  rating rec.
   else nn = fprintf(fid,"stars: 0\n");
   if (! nn) goto file_err;

   if (sxrec->size[0])
      nn = fprintf(fid,"size: %s\n",sxrec->size);
   else nn = fprintf(fid,"size: null\n");

   if (sxrec->tags)
      nn = fprintf(fid,"tags: %s\n",sxrec->tags);                          //  tags rec.
   else nn = fprintf(fid,"tags: null, ""\n");
   if (! nn) goto file_err;

   if (sxrec->capt)
      nn = fprintf(fid,"capt: %s\n",sxrec->capt);                          //  caption rec.
   else nn = fprintf(fid,"capt: null\n");
   if (! nn) goto file_err;

   if (sxrec->comms)
      nn = fprintf(fid,"comms: %s\n",sxrec->comms);                        //  comments rec.
   else nn = fprintf(fid,"comms: null\n");
   if (! nn) goto file_err;

   if (sxrec->gtags)
      nn = fprintf(fid,"gtags: %s\n",sxrec->gtags);                        //  geotags rec.
   else nn = fprintf(fid,"gtags: null^ null^ null^ null\n");
   if (! nn) goto file_err;

   nn = fprintf(fid,"\n");                                                 //  EOL blank rec.
   if (! nn) goto file_err;

   return 0;

file_err:
   zmessLogACK(Mwin,"image index write error 4\n %s",strerror(errno));
   if (fid) fclose(fid);
   fid = 0;
   return 3;
}


//  file name compare in image index sequence
//  use case blind compare first, then normal compare as a tiebreaker

int image_fcomp(cchar *file1, cchar *file2)
{
   int      nn;
   nn = strcasecmp(file1,file2);                                           //  compare case blind
   if (nn != 0) return nn;
   nn = strcmp(file1,file2);                                               //  if equal, use utf8 compare
   return nn;
}



