/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007-2015 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix@posteo.de

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************

   Fotoxx image edit - Tools menu functions

   ----------------------------------------------------------------------  F-view
   m_index              dialog to create/update image index file
   index_rebuild        create/update image index file
   m_options            user options dialog
   m_KBshortcuts        edit keyboard shortcuts
   KBshorts_load        load KB shortcuts file into memory
   m_show_brdist        show brightness distribution graph
   m_gridlines          setup for grid lines
   m_fg_color           choose color for foreground lines
   m_show_RGB           show RGB values for clicked image positions
   m_magnify            magnify the image within a radius of the mouse
   m_darkbright         highlight darkest and brightest pixels
   m_moncolor           monitor color and contrast check
   m_mongamma           monitor gamma check and adjust
   m_changelang         choose GUI language
   m_untranslated       report misting translations
   m_resources          print memory allocated and CPU time used
   m_zappcrash          zappcrash test
   ----------------------------------------------------------------------  G-view
   m_edit_bookmarks     edit bookmarks
   m_goto_bookmark      jump to a gallery/bookmark position
   m_batch_convert      batch rename, convert, resize, move
   batch_sharp_func     callable sharpen function
   m_batch_upright      find rotated files and upright them
   m_batch_deltrash     delete or trash selected files
   m_batch_dcraw        convert RAW files using DCraw
   m_batch_rawtherapee  convert RAW files using Raw Therapee
   m_duplicates         find all duplicated images
   m_burn               burn images to DVD/Blue-Ray disc

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"                                                        //  (variables in fotoxx.h are refs)
#include <sys/wait.h>

/**************************************************************************/


//  Index Image Files menu function
//  Manual index request, or missing information, or new install.
//  Dialog to supply top image directories and thumbnails directory.
//  Update the corresponding config. files and generate new image index.

void m_index(GtkWidget *, cchar *menu)
{
   void index_clickfunc(GtkWidget *widget, int line, int pos);
   int index_dialog_event(zdialog *zd, cchar *event);

   zdialog        *zd;
   FILE           *fid;
   char           filespec[200], buff[200], thumbdirk[200];
   char           *pp;
   GtkWidget      *widget;
   int            ftf, cc, zstat;
   cchar          *greet1 = ZTX("Select directories containing image files \n"
                                "(subdirectories are included automatically).");
   cchar          *greet2 = ZTX("Select to add, click on X to delete.");
   cchar          *greet3 = ZTX("Select directory for thumbnails.");
   cchar          *termmess = ZTX("Index function terminated. Fotoxx will close. \n" 
                                  "Indexing is required for search and map functions \n"
                                  "and to make the gallery pages acceptably fast.");

   F1_help_topic = "index_files";

   if (! strmatch(menu,"retry")) {                                         //  if initial index attempt
      if (checkpend("all")) return;                                        //  check nothing pending
      Fblock = 1;
   }

/***
       _____________________________________________________
      |              Index Image Files                      |              //  v.15.06
      |                                                     |
      |  Select directories containing image files.         |
      |  (subdirectories are included automatically).       |
      |  [Select] Select to add, click on X to delete.      |
      |  _________________________________________________  |
      | | X  /home/<user>/Pictures                        | |
      | | X  /home/<user>/...                             | |
      | |                                                 | |
      | |                                                 | |
      | |                                                 | |
      | |_________________________________________________| |
      |                                                     |
      | [Select] Select directory for thumbnails.           |
      | [_________________________________________________] |
      |                                                     |
      |                           [Help] [Proceed] [Cancel] |
      |_____________________________________________________|

***/

   zd = zdialog_new(ZTX("Index Image Files"),Mwin,Bhelp,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbgreet1","dialog");
   zdialog_add_widget(zd,"label","labgreet1","hbgreet1",greet1,"space=5");
   zdialog_add_widget(zd,"hbox","hbtop","dialog");
   zdialog_add_widget(zd,"button","browsetop","hbtop",Bselect,"space=5");        //  browsetop button
   zdialog_add_widget(zd,"label","labgreet2","hbtop",greet2);
   zdialog_add_widget(zd,"hbox","hbtop2","dialog",0,"expand");
   zdialog_add_widget(zd,"label","space","hbtop2",0,"space=3");
   zdialog_add_widget(zd,"vbox","vbtop2","hbtop2",0,"expand");
   zdialog_add_widget(zd,"label","space","hbtop2",0,"space=3");
   zdialog_add_widget(zd,"frame","frtop","vbtop2",0,"expand");
   zdialog_add_widget(zd,"scrwin","scrtop","frtop",0,"expand");
   zdialog_add_widget(zd,"text","topdirks","scrtop");                            //  topdirks text
   zdialog_add_widget(zd,"vbox","vbspace","dialog",0,"space=5");
   zdialog_add_widget(zd,"hbox","hbthumb1","dialog");
   zdialog_add_widget(zd,"button","browsethumb","hbthumb1",Bselect,"space=3");   //  browsethumb button
   zdialog_add_widget(zd,"label","labgreet3","hbthumb1",greet3,"space=5");

   zdialog_add_widget(zd,"hbox","hbthumb2","dialog");
   zdialog_add_widget(zd,"frame","frthumb","hbthumb2",0,"space=5|expand");
   zdialog_add_widget(zd,"text","thumbdirk","frthumb");                          //  thumbdirk text

   widget = zdialog_widget(zd,"topdirks");                                 //  set click function for top directories
   textwidget_set_clickfunc(widget,index_clickfunc);                       //    text window

   snprintf(thumbdirk,200,"%s/thumbnails",get_zuserdir());                 //  stuff default thumbnails directory
   zdialog_stuff(zd,"thumbdirk",thumbdirk);                                //    /home/<user>/.fotoxx/thumbnails

   snprintf(filespec,200,"%s/top_directories",index_dirk);                 //  read top image directories file,
   widget = zdialog_widget(zd,"topdirks");                                 //    stuff data into dialog widgets

   fid = fopen(filespec,"r");
   if (fid) {
      zdialog_stuff(zd,"topdirks","");                                     //  clear text list
      while (true) {
         pp = fgets_trim(buff,200,fid,1);
         if (! pp) break;
         if (strmatchN(buff,"thumbnails directory: /",23)) {               //  if "thumbnails directory: /..."
            zdialog_stuff(zd,"thumbdirk",buff+22);                         //  stuff thumbnails directory
            break;                                                         //  last record
         }
         else wprintf(widget," X  %s\n",buff);                             //  stuff " X  /dir1/dir2..."
      }
      fclose(fid);
   }
   else wprintf(widget," X  %s/Pictures\n",getenv("HOME"));                //  default  /home/<user>/Pictures     v.15.07

   zdialog_resize(zd,400,350);                                             //  run dialog
   zdialog_run(zd,index_dialog_event,"parent");
   zstat = zdialog_wait(zd);                                               //  wait for completion

   if (zstat != 2) {                                                       //  canceled
      zdialog_free(zd);
      if (Findexbusy) {                                                    //  index started and not finished
         zmessageACK(Mwin,termmess);                                       //  cannot continue
         Fblock = Findexbusy = 0;                                          //  v.15.07
         m_quit(0,0);
      }
      Fblock = 0;
      return;
   }

   snprintf(filespec,200,"%s/top_directories",index_dirk);                 //  open/write top directories file
   fid = fopen(filespec,"w");
   if (! fid) {                                                            //  fatal bug
      zmessageACK(Mwin,"top directories file: \n %s",strerror(errno));
      m_quit(0,0);
   }

   widget = zdialog_widget(zd,"topdirks");                                 //  get top directories from dialog widget

   ftf = 1;
   while (true) {
      pp = wscanf(widget,ftf);                                             //  loop widget text lines
      if (! pp) break;
      pp += 4;                                                             //  skip " X  "
      if (*pp != '/') continue;
      strncpy0(buff,pp,200);                                               //  /dir1/dir2/...
      cc = strlen(buff);
      if (cc < 5) continue;                                                //  ignore blanks or rubbish
      if (buff[cc-1] == '/') buff[cc-1] = 0;                               //  remove trailing '/'
      fprintf(fid,"%s\n",buff);                                            //  write top directory to output file
   }

   zdialog_fetch(zd,"thumbdirk",buff,200);                                 //  get thumbnails directory from dialog
   strTrim2(buff);                                                         //  remove surrounding blanks
   cc = strlen(buff);
   if (cc && buff[cc-1] == '/') buff[cc-1] = 0;                            //  remove trailing '/'
   if (cc > 10)
      fprintf(fid,"thumbnails directory: %s\n",buff);                      //  write thumbnails directory to output file
   fclose(fid);

   zdialog_free(zd);                                                       //  close dialog

   index_rebuild(1);                                                       //  build image index and thumbnail files
   if (! FindexOK) m_index(0,"retry");                                     //  failed
   return;
}


//  mouse click function for top directories text window
//  remove directory from list where "X" is clicked

void index_clickfunc(GtkWidget *widget, int line, int pos)                 //  v.15.07
{
   GtkTextBuffer  *textbuff;
   GdkWindow      *gdkwin;
   char     *pp;
   char     *dirlist[maxtopdirks];
   int      ftf = 1, ii, jj;
   
   gdkwin = gtk_widget_get_window(widget);
   gdk_window_freeze_updates(gdkwin);
   
   for (ii = jj = 0; ii < maxtopdirks; ii++)                               //  loop text lines in widget
   {                                                                       //    " X  /dir1/dir2/... "
      pp = wscanf(widget,ftf);
      if (! pp) break;
      if (ii == line && pos < 3) continue;                                 //  if "X" clicked, skip deleted line
      dirlist[jj] = zstrdup(pp);
      jj++;
   }
   
   textbuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));             //  clear widget text
   gtk_text_buffer_set_text(textbuff,"",0);

   for (ii = 0; ii < jj; ii++)                                             //  stuff remaining lines back into widget
   {
      wprintf(widget,"%s\n",dirlist[ii]);
      zfree(dirlist[ii]);
   }

   gdk_window_thaw_updates(gdkwin);
   return;
}


//  index dialog event and completion function

int index_dialog_event(zdialog *zd, cchar *event)
{
   GtkWidget   *widget;
   char        **flist, *thumbdirk;
   int         ii;
   cchar       *topmess = ZTX("Choose top image directories");
   cchar       *thumbmess = ZTX("Choose thumbnail directory");

   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"browsetop")) {                                      //  [browse] top directories
      flist = zgetfiles(topmess,MWIN,"folders",getenv("HOME"));            //  get top directories from user
      if (! flist) return 0;
      widget = zdialog_widget(zd,"topdirks");                              //  add to dialog list
      for (ii = 0; flist[ii]; ii++) {
         wprintf(widget," X  %s\n",flist[ii]);                             //  " X  /dir1/dir2/..."
         zfree(flist[ii]);
      }
      zfree(flist);
   }

   if (strmatch(event,"browsethumb")) {                                    //  [browse] thumbnail directory
      thumbdirk = zgetfile(thumbmess,MWIN,"folder",getenv("HOME"));
      if (! thumbdirk) return 0;
      thumbdirk = zstrdup(thumbdirk,12);
      if (! strstr(thumbdirk,"/thumbnails"))                               //  if not containing /thumbnails,
         strcat(thumbdirk,"/thumbnails");                                  //    append /thumbnails
      zdialog_stuff(zd,"thumbdirk",thumbdirk);
      zfree(thumbdirk);
   }

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat == 1) {                                                   //  [help]
      zd->zstat = 0;
      showz_userguide("index_files");
      return 1;
   }

   return 1;                                                               //  [proceed] or cancel status
}


//  Rebuild the image index using top directories
//  and other information in configuration files.
//  Called from main() when Fotoxx is started (menu = 0)
//  Called from menu function m_index() (menu = 1)

zdialog  *zd_indexlog;

void index_rebuild(int menu)
{
   int    indexlog_dialog_event(zdialog *zd, cchar *event);
   int    index_compare(cchar *rec1, cchar *rec2);

   GtkWidget   *wlog;
   FILE        *fid;
   int         ii, jj, ntop, nthumb, err;
   int         ftf, cc, NF, orec, nrec, comp;
   char        *pp, filespec[200];
   char        buff[XFCC+500];
   char        **flist, *file, *thumbfile;
   char        *Fupdate = 0;
   STATB       statdat;
   sxrec_t     *sxrec_old = 0, *sxrec_new = 0;
   int         Nold, Nnew, imagesfound, updatesreq;

   cchar    *exifkeys[11] = { "FileName", exif_date_key, iptc_keywords_key,
            iptc_rating_key, exif_size_key, exif_comment_key, iptc_caption_key,
            exif_city_key, exif_country_key, exif_latitude_key, exif_longitude_key };

   char     *ppv[11], *filename;
   char     *exifdate, *iptctags, *iptcrating;
   char     *exifsize, *iptccapt, *exifcomms;
   char     *exifcity, *exifcountry, *exiflat, *exiflong;
   char     city2[100], country2[100], lat2[20], long2[20], gtags2[200];
   int      indexupdates, thumbupdates, thumbdeletes, sumupdates;
   char     statsmess[100] = "index updates: %d  thumbnails: %d  deleted: %d \n";

   cchar *indexmess = ZTX("No image file index was found.\n"               //  v.15.06
                          "An image file index will be created.\n"
                          "Your image files will not be changed.\n" 
                          "This may need considerable time if you \n"
                          "have many thousands of image files.");

   cchar *indexerr = ZTX("Invalid directory: \n  %s \n"
                         "Please remove.");
   
   cchar *thumberr = ZTX("Thumbnails directory: \n  %s \n"
                         "must be named .../thumbnails");
   
   cchar *duperr = ZTX("Duplicate directory: \n  %s \n"
                       "Please remove.");

   PangoFontDescription    *monofont;

   FindexOK = 0;
   Findexbusy = 1;                                                         //  must run to completion
   Fblock = 1;

   //  if index update is not wanted, exit now (-noindex parameter)

   if (Fnoindex && menu == 0) {
      printz("metadata index update suppressed \n");
      Findexbusy = 0;
      Fblock = 0;
      return;
   }

   //  get current top image directories and thumbnails directory
   //  from /home/<user>/.fotoxx/image_index/top_directories

   ntop = nthumb = 0;
   snprintf(filespec,200,"%s/top_directories",index_dirk);                 //  read top directories file

   fid = fopen(filespec,"r");
   if (fid) {
      while (true) {                                                       //  get top image directories
         pp = fgets_trim(buff,200,fid,1);
         if (! pp) break;
         if (strmatchN(buff,"thumbnails directory: /",23)) {
            if (thumbdirk) zfree(thumbdirk);
            thumbdirk = zstrdup(buff+22);
            nthumb = 1;
            break;
         }
         if (topdirks[ntop]) zfree(topdirks[ntop]);
         topdirks[ntop] = zstrdup(buff);                                   //  top directories list in global space
         if (++ntop == maxtopdirks) break;
      }
      fclose(fid);
   }

   Ntopdirks = ntop;

   if (! ntop) {                                                           //  if nothing found, must ask user
      zmessageACK(Mwin,indexmess);
      goto retry;
   }

   for (ii = 0; ii < ntop; ii++) {                                         //  validate top directories
      err = stat(topdirks[ii],&statdat);
      if (err || ! S_ISDIR(statdat.st_mode)) {
         zmessageACK(Mwin,indexerr,topdirks[ii]);                          //  v.15.07
         goto retry;
      }
   }

   if (! nthumb) {                                                         //  no thumbnails directory, must ask user
      zmessageACK(Mwin,ZTX("no thumbnails directory defined"));
      goto retry;
   }

   cc = strlen(thumbdirk) - 11 ;                                           //  check /thumbnails name             v.15.07
   if (! strmatch(thumbdirk+cc,"/thumbnails")) {
      zmessageACK(Mwin,thumberr,thumbdirk);
      goto retry;
   }

   err = stat(thumbdirk,&statdat);                                         //  create thumbnails directory if needed
   if (err || ! S_ISDIR(statdat.st_mode))
      err = mkdir(thumbdirk,0750);
   if (err) {
      zmessageACK(Mwin,"%s \n %s",thumbdirk,strerror(errno));
      goto retry;
   }
   
   for (ii = 0; ii < ntop; ii++) {                                         //  disallow top dir = thumbnail dir   v.15.06
      if (strmatch(topdirks[ii],thumbdirk)) {
         zmessageACK(Mwin,indexerr,topdirks[ii]);                          //  v.15.07
         goto retry;
      }
   }

   for (ii = 0; ii < ntop; ii++)                                           //  check for duplicate directories    v.15.07
   for (jj = ii+1; jj < ntop; jj++) {
      if (strmatch(topdirks[ii],topdirks[jj])) {
         zmessageACK(Mwin,duperr,topdirks[jj]);
         goto retry;
      }
   }

   //  create log window for reporting status and statistics

   if (zd_indexlog) zdialog_free(zd_indexlog);                             //  make dialog for output log
   zd_indexlog = zdialog_new(0,Mwin,BOK,Bcancel,null);
   zdialog_set_decorated(zd_indexlog,0);
   zdialog_set_modal(zd_indexlog);
   zdialog_add_widget(zd_indexlog,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd_indexlog,"scrwin","scrwin","frame");
   zdialog_add_widget(zd_indexlog,"text","text","scrwin");

   wlog = zdialog_widget(zd_indexlog,"text");
   monofont = pango_font_description_from_string("monospace 9");
   gtk_widget_override_font(wlog,monofont);

   zdialog_resize(zd_indexlog,500,300);
   zdialog_run(zd_indexlog,indexlog_dialog_event,"parent");

   wprintf(wlog,"top image directories:\n");                               //  log top image directories
   for (ii = 0; ii < Ntopdirks; ii++)
      wprintf(wlog," %s\n",topdirks[ii]);

   wprintf(wlog,"thumbnails directory: \n %s \n",thumbdirk);               //  and thumbnails directory
   zmainloop();

   //  read image index file and build "old list" of index recs

   cc = maximages * sizeof(sxrec_t);
   sxrec_old = (sxrec_t *) zmalloc(cc);                                    //  "old" image index recs

   Nold = 0;
   ftf = 1;

   while (true)
   {
      err = read_sxrec_seq(sxrec_old[Nold],ftf);                           //  read curr. index recs
      if (err) break;
      if (++Nold == maximages) {
         printz("exceeded max. images: %d \n",maximages);
         m_quit(0,0);
      }
   }

   //  find all image files and create "new list" of index recs

   cc = maximages * sizeof(sxrec_t);
   sxrec_new = (sxrec_t *) zmalloc(cc);                                    //  "new" image index recs

   Fupdate = (char *) zmalloc(maximages);                                  //  flag, index update needed

   Nnew = 0;

   for (ii = 0; ii < Ntopdirks; ii++)
   {
      err = find_imagefiles(topdirks[ii],flist,NF,0);
      if (err) {
         printz("find_imagefiles() failure \n");
         m_quit(0,0);
      }

      if (Nnew + NF > maximages) {
         printz("exceeded max. images: %d \n",maximages);
         m_quit(0,0);
      }

      for (jj = 0; jj < NF; jj++)
      {
         file = flist[jj];
         nrec = Nnew++;
         sxrec_new[nrec].file = zstrdup(file);                             //  image filespec
         stat(file,&statdat);
         compact_time(statdat.st_mtime,sxrec_new[nrec].fdate);             //  file mod date
         sxrec_new[nrec].pdate[0] = 0;                                     //  image date = empty
         strcpy(sxrec_new[nrec].rating,"0");                               //  stars = "0"
         strcpy(sxrec_new[nrec].size,"0x0");                               //  size = "0x0"
         sxrec_new[nrec].tags = 0;                                         //  tags = empty
         sxrec_new[nrec].capt = 0;                                         //  caption = empty
         sxrec_new[nrec].comms = 0;                                        //  comments = empty
         sxrec_new[nrec].gtags = 0;                                        //  geotags = empty
         zfree(file);
      }

      if (NF) zfree(flist);
      if (! zd_indexlog) goto retry;                                       //  check for user abort
   }

   imagesfound = Nnew;                                                     //  total images found
   wprintf(wlog,"images found: %d \n",imagesfound);

   //  merge and compare lists
   //  if filespecs match and have the same date, then "old" record is OK
   //  otherwise flag the file for index record update

   if (Nold)
      HeapSort((char *) sxrec_old,sizeof(sxrec_t),Nold,index_compare);
   if (Nnew)
      HeapSort((char *) sxrec_new,sizeof(sxrec_t),Nnew,index_compare);
   
   updatesreq = imagesfound;

   for (orec = nrec = 0; nrec < Nnew; nrec++)                              //  loop all image files
   {
      Fupdate[nrec] = 1;                                                   //  assume index update is needed

      if (orec == Nold) continue;                                          //  no more old recs

      while (true)
      {
         if (orec == Nold) break;
         comp = strcasecmp(sxrec_old[orec].file,sxrec_new[nrec].file);     //  compare orec file to nrec file
         if (comp == 0)
            comp = strcmp(sxrec_old[orec].file,sxrec_new[nrec].file);
         if (comp >= 0) break;                                             //  orec >= nrec
         orec++;                                                           //  orec < nrec, next orec
         continue;
      }

      if (comp == 0)                                                       //  orec = nrec (same image file)
      {
         if (strmatch(sxrec_new[nrec].fdate,sxrec_old[orec].fdate))        //  file dates match
         {
            Fupdate[nrec] = 0;                                             //  index update not needed
            updatesreq--;
            strncpy0(sxrec_new[nrec].pdate,sxrec_old[orec].pdate,15);      //  copy data from old to new
            strncpy0(sxrec_new[nrec].rating,sxrec_old[orec].rating,2);
            strncpy0(sxrec_new[nrec].size,sxrec_old[orec].size,15);
            sxrec_new[nrec].tags = sxrec_old[orec].tags;
            sxrec_old[orec].tags = 0;
            sxrec_new[nrec].capt = sxrec_old[orec].capt;
            sxrec_old[orec].capt = 0;
            sxrec_new[nrec].comms = sxrec_old[orec].comms;
            sxrec_old[orec].comms = 0;
            sxrec_new[nrec].gtags = sxrec_old[orec].gtags;
            sxrec_old[orec].gtags = 0;
         }
         orec++;                                                           //  next orec
      }
   }

   wprintf(wlog,"index updates needed: %d \n",updatesreq);

   for (orec = 0; orec < Nold; orec++)                                     //  release old list memory
   {
      zfree(sxrec_old[orec].file);
      if (sxrec_old[orec].tags) zfree(sxrec_old[orec].tags);
      if (sxrec_old[orec].capt) zfree(sxrec_old[orec].capt);
      if (sxrec_old[orec].comms) zfree(sxrec_old[orec].comms);
      if (sxrec_old[orec].gtags) zfree(sxrec_old[orec].gtags);
   }

   zfree(sxrec_old);
   sxrec_old = 0;

   if (! zd_indexlog) goto retry;                                          //  check for user abort


//  Process entries needing update in new index list
//  (new files or files dated later than image index date).
//  Get updated metadata from image file EXIF/IPTC data.

   wprintf(wlog,"updating image index and thumbnails ... \n");

   indexupdates = thumbupdates = thumbdeletes = sumupdates = 0;

   for (nrec = 0; nrec < Nnew; nrec++)                                     //  loop all image files
   {
      file = sxrec_new[nrec].file;                                         //  image file
      if (! Fupdate[nrec]) goto thumbupdate;                               //  no update needed

      err = exif_get(file,exifkeys,ppv,11);                                //  get iptc/exif metadata
      if (err) {
         printz("exif_get() failure: %s \n",file);                         //  metadata unreadable
         goto thumbupdate;
      }

      pp = strrchr(file,'/');
      if (! ppv[0] || ! strmatch(ppv[0],pp+1)) {                           //  image file has no metadata
         printz("exif_get() no data: %s \n",file);
         goto thumbupdate;
      }

      filename = ppv[0];
      exifdate = ppv[1];
      iptctags = ppv[2];
      iptcrating = ppv[3];
      exifsize = ppv[4];
      exifcomms = ppv[5];
      iptccapt = ppv[6];
      exifcity = ppv[7];
      exifcountry = ppv[8];
      exiflat = ppv[9];
      exiflong = ppv[10];

      if (exifdate && strlen(exifdate) > 3)                                //  exif date (photo date)
         exif_tagdate(exifdate,sxrec_new[nrec].pdate);
      else strcpy(sxrec_new[nrec].pdate,"null");                           //  not present

      if (iptcrating && strlen(iptcrating)) {                              //  iptc rating
         sxrec_new[nrec].rating[0] = *iptcrating;
         sxrec_new[nrec].rating[1] = 0;
      }
      else strcpy(sxrec_new[nrec].rating,"0");                             //  not present

      if (exifsize && strlen(exifsize))                                    //  exif size
         strncpy0(sxrec_new[nrec].size,exifsize,15);

      if (iptctags && strlen(iptctags)) {                                  //  iptc tags
         sxrec_new[nrec].tags = iptctags;
         iptctags = 0;
      }

      if (iptccapt && strlen(iptccapt)) {                                  //  iptc caption
         sxrec_new[nrec].capt = iptccapt;
         iptccapt = 0;
      }

      if (exifcomms && strlen(exifcomms)) {                                //  exif comments
         sxrec_new[nrec].comms = exifcomms;
         exifcomms = 0;
      }

      strcpy(city2,"null");                                                //  geotags = empty
      strcpy(country2,"null");
      strcpy(lat2,"null");
      strcpy(long2,"null");

      if (exifcity) strncpy0(city2,exifcity,99);                           //  get from exif if any
      if (exifcountry) strncpy0(country2,exifcountry,99);
      if (exiflat) strncpy0(lat2,exiflat,10);
      if (exiflong) strncpy0(long2,exiflong,10);

      *gtags2 = 0;
      strncatv(gtags2,200,city2,"^ ",country2,"^ ",lat2,"^ ",long2,null);
      sxrec_new[nrec].gtags = zstrdup(gtags2);

      zfree(filename);
      if (exifdate) zfree(exifdate);                                       //  free EXIF data
      if (iptctags) zfree(iptctags);
      if (iptcrating) zfree(iptcrating);
      if (exifsize) zfree(exifsize);
      if (exifcomms) zfree(exifcomms);
      if (iptccapt) zfree(iptccapt);
      if (exifcity) zfree(exifcity);
      if (exifcountry) zfree(exifcountry);
      if (exiflat) zfree(exiflat);
      if (exiflong) zfree(exiflong);

      indexupdates++;                                                      //  count index updates

   thumbupdate:                                                            //  back to combined processing        v.15.05

      thumbfile = image_thumbfile(file,&ii);                               //  update thumbnail if missing or stale
      if (thumbfile) zfree(thumbfile);
      if (ii) thumbupdates++;                                              //  count thumbnail updates

      if (! zd_indexlog) goto retry;                                       //  check for user abort

      if (indexupdates + thumbupdates > sumupdates)                        //  update progress counters
         wprintf(wlog,-2,statsmess,indexupdates,thumbupdates,thumbdeletes);
      sumupdates = indexupdates + thumbupdates;
   }

   //  write updated index records to index file

   ftf = 1;
   for (nrec = 0; nrec < Nnew; nrec++)
   {
      err = write_sxrec_seq(&sxrec_new[nrec],ftf);
      if (err) zappcrash("update image index: %s",strerror(errno));        //  v.15.03
   }

   write_sxrec_seq(null,ftf);                                              //  close output

   //  find orphan thumbnails and delete them

   snprintf(command,CCC,"find %s -type f",thumbdirk);
   fid = popen(command,"r");
   if (! fid) zappcrash("find command: %s",strerror(errno));               //  v.15.03

   while (true)
   {
      thumbfile = fgets_trim(buff,XFCC,fid);
      if (! thumbfile) break;
      file = thumb2imagefile(thumbfile);
      if (file) { zfree(file); continue; }
      remove(thumbfile);
      thumbdeletes++;
   }

   pclose(fid);

   //  free "new list" memory

   for (nrec = 0; nrec < Nnew; nrec++)
   {
      zfree(sxrec_new[nrec].file);
      if (sxrec_new[nrec].tags) zfree(sxrec_new[nrec].tags);
      if (sxrec_new[nrec].capt) zfree(sxrec_new[nrec].capt);
      if (sxrec_new[nrec].comms) zfree(sxrec_new[nrec].comms);
      if (sxrec_new[nrec].gtags) zfree(sxrec_new[nrec].gtags);
   }

   zfree(sxrec_new);
   zfree(Fupdate);
   sxrec_new = 0;
   Fupdate = 0;

   printz(statsmess,indexupdates,thumbupdates,thumbdeletes);

   wprintf(wlog,-2,statsmess,indexupdates,thumbupdates,thumbdeletes);
   wprintf(wlog,"%s\n",ZTX("COMPLETED"));                                  //  index complete and OK

   for (ii = 0; ii < 5; ii++) {                                            //  push last message out and
      zmainloop();                                                         //    delay a bit before killing
      zsleep(0.1);                                                         //       the log window
   }

   if (! menu && zd_indexlog)                                              //  kill log window if not manual run
      zdialog_send_event(zd_indexlog,"kill");

   Findexbusy = 0;                                                         //  completed OK
   FindexOK = 1;
   Fnoindex = 0;                                                           //  -noindex option moot
   Fblock = 0;
   return;

retry:   //  error or user cancel, start over (user can try again or quit)

   if (zd_indexlog) zdialog_send_event(zd_indexlog,"kill");
   if (sxrec_old) zfree(sxrec_old);
   if (sxrec_new) zfree(sxrec_new);
   if (Fupdate) zfree(Fupdate);
   
   m_index(0,"retry");
   return;
}


//  index log window dialog response function

int indexlog_dialog_event(zdialog *zd, cchar *event)
{
   int      nn;
   cchar    *canmess = ZTX("Indexing is required for first-time startup.");
   
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (strmatch(event,"kill")) {                                           //  auto-kill from index_rebuild()
      zdialog_free(zd);
      zd_indexlog = 0;
      return 1;
   }

   if (zd->zstat == 1) {                                                   //  [OK] button
      if (FindexOK) {     
         zdialog_free(zd);                                                 //  index completed, kill dialog
         zd_indexlog = 0;
         return 1;
      }
   }
   
   if (zd->zstat == 2) {                                                   //  [Cancel] button
      if (FindexOK) {
         zdialog_free(zd);                                                 //  index completed, kill dialog
         zd_indexlog = 0;
         return 1;
      }
      nn = zdialog_choose(Mwin,canmess,Bcontinue,Bcancel,null);            //  ask for confirmation               v.15.07
      if (nn == 1) {
         zd->zstat = 0;                                                    //  continue
         return 1;
      }
      zdialog_free(zd);                                                    //  cancel
      zd_indexlog = 0;
      return 1;
   }

   zd->zstat = 0;                                                          //  keep dialog open while index busy
   return 1;
}


//  sort compare function - compare index record filespecs and return
//   <0 | 0 | >0   for   file1 < | == | > file2

int index_compare(cchar *rec1, cchar *rec2)
{
   char * file1 = ((sxrec_t *) rec1)->file;
   char * file2 = ((sxrec_t *) rec2)->file;
   int nn = strcasecmp(file1,file2);                                       //  compare ignoring case
   if (nn) return nn;
   nn = strcmp(file1,file2);                                               //  if equal, use utf8 compare
   return nn;
}


/**************************************************************************/

//  user options dialog

namespace useroptions
{
   int      options_fclone;

   cchar    *startopt[7][2] = {  
               "recent", ZTX("Recent Files Gallery"),                      //  fotoxx startup display options
               "newest", ZTX("Newest Files Gallery"),
               "prevG",  ZTX("Previous Gallery"),
               "prevI",  ZTX("Previous Image"),
               "blank",  ZTX("Blank Window"),
               "dirk",   ZTX("Directory"),
               "file",   ZTX("Image File")  };
}


//  menu function

void m_options(GtkWidget *, cchar *)
{
   using namespace useroptions;
   
   int   options_dialog_event(zdialog *zd, cchar *event);

   int      ii;   
   zdialog  *zd;
   char     thumbsize[8];

/***
       ____________________________________________________________________
      |                          User Options                              |
      |                                                                    |
      |  Startup Display [ previous image |v]                              |
      |  [Browse] [______________________________________________________] |
      |                                                                    |
      |  Menu Style  (o) Icons  (o) Icons + Text     Icon Size [___|-+]    |
      |  Image Pan/Scroll: [drag magnified |v]      Zooms for 2x [___|-+]  |
      |  JPEG save quality [____|-+]     Thumbnail size [___|v]            |
      |  [x] show hidden directories in gallery view                       |
      |  [x] prev/next shows last file version only                        |
      |  RAW command   [_________________________________________________] |
      |  RAW file types  [_______________________________________________] |
      |                                                                    |
      |                                                             [done] |
      |____________________________________________________________________|

***/

   F1_help_topic = "user_options";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock++;

   zd = zdialog_new(ZTX("User Options"),Mwin,Bdone,null);

   zdialog_add_widget(zd,"hbox","hbstart","dialog");
   zdialog_add_widget(zd,"label","labstart","hbstart",ZTX("Startup Display"),"space=5");
   zdialog_add_widget(zd,"combo","startopt","hbstart",0);
   
   for (ii = 0; ii < 7; ii++)
      zdialog_cb_app(zd,"startopt",startopt[ii][1]);                       //  startup display option             v.15.07

   zdialog_add_widget(zd,"hbox","hbbrowse","dialog","space=3");
   zdialog_add_widget(zd,"button","browse","hbbrowse",Bbrowse,"space=5");
   zdialog_add_widget(zd,"entry","browsefile","hbbrowse",0,"expand");

   zdialog_add_widget(zd,"hbox","hbmenu","dialog");
   zdialog_add_widget(zd,"label","labms","hbmenu",ZTX("Menu Style"),"space=5");
   zdialog_add_widget(zd,"radio","icons","hbmenu",ZTX("Icons"),"space=3");
   zdialog_add_widget(zd,"radio","both","hbmenu",ZTX("Icons + Text"),"space=5");
   zdialog_add_widget(zd,"label","space","hbmenu",0,"space=10");
   zdialog_add_widget(zd,"label","labis","hbmenu",ZTX("Icon size"));
   zdialog_add_widget(zd,"spin","iconsize","hbmenu","32|48|1|40","space=3");
   
   zdialog_add_widget(zd,"hbox","hbips","dialog");
   zdialog_add_widget(zd,"label","labips","hbips",ZTX("Image Pan/scroll:"),"space=5");
   zdialog_add_widget(zd,"combo","panscroll","hbips",0);
   zdialog_add_widget(zd,"label","space","hbips",0,"space=10");
   zdialog_add_widget(zd,"label","labzoom","hbips",ZTX("Zooms for 2x"),"space=5");
   zdialog_add_widget(zd,"spin","zoomcount","hbips","1|4|1|2");
   
   zdialog_cb_app(zd,"panscroll","drag");                                  //  drag/scroll option                 v.15.07
   zdialog_cb_app(zd,"panscroll","scroll");
   zdialog_cb_app(zd,"panscroll","drag magnified");
   zdialog_cb_app(zd,"panscroll","scroll magnified");

   zdialog_add_widget(zd,"hbox","hbjpeg","dialog");
   zdialog_add_widget(zd,"label","labqual","hbjpeg",ZTX("JPEG save quality"),"space=5");
   zdialog_add_widget(zd,"spin","quality","hbjpeg","1|100|1|90");
   zdialog_add_widget(zd,"label","space","hbjpeg",0,"space=10");
   zdialog_add_widget(zd,"label","labthumb","hbjpeg",ZTX("Thumbnail size"),"space=5");
   zdialog_add_widget(zd,"combo","thumbsize","hbjpeg",0,"space=5");

   zdialog_add_widget(zd,"hbox","hbhidden","dialog");
   zdialog_add_widget(zd,"check","showhidden","hbhidden",ZTX("show hidden directories in gallery view"),"space=5");

   zdialog_add_widget(zd,"hbox","hblastver","dialog");
   zdialog_add_widget(zd,"check","lastver","hblastver",ZTX("prev/next shows last file version only"),"space=5");

   zdialog_add_widget(zd,"hbox","hbrawcomm","dialog");
   zdialog_add_widget(zd,"label","rawlab1","hbrawcomm",ZTX("RAW command"),"space=5");
   zdialog_add_widget(zd,"entry","rawcomm","hbrawcomm","dcraw -w -q 0 -T -6","expand");

   zdialog_add_widget(zd,"hbox","hbrawfile","dialog");
   zdialog_add_widget(zd,"label","rawlab2","hbrawfile",ZTX("RAW file types"),"space=5");
   zdialog_add_widget(zd,"entry","rawtypes","hbrawfile",".raw .rw2","expand");


   for (ii = 0; ii < 7; ii++) {
      if (strmatch(startdisplay,startopt[ii][0]))                          //  set startup display option         v.15.07
         zdialog_stuff(zd,"startopt",startopt[ii][1]);
   }

   if (strmatch(startdisplay,"dirk"))                                      //  if directory or file option,       v.15.07
      zdialog_stuff(zd,"browsefile",startdirk);                            //    set current directory or file
   if (strmatch(startdisplay,"file")) 
      zdialog_stuff(zd,"browsefile",startfile);
   
   zdialog_stuff(zd,"icons",0);                                            //  menu style
   zdialog_stuff(zd,"both",0);

   if (strmatch(menu_style,"icons"))
      zdialog_stuff(zd,"icons",1);
   else zdialog_stuff(zd,"both",1);

   zdialog_stuff(zd,"iconsize",iconsize);                                  //  icon size                          v.14.08
   
   if (Fdragopt == 1) zdialog_stuff(zd,"panscroll","drag");                //  set pan/scroll option              v.15.07
   if (Fdragopt == 2) zdialog_stuff(zd,"panscroll","scroll");
   if (Fdragopt == 3) zdialog_stuff(zd,"panscroll","drag magnified");
   if (Fdragopt == 4) zdialog_stuff(zd,"panscroll","scroll magnified");

   if (zoomcount >= 1 && zoomcount <= 4)                                   //  zooms for 2x increase
      zdialog_stuff(zd,"zoomcount",zoomcount);

   zdialog_stuff(zd,"quality",jpeg_def_quality);                           //  default jpeg file save quality

   zdialog_cb_app(zd,"thumbsize","128");                                   //  thumbnail file pixel size          v.15.02
   zdialog_cb_app(zd,"thumbsize","256");
   zdialog_cb_app(zd,"thumbsize","512");
   snprintf(thumbsize,8,"%d",thumbfilesize);                               //  preload with current value
   zdialog_stuff(zd,"thumbsize",thumbsize);

   zdialog_stuff(zd,"showhidden",Fshowhidden);                             //  show hidden dirs in gallery view   v.15.04
   zdialog_stuff(zd,"lastver",Flastversion);                               //  prev/next shows last version only  v.14.12

   zdialog_stuff(zd,"rawcomm",DCrawcommand);                               //  RAW file conversion command
   zdialog_stuff(zd,"rawtypes",RAWfiletypes);                              //  RAW file types

   options_fclone = 0;

   zdialog_run(zd,options_dialog_event);                                   //  run dialog and wait for completion
   zdialog_wait(zd);
   zdialog_free(zd);

   Fblock = 0;

   if (options_fclone) {                                                   //  start new and exit current fotoxx
      m_clone(0,0);                                                        //  v.14.06
      m_quit(0,0);
   }

   return;
}


//  options dialog event function

int options_dialog_event(zdialog *zd, cchar *event)
{
   using namespace useroptions;

   int         ii, jj, nn;
   char        *pp, temp[200];

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (strmatch(event,"thumbsize")) {                                      //  warn if thumbfilesize is changed   v.15.02
      zdialog_fetch(zd,"thumbsize",nn);
      if (nn != thumbfilesize)
         zmessageACK(Mwin,ZTX("Delete present thumbnails to make effective"));
   }

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [done]

   if (zd->zstat && zd->zstat != 1) return 0;                              //  cancel

   if (strmatch(event,"browse"))                                           //  browse for startup directory or file
   {
      zdialog_fetch(zd,"startopt",temp,200);                               //  get startup option
      for (ii = 0; ii < 7; ii++) {
         if (strmatch(temp,startopt[ii][1])) {
            zfree(startdisplay);
            startdisplay = zstrdup(startopt[ii][0]);
         }
      }

      if (strmatch(startdisplay,"dirk")) {
         zdialog_fetch(zd,"browsefile",temp,200);
         pp = zgetfile(ZTX("Select startup directory"),MWIN,"folder",temp);
         if (! pp) return 0;
         zdialog_stuff(zd,"browsefile",pp);
         zfree(pp);
      }

      if (strmatch(startdisplay,"file")) {
         zdialog_fetch(zd,"browsefile",temp,200);
         pp = zgetfile(ZTX("Select startup image file"),MWIN,"file",temp);
         if (! pp) return 0;
         zdialog_stuff(zd,"browsefile",pp);
         zfree(pp);
      }
   }

   if (zd->zstat == 1)                                                     //  done, check inputs are valid
   {
      zdialog_fetch(zd,"startopt",temp,200);                               //  get startup option
      for (ii = 0; ii < 7; ii++) {
         if (strmatch(temp,startopt[ii][1])) {
            zfree(startdisplay);
            startdisplay = zstrdup(startopt[ii][0]);
         }
      }

      if (strmatch(startdisplay,"dirk")) {                                 //  startup option = directory 
         zdialog_fetch(zd,"browsefile",temp,200);                          //  startup directory
         if (startdirk) zfree(startdirk);
         startdirk = zstrdup(temp);
         if (image_file_type(startdirk) != 1) {
            zmessageACK(Mwin,ZTX("startup directory is invalid"));
            zd->zstat = 0;
            return 0;
         }
      }

      if (strmatch(startdisplay,"file")) {                                 //  startup option = image file
         zdialog_fetch(zd,"browsefile",temp,200);                          //  startup file
         if (startfile) zfree(startfile);
         startfile = zstrdup(temp);
         if (image_file_type(startfile) != 2) {
            zmessageACK(Mwin,ZTX("startup file is invalid"));
            zd->zstat = 0;
            return 0;
         }
      }

      zdialog_fetch(zd,"icons",nn);                                        //  menu type = icons
      if (nn) {
         if (! strmatch(menu_style,"icons")) {
            zfree(menu_style);
            menu_style = zstrdup("icons");
            options_fclone++;
         }
      }

      zdialog_fetch(zd,"both",nn);                                         //  menu type = icons + text
      if (nn) {
         if (! strmatch(menu_style,"both")) {
            zfree(menu_style);
            menu_style = zstrdup("both");
            options_fclone++;
         }
      }

      zdialog_fetch(zd,"iconsize",nn);                                     //  icon size                          v.14.08
      if (nn != iconsize) {
         iconsize = nn;
         options_fclone++;
      }

      zdialog_fetch(zd,"panscroll",temp,200);                              //  get drag/scroll option             v.15.07
      if strmatch(temp,"drag") Fdragopt = 1;
      if strmatch(temp,"scroll") Fdragopt = 2;
      if strmatch(temp,"drag magnified") Fdragopt = 3;
      if strmatch(temp,"scroll magnified") Fdragopt = 4;

      zdialog_fetch(zd,"zoomcount",zoomcount);                             //  zooms for 2x image, 1-4
      zoomratio = pow( 2.0, 1.0 / zoomcount);                              //  2.0, 1.4142, 1.2599, 1.1892

      zdialog_fetch(zd,"quality",jpeg_def_quality);                        //  default jpeg file save quality

      zdialog_fetch(zd,"thumbsize",thumbfilesize);                         //  thumbnail file pixel size          v.15.02

      zdialog_fetch(zd,"showhidden",Fshowhidden);                          //  show hidden dirs in gallery view   v.15.04

      zdialog_fetch(zd,"lastver",Flastversion);                            //  prev/next shows last version only  v.14.12

      zdialog_fetch(zd,"rawcomm",temp,200);                                //  command for RAW to tiff using DCraw
      if (DCrawcommand) zfree(DCrawcommand);
      DCrawcommand = zstrdup(temp);

      zdialog_fetch(zd,"rawtypes",temp,200);                               //  RAW file types, .raw .rw2 ...
      pp = zstrdup(temp,100);

      for (ii = jj = 0; temp[ii]; ii++) {                                  //  insure blanks between .raw types
         if (temp[ii] == '.' && temp[ii-1] != ' ') pp[jj++] = ' ';
         pp[jj++] = temp[ii];
      }
      if (pp[jj-1] != ' ') pp[jj++] = ' ';                                 //  insure 1 final blank               v.14.07
      pp[jj] = 0;

      if (RAWfiletypes) zfree(RAWfiletypes);
      RAWfiletypes = pp;

      save_params();                                                       //  done, save modified parameters
      return 0;
   }

   return 0;
}


/**************************************************************************/

//  edit keyboard shortcuts

namespace KBshortcutnames
{
   zdialog     *zd;

   char        *shortcutkey2[maxshortcuts];                                //  KB shortcut keys (or combinations)
   char        *shortcutmenu2[maxshortcuts];                               //  corresponding menu names
   int         Nshortcuts2;                                                //  count of entries

   int         Nreserved = 10;                                             //  reserved shortcuts (hard coded)
   cchar       *reserved[10] = { "F", "G", "W", "F1", "+", "=", "-",       //  use upper case for this table
                                 "Z", "S", "M"  };                         //  (upper case X is Shift+X) 
   int         Neligible = 25;
   cchar       *eligible[25];                                              //  allowed shortcuts
   cchar       *eligible_en[25] = {
               "Open Image File", "Open Previous File", "Rename Image File",
               "Save to Disk", "Trash Image File", "Quit Fotoxx",
               "Show RGB", "Grid Lines", "Magnify Image",
               "Edit Metadata", "View Metadata (short)", "Edit Geotags",
               "Select", "Trim/Rotate", "Resize", "Add Text",
               "Voodoo Enhance", "Retouch Combo", "Tone Mapping", "Brightness Dist.",
               "Sharpen", "Red Eyes", "Undo", "Redo", "Keyboard Shortcuts" };
}


//  KB shortcuts menu function

void m_KBshortcuts(GtkWidget *, cchar *)
{
   using namespace KBshortcutnames;

   void KBshorts_clickfunc(GtkWidget *widget, int line, int pos);
   int  KBshorts_keyfunc(GtkWidget *dialog, GdkEventKey *event);
   int  KBshorts_dialog_event(zdialog *zd, cchar *event);

   int         ii, cc;
   GtkWidget   *widget;
   char        textline[100];

/***
          _______________________________________________
         |              Edit KB Shortcuts                |
         |_______________________________________________|
         | Alt+G                  Grid Lines             |
         | Alt+U                  Undo                   |
         | Alt+R                  Redo                   |
         | T                      Trim/Rotate            |
         | ...                    ...                    |
         |_______________________________________________|
         |                                               |
         | shortcut key: [ CTRL+G ] [ (dropdown list)  ] |
         |                                               |
         |           [reserved] [delete] [done] [cancel] |
         |_______________________________________________|

***/

   F1_help_topic = "KB_shortcuts";

   zd = zdialog_new(ZTX("Edit KB Shortcuts"),Mwin,Breserved,Bdelete,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hblist","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frlist","hblist",0,"space=5|expand");
   zdialog_add_widget(zd,"text","shortlist","frlist",0,"expand");
   zdialog_add_widget(zd,"hbox","hbshort","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labshort","hbshort",ZTX("shortcut key:"),"space=5");
   zdialog_add_widget(zd,"label","shortkey","hbshort",ZTX("(enter key)"));
   zdialog_add_widget(zd,"label","space","hbshort",0,"expand");
   zdialog_add_widget(zd,"comboE","shortmenu","hbshort",0,"space=5");

   widget = zdialog_widget(zd,"shortlist");
   textwidget_set_clickfunc(widget,KBshorts_clickfunc);

   cc = maxshortcuts * sizeof(char *);
   memset(shortcutkey2,0,cc);
   memset(shortcutmenu2,0,cc);

   for (ii = 0; ii < Neligible; ii++)                                      //  use translated menu names
      eligible[ii] = ZTX(eligible_en[ii]);

   for (ii = 0; ii < Nshortcuts; ii++) {                                   //  copy global shortcuts list
      shortcutkey2[ii] = zstrdup(shortcutkey[ii]);                         //    for subsequent editing
      shortcutmenu2[ii] = zstrdup(shortcutmenu[ii]);
   }
   Nshortcuts2 = Nshortcuts;

   wclear(widget);

   for (int ii = 0; ii < Nshortcuts2; ii++) {                              //  show shortcuts list in dialog
      snprintf(textline,100,"%-20s %s \n",shortcutkey2[ii],shortcutmenu2[ii]);
      wprintx(widget,0,textline,"monospace 9");
   }

   for (int ii = 0; ii < Neligible; ii++)                                  //  add eligible menu names to
      zdialog_cb_app(zd,"shortmenu",eligible[ii]);                         //    combobox dropdown list

   widget = zdialog_widget(zd,"dialog");
   G_SIGNAL(widget,"key-press-event",KBshorts_keyfunc,0);
   G_SIGNAL(widget,"key-release-event",KBshorts_keyfunc,0);

   zdialog_resize(zd,420,0);
   zdialog_run(zd,KBshorts_dialog_event,"save");
   return;
}


//  mouse click functions to select existing shortcut from list

void KBshorts_clickfunc(GtkWidget *widget, int line, int pos)
{
   using namespace KBshortcutnames;

   char     *txline;
   char     shortkey[20];
   char     shortmenu[60];

   txline = textwidget_get_line(widget,line,1);
   if (! txline) return;

   strncpy0(shortkey,txline,20);
   strncpy0(shortmenu,txline+21,60);
   zfree(txline);

   strTrim(shortkey);
   strTrim(shortmenu);

   zdialog_stuff(zd,"shortkey",shortkey);
   zdialog_stuff(zd,"shortmenu",shortmenu);

   return;
}


//  intercept KB key events

int KBshorts_keyfunc(GtkWidget *dialog, GdkEventKey *event)
{
   using namespace KBshortcutnames;

   static int  Ctrl = 0, Alt = 0, Shift = 0;
   int         key, type, ii, cc;
   char        keyname[20];

   key = event->keyval;
   type = event->type;

   if (type == GDK_KEY_PRESS) type = 1;
   else type = 0;

   if (key == GDK_KEY_Control_L || key == GDK_KEY_Control_R) {             //  Ctrl key
       Ctrl = type;
       return 1;
   }

   if (key == GDK_KEY_Alt_L || key == GDK_KEY_Alt_R) {                     //  Alt key
       Alt = type;
       return 1;
   }

   if (key == GDK_KEY_Shift_L || key == GDK_KEY_Shift_R) {                 //  Shift key
       Shift = type;
       return 1;
   }

   if (type) return 1;                                                     //  wait for key release

   if (key == GDK_KEY_F1) {                                                //  key is F1 (context help)
      KBstate(event,0);                                                    //  send to main app
      return 1;
   }

   if (key >= GDK_KEY_F2 && key <= GDK_KEY_F9) {                           //  key is F2 to F9
      ii = key - GDK_KEY_F1;
      strcpy(keyname,"F1");
      keyname[1] += ii;
   }

   else if (key > 255) return 1;                                           //  not a simple Ascii key

   else {
      *keyname = 0;                                                        //  build input key combination
      if (Ctrl) strcat(keyname,"Ctrl+");                                   //  [Ctrl+] [Alt+] [Shift+] key
      if (Alt) strcat(keyname,"Alt+");
      if (Shift) strcat(keyname,"Shift+");
      cc = strlen(keyname);
      keyname[cc] = toupper(key);                                          //  x --> X, Ctrl+x --> Ctrl+X
      keyname[cc+1] = 0;
   }

   for (ii = 0; ii < Nreserved; ii++)
      if (strmatch(keyname,reserved[ii])) break;
   if (ii < Nreserved) {
      zmessageACK(Mwin,ZTX("\"%s\"  Reserved, cannot be used"),keyname);
      Ctrl = Alt = 0;
      return 1;
   }

   zdialog_stuff(zd,"shortkey",keyname);                                   //  stuff key name into dialog
   zdialog_stuff(zd,"shortmenu","");                                       //  clear menu choice

   return 1;
}


//  dialog event and completion function

int KBshorts_dialog_event(zdialog *zd, cchar *event)
{
   using namespace KBshortcutnames;

   int         ii, jj, err;
   GtkWidget   *widget;
   char        textline[100];
   char        shortkey[20];
   char        shortmenu[60];
   char        kbfile[200];
   FILE        *fid = 0;

   if (strmatch(event,"escape")) zd->zstat = 4;                            //  escape = cancel                    v.15.07

   if (strmatch(event,"shortmenu"))                                        //  menu choice made
   {
      zd->zstat = 0;                                                       //  keep dialog active

      zdialog_fetch(zd,"shortkey",shortkey,20);                            //  get shortcut key and menu
      zdialog_fetch(zd,"shortmenu",shortmenu,60);                          //    from dialog widgets
      if (*shortkey <= ' ' || *shortmenu <= ' ') return 0;

      for (ii = 0; ii < Nshortcuts2; ii++)                                 //  find matching shortcut key in list
         if (strmatch(shortcutkey2[ii],shortkey)) break;

      if (ii < Nshortcuts2) {                                              //  if found, remove from list
         zfree(shortcutkey2[ii]);
         zfree(shortcutmenu2[ii]);
         for (jj = ii; jj < Nshortcuts2; jj++) {
            shortcutkey2[jj] = shortcutkey2[jj+1];
            shortcutmenu2[jj] = shortcutmenu2[jj+1];
         }
         --Nshortcuts2;
      }

      for (ii = 0; ii < Nshortcuts2; ii++)                                 //  find matching shortcut menu in list
         if (strmatch(shortcutmenu2[ii],shortmenu)) break;

      if (ii < Nshortcuts2) {                                              //  if found, remove from list
         zfree(shortcutkey2[ii]);
         zfree(shortcutmenu2[ii]);
         for (jj = ii; jj < Nshortcuts2; jj++) {
            shortcutkey2[jj] = shortcutkey2[jj+1];
            shortcutmenu2[jj] = shortcutmenu2[jj+1];
         }
         --Nshortcuts2;
      }

      if (Nshortcuts2 == maxshortcuts) {
         zmessageACK(Mwin,"exceed %d shortcuts",maxshortcuts);
         return 1;
      }

      ii = Nshortcuts2++;                                                  //  add new shortcut to end of list
      shortcutkey2[ii] = zstrdup(shortkey);
      shortcutmenu2[ii] = zstrdup(shortmenu);

      widget = zdialog_widget(zd,"shortlist");
      wclear(widget);

      for (ii = 0; ii < Nshortcuts2; ii++) {                               //  show existing shortcuts in dialog
         snprintf(textline,100,"%-20s %s \n",shortcutkey2[ii],shortcutmenu2[ii]);
         wprintx(widget,0,textline,"monospace 9");
      }

      return 1;
   }

   if (! zd->zstat) return 1;                                              //  wait for completion
   
   if (zd->zstat == 1)                                                     //  show reserved shortcuts
   {
      zd->zstat = 0;                                                       //  keep dialog active
      
      write_popup_text("open","Reserved Shortcuts",300,200,Mwin);
      write_popup_text("write","F   image File view mode");
      write_popup_text("write","G   Gallery view mode");
      write_popup_text("write","W   World maps view mode");
      write_popup_text("write","F1  User Guide topic for current function");
      write_popup_text("write","+   zoom-in");
      write_popup_text("write","=   zoom-in");
      write_popup_text("write","-   zoom-out");
      write_popup_text("write","Z   toggle image full size / fit window");
      write_popup_text("write","S   sync gallery to current image file");
      write_popup_text("write","M   magnify image at mouse drag");
      write_popup_text("write","");
      write_popup_text("write","These may only be used for custom shortcuts");
      write_popup_text("write","in combination with Ctrl/Alt/Shift keys.");
      return 1;
   }

   if (zd->zstat == 2)                                                     //  delete shortcut
   {
      zd->zstat = 0;                                                       //  keep dialog active

      zdialog_fetch(zd,"shortkey",shortkey,20);                            //  get shortcut key
      if (*shortkey <= ' ') return 0;

      for (ii = 0; ii < Nshortcuts2; ii++)                                 //  find matching shortcut key in list
         if (strmatch(shortcutkey2[ii],shortkey)) break;

      if (ii < Nshortcuts2) {                                              //  if found, remove from list
         zfree(shortcutkey2[ii]);
         zfree(shortcutmenu2[ii]);
         for (jj = ii; jj < Nshortcuts2; jj++) {
            shortcutkey2[jj] = shortcutkey2[jj+1];
            shortcutmenu2[jj] = shortcutmenu2[jj+1];
         }
         --Nshortcuts2;
      }

      widget = zdialog_widget(zd,"shortlist");
      wclear(widget);

      for (ii = 0; ii < Nshortcuts2; ii++) {                               //  show existing shortcuts in dialog
         snprintf(textline,100,"%-20s %s \n",shortcutkey2[ii],shortcutmenu2[ii]);
         wprintx(widget,0,textline,"monospace 9");
      }

      zdialog_stuff(zd,"shortkey","");
      zdialog_stuff(zd,"shortmenu","");
      return 1;
   }

   if (zd->zstat == 3)                                                     //  done - save the shortcuts
   {
      zdialog_free(zd);

      for (ii = 0; ii < Nshortcuts2; ii++) {                               //  convert menunames back to English
         for (jj = 0; jj < Neligible; jj++)
            if (strmatch(eligible[jj],shortcutmenu2[ii])) break;
         if (jj < Neligible) {
            zfree(shortcutmenu2[ii]);
            shortcutmenu2[ii] = zstrdup(eligible_en[jj]);
         }
      }

      err = locale_filespec("user","KB-shortcuts",kbfile);                 //  write shortcuts file
      if (! err) fid = fopen(kbfile,"w");
      if (err || ! fid) {
         zmessageACK(Mwin,ZTX("unable to save KB-shortcuts file"));
         return 1;
      }
      for (ii = 0; ii < Nshortcuts2; ii++)
         fprintf(fid,"%-20s %s \n",shortcutkey2[ii],shortcutmenu2[ii]);
      fclose(fid);

      for (ii = 0; ii < Nshortcuts; ii++) {                                //  clear global shortcuts data
         if (shortcutkey[ii]) zfree(shortcutkey[ii]);
         if (shortcutmenu[ii]) zfree(shortcutmenu[ii]);
         shortcutkey[ii] = shortcutmenu[ii] = 0;
      }

      KBshortcuts_load();                                                  //  reload and activate
      return 1;
   }

   zdialog_free(zd);                                                       //  cancel or [x]
   return 1;
}


//  read KB-shortcuts file and load shortcuts table in memory
//  also called from initzfunc() at fotoxx startup

int KBshortcuts_load()
{
   int         ii, err;
   char        kbfile[200], buff[200];
   char        *pp1, *pp2;
   FILE        *fid;

   Nshortcuts = 0;
   err = locale_filespec("user","KB-shortcuts",kbfile);
   if (err) return 0;
   ii = 0;
   fid = fopen(kbfile,"r");
   if (! fid) return 0;
   while (true) {
      pp1 = fgets_trim(buff,200,fid,1);                                    //  next record
      if (! pp1) break;
      if (*pp1 == '#') continue;                                           //  comment
      if (*pp1 <= ' ') continue;                                           //  blank
      pp2 = strchr(pp1,' ');
      if (! pp2) continue;
      if (pp2 - pp1 > 20) continue;
      *pp2 = 0;
      shortcutkey[ii] = zstrdup(pp1);                                      //  shortcut key
      pp1 = pp2 + 1;
      while (*pp1 && *pp1 == ' ') pp1++;
      if (! *pp1) continue;
      shortcutmenu[ii] = zstrdup(ZTX(pp1));                                //  menu can be English or translation
      if (++ii == maxshortcuts) break;
   }

   fclose(fid);
   Nshortcuts = ii;
   return 0;
}


/**************************************************************************/

//  create or update brightness distribution graph

namespace brdist_names
{
   GtkWidget   *drawwin_dist, *drawwin_scale;                              //  brightness distribution graph widgets
   int         RGBW[4] = { 0, 0, 0, 1 };                                   //     "  colors: red/green/blue/white (all)
}


//  menu function

void m_show_brdist(GtkWidget *, cchar *menu)                               //  menu function
{
   using namespace brdist_names;

   int  show_brdist_dialog_event(zdialog *zd, cchar *event);

   GtkWidget   *frdist, *frscale;
   zdialog     *zd;

   if (! Fpxb) return;

   if (menu && strmatch(menu,"kill")) {
      if (zdbrdist) zdialog_destroy(zdbrdist);
      zdbrdist = 0;
      return;
   }

   if (zdbrdist) {                                                         //  dialog already present
      gtk_widget_queue_draw(drawwin_dist);                                 //  refresh drawing windows
      return;
   }

   if (menu) F1_help_topic = "show_brdist";

   zd = zdialog_new(ZTX("Brightness Distribution"),Mwin,null);
   zdialog_add_widget(zd,"frame","frdist","dialog",0,"expand");            //  frames for 2 drawing areas
   zdialog_add_widget(zd,"frame","frscale","dialog");
   frdist = zdialog_widget(zd,"frdist");
   frscale = zdialog_widget(zd,"frscale");

   drawwin_dist = gtk_drawing_area_new();                                  //  histogram drawing area
   gtk_container_add(GTK_CONTAINER(frdist),drawwin_dist);
   G_SIGNAL(drawwin_dist,"draw",brdist_drawgraph,RGBW);

   drawwin_scale = gtk_drawing_area_new();                                 //  brightness scale under histogram
   gtk_container_add(GTK_CONTAINER(frscale),drawwin_scale);
   gtk_widget_set_size_request(drawwin_scale,300,12);
   G_SIGNAL(drawwin_scale,"draw",brdist_drawscale,0);

   zdialog_add_widget(zd,"hbox","hbcolors","dialog");
   zdialog_add_widget(zd,"check","all","hbcolors",Ball,"space=5");
   zdialog_add_widget(zd,"check","red","hbcolors",Bred,"space=5");
   zdialog_add_widget(zd,"check","green","hbcolors",Bgreen,"space=5");
   zdialog_add_widget(zd,"check","blue","hbcolors",Bblue,"space=5");

   zdialog_stuff(zd,"red",RGBW[0]);
   zdialog_stuff(zd,"green",RGBW[1]);
   zdialog_stuff(zd,"blue",RGBW[2]);
   zdialog_stuff(zd,"all",RGBW[3]);

   zdialog_resize(zd,300,250);
   zdialog_run(zd,show_brdist_dialog_event,"save");

   zdbrdist = zd;
   return;
}


//  dialog event and completion function

int show_brdist_dialog_event(zdialog *zd, cchar *event)
{
   using namespace brdist_names;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (zd->zstat) {
      zdialog_free(zd);
      zdbrdist = 0;
      return 0;
   }

   if (strstr("all red green blue",event)) {                               //  update chosen colors
      zdialog_fetch(zd,"red",RGBW[0]);
      zdialog_fetch(zd,"green",RGBW[1]);
      zdialog_fetch(zd,"blue",RGBW[2]);
      zdialog_fetch(zd,"all",RGBW[3]);
      gtk_widget_queue_draw(drawwin_dist);                                 //  refresh drawing window
   }

   return 0;
}


//  draw the histogram and the brightness scale underneath

void brdist_drawgraph(GtkWidget *drawin, cairo_t *cr, int *rgbw)           //  v.14.04
{
   int         bin, Nbins = 85, brdist[85][4];                             //  bin count, R/G/B/all
   int         px, py, dx, dy;
   int         ww, hh, winww, winhh;
   int         ii, rgb, maxdist, bright;
   float       *pixf, pixg[3];
   uint8       *pixi;

   if (! Fpxb) return;
   if (rgbw[0]+rgbw[1]+rgbw[2]+rgbw[3] == 0) return;

   if (! resource_lock(Fpaintlock)) return;                                //  15.02
   gdk_window_freeze_updates(gdkwin);

   winww = gtk_widget_get_allocated_width(drawin);                         //  drawing window size
   winhh = gtk_widget_get_allocated_height(drawin);

   for (bin = 0; bin < Nbins; bin++)                                       //  clear brightness distribution
   for (rgb = 0; rgb < 4; rgb++)
      brdist[bin][rgb] = 0;

   ww = Fpxb->ww;
   hh = Fpxb->hh;

   for (ii = 0; ii < ww * hh; ii++)                                        //  overhaul
   {
      if (sa_stat == 3 && ! sa_pixmap[ii]) continue;                       //  stay within active select area

      py = ii / ww;                                                        //  image pixel
      px = ii - ww * py;

      dx = Mscale * px - Morgx + Dorgx;                                    //  stay within visible window
      if (dx < 0 || dx > Dww-1) continue;                                  //    for zoomed image
      dy = Mscale * py - Morgy + Dorgy;
      if (dy < 0 || dy > Dhh-1) continue;

      pixi = PXBpix(Fpxb,px,py);                                           //  use displayed image
      pixg[0] = pixi[0];                                                   //  convert uint8 RGB to float
      pixg[1] = pixi[1];
      pixg[2] = pixi[2];
      pixf = pixg;

      for (rgb = 0; rgb < 3; rgb++) {                                      //  get R/G/B brightness levels
         bright = pixf[rgb] * Nbins / 256.0;                               //  scale 0 to Nbins-1
         if (bright < 0 || bright > 255) {
            printz("pixel %d/%d: %.2f %.2f %.2f \n",px,py,pixf[0],pixf[1],pixf[2]);
            resource_unlock(Fpaintlock);
            gdk_window_thaw_updates(gdkwin);
            return;
         }
         ++brdist[bright][rgb];
      }

      bright = (pixf[0] + pixf[1] + pixf[2]) * 0.333 * Nbins / 256.0;      //  R+G+B, 0 to Nbins-1
      ++brdist[bright][3];
   }

   resource_unlock(Fpaintlock);
   gdk_window_thaw_updates(gdkwin);

   maxdist = 0;

   for (bin = 1; bin < Nbins-1; bin++)                                     //  find max. bin over all RGB
   for (rgb = 0; rgb < 3; rgb++)                                           //    omit bins 0 and last
      if (brdist[bin][rgb] > maxdist)                                      //      which can be huge
         maxdist = brdist[bin][rgb];

   for (rgb = 0; rgb < 4; rgb++)                                           //  R/G/B/white (all)
   {
      if (! rgbw[rgb]) continue;                                           //  color not selected for graph
      if (rgb == 0) cairo_set_source_rgb(cr,1,0,0);
      if (rgb == 1) cairo_set_source_rgb(cr,0,1,0);
      if (rgb == 2) cairo_set_source_rgb(cr,0,0,1);
      if (rgb == 3) cairo_set_source_rgb(cr,0,0,0);                        //  color "white" = R+G+B uses black line

      cairo_move_to(cr,0,winhh-1);                                         //  start at (0,0)

      for (px = 0; px < winww; px++)                                       //  x from 0 to window width
      {
         bin = Nbins * px / winww;                                         //  bin = 0-Nbins for x = 0-width
         py = 0.9 * winhh * brdist[bin][rgb] / maxdist;                    //  height of bin in window
         py = winhh * sqrt(1.0 * py / winhh);
         py = winhh - py - 1;
         cairo_line_to(cr,px,py);                                          //  draw line from bin to bin
      }

      cairo_stroke(cr);
   }

   return;
}


/**************************************************************************/

//  Paint a horizontal stripe drawing area with a color progressing from
//  black to white. This represents a brightness scale from 0 to 255.

void brdist_drawscale(GtkWidget *drawarea, cairo_t *cr, int *)
{
   int      px, ww, hh;
   float    fbright;

   ww = gtk_widget_get_allocated_width(drawarea);                          //  drawing area size
   hh = gtk_widget_get_allocated_height(drawarea);

   for (px = 0; px < ww; px++)                                             //  draw brightness scale
   {
      fbright = 1.0 * px / ww;
      cairo_set_source_rgb(cr,fbright,fbright,fbright);
      cairo_move_to(cr,px,0);
      cairo_line_to(cr,px,hh-1);
      cairo_stroke(cr);
   }

   return;
}


/**************************************************************************/

//  setup x and y grid lines - count/spacing, enable/disable, offsets

void m_gridlines(GtkWidget *widget, cchar *menu)
{
   int gridlines_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         G;

   if (menu && strmatchN(menu,"grid ",5))                                  //  grid N from some edit functions
      currgrid = menu[5] - '0';

   else if (! widget) {                                                    //  from KB shortcut
      toggle_grid(2);                                                      //  toggle grid lines on/off
      return;
   }

   else currgrid = 0;                                                      //  must be menu call

   if (currgrid == 0)
      F1_help_topic = "grid_lines";

   zd = zdialog_new(ZTX("Grid Lines"),Mwin,Bdone,null);

   zdialog_add_widget(zd,"hbox","hb0","dialog",0,"space=10");
   zdialog_add_widget(zd,"vbox","vb1","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb0",0,"homog");
   zdialog_add_widget(zd,"vbox","vbspace","hb0",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb3","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb4","hb0",0,"homog");

   zdialog_add_widget(zd,"label","lab1x","vb1",ZTX("x-spacing"));
   zdialog_add_widget(zd,"label","lab2x","vb1",ZTX("x-count"));
   zdialog_add_widget(zd,"label","lab4x","vb1",ZTX("x-enable"));

   zdialog_add_widget(zd,"spin","spacex","vb2","10|200|1|50","space=2");
   zdialog_add_widget(zd,"spin","countx","vb2","0|100|1|2","space=2");
   zdialog_add_widget(zd,"check","enablex","vb2",0);

   zdialog_add_widget(zd,"label","lab1y","vb3",ZTX("y-spacing"));
   zdialog_add_widget(zd,"label","lab2y","vb3",ZTX("y-count"));
   zdialog_add_widget(zd,"label","lab4y","vb3",ZTX("y-enable"));

   zdialog_add_widget(zd,"spin","spacey","vb4","10|200|1|50");
   zdialog_add_widget(zd,"spin","county","vb4","0|100|1|2");
   zdialog_add_widget(zd,"check","enabley","vb4",0);

   zdialog_add_widget(zd,"hbox","hboffx","dialog");
   zdialog_add_widget(zd,"label","lab3x","hboffx",Bxoffset,"space=7");
   zdialog_add_widget(zd,"hscale","offsetx","hboffx","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","space","hboffx",0,"space=20");

   zdialog_add_widget(zd,"hbox","hboffy","dialog");
   zdialog_add_widget(zd,"label","lab3y","hboffy",Byoffset,"space=7");
   zdialog_add_widget(zd,"hscale","offsety","hboffy","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","space","hboffy",0,"space=20");

   G = currgrid;                                                           //  grid logic overhaul

   if (! gridsettings[G][GXS])                                             //  if [G] never set, get defaults
      for (int ii = 0; ii < 9; ii++)
         gridsettings[G][ii] = gridsettings[0][ii];

   zdialog_stuff(zd,"enablex",gridsettings[G][GX]);                        //  current settings >> dialog widgets
   zdialog_stuff(zd,"enabley",gridsettings[G][GY]);
   zdialog_stuff(zd,"spacex",gridsettings[G][GXS]);
   zdialog_stuff(zd,"spacey",gridsettings[G][GYS]);
   zdialog_stuff(zd,"countx",gridsettings[G][GXC]);
   zdialog_stuff(zd,"county",gridsettings[G][GYC]);
   zdialog_stuff(zd,"offsetx",gridsettings[G][GXF]);
   zdialog_stuff(zd,"offsety",gridsettings[G][GYF]);

   zdialog_run(zd,gridlines_dialog_event);
   zdialog_wait(zd);
   zdialog_free(zd);
   return;
}


//  dialog event function

int gridlines_dialog_event(zdialog *zd, cchar *event)
{
   int      G = currgrid;

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [done]  v.14.03
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (zd->zstat)                                                          //  done or cancel
   {
      if (zd->zstat != 1) return 1;
      if (gridsettings[G][GX] || gridsettings[G][GY])
         gridsettings[G][GON] = 1;
      else gridsettings[G][GON] = 0;
      Fpaint2();
      return 1;
   }

   if (strmatch(event,"enablex"))                                          //  x/y grid enable or disable
      zdialog_fetch(zd,"enablex",gridsettings[G][GX]);

   if (strmatch(event,"enabley"))
      zdialog_fetch(zd,"enabley",gridsettings[G][GY]);

   if (strmatch(event,"spacex"))                                           //  x/y grid spacing (if counts == 0)
      zdialog_fetch(zd,"spacex",gridsettings[G][GXS]);

   if (strmatch(event,"spacey"))
      zdialog_fetch(zd,"spacey",gridsettings[G][GYS]);

   if (strmatch(event,"countx"))                                           //  x/y grid line counts
      zdialog_fetch(zd,"countx",gridsettings[G][GXC]);

   if (strmatch(event,"county"))
      zdialog_fetch(zd,"county",gridsettings[G][GYC]);

   if (strmatch(event,"offsetx"))                                          //  x/y grid starting offsets
      zdialog_fetch(zd,"offsetx",gridsettings[G][GXF]);

   if (strmatch(event,"offsety"))
      zdialog_fetch(zd,"offsety",gridsettings[G][GYF]);

   if (gridsettings[G][GX] || gridsettings[G][GY])                         //  if either grid enabled, show grid
      gridsettings[G][GON] = 1;

   Fpaint2();
   return 1;
}


//  toggle grid lines on or off
//  action: 0 = off, 1 = on, 2 = toggle: on > off, off > on

void toggle_grid(int action)
{
   int   G = currgrid;

   if (action == 0) gridsettings[G][GON] = 0;                              //  grid off
   if (action == 1) gridsettings[G][GON] = 1;                              //  grid on
   if (action == 2) gridsettings[G][GON] = 1 - gridsettings[G][GON];       //  toggle grid

   if (gridsettings[G][GON])
      if (! gridsettings[G][GX] && ! gridsettings[G][GY])                  //  if grid on and x/y both off,
         gridsettings[G][GX] = gridsettings[G][GY] = 1;                    //    set both grids on

   Fpaint2();
   return;
}


/**************************************************************************/

//  choose color for foreground lines
//  (area outline, mouse circle, trim rectangle ...)

void  m_fg_color(GtkWidget *, cchar *menu)
{
   int fg_color_dialog_event(zdialog *zd, cchar *event);

   zdialog  *zd;

   if (strmatch(menu,ZTX("Line Color")))
      F1_help_topic = "line_color";
   if (strmatch(menu,ZTX("Area Color")))
      F1_help_topic = "area_color";

   zd = zdialog_new(ZTX("Line Color"),Mwin,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio",Bblack,"hb1",Bblack,"space=3");           //  add radio button per color
   zdialog_add_widget(zd,"radio",Bwhite,"hb1",Bwhite,"space=3");
   zdialog_add_widget(zd,"radio",Bred,"hb1",Bred,"space=3");
   zdialog_add_widget(zd,"radio",Bgreen,"hb1",Bgreen,"space=3");

   zdialog_stuff(zd,Bblack,0);                                             //  all are initially off
   zdialog_stuff(zd,Bwhite,0);
   zdialog_stuff(zd,Bred,0);
   zdialog_stuff(zd,Bgreen,0);

   if (FG_COLOR == BLACK)                                                  //  set current color on
      zdialog_stuff(zd,Bblack,1);
   if (FG_COLOR == WHITE)
      zdialog_stuff(zd,Bwhite,1);
   if (FG_COLOR == RED)
      zdialog_stuff(zd,Bred,1);
   if (FG_COLOR == GREEN)
      zdialog_stuff(zd,Bgreen,1);

   zdialog_run(zd,fg_color_dialog_event,"save");                           //  run dialog, parallel
   return;
}


//  dialog event and completion function

int fg_color_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (strstr("Black White Red Green",event))
   {
      if (strmatch(event,"Black")) FG_COLOR = BLACK;                       //  set selected color
      if (strmatch(event,"White")) FG_COLOR = WHITE;
      if (strmatch(event,"Red"))   FG_COLOR = RED;
      if (strmatch(event,"Green")) FG_COLOR = GREEN;
      if (CEF && CEF->zd) zdialog_send_event(CEF->zd,"fg_color");
      Fpaint2();
   }

   if (zd->zstat) zdialog_free(zd);                                        // [x] button
   return 1;
}


/**************************************************************************/

//  Show RGB values for 1-10 pixels selected with mouse-clicks.

void  show_RGB_mousefunc();
int   show_RGB_timefunc(void *);

zdialog     *RGBSzd;
int         RGBSpixel[10][2];                                              //  last 10 pixels clicked
int         RGBSnpix;                                                      //  count of pixels
int         RGBSmetric = 1;                                                //  1/2/3 = /RGB/EV/OD
int         RGBSdelta = 0;                                                 //  abs/delta mode
int         RGBSlabels = 0;                                                //  pixel labels on/off
int         RGBSupdate = 0;                                                //  window update needed


void m_show_RGB(GtkWidget *, cchar *menu)
{
   int   show_RGB_event(zdialog *zd, cchar *event);

   PangoFontDescription    *pfontdesc;
   GtkWidget   *widget;
   cchar       *mess = ZTX("Click image to select pixels.");
   cchar       *header = " Pixel            Red     Green   Blue";
   char        hbx[8] = "hbx", pixx[8] = "pixx";                           //  last char. is '0' to '9'
   int         ii;

   F1_help_topic = "show_RGB";

   if (! curr_file) return;                                                //  no image file
   RGBSnpix = 0;                                                           //  no pixels yet
   pfontdesc = pango_font_description_from_string("Monospace 9");          //  monospace font

/***
    _________________________________________
   |                                         |
   |  Click image to select pixels.          |
   |  [x] delta  [x] labels                  |
   |   Metric: (o) RGB  (o) EV               |
   |   Pixel           Red    Green  Blue    |
   |   A xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   B xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   C xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   D xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   E xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   F xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   G xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   H xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   I xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |   J xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |     xxxxx xxxxx   xxxxx  xxxxx  xxxxx   |
   |                                         |
   |                        [clear] [done]   |
   |_________________________________________|

***/

   if (RGBSzd) zdialog_free(RGBSzd);                                       //  delete previous if any
   zdialog *zd = zdialog_new(ZTX("Show RGB"),Mwin,Bclear,Bdone,null);
   RGBSzd = zd;

   zdialog_add_widget(zd,"hbox","hbmess","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmess","hbmess",mess,"space=5");

   zdialog_add_widget(zd,"hbox","hbmym","dialog");
   zdialog_add_widget(zd,"check","delta","hbmym","delta","space=8");
   zdialog_add_widget(zd,"check","labels","hbmym","labels","space=8");

   if (RGBSdelta && E3pxm) zdialog_stuff(zd,"delta",1);

   zdialog_add_widget(zd,"hbox","hbmetr","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmetr","hbmetr","Metric:","space=5");
   zdialog_add_widget(zd,"radio","radRGB","hbmetr","RGB","space=3");
   zdialog_add_widget(zd,"radio","radEV","hbmetr","EV","space=3");

   if (RGBSmetric == 1) zdialog_stuff(zd,"radRGB",1);                      //  get which metric to use
   if (RGBSmetric == 2) zdialog_stuff(zd,"radEV",1);

   zdialog_add_widget(zd,"vbox","vbdat","dialog");                         //  vbox for current pixel values
   zdialog_add_widget(zd,"hbox","hbpix","vbdat");
   zdialog_add_widget(zd,"label","labheader","hbpix",header);              //  Pixel        Red    Green  Blue
   widget = zdialog_widget(zd,"labheader");
   gtk_widget_override_font(widget,pfontdesc);

   for (ii = 0; ii < 10; ii++)
   {                                                                       //  10 labels for pixel data output
      hbx[2] = '0' + ii;
      pixx[3] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbx,"vbdat");
      zdialog_add_widget(zd,"label",pixx,hbx);
      widget = zdialog_widget(zd,pixx);                                    //  use monospace font
      gtk_widget_override_font(widget,pfontdesc);
   }

   zdialog_run(zd,show_RGB_event,"save");                                  //  run dialog
   takeMouse(show_RGB_mousefunc,dragcursor);                               //  connect mouse function
   g_timeout_add(200,show_RGB_timefunc,0);                                 //  start timer function, 200 ms

   return;
}


//  dialog event function

int show_RGB_event(zdialog *zd, cchar *event)
{
   int      button;

   if (strmatch(event,"enter")) zd->zstat = 2;                             //  [done]
   if (strmatch(event,"escape")) zd->zstat = 3;                            //  escape = cancel                    v.15.07

   if (zd->zstat) {
      if (zd->zstat == 1) {                                                //  clear
         zd->zstat = 0;                                                    //  keep dialog active
         RGBSnpix = 0;                                                     //  point count = 0
         erase_toptext(102);                                               //  erase labels on image
         RGBSupdate++;                                                     //  update display window
      }
      else {                                                               //  done or kill
         freeMouse();                                                      //  disconnect mouse function
         zdialog_free(RGBSzd);                                             //  kill dialog
         RGBSzd = 0;
         erase_toptext(102);
      }
      Fpaint2();
   }

   if (strmatch(event,"focus"))                                            //  toggle mouse capture
      takeMouse(show_RGB_mousefunc,dragcursor);                            //  connect mouse function

   if (strmatch(event,"delta")) {                                          //  set absolute/delta mode
      zdialog_fetch(zd,"delta",RGBSdelta);
      if (RGBSdelta && ! E3pxm) {
         RGBSdelta = 0;                                                    //  block delta mode if no edit underway
         zdialog_stuff(zd,"delta",0);
         zmessageACK(Mwin,ZTX("Edit function must be active"));            //  13.02.1
      }
   }

   if (strmatch(event,"labels")) {                                         //  get labels on/off
      zdialog_fetch(zd,"labels",RGBSlabels);
      RGBSupdate++;                                                        //  update window
   }

   if (strmatch(event,"radRGB")) {                                         //  metric = RGB
      zdialog_fetch(zd,event,button);
      if (button) RGBSmetric = 1;
   }

   if (strmatch(event,"radEV")) {                                          //  metric = EV
      zdialog_fetch(zd,event,button);
      if (button) RGBSmetric = 2;
   }

   return 0;
}


//  mouse function
//  table positions [0] to [RGBSnpix-1] for labeled pixel positions, fixed
//  table position [RGBSnpix] = values from current mouse position

void show_RGB_mousefunc()                                                  //  mouse function
{
   int      ii;
   PXM      *pxm;

   if (E3pxm) pxm = E3pxm;                                                 //  report image being edited
   else if (E1pxm) pxm = E1pxm;
   else if (E0pxm) pxm = E0pxm;
   else {
      if (! resource_lock(Fpaintlock)) return;                             //  v.15.02
      E0pxm = PXM_load(curr_file,1);                                       //  never edited,
      resource_unlock(Fpaintlock);
      if (! E0pxm) return;                                                 //  create E0 image for my use
      pxm = E0pxm;
   }

   if (Mxposn < 0 || Mxposn > pxm->ww-1) return;                           //  mouse outside image, ignore
   if (Myposn < 0 || Myposn > pxm->hh-1) return;

   if (LMclick)                                                            //  left click, add labeled position
   {
      LMclick = 0;

      if (RGBSnpix == 9) {                                                 //  if all 9 labeled positions filled,
         for (ii = 1; ii < 9; ii++) {                                      //    remove first (oldest) and
            RGBSpixel[ii-1][0] = RGBSpixel[ii][0];                         //      push the rest back
            RGBSpixel[ii-1][1] = RGBSpixel[ii][1];
         }
         RGBSnpix = 8;                                                     //  position for newest
      }

      ii = RGBSnpix;                                                       //  next position to fill
      RGBSpixel[ii][0] = Mxclick;                                          //  save newest pixel
      RGBSpixel[ii][1] = Myclick;
      RGBSnpix++;
   }

   ii = RGBSnpix;                                                          //  fill next position from active mouse
   RGBSpixel[ii][0] = Mxposn;                                              //  v.14.01
   RGBSpixel[ii][1] = Myposn;

   RGBSupdate++;                                                           //  update window
   return;
}


//  timer function - continuously display RGB values for selected pixels

int show_RGB_timefunc(void *arg)                                           //  up to 10 pixels, live update
{
   static char    label[10][4] = { " A ", " B ", " C ", " D ", " E ",
                                   " F ", " G ", " H ", " I ", " J " };
   PXM         *pxm = 0;
   int         ii, jj, px, py, delta;
   int         ww, hh;
   float       red1, green1, blue1;
   float       red3, green3, blue3;
   float       *ppixa, *ppixb;
   char        text[100], pixx[8] = "pixx";

   #define EVfunc(rgb) (log2(rgb) - 7)                                     //  RGB units to EV (128 == 0 EV)

   if (! RGBSzd) return 0;                                                 //  user quit, cancel timer            v.14.01

   if (RGBSdelta && E3pxm) delta = 1;                                      //  delta mode only if edit underway
   else delta = 0;

   if (E3pxm) pxm = E3pxm;                                                 //  report image being edited
   else if (E1pxm) pxm = E1pxm;
   else if (E0pxm) pxm = E0pxm;
   else return 1;

   ww = pxm->ww;
   hh = pxm->hh;

   if (RGBSnpix)                                                           //  some labeled positions to report
   {
      for (ii = 0; ii < 10; ii++)
      {
         px = RGBSpixel[ii][0];                                            //  next pixel to report
         py = RGBSpixel[ii][1];
         if (px >= 0 && px < ww && py >= 0 && py < hh) continue;           //  within image limits
         for (jj = ii+1; jj < 10; jj++) {
            RGBSpixel[jj-1][0] = RGBSpixel[jj][0];                         //  remove pixel outside limits
            RGBSpixel[jj-1][1] = RGBSpixel[jj][1];                         //    and pack the remaining down
         }
         RGBSpixel[9][0] = RGBSpixel[9][1] = 0;                            //  last position is empty
         RGBSnpix--;
         ii--;
         RGBSupdate++;                                                     //  update window
      }
   }

   if (RGBSupdate)                                                         //  refresh window labels
   {
      RGBSupdate = 0;
      erase_toptext(102);

      if (RGBSlabels) {
         for (ii = 0; ii < RGBSnpix; ii++) {                               //  show pixel labels on image
            px = RGBSpixel[ii][0];
            py = RGBSpixel[ii][1];
            add_toptext(102,px,py,label[ii],"Sans 8");
         }
      }

      Fpaint2();
   }

   red1 = green1 = blue1 = 0;

   for (ii = 0; ii < 10; ii++)                                             //  loop positons 0 to 9
   {
      pixx[3] = '0' + ii;                                                  //  widget names "pix0" ... "pix9"

      px = RGBSpixel[ii][0];                                               //  next pixel to report
      py = RGBSpixel[ii][1];
      if (ii > RGBSnpix) {                                                 //  no pixel there yet
         zdialog_stuff(RGBSzd,pixx,"");                                    //  blank report line
         continue;
      }

      ppixa = PXMpix(pxm,px,py);                                           //  get pixel RGB values
      red3 = ppixa[0];
      green3 = ppixa[1];
      blue3 = ppixa[2];

      if (delta && E1pxm && px < E1pxm->ww && py < E1pxm->hh) {            //  delta RGB for ongoing edited image
         ppixb = PXMpix(E1pxm,px,py);                                      //  "before" image E1
         red1 = ppixb[0];
         green1 = ppixb[1];
         blue1 = ppixb[2];
      }

      if (ii == RGBSnpix)
         snprintf(text,100,"   %5d %5d  ",px,py);                          //  mouse pixel, format "   xxxx yyyy"
      else snprintf(text,100," %c %5d %5d  ",'A'+ii,px,py);                //  fixed pixel, format " A xxxx yyyy"

      if (RGBSmetric == 1) {                                               //  output RGB values
         if (delta) {
            red3 -= red1;                                                  //  delta RGB
            green3 -= green1;
            blue3 -= blue1;
         }
         snprintf(text+14,86,"   %6.2f  %6.2f  %6.2f ",red3,green3,blue3); //  xxx.xx
      }

      if (RGBSmetric == 2) {                                               //  output EV values
         red3 = EVfunc(red3);
         green3 = EVfunc(green3);
         blue3 = EVfunc(blue3);
         if (delta) {
            red3 -= EVfunc(red1);                                          //  delta EV
            green3 -= EVfunc(green1);
            blue3 -= EVfunc(blue1);
         }
         snprintf(text+14,86,"   %6.3f  %6.3f  %6.3f ",red3,green3,blue3); //  x.xxx
      }

      zdialog_stuff(RGBSzd,pixx,text);                                     //  pixel and RGB values >> dialog
   }

   return 1;
}


/**************************************************************************/

//  magnify image within a given radius of dragged mouse                   //  v.14.12

namespace magnify_names
{
   int   magnify_dialog_event(zdialog* zd, cchar *event);
   void  magnify_mousefunc();
   void  magnify_dopixels(int ftf);

   float       Xsize;                                                      //  magnification, 1 - 10x
   int         Mxpos, Mypos;                                               //  mouse location, image space
   int         Drad;                                                       //  mouse radius, window space
   uint8       kernel[601][601];                                           //  up to radius 300
}


//  menu function

void m_magnify(GtkWidget *, cchar *)
{
   using namespace magnify_names;

   F1_help_topic = "magnify";

   cchar *   mess = ZTX("Drag mouse on image. \n"
                        "Left click to cancel. \n"
                        "Key M to toggle dialog.");

   /***
             __________________________
            |    Magnify Image         |
            |                          |
            |  Drag mouse on image.    |
            |  Left click to cancel.   |
            |  Key M to toggle dialog. |
            |                          |
            |  radius  [_____|-+]      |
            |  X-size  [_____|-+]      |
            |                          |
            |                [cancel]  |
            |__________________________|

   ***/

   if (zdmagnify) {                                                        //  toggle magnify mode                v.15.01
      zdialog_send_event(zdmagnify,"kill");
      return;
   }

   else {
      zdialog *zd = zdialog_new(ZTX("Magnify Image"),Mwin,Bcancel,null);
      zdmagnify = zd;

      zdialog_add_widget(zd,"label","labdrag","dialog",mess,"space=5");

      zdialog_add_widget(zd,"hbox","hbr","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labr","hbr",Bradius,"space=5");
      zdialog_add_widget(zd,"spin","Drad","hbr","100|300|10|100");
      zdialog_add_widget(zd,"hbox","hbx","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labx","hbx",ZTX("X-size"),"space=5");
      zdialog_add_widget(zd,"spin","Xsize","hbx","1.5|5|0.1|2");

      zdialog_fetch(zd,"Drad",Drad);                                       //  initial mouse radius
      zdialog_fetch(zd,"Xsize",Xsize);                                     //  initial magnification

      zdialog_resize(zd,200,0);
      zdialog_restore_inputs(zd);                                          //  preload prior user inputs
      zdialog_run(zd,magnify_dialog_event,"save");                         //  run dialog, parallel

      zdialog_send_event(zd,"Drad");                                       //  initializations
      zdialog_send_event(zd,"Xsize");
   }

   takeMouse(magnify_mousefunc,dragcursor);                                //  connect mouse function

   gtk_window_present(MWIN);                                               //  keep focus on main window          v.15.01
   return;
}


//  dialog event and completion callback function

int magnify_names::magnify_dialog_event(zdialog *zd, cchar *event)
{
   using namespace magnify_names;

   int      dx, dy, rad, kern;

   if (strmatch(event,"escape")) event = "kill";                           //  escape = cancel                    v.15.07

   draw_mousearc(0,0,0,0,1);                                               //  erase mouse ellipse

   if (strmatch(event,"kill")) zd->zstat = 1;                              //  from slide show                    v.14.12

   if (zd->zstat) {                                                        //  terminate
      zdmagnify = 0;
      zdialog_free(zd);
      freeMouse();
      return 1;
   }

   if (strmatch(event,"focus"))                                            //  toggle mouse capture
      takeMouse(magnify_mousefunc,dragcursor);

   if (strmatch(event,"Drad"))
   {
      zdialog_fetch(zd,"Drad",Drad);                                       //  new mouse radius

      for (dy = -Drad; dy <= Drad; dy++)                                   //  build kernel
      for (dx = -Drad; dx <= Drad; dx++)
      {
         rad = sqrt(dx*dx + dy*dy);
         if (rad < Drad-3) kern = 1;                                       //  margin for circle                  v.15.01
         else kern = 0;
         kernel[dx+Drad][dy+Drad] = kern;
      }

      return 1;
   }

   if (strmatch(event,"Xsize"))
   {
      zdialog_fetch(zd,"Xsize",Xsize);                                     //  new magnification
      return 1;
   }

   return 1;
}


//  pixel paint mouse function

void magnify_names::magnify_mousefunc()
{
   using namespace magnify_names;

   static int     ftf;

   if (! curr_file) return;
   if (FGW != 'F') return;

   if (Mxdown) Fpaint2();                                                  //  drag start, erase prior if any
   Mxdown = 0;

   if (Mxdrag || Mydrag)                                                   //  drag in progress
   {
      Mxpos = Mxdrag;                                                      //  save mouse position
      Mypos = Mydrag;
      Mxdrag = Mydrag = 0;
      magnify_dopixels(ftf);                                               //  magnify pixels inside mouse
      gdk_window_set_cursor(gdkwin,blankcursor);
      ftf = 0;
   }

   else
   {
      ftf = 1;
      gdk_window_set_cursor(gdkwin,dragcursor);
   }

   return;
}


//  get pixels from mouse circle within full size image
//  scale and move into pixbuf and write to window

void magnify_names::magnify_dopixels(int ftf)
{
   using namespace magnify_names;

   static PIXBUF   *pxb1 = 0, *pxb2 = 0;

   int         Srad, drad, Dxpos, Dypos, rx, ry;
   static int  org1x, org1y, org2x, org2y;
   static int  fww, fhh, ww1, hh1, ww2, hh2;
   uint8       *pixels1, *pixels2, *pix1, *pix2;
   int         rs1, rs2, px1, py1, px2, py2;
   float       scale;
   cairo_t     *cr;

   if (! ftf && pxb2)                                                      //  continuation of mouse drag
   {
      org1x = org2x - Dorgx + Morgx;                                       //  replace prior display area
      if (org1x < 0) org1x = 0;                                            //    with original image pixels
      if (org1x + ww2 > Mpxb->ww) ww2 = Mpxb->ww - org1x;

      org1y = org2y - Dorgy + Morgy;
      if (org1y < 0) org1y = 0;
      if (org1y + hh2 > Mpxb->hh) hh2 = Mpxb->hh - org1y;

      g_object_unref(pxb2);
      pxb2 = gdk_pixbuf_new_subpixbuf(Mpxb->pixbuf,org1x,org1y,ww2,hh2);
      if (! pxb2) return;
      cr = gdk_cairo_create(gdkwin);
      gdk_cairo_set_source_pixbuf(cr,pxb2,org2x,org2y);
      cairo_paint(cr);
      cairo_destroy(cr);
   }

   fww = Fpxb->ww;                                                         //  curr. image size
   fhh = Fpxb->hh;

   scale = Mscale * Xsize;                                                 //  image to window pixels scale
   Srad = Drad / scale;                                                    //  mouse radius scaled to 1x image

   if (Srad > fww/3) Srad = fww/3;                                         //  restrict to 1/3 image size
   if (Srad > fhh/3) Srad = fhh/3;

   drad = Srad * scale;                                                    //  restricted mouse radius
   drad = drad / 2 * 2 + 1;                                                //  make odd

   if (Mxpos < Srad) Mxpos = Srad;                                         //  restrict within image limits
   if (Mxpos > fww-1-Srad) Mxpos = fww-1-Srad;

   if (Mypos < Srad) Mypos = Srad;
   if (Mypos > fhh-1-Srad) Mypos = fhh-1-Srad;

   Dxpos = Mxpos * Mscale - Morgx + Dorgx;                                 //  corresp. mouse position in window
   Dypos = Mypos * Mscale - Morgy + Dorgy;

   org1x = Mxpos - Srad;                                                   //  source pixels in 1x image
   org1y = Mypos - Srad;

   ww1 = hh1 = Srad * 2;                                                   //  enclosing square

   pxb1 = gdk_pixbuf_new_subpixbuf(Fpxb->pixbuf,org1x,org1y,ww1,hh1);      //  pixbuf with source pixels
   if (! pxb1) return;

   ww2 = hh2 = drad * 2 + 1;                                               //  enclosing square for scaled pixels

   if (pxb2) g_object_unref(pxb2);
   pxb2 = gdk_pixbuf_scale_simple(pxb1,ww2,hh2,BILINEAR);                  //  create pixbuf with scaled pixels

   g_object_unref(pxb1);
   if (! pxb2) return;

   org2x = Dxpos - (Mxpos - org1x) * scale;                                //  pixbuf position in window
   org2y = Dypos - (Mypos - org1y) * scale;

   if (org2x < Dorgx) org2x = Dorgx;                                       //  keep within window
   if (org2x + ww2 > Dww - Dorgx)
      org2x = Dww - Dorgx - ww2;

   if (org2y < Dorgy) org2y = Dorgy;
   if (org2y + hh2 > Dhh - Dorgy)
      org2y = Dhh - Dorgy - hh2;

   pixels1 = gdk_pixbuf_get_pixels(Mpxb->pixbuf);                          //  image pixels at window scale
   rs1 = gdk_pixbuf_get_rowstride(Mpxb->pixbuf);                           //    corresponding to square area

   pixels2 = gdk_pixbuf_get_pixels(pxb2);                                  //  magnified pixels to display
   rs2 = gdk_pixbuf_get_rowstride(pxb2);

   for (ry = -drad; ry <= drad; ry++)                                      //  loop all pixels
   for (rx = -drad; rx <= drad; rx++)
   {
      if (kernel[rx+drad][ry+drad]) continue;                              //  within mouse circle, no changes

      px1 = rx + drad + org2x - Dorgx + Morgx;                             //  outside mouse circle
      py1 = ry + drad + org2y - Dorgy + Morgy;
      pix1 = pixels1 + py1 * rs1 + px1 * 3;                                //  normal image pixel

      px2 = rx + drad;
      py2 = ry + drad;
      pix2 = pixels2 + py2 * rs2 + px2 * 3;                                //  magnified pixel

      memcpy(pix2,pix1,3);                                                 //  normal pixel replaces magnified pixel
   }

   cr = gdk_cairo_create(gdkwin);
   gdk_cairo_set_source_pixbuf(cr,pxb2,org2x,org2y);
   cairo_paint(cr);

   Dxpos = org2x + drad + 2;                                               //  draw circle around magnified area  v.15.01
   Dypos = org2y + drad + 2;
   cairo_set_source_rgb(cr,0,0,0);
   cairo_set_line_width(cr,2);
   cairo_arc(cr,Dxpos,Dypos,drad-2,0,2*PI);
   cairo_stroke(cr);
   cairo_destroy(cr);

   return;
}


/**************************************************************************/

//  dark_brite menu function
//  highlight darkest and brightest pixels

namespace darkbrite {
   float    darklim = 0;
   float    brightlim = 255;
}

void m_darkbrite(GtkWidget *, const char *)
{
   using namespace darkbrite;

   int    darkbrite_dialog_event(zdialog* zd, const char *event);

   cchar    *title = ZTX("Darkest and Brightest Pixels");

   F1_help_topic = "darkbrite_pixels";

   if (! curr_file) return;                                                //  no image file

/**
       ______________________________________
      |    Darkest and Brightest Pixels      |
      |                                      |
      |  Dark Limit   ===[]============ NNN  |
      |  Bright Limit ============[]=== NNN  |
      |                                      |
      |                              [Done]  |
      |______________________________________|

**/

   zdialog *zd = zdialog_new(title,Mwin,Bdone,null);                       //  darkbrite dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=3|expand");
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"space=3");
   zdialog_add_widget(zd,"label","labD","vb1",ZTX("Dark Limit"));
   zdialog_add_widget(zd,"label","labB","vb1",ZTX("Bright Limit"));
   zdialog_add_widget(zd,"hscale","limD","vb2","0|255|1|0","expand");
   zdialog_add_widget(zd,"hscale","limB","vb2","0|255|1|255","expand");
   zdialog_add_widget(zd,"label","valD","vb3");
   zdialog_add_widget(zd,"label","valB","vb3");

   zdialog_rescale(zd,"limD",0,0,255);                                     //  v.15.01
   zdialog_rescale(zd,"limB",0,255,255);

   zdialog_stuff(zd,"limD",darklim);                                       //  start with prior values
   zdialog_stuff(zd,"limB",brightlim);

   zdialog_resize(zd,300,0);
   zdialog_run(zd,darkbrite_dialog_event,"save");                          //  run dialog - parallel
   zddarkbrite = zd;                                                       //  global pointer for Fpaint*()

   zdialog_send_event(zd,"limD");                                          //  initz. NNN labels
   zdialog_send_event(zd,"limB");

   return;
}


//  darkbrite dialog event and completion function

int darkbrite_dialog_event(zdialog *zd, const char *event)                 //  darkbrite dialog event function
{
   using namespace darkbrite;

   char     text[8];

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [done]
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07

   if (zd->zstat)
   {
      zdialog_free(zd);
      zddarkbrite = 0;
      Fpaint2();
      return 0;
   }

   if (strmatch(event,"limD")) {
      zdialog_fetch(zd,"limD",darklim);
      snprintf(text,8,"%.0f",darklim);
      zdialog_stuff(zd,"valD",text);
   }

   if (strmatch(event,"limB")) {
      zdialog_fetch(zd,"limB",brightlim);
      snprintf(text,8,"%.0f",brightlim);
      zdialog_stuff(zd,"valB",text);
   }

   Fpaint2();
   return 0;
}


//  this function called by Fpaint() if zddarkbrite dialog active

void darkbrite_paint()                                                     //  v.14.05
{
   using namespace darkbrite;

   int         px, py;
   uint8       *pix;
   float       P, D = darklim, B = brightlim;

   for (py = 0; py < Mpxb->hh; py++)                                       //  loop all image pixels
   for (px = 0; px < Mpxb->ww; px++)
   {
      pix = PXBpix(Mpxb,px,py);
      P = pixbright(pix);
      if (P < D) pix[0] = 100;                                             //  dark pixel = mostly red
      else if (P > B) pix[0] = 0;                                          //  bright pixel = mostly green/blue
   }

   return;                                                                 //  not executed, avoid gcc warning
}


/**************************************************************************/

//  monitor color and contrast test function

void m_moncolor(GtkWidget *, cchar *)
{
   int  moncolor_dialog_event(zdialog *zd, cchar *event);

   char        file[200];
   int         err;
   char        *savecurrfile = 0;
   char        *savecurrdirk = 0;
   zdialog     *zd;
   cchar       *message = ZTX("Brightness should show a gradual ramp \n"
                              "extending all the way to the edges.");

   F1_help_topic = "monitor_color";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

   if (curr_file) savecurrfile = zstrdup(curr_file);
   if (curr_dirk) savecurrdirk = zstrdup(curr_dirk);

   snprintf(file,200,"%s/images/colorchart.png",get_zdatadir());           //  color chart .png file

   err = f_open(file,0,0,1);
   if (err) goto restore;

   Fzoom = 1;
   gtk_window_set_title(MWIN,"check monitor");

   zd = zdialog_new("check monitor",Mwin,Bdone,null);                      //  start user dialog
   if (message) {
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","lab1","hb1",message,"space=5");
   }

   zdialog_resize(zd,300,0);
   zdialog_run(zd,moncolor_dialog_event,"0/0");
   zdialog_wait(zd);                                                       //  wait for dialog complete
   zdialog_free(zd);

restore:

   if (savecurrdirk) {
      curr_dirk = zstrdup(savecurrdirk);
      zfree(savecurrdirk);
   }

   if (savecurrfile) {
      f_open(savecurrfile);
      zfree(savecurrfile);
   }
   else {
      curr_file = 0;
      if (curr_dirk) gallery(curr_dirk,"init",-1);
   }

   Fblock = 0;
   return;
}


//  dialog event and completion function

int moncolor_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   return 1;
}


/**************************************************************************/

//  check and adjust monitor gamma

void m_mongamma(GtkWidget *, cchar *)
{
   int   mongamma_dialog_event(zdialog *zd, cchar *event);

   int         err;
   char        gammachart[200];
   zdialog     *zd;
   char        *savecurrent = 0;

   cchar       *permit = "Chart courtesy of Norman Koren";
   cchar       *website = "http://www.normankoren.com/makingfineprints1A.html#gammachart";

   F1_help_topic = "monitor_gamma";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 0;

   err = shell_quiet("which xgamma");                                      //  check for xgamma
   if (err) {
      zmessageACK(Mwin,"xgamma program is not installed");
      return;
   }

   if (curr_file) savecurrent = zstrdup(curr_file);

   snprintf(gammachart,200,"%s/images/gammachart2.png",get_zdatadir());    //  gamma chart .png file
   err = f_open(gammachart);
   if (err) goto restore;

   Fzoom = 1;                                                              //  scale 100% (required)
   gtk_window_set_title(MWIN,"adjust gamma chart");

   zd = zdialog_new(ZTX("Monitor Gamma"),Mwin,Bdone,null);                 //  start user dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labgamma","hb1","gamma","space=5");
   zdialog_add_widget(zd,"hscale","gamma","hb1","0.6|1.4|0.02|1.0","expand");
   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"link",permit,"hb2",website);

   zdialog_resize(zd,200,0);
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs
   zdialog_run(zd,mongamma_dialog_event);
   zdialog_wait(zd);                                                       //  wait for dialog complete
   zdialog_free(zd);

restore:

   if (savecurrent) {
      f_open(savecurrent);                                                 //  back to current file
      zfree(savecurrent);
   }

   Fblock = 0;
   return;
}


//  dialog event function

int mongamma_dialog_event(zdialog *zd, cchar *event)
{
   double   gamma;

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [done]  v.14.03
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"gamma")) {
      zdialog_fetch(zd,"gamma",gamma);
      shell_ack("xgamma -quiet -gamma %.2f",gamma);
   }

   return 0;
}


/**************************************************************************/

//  set GUI language

void  m_changelang(GtkWidget *, cchar *)
{
   int  changelang_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         ii, cc, val, zstat;
   char        progexe[300], lang1[8], *pp;

   cchar  *langs[10] = { "en English",                                     //  english first
                         "ca Catalan", "de German", "es Spanish",
                         "fr French", "it Italian", "nl Dutch",
                         "pt Portuguese", "ru Russian", null  };

   cchar  *title = ZTX("Available Translations");

   F1_help_topic = "change_language";

   zd = zdialog_new(ZTX("Set Language"),Mwin,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"label","title","dialog",title,"space=5");
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1");

   for (ii = 0; langs[ii]; ii++)                                           //  make radio button per language
      zdialog_add_widget(zd,"radio",langs[ii],"vb1",langs[ii]);

   cc = strlen(zfuncs::zlang);                                             //  current language
   for (ii = 0; langs[ii]; ii++)                                           //  match on lc_RC
      if (strmatchN(zfuncs::zlang,langs[ii],cc)) break;
   if (! langs[ii])
      for (ii = 0; langs[ii]; ii++)                                        //  failed, match on lc alone
         if (strmatchN(zfuncs::zlang,langs[ii],2)) break;
   if (! langs[ii]) ii = 0;                                                //  failed, default english
   zdialog_stuff(zd,langs[ii],1);

   zdialog_resize(zd,200,0);
   zdialog_run(zd,changelang_dialog_event);                                //  run dialog
   zstat = zdialog_wait(zd);

   for (ii = 0; langs[ii]; ii++) {                                         //  get active radio button
      zdialog_fetch(zd,langs[ii],val);
      if (val) break;
   }

   zdialog_free(zd);                                                       //  kill dialog

   if (zstat != 1) return;                                                 //  user cancel
   if (! val) return;                                                      //  no selection

   strncpy0(lang1,langs[ii],8);
   pp = strchr(lang1,' ');                                                 //  isolate lc_RC part
   *pp = 0;

   cc = readlink("/proc/self/exe",progexe,300);                            //  get own program path
   if (cc <= 0) {
      zmessageACK(Mwin,"cannot get /proc/self/exe");
      return;
   }
   progexe[cc] = 0;

   shell_ack("%s -l %s -p &",progexe,lang1);                               //  start new fotoxx with language

   m_quit(0,0);                                                            //  exit
   return;
}


//  dialog event and completion function

int changelang_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   return 1;
}


/**************************************************************************/

//  report missing translations in a popup window

void  m_untranslated(GtkWidget *, cchar *)
{
   int      ftf = 1, Nmiss = 0;
   cchar    *missing;
   char     text[40];

   F1_help_topic = "missing_translations";

   if (strmatchN(zfuncs::zlang,"en",2)) {
      zmessageACK(Mwin,"use a non-English locale");                        //  v.15.04
      return;
   }

   write_popup_text("open","missing translations",400,200,Mwin);

   while (true)
   {
      missing = ZTX_missing(ftf);
      if (! missing) break;
      write_popup_text("write",missing);
      Nmiss++;
   }

   snprintf(text,40,"%d missing translations",Nmiss);
   write_popup_text("write",text);
   write_popup_text("top",0);

   return;
}


/**************************************************************************/

//  Dump CPU usage, zdialog statistics, memory statistics
//  counters are reset

void m_resources(GtkWidget *, cchar *)
{
   F1_help_topic = "resources";                                            //  v.15.07
   zmalloc_report();
   return;
}


/**************************************************************************/

//  zappcrash test - make a segment fault

char     *zapptext = null;

void m_zappcrash(GtkWidget *, cchar *)                                     //  v.14.01
{
   int      ii, jj;

   for (ii = 0; ii < 100; ii++) {                                          //  fake-out compiler diagnostics
      jj = 1000.0 * drandz();
      zapptext[jj] = 'a' + ii / 10;                                        //  write to address 0 + jj
   }

   printz("zapptext: %s \n",zapptext);
   return;
}


/**************************************************************************/

//  edit bookmarks

namespace bookmarknames
{
   #define     maxbmks 50
   char        *bookmarks[maxbmks];                                        //  bookmark names and files
   int         Nbmks;                                                      //  count of entries
   int         bmkposn;                                                    //  current entry, 0-last
   GtkWidget   *bmkswidget;
   zdialog     *zd_goto_bookmark;
}

void bookmarks_load();
void bookmarks_listclick(GtkWidget *widget, int line, int pos);
int  bookmarks_dialog_event(zdialog *zd, cchar *event);
void bookmarks_refresh();


void m_edit_bookmarks(GtkWidget *, cchar *)
{
   using namespace bookmarknames;

   zdialog     *zd;
   cchar       *bmk_add = ZTX("Click list position. Click thumbnail to add.");

/***
          _______________________________________________
         |    Edit Bookmarks                             |
         |                                               |
         | Click list position. Click thumbnail to add.  |
         |-----------------------------------------------|
         | bookmarkname1      /topdir/.../filename1.jpg  |
         | bookmarkname2      /topdir/.../filename2.jpg  |
         | bookmarkname3      /topdir/.../filename3.jpg  |
         | bookmarkname4      /topdir/.../filename4.jpg  |
         | bookmarkname5      /topdir/.../filename5.jpg  |
         | bookmarkname6      /topdir/.../filename6.jpg  |
         |-----------------------------------------------|
         | [bookmarkname...] [rename] [delete]           |
         |                                        [done] |
         |_______________________________________________|

***/

   F1_help_topic = "bookmarks";
   if (zd_edit_bookmarks) return;                                          //  already busy
   if (zd_goto_bookmark) return;
   if (checkpend("all")) return;

   zd = zdialog_new(ZTX("Edit Bookmarks"),Mwin,Bdone,null);
   zd_edit_bookmarks = zd;
   zdialog_add_widget(zd,"hbox","hbtip","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labtip","hbtip",bmk_add,"space=5");
   zdialog_add_widget(zd,"hbox","hbbmk","dialog",0,"space=5|expand");
   zdialog_add_widget(zd,"frame","frbmk","hbbmk",0,"space=5|expand");
   zdialog_add_widget(zd,"text","bmklist","frbmk",0,"expand");
   zdialog_add_widget(zd,"hbox","hbname","dialog",0,"space=5");
   zdialog_add_widget(zd,"entry","bmkname","hbname",0,"space=5|size=30");
   zdialog_add_widget(zd,"button","rename","hbname",Brename,"space=5");
   zdialog_add_widget(zd,"button","delete","hbname",Bdelete,"space=5");

   bmkswidget = zdialog_widget(zd,"bmklist");                              //  connect mouse to bookmark list
   textwidget_set_clickfunc(bmkswidget,bookmarks_listclick);

   bookmarks_load();                                                       //  load bookmarks from bookmarks file
   bookmarks_refresh();                                                    //  update bookmarks list in dialog

   zdialog_resize(zd,300,300);
   zdialog_run(zd,bookmarks_dialog_event,"save");                          //  run dialog, parallel

   m_viewmode(0,"G");                                                      //  show current gallery
   return;
}


//  load bookmarks list from bookmarks file

void bookmarks_load()                                                      //  v.15.05
{
   using namespace bookmarknames;

   int         err;
   char        buff[XFCC], bmkfile[200];
   char        *pp, *pp2;
   FILE        *fid;
   STATB       stbuff;

   Nbmks = 0;
   err = locale_filespec("user","bookmarks",bmkfile);                      //  read bookmarks file
   if (! err) {
      fid = fopen(bmkfile,"r");
      if (fid) {
         while (true) {
            pp = fgets_trim(buff,XFCC,fid,1);                              //  next bookmark rec.
            if (! pp) break;
            if (strlen(pp) < 40) continue;
            pp2 = strchr(pp+32,'/');                                       //  verify bookmark                    v.15.05
            if (! pp2) continue;
            err = stat(pp2,&stbuff);
            if (err) continue;
            bookmarks[Nbmks] = zstrdup(pp);                                //  fill bookmark list
            if (++Nbmks == maxbmks) break;
         }
         fclose(fid);
      }
   }
   
   bmkposn = Nbmks;                                                        //  next free position                 v.15.05
   return;
}


//  mouse click function to select existing bookmark from list

void bookmarks_listclick(GtkWidget *widget, int line, int pos)
{
   using namespace bookmarknames;

   char     bookmarkname[32];

   if (! zd_edit_bookmarks) return;
   if (line < 0) line = 0;
   if (line > Nbmks-1) line = Nbmks-1;
   bmkposn = line;
   strncpy0(bookmarkname,bookmarks[bmkposn],31);
   strTrim(bookmarkname);
   zdialog_stuff(zd_edit_bookmarks,"bmkname",bookmarkname);
   return;
}


//  mouse click function to receive clicked bookmarks

void bookmarks_Lclick_func(int Nth)
{
   using namespace bookmarknames;

   char     *imagefile, *newbookmark;
   char     *pp, bookmarkname[32];
   int      cc;

   if (! zd_edit_bookmarks) return;
   if (Nth < 0) return;                                                    //  gallery gone ?
   imagefile = gallery(0,"find",Nth);                                      //  get file at clicked position
   if (! imagefile) return;

   pp = strrchr(imagefile,'/');                                            //  get file name or last subdirk name
   if (! pp) return;                                                       //    to use as default bookmark name
   strncpy0(bookmarkname,pp+1,31);                                         //  max. 30 chars. + null

   cc = strlen(imagefile) + 34;                                            //  construct bookmark record:
   newbookmark = (char *) zmalloc(cc);                                     //    filename  /directories.../filename
   snprintf(newbookmark,cc,"%-30s  %s",bookmarkname,imagefile);
   zfree(imagefile);

   if (Nbmks == maxbmks) {                                                 //  if list full, remove first
      zfree(bookmarks[0]);
      Nbmks--;
      for (int ii = 0; ii < Nbmks; ii++)
         bookmarks[ii] = bookmarks[ii+1];
   }
   
   if (Nbmks == 0) bmkposn = 0;                                            //  1st bookmark --> 0
   else bmkposn++;                                                         //  else clicked position + 1          v.15.05
   if (bmkposn < 0) bmkposn = 0;
   if (bmkposn > Nbmks) bmkposn = Nbmks;

   for (int ii = Nbmks; ii > bmkposn; ii--)                                //  make hole to insert new bookmark
      bookmarks[ii] = bookmarks[ii-1];

   bookmarks[bmkposn] = newbookmark;                                       //  insert
   Nbmks++;

   bookmarks_refresh();                                                    //  update bookmarks list in dialog

   zdialog_stuff(zd_edit_bookmarks,"bmkname",bookmarkname);

   return;
}


//  dialog event and completion function

int bookmarks_dialog_event(zdialog *zd, cchar *event)
{
   using namespace bookmarknames;

   char        bmkfile[200];
   char        bookmarkname[32];
   FILE        *fid;
   int         cc;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"delete"))                                           //  delete bookmark at position
   {
      if (bmkposn < 0 || bmkposn > Nbmks-1) return 1;
      for (int ii = bmkposn; ii < Nbmks-1; ii++)
         bookmarks[ii] = bookmarks[ii+1];
      Nbmks--;
      zdialog_stuff(zd,"bmkname","");                                      //  clear name field
      bookmarks_refresh();                                                 //  update bookmarks list in dialog
   }

   if (strmatch(event,"rename"))                                           //  apply new name to bookmark
   {
      if (bmkposn < 0 || bmkposn > Nbmks-1) return 1;
      zdialog_fetch(zd,"bmkname",bookmarkname,31);                         //  get name from dialog
      cc = strlen(bookmarkname);
      if (cc < 30) memset(bookmarkname+cc,' ',30-cc);                      //  blank pad to 30 chars.
      bookmarkname[30] = 0;
      strncpy(bookmarks[bmkposn],bookmarkname,30);                         //  replace name in bookmarks list
      bookmarks_refresh();                                                 //  update bookmarks list in dialog
   }

   if (strmatch(event,"enter")) zd->zstat = 1;                             //  [done]  v.14.03

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat == 1)                                                     //  done
   {
      locale_filespec("user","bookmarks",bmkfile);                         //  write bookmarks file
      fid = fopen(bmkfile,"w");
      if (! fid)
         zmessageACK(Mwin,ZTX("unable to save bookmarks file"));
      else {
         for (int ii = 0; ii < Nbmks; ii++)
            fprintf(fid,"%s\n",bookmarks[ii]);
         fclose(fid);
      }
   }

   for (int ii = 0; ii < Nbmks; ii++)                                      //  free memory
      zfree(bookmarks[ii]);

   zdialog_free(zd);
   zd_edit_bookmarks = 0;
   return 1;
}


//  private function to update dialog widget with new bookmarks list

void bookmarks_refresh()
{
   using namespace bookmarknames;
   
   char     bookmarkline[XFCC+32];
   char     blanks[33] = "                                ";
   int      cc;

   if (! zd_edit_bookmarks && ! zd_goto_bookmark) return;
   wclear(bmkswidget);                                                     //  clear bookmarks list
   for (int ii = 0; ii < Nbmks; ii++) {                                    //  write bookmarks list
      strncpy0(bookmarkline,bookmarks[ii],31);
      cc = utf8len(bookmarkline);                                          //  compensate multibyte chars.        v.15.05
      strncat(bookmarkline,blanks,32-cc);
      strcat(bookmarkline,bookmarks[ii]+32);
      wprintx(bmkswidget,0,bookmarkline,"monospace 9");
      wprintx(bmkswidget,0,"\n");
   }
   return;
}


//  select a bookmark and jump gallery to selected bookmark
//  connected to gallery menu "GoTo" button

void m_goto_bookmark(GtkWidget *, cchar *)
{
   using namespace bookmarknames;

   void goto_bookmark_listclick(GtkWidget *widget, int line, int pos);
   int  goto_bookmark_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;

/***
                  Gallery Button: [GoTo]
          _______________________________________________
         |             Go to Bookmark                    |
         |-----------------------------------------------|
         | bookmarkname1      /topdir/.../filename1.jpg  |
         | bookmarkname2      /topdir/.../filename2.jpg  |
         | bookmarkname3      /topdir/.../filename3.jpg  |
         | bookmarkname4      /topdir/.../filename4.jpg  |
         | bookmarkname5      /topdir/.../filename5.jpg  |
         | bookmarkname6      /topdir/.../filename6.jpg  |
         |                                               |
         |                              [edit bookmarks] |
         |_______________________________________________|
***/

   F1_help_topic = "bookmarks";
   if (zd_edit_bookmarks) return;                                          //  already busy
   if (zd_goto_bookmark) return;

   zd = zdialog_new(ZTX("Go To Bookmark"),Mwin,ZTX("Edit Bookmarks"),null);
   zd_goto_bookmark = zd;

   zdialog_add_widget(zd,"hbox","hbbmk","dialog",0,"space=5|expand");
   zdialog_add_widget(zd,"frame","frbmk","hbbmk",0,"space=5|expand");
   zdialog_add_widget(zd,"text","bmklist","frbmk",0,"expand");

   bmkswidget = zdialog_widget(zd,"bmklist");                              //  connect mouse to bookmark list
   textwidget_set_clickfunc(bmkswidget,goto_bookmark_listclick);

   bookmarks_load();                                                       //  get bookmarks from bookmarks file  v.15.05
   bookmarks_refresh();                                                    //  update bookmarks list in dialog

   zdialog_resize(zd,300,0);
   zdialog_run(zd,goto_bookmark_dialog_event,"mouse");                     //  run dialog
   return;
}


//  dialog event and completion function

int goto_bookmark_dialog_event(zdialog *zd, cchar *event)
{
   using namespace bookmarknames;
   
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   int zstat = zd->zstat;
   if (! zstat) return 1;                                                  //  wait for completion
   zdialog_free(zd);
   zd_goto_bookmark = 0;
   if (zstat == 1) m_edit_bookmarks(0,0);                                  //  [edit bookmarks] button
   return 1;
}


//  mouse click function to receive clicked bookmarks

void goto_bookmark_listclick(GtkWidget *widget, int line, int pos)
{
   using namespace bookmarknames;

   char     *file;
   int      err;
   STATB    sbuff;

   if (checkpend("all")) return;

   if (! zd_goto_bookmark) return;
   zdialog_free(zd_goto_bookmark);                                         //  kill dialog
   zd_goto_bookmark = 0;

   bmkposn = line;                                                         //  get clicked line
   if (bmkposn < 0 || bmkposn > Nbmks-1) return;
   file = bookmarks[bmkposn] + 32;
   err = stat(file,&sbuff);
   if (err) {
      zmessageACK(Mwin,ZTX("file not found"));
      return;
   }

   gallery(file,"init");                                                   //  go to gallery and file position
   m_viewmode(0,"G");
   gallery(file,"paint");

   return;
}


/**************************************************************************/

//  Batch file rename, convert, resize, move

namespace batch_convert                                                    //  overhauled v.14.02
{
   char     **filelist;
   int      Fsametype, Fsamesize, maxww, maxhh;
   int      Fdelete, Fcopymeta, Fupright, Fsharpen;
   char     newloc[400], newname[200], newext[8];
   int      filecount, baseseq, addseq;
   int      amount, thresh;
};


void m_batch_convert(GtkWidget *, cchar *)
{
   using namespace batch_convert;

   int  batch_convert_dialog_event(zdialog *zd, cchar *event);
   int  batch_sharp_func(PXM *pxm, int amount, int thresh);

   zdialog  *zd;
   int      zstat;
   char     *infile, *outfile, *tempfile;
   char     *inloc, *inname, *inext;
   char     *outloc, *outname, *outext;
   char     **oldfiles, **newfiles;
   char     seqnum[20];
   char     *pp1, *pp2, *ppv[2];
   int      ii, jj, cc, err;
   int      ww, hh, bpc, delinput;
   int      Noldnew;
   char     orientation;
   float    scale, wscale, hscale;
   PXM      *pxmin, *pxmout;
   STATB    statdat;
   cchar    *exifkey[1] = { exif_orientation_key };
   cchar    *exifdata[1];

   F1_help_topic = "batch_convert";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       ____________________________________________________________________
      |                    Batch Convert                                   |
      |                                                                    |
      |  [Select Files]  N files selected                                  |
      |                                                                    |
      |  New Name [____________________________]  base [____]  adder [__]  |
      |  New Location [________________________________________] [browse]  |
      |  New File Type  (o) JPG  (o) PNG  (o) TIF  (o) no change           |
      |  Max. Width [____]  Height [____]  [x] no change                   |
      |  [x] Delete Originals  [x] Copy Metadata  [x] Upright              |
      |  [x] Sharpen   amount [___]  threshold [___]                       |
      |                                                                    |
      |                                               [proceed] [cancel]   |
      |____________________________________________________________________|

***/

   zd = zdialog_new(ZTX("Batch Convert"),Mwin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf",Bnofileselected,"space=10");

   zdialog_add_widget(zd,"hbox","hbname","dialog");
   zdialog_add_widget(zd,"label","labname","hbname",ZTX("New Name"),"space=5");
   zdialog_add_widget(zd,"entry","newname","hbname",0,"expand");
   zdialog_add_widget(zd,"label","labbase","hbname",ZTX("base"),"space=5");
   zdialog_add_widget(zd,"entry","baseseq","hbname",0,"size=5");
   zdialog_add_widget(zd,"label","labadder","hbname",ZTX("adder"),"space=5");
   zdialog_add_widget(zd,"entry","addseq","hbname",0,"size=3");

   zdialog_add_widget(zd,"hbox","hbloc","dialog");
   zdialog_add_widget(zd,"label","labloc","hbloc",ZTX("New Location"),"space=5");
   zdialog_add_widget(zd,"entry","newloc","hbloc",0,"expand");
   zdialog_add_widget(zd,"button","browse","hbloc",Bbrowse,"space=5");

   zdialog_add_widget(zd,"hbox","hbtyp","dialog");
   zdialog_add_widget(zd,"label","labtyp","hbtyp",ZTX("New File Type"),"space=5");
   zdialog_add_widget(zd,"radio","jpg","hbtyp","JPG","space=6");
   zdialog_add_widget(zd,"radio","png","hbtyp","PNG","space=6");
   zdialog_add_widget(zd,"radio","tif","hbtyp","TIF","space=6");
   zdialog_add_widget(zd,"radio","sametype","hbtyp",ZTX("no change"),"space=6");

   zdialog_add_widget(zd,"hbox","hbwh","dialog");
   zdialog_add_widget(zd,"label","labw","hbwh",ZTX("max. Width"),"space=5");
   zdialog_add_widget(zd,"entry","maxww","hbwh","1000","size=5");
   zdialog_add_widget(zd,"label","space","hbwh",0,"space=5");
   zdialog_add_widget(zd,"label","labh","hbwh",ZTX("Height"),"space=5");
   zdialog_add_widget(zd,"entry","maxhh","hbwh","700","size=5");
   zdialog_add_widget(zd,"check","samesize","hbwh",ZTX("no change"),"space=12");

   zdialog_add_widget(zd,"hbox","hbopts","dialog");
   zdialog_add_widget(zd,"check","delete","hbopts",ZTX("Delete Originals"),"space=3");
   zdialog_add_widget(zd,"check","copymeta","hbopts",ZTX("Copy Metadata"),"space=3");
   zdialog_add_widget(zd,"check","upright","hbopts",ZTX("Upright"),"space=3");

   zdialog_add_widget(zd,"hbox","hbsharp","dialog");
   zdialog_add_widget(zd,"check","sharpen","hbsharp",ZTX("Sharpen"),"space=3");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labamount","hbsharp",ZTX("amount"),"space=3");
   zdialog_add_widget(zd,"spin","amount","hbsharp","0|400|1|100");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labthresh","hbsharp",ZTX("threshold"),"space=3");
   zdialog_add_widget(zd,"spin","thresh","hbsharp","0|100|1|20");

   zdialog_stuff(zd,"sametype",1);                                         //  same file type
   zdialog_stuff(zd,"samesize",1);                                         //  same size
   zdialog_stuff(zd,"delete",0);                                           //  delete originals - no
   zdialog_stuff(zd,"copymeta",0);                                         //  copy metadata - no
   zdialog_stuff(zd,"upright",1);                                          //  upright rotation - yes
   zdialog_stuff(zd,"sharpen",0);                                          //  sharpen - no
   zdialog_stuff(zd,"amount",100);
   zdialog_stuff(zd,"thresh",20);

   filelist = 0;
   filecount = 0;
   *newloc = 0;
   oldfiles = newfiles = 0;
   Noldnew = 0;

   zdialog_restore_inputs(zd);                                             //  preload prior user inputs
   zdialog_run(zd,batch_convert_dialog_event,"parent");                    //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   if (zstat != 1) goto cleanup;                                           //  canceled
   if (! filecount) goto cleanup;

   free_resources();                                                       //  no curr. file                      v.15.03

   m_viewmode(0,"F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   if (Fdelete) {
      cc = filecount * sizeof(char *);                                     //  input files are to be deleted:     v.14.09
      oldfiles = (char **) zmalloc(cc);                                    //    reserve space to hold list of
      newfiles = (char **) zmalloc(cc);                                    //    old/new filespecs to update albums
   }

   write_popup_text("open","Processing files",500,200,Mwin);               //  status monitor popup window

   for (ii = 0; ii < filecount; ii++)                                      //  loop selected files
   {
      infile = filelist[ii];                                               //  input file

      parsefile(infile,&inloc,&inname,&inext);                             //  parse directory, filename, .ext
      if (! inloc || ! inname || ! inext) continue;

      outloc = zstrdup(inloc);                                             //  initial output = input file
      outname = zstrdup(inname);
      outext = (char *) zmalloc(8);

      cc = strlen(outloc) - 1;                                             //  remove trailing '/'
      if (outloc[cc] == '/') outloc[cc] = 0;

      if (*newname)                                                        //  new file name was given
      {
         pp1 = strchr(newname,'#');                                        //  .....#####.....
         pp2 = pp1 + 1;                                                    //       |    |
         while (*pp2 == '#') pp2++;                                        //       pp1  pp2
         cc = pp2 - pp1;
         jj = baseseq + ii * addseq;                                       //  new sequence number nnnnnn
         snprintf(seqnum,20,"%0*d",cc,jj);
         zfree(outname);
         outname = zstrdup(newname,8);                                     //  .....#####.....
         pp1 = strchr(outname,'#');
         strcpy(pp1,seqnum);
         strcat(pp1,pp2);                                                  //  .....nnnnn.....
      }

      if (*newloc) {                                                       //  new location was given
         zfree(outloc);
         outloc = zstrdup(newloc);
      }

      if (Fsametype) {
         if (strcasestr(".jpg .jpeg",inext)) strcpy(outext,".jpg");        //  new .ext from existing .ext
         else if (strcasestr(".png",inext)) strcpy(outext,".png");
         else if (strcasestr(".tif .tiff",inext)) strcpy(outext,".tif");
         else strcpy(outext,".jpg");                                       //  unknown >> .jpg
      }
      else strcpy(outext,newext);                                          //  new .ext was given

      cc = strlen(outloc) + strlen(outname) + strlen(outext) + 4;
      outfile = (char *) zmalloc(cc);
      snprintf(outfile,cc,"%s/%s%s",outloc,outname,outext);

      zfree(outloc);
      zfree(outname);
      zfree(outext);

      write_popup_text("write",outfile);                                   //  log each output file
      zmainloop();

      if (*newloc) {
         err = stat(outfile,&statdat);                                     //  check if file exists in new location
         if (! err) {
            write_popup_text("write",Bfileexists);
            zfree(outfile);
            continue;
         }
      }

      pxmin = PXM_load(infile,0);                                          //  read input file
      if (! pxmin) {
         write_popup_text("write",ZTX("file type not supported"));
         zfree(outfile);
         continue;
      }

      if (Fupright)                                                        //  upright image if turned
      {
         exif_get(infile,exifkey,ppv,1);                                   //  get EXIF: Orientation
         if (ppv[0]) {
            orientation = *ppv[0];                                         //  single character
            zfree(ppv[0]);
         }
         else orientation = '1';                                           //  if missing assume unrotated

         pxmout = 0;
         if (orientation == '6')
            pxmout = PXM_rotate(pxmin,+90);                                //  rotate clockwise 90 deg.
         if (orientation == '8')
            pxmout = PXM_rotate(pxmin,-90);                                //  counterclockwise 90 deg.
         if (orientation == '3')
            pxmout = PXM_rotate(pxmin,180);                                //  180 deg.                           v.15.03
         if (pxmout) {
            PXM_free(pxmin);                                               //  input image unrotated
            pxmin = pxmout;
         }
      }

      bpc = f_load_bpc;                                                    //  input file bits/color
      ww = pxmin->ww;                                                      //  input file size
      hh = pxmin->hh;

      if (Fsamesize) pxmout = pxmin;                                       //  same size, output = input
      else {
         wscale = hscale = 1.0;
         if (ww > maxww) wscale = 1.0 * maxww / ww;                        //  compute new size
         if (hh > maxhh) hscale = 1.0 * maxhh / hh;
         if (wscale < hscale) scale = wscale;
         else scale = hscale;
         if (scale > 0.999) pxmout = pxmin;                                //  no change
         else {
            ww = ww * scale;
            hh = hh * scale;
            pxmout = PXM_rescale(pxmin,ww,hh);                             //  rescaled output file
            PXM_free(pxmin);                                               //  free memory
         }
      }

      if (Fsharpen)                                                        //  auto sharpen output image          v.14.06
         batch_sharp_func(pxmout,amount,thresh);

      tempfile = zstrdup(infile,12);                                       //  temp file needed for EXIF/IPTC copy
      pp1 = strrchr(outfile,'.');
      pp2 = strrchr(tempfile,'.');
      strcpy(pp2,"-temp");
      strcpy(pp2+5,pp1);

      if (strmatch(".tif",pp1))                                            //  write output file to temp file
         err = PXM_TIFF_save(pxmout,tempfile,bpc);
      else if (strmatch(".png",pp1))
         err = PXM_PNG_save(pxmout,tempfile,bpc);
      else
         err = PXM_ANY_save(pxmout,tempfile);

      if (err) {
         write_popup_text("write",ZTX("cannot create new file"));
         zfree(outfile);
         zfree(tempfile);
         PXM_free(pxmout);
         continue;
      }

      if (Fcopymeta)                                                       //  copy EXIF/IPTC if requested
      {
         if (Fupright) {                                                   //  if image possibly uprighted
            exifdata[0] = "";                                              //    remove exif:orientation
            exif_copy(infile,tempfile,exifkey,exifdata,1);                 //  (set = 1 does not work)
         }
         else exif_copy(infile,tempfile,0,0,0);                            //  else no EXIF change
      }

      err = shell_ack("cp -p \"%s\" \"%s\" ",tempfile,outfile);            //  copy tempfile to output file
      if (err) write_popup_text("write",wstrerror(err));
      remove(tempfile);                                                    //  remove tempfile

      delinput = 0;                                                        //  figure out if input file can be deleted
      if (Fdelete) delinput = 1;                                           //  user says yes
      if (err) delinput = 0;                                               //  not if error
      if (strmatch(infile,outfile)) delinput = 0;                          //  not if overwritten by output

      if (delinput) {                                                      //  delete input file
         remove(infile);
         delete_image_index(infile);                                       //  remove from image index
      }

      if (! err) {
         load_filemeta(outfile);                                           //  update image index for output file
         update_image_index(outfile);
      }

      if (delinput) {                                                      //  if input file was deleted,         v.14.09
         oldfiles[Noldnew] = zstrdup(infile);                              //    mark for updating albums
         newfiles[Noldnew] = zstrdup(outfile);
         Noldnew++;
      }

      zfree(outfile);
      zfree(tempfile);
      PXM_free(pxmout);
   }

   if (Noldnew) {                                                          //  update albums for renamed/moved files
      write_popup_text("write",ZTX("updating albums ..."));                //  v.14.09
      conv_albums(oldfiles,newfiles,Noldnew);
   }

   write_popup_text("write","COMPLETED");

   if (navi::gallerytype == 1)                                             //  update gallery file list           v.15.03
      gallery(0,"init");

   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:

   if (filecount) {                                                        //  free memory
      for (ii = 0; ii < filecount; ii++)
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
      filecount = 0;
   }

   if (Noldnew) {                                                          //  v.14.09
      for (ii = 0; ii < Noldnew; ii++) {
         zfree(oldfiles[ii]);
         zfree(newfiles[ii]);
      }
      zfree(oldfiles);
      zfree(newfiles);
      Noldnew = 0;
   }

   Fblock = 0;
   return;
}


//  dialog event and completion callback function

int batch_convert_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batch_convert;

   char         countmess[60];
   char         *ploc;
   int          ii, cc, yn, err;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"files"))                                            //  select images to convert
   {
      if (filelist) {                                                      //  free prior list
         for (ii = 0; ii < filecount; ii++)
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
         filecount = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      ii = 0;
      if (filelist)                                                        //  count files selected
         for (ii = 0; filelist[ii]; ii++);
      filecount = ii;

      snprintf(countmess,60,Bfileselected,ii);                             //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
   }

   if (strmatch(event,"browse")) {
      ploc = zgetfile(ZTX("Select directory"),MWIN,"folder",curr_dirk);    //  new location browse
      if (! ploc) return 1;
      zdialog_stuff(zd,"newloc",ploc);
      zfree(ploc);
   }

   if (strstr("maxhh maxww",event))                                        //  if max width/height changed,
      zdialog_stuff(zd,"samesize",0);                                      //    reset "no change"

   if (strmatch(event,"samesize")) {                                       //  if "no change" set,
      zdialog_fetch(zd,"samesize",Fsamesize);                              //    clear max width / height
      if (Fsamesize) {
         zdialog_stuff(zd,"maxww","");
         zdialog_stuff(zd,"maxhh","");
      }
   }

   if (zd->zstat != 1) return 1;                                           //  wait for [proceed]
   zd->zstat = 0;                                                          //  dialog active until inputs OK

   zdialog_fetch(zd,"newname",newname,100);                                //  new file name
   zdialog_fetch(zd,"baseseq",baseseq);                                    //  base sequence number
   zdialog_fetch(zd,"addseq",addseq);                                      //  sequence number adder
   zdialog_fetch(zd,"newloc",newloc,400);                                  //  new location (directory)
   zdialog_fetch(zd,"maxww",maxww);                                        //  new max width
   zdialog_fetch(zd,"maxhh",maxhh);                                        //  new max height
   zdialog_fetch(zd,"samesize",Fsamesize);                                 //  keep same width/height
   zdialog_fetch(zd,"sametype",Fsametype);                                 //  keep same file type
   zdialog_fetch(zd,"delete",Fdelete);                                     //  delete originals
   zdialog_fetch(zd,"copymeta",Fcopymeta);                                 //  copy metadata
   zdialog_fetch(zd,"upright",Fupright);                                   //  upright rotation
   zdialog_fetch(zd,"sharpen",Fsharpen);                                   //  auto sharpen                       v.14.02
   zdialog_fetch(zd,"amount",amount);                                      //  sharpen amount                     v.14.06
   zdialog_fetch(zd,"thresh",thresh);                                      //  sharpen threshold

   if (! filecount) {
      zmessageACK(Mwin,Bnofileselected);
      return 1;
   }

   strTrim2(newname);                                                      //  check new name, baseseq, addseq
   if (! blank_null(newname)) {
      err = 0;
      if (strlen(newname) < 2 || baseseq < 1 || addseq < 1) err++;
      if (! strchr(newname,'#')) err++;
      if (err) {
         zmessageACK(Mwin,ZTX("new name/base/adder unreasonable"
                              "\n e.g. newname ###   100  10"));
         return 1;
      }
   }

   strTrim2(newloc);                                                       //  check location
   if (! blank_null(newloc)) {
      cc = strlen(newloc) - 1;
      if (newloc[cc] == '/') newloc[cc] = 0;                               //  remove trailing '/'
      err = check_create_dir(newloc);                                      //  create if needed                   v.14.03
      if (err) return 1;
   }

   if (! Fsametype) {                                                      //  file type change
      strcpy(newext,".jpg");                                               //  get new .ext
      zdialog_fetch(zd,"png",ii);
      if (ii) strcpy(newext,".png");
      zdialog_fetch(zd,"tif",ii);
      if (ii) strcpy(newext,".tif");
   }

   if (! Fsamesize && (maxww < 20 || maxhh < 20)) {
      zmessageACK(Mwin,ZTX("max. size %d x %d is not reasonable"),maxww,maxhh);
      return 1;
   }

   /**   Convert NN image files                    0
           Rename to xxxxxxx                       1
           Convert to .ext                         2
           Resize within NNxNN                     3
           Output to /.../...                      4
           Copy Metadata  Upright  Sharpen         5
           Delete Originals                        6
         PROCEED?                                  7
   **/

   char  mess0[60], mess1[100], mess2[60], mess3[60], mess4[200], mess5[80], mess6[60], mess7[40];
   char  warnmess[600];

   *mess0 = *mess1 = *mess2 = *mess3 = *mess4 = *mess5 = *mess6 = *mess7 = 0;

   snprintf(mess0,60,ZTX("Convert %d image files"),filecount);
   if (*newname) snprintf(mess1,100,"\n  %s %s",ZTX("Rename to"),newname);
   if (! Fsametype) snprintf(mess2,60,"\n  %s %s",ZTX("Convert to"),newext);
   if (! Fsamesize) snprintf(mess3,60,"\n  %s %dx%d",ZTX("Resize within"),maxww,maxhh);
   if (*newloc) snprintf(mess4,200,"\n  %s  %s",ZTX("Output to"),newloc);
   if (Fcopymeta || Fupright || Fsharpen) strcat(mess5,"\n  ");
   if (Fcopymeta) { strcat(mess5,ZTX("Copy Metadata")); strcat(mess5,"  "); }
   if (Fupright) { strcat(mess5,ZTX("Upright")); strcat(mess5,"  "); }
   if (Fsharpen) { strcat(mess5,ZTX("Sharpen")); strcat(mess5,"  "); }
   if (Fdelete) snprintf(mess6,60,"\n  %s",ZTX("Delete Originals"));
   snprintf(mess7,40,"\n\n%s",ZTX("PROCEED?"));

   snprintf(warnmess,600,"%s %s %s %s %s %s %s %s",mess0,mess1,mess2,mess3,mess4,mess5,mess6,mess7);

   yn = zmessageYN(Mwin,warnmess);
   if (! yn) return 1;

   zd->zstat = 1;                                                          //  [proceed]
   return 1;
}


/**************************************************************************/

//  callable sharpen function for batch_convert and batch_RAW_convert
//  amount: 0 to 400    strength of applied algorithm
//  thresh: 0 to 100    contrast level below which sharpen is diminished

namespace batch_sharp_names
{
   PXM      *pxm1, *pxm2;
   int      BSamount, BSthresh;
}


int batch_sharp_func(PXM *pxm, int amount, int thresh)                     //  14.06
{
   using namespace batch_sharp_names;

   void * batch_sharp_wthread(void *arg);

   pxm2 = pxm;                                                             //  output
   pxm1 = PXM_copy(pxm2);                                                  //  input

   BSamount = amount;
   BSthresh = thresh;

   for (int ii = 0; ii < NWT; ii++)                                        //  start worker threads
      start_wthread(batch_sharp_wthread,&Nval[ii]);
   wait_wthreads();                                                        //  wait for completion

   PXM_free(pxm1);
   return 1;
}


void * batch_sharp_wthread(void *arg)                                      //  worker thread function
{
   using namespace batch_sharp_names;

   float       *pix1, *pix2;
   int         px, py;
   float       amount, thresh;
   float       b1, b1x, b1y, b2x, b2y, b2, bf, f1, f2;
   float       red1, green1, blue1, red2, green2, blue2;

   int         index = *((int *) arg);

   amount = 1 + 0.01 * BSamount;                                           //  1.0 - 5.0
   thresh = BSthresh;                                                      //  0 - 100

   for (py = index + 1; py < pxm1->hh; py += NWT)                          //  loop all image pixels
   for (px = 1; px < pxm1->ww; px++)
   {
      pix1 = PXMpix(pxm1,px,py);                                           //  input pixel
      pix2 = PXMpix(pxm2,px,py);                                           //  output pixel

      b1 = pixbright(pix1);                                                //  pixel brightness, 0 - 256
      if (b1 == 0) continue;                                               //  black, don't change
      b1x = b1 - pixbright(pix1-3);                                        //  horiz. brightness gradient
      b1y = b1 - pixbright(pix1-3 * pxm1->ww);                             //  vertical
      f1 = fabsf(b1x + b1y);

      if (f1 < thresh)                                                     //  moderate brightness change for
         f1 = f1 / thresh;                                                 //    pixels below threshold gradient
      else  f1 = 1.0;
      f2 = 1.0 - f1;

      b1x = b1x * amount;                                                  //  amplified gradient
      b1y = b1y * amount;

      b2x = pixbright(pix1-3) + b1x;                                       //  + prior pixel brightness
      b2y = pixbright(pix1-3 * pxm2->ww) + b1y;                            //  = new brightness
      b2 = 0.5 * (b2x + b2y);

      b2 = f1 * b2 + f2 * b1;                                              //  possibly moderated

      bf = b2 / b1;                                                        //  ratio of brightness change
      if (bf < 0) bf = 0;
      if (bf > 4) bf = 4;

      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];

      red2 = bf * red1;                                                    //  output RGB
      if (red2 > 255.9) red2 = 255.9;
      green2 = bf * green1;
      if (green2 > 255.9) green2 = 255.9;
      blue2 = bf * blue1;
      if (blue2 > 255.9) blue2 = 255.9;

      pix2[0] = red2;
      pix2[1] = green2;
      pix2[2] = blue2;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Batch upright image files.
//  Look for files rotated 90˚ (according to EXIF) and upright them.

char     **bup_filelist;
int      bup_filecount;
int      bup_allfiles;

void m_batch_upright(GtkWidget *, cchar *)
{
   int  batch_upright_dialog_event(zdialog *zd, cchar *event);

   zdialog        *zd;
   int            zstat;
   char           *infile, *tempfile;
   char           *pp1, *pp2, *ppv[1];
   int            ii, jj, err, bpc;
   char           orientation;
   PXM            *pxmin, *pxmout;
   cchar          *exifkey[1] = { exif_orientation_key };
   cchar          *exifdata[1];
   char           **allfiles;
   int            NAF;

   F1_help_topic = "batch_upright";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       ___________________________________
      |       Batch Upright               |
      |                                   |
      |  [Select Files]  N files selected |
      |  [x] Survey all files             |
      |                                   |
      |               [proceed] [cancel]  |
      |___________________________________|

***/

   zd = zdialog_new(ZTX("Batch Upright"),Mwin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf",Bnofileselected,"space=10");
   zdialog_add_widget(zd,"hbox","hbaf","dialog");
   zdialog_add_widget(zd,"check","allfiles","hbaf",ZTX("Survey all files"),"space=5");

   bup_filelist = 0;
   bup_filecount = 0;

   zdialog_run(zd,batch_upright_dialog_event,"parent");                    //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   if (zstat != 1) goto cleanup;                                           //  canceled
   if (! bup_allfiles && ! bup_filecount) goto cleanup;                    //  nothing selected

   free_resources();                                                       //  no curr. file                      v.15.03
   m_viewmode(0,"F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Processing files",500,200,Mwin);               //  status monitor popup window
   
   if (bup_allfiles)                                                       //  "survey all files" selected        v.15.07
   {
      bup_filelist = (char **) zmalloc(maximages * sizeof(char *));
      bup_filecount = 0;

      for (ii = 0; ii < Ntopdirks; ii++)
      {
         err = find_imagefiles(topdirks[ii],allfiles,NAF,0);
         if (err) {
            write_popup_text("write","find_imagefiles() failure");
            goto cleanup;
         }
         
         if (bup_filecount + NAF >= maximages) {
            write_popup_text("write","exceed max images: %d",maximages);
            goto cleanup;
         }

         for (jj = 0; jj < NAF; jj++)
            bup_filelist[bup_filecount+jj] = allfiles[jj];
         bup_filecount += NAF;
      }
   }

   for (ii = 0; ii < bup_filecount; ii++)                                  //  loop selected files
   {
      infile = bup_filelist[ii];                                           //  input file
      write_popup_text("write",infile);                                    //  log each output file
      zmainloop();

      exif_get(infile,exifkey,ppv,1);                                      //  get EXIF: Orientation
      if (! ppv[0]) continue;
      orientation = *ppv[0];
      if (orientation == '1') continue;                                    //  not rotated

      pxmin = PXM_load(infile,0);                                          //  read input file
      if (! pxmin) {
         write_popup_text("write",ZTX("file cannot be read"));
         continue;
      }

      pxmout = 0;
      if (orientation == '6')
         pxmout = PXM_rotate(pxmin,+90);                                   //  rotate clockwise 90 deg.
      if (orientation == '8')
         pxmout = PXM_rotate(pxmin,-90);                                   //  counterclockwise
      if (orientation == '3')
         pxmout = PXM_rotate(pxmin,180);                                   //  180 deg.                           v.15.03

      PXM_free(pxmin);
      if (! pxmout) continue;                                              //  not rotated

      tempfile = zstrdup(infile,12);                                       //  temp file needed for EXIF/IPTC copy
      pp1 = strrchr(infile,'.');
      pp2 = strrchr(tempfile,'.');
      strcpy(pp2,"-temp");
      strcpy(pp2+5,pp1);

      bpc = f_load_bpc;                                                    //  input file bits/color

      if (strmatch(f_load_type,"tif"))
         err = PXM_TIFF_save(pxmout,tempfile,bpc);
      else if (strmatch(f_load_type,"png"))
         err = PXM_PNG_save(pxmout,tempfile,bpc);
      else
         err = PXM_ANY_save(pxmout,tempfile);

      PXM_free(pxmout);

      if (err) {
         write_popup_text("write","*** upright failed");
         zfree(tempfile);
         continue;
      }

      exifdata[0] = "";                                                    //  remove exif:orientation
      exif_copy(infile,tempfile,exifkey,exifdata,1);                       //  (set = 1 does not work)

      err = shell_ack("cp -p \"%s\" \"%s\" ",tempfile,infile);             //  copy temp file to input file
      if (err) write_popup_text("write",wstrerror(err));
      remove(tempfile);                                                    //  remove tempfile
      zfree(tempfile);

      if (! err) {
         load_filemeta(infile);                                            //  update image index
         update_image_index(infile);
         write_popup_text("write","  uprighted");
      }
   }

   write_popup_text("write","COMPLETED");

   if (navi::gallerytype == 1)                                             //  refresh gallery                    v.15.03
      gallery(0,"init");

   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:

   if (bup_filecount) {                                                    //  free memory
      for (ii = 0; ii < bup_filecount; ii++)
         zfree(bup_filelist[ii]);
      zfree(bup_filelist);
      bup_filelist = 0;
      bup_filecount = 0;
   }

   Fblock = 0;
   return;
}


//  dialog event and completion callback function

int batch_upright_dialog_event(zdialog *zd, cchar *event)
{
   char         countmess[50];
   int          ii;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"files"))                                            //  select images to convert
   {
      if (bup_filelist) {                                                  //  free prior list
         for (ii = 0; ii < bup_filecount; ii++)
            zfree(bup_filelist[ii]);
         zfree(bup_filelist);
         bup_filelist = 0;
         bup_filecount = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      bup_filelist = gallery_getfiles();                                   //  get list of files to convert
      zdialog_show(zd,1);

      ii = 0;
      if (bup_filelist)                                                    //  count files selected
         for (ii = 0; bup_filelist[ii]; ii++);
      bup_filecount = ii;

      snprintf(countmess,50,Bfileselected,ii);                             //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
      zdialog_stuff(zd,"allfiles",0);
   }

   if (zd->zstat != 1) return 1;                                           //  wait for [proceed]
   
   zdialog_fetch(zd,"allfiles",bup_allfiles);                              //  get "survey all" option

   if (! bup_allfiles && ! bup_filecount) {                                //  nothing selected
      zmessageACK(Mwin,Bnofileselected);
      zd->zstat = 0;                                                       //  keep dialog active
   }

   if (bup_allfiles && bup_filecount) {
      zmessageACK(Mwin,ZTX("cannot select both options"));
      zd->zstat = 0;
   }

   return 1;
}


/**************************************************************************/

//  Batch delete or trash image files.

char     **bdt_filelist;
int      bdt_filecount;
int      bdt_option;                                                       //  1/2 = delete/trash

void m_batch_deltrash(GtkWidget *, cchar *)                                //  v.15.07
{
   int  batch_deltrash_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         zstat, ii, err, gstat;
   char        *file;
   GError      *gerror = 0;
   GFile       *gfile = 0;
   STATB       statb;

   F1_help_topic = "batch_delete_trash";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       ___________________________________
      |       Batch Delete/Trash          |
      |                                   |
      |  [Select Files]  N files selected |
      |  (o) delete    (o) trash          |
      |                                   |
      |               [proceed] [cancel]  |
      |___________________________________|

***/

   zd = zdialog_new(ZTX("Batch Delete/Trash"),Mwin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf",Bnofileselected,"space=10");
   zdialog_add_widget(zd,"hbox","hbdt","dialog");
   zdialog_add_widget(zd,"label","labdel","hbdt",ZTX("delete"),"space=5");
   zdialog_add_widget(zd,"radio","delete","hbdt",0);
   zdialog_add_widget(zd,"label","space","hbdt",0,"space=10");
   zdialog_add_widget(zd,"label","labtrash","hbdt",ZTX("trash"),"space=5");
   zdialog_add_widget(zd,"radio","trash","hbdt",0);
   
   bdt_filelist = 0;
   bdt_filecount = 0;
   bdt_option = 2;

   zdialog_stuff(zd,"delete",0);
   zdialog_stuff(zd,"trash",1);

   zdialog_run(zd,batch_deltrash_dialog_event,"parent");                   //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   
   zdialog_fetch(zd,"delete",bdt_option);                                  //  get delete/trash option
   if (! bdt_option) bdt_option = 2;

   zdialog_free(zd);
   if (zstat != 1) goto cleanup;                                           //  canceled
   if (! bdt_filecount) goto cleanup;

   free_resources();                                                       //  no curr. file

   m_viewmode(0,"F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Processing files",500,200,Mwin);               //  status monitor popup window

   for (ii = 0; ii < bdt_filecount; ii++)                                  //  loop selected files
   {
      file = bdt_filelist[ii];                                             //  log each file
      write_popup_text("write",file);
      zmainloop();
   
      err = stat(file,&statb);                                             //  file exists?
      if (err || ! S_ISREG(statb.st_mode)) {
         write_popup_text("write","file not found");
         continue;
      }

      if (bdt_option == 1) {                                               //  delete file
         err = remove(file);
         if (err) {
            write_popup_text("write",wstrerror(err));
            continue;
         }
      }
      
      if (bdt_option == 2) {                                               //  move file to trash
         gfile = g_file_new_for_path(file);
         gstat = g_file_trash(gfile,0,&gerror);                            //  move file to trash
         g_object_unref(gfile);
         if (! gstat) {
            write_popup_text("write",gerror->message);
            continue;
         }
      }

      delete_image_index(file);                                            //  delete file in image index
   }

   write_popup_text("write","COMPLETED");

cleanup:

   if (bdt_filecount) {                                                    //  free memory
      for (ii = 0; ii < bdt_filecount; ii++)
         zfree(bdt_filelist[ii]);
      zfree(bdt_filelist);
      bdt_filelist = 0;
      bdt_filecount = 0;
   }

   if (navi::gallerytype == 1)                                             //  refresh gallery                    v.15.03
      gallery(0,"init");

   gallery_monitor("start");                                               //  resume tracking gallery updates

   Fblock = 0;
   return;
}


//  dialog event and completion callback function

int batch_deltrash_dialog_event(zdialog *zd, cchar *event)
{
   char         countmess[50];
   int          ii;
   
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"files"))                                            //  select images to convert
   {
      if (bdt_filelist) {                                                  //  free prior list
         for (ii = 0; ii < bdt_filecount; ii++)
            zfree(bdt_filelist[ii]);
         zfree(bdt_filelist);
         bdt_filelist = 0;
         bdt_filecount = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      bdt_filelist = gallery_getfiles();                                   //  get list of files to convert
      zdialog_show(zd,1);

      ii = 0;
      if (bdt_filelist)                                                    //  count files selected
         for (ii = 0; bdt_filelist[ii]; ii++);
      bdt_filecount = ii;

      snprintf(countmess,50,Bfileselected,ii);                             //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (strmatch(event,"delete")) {                                         //  delete radio button
      zdialog_fetch(zd,"delete",ii);
      if (ii) bdt_option = 1;
      zdialog_stuff(zd,"trash",0);
   }

   if (strmatch(event,"trash")) {                                          //  trash radio button
      zdialog_fetch(zd,"trash",ii);
      if (ii) bdt_option = 2;
      zdialog_stuff(zd,"delete",0);
   }

   if (zd->zstat != 1) return 1;                                           //  wait for [proceed]

   if (! bdt_filecount) {                                                  //  nothing selected
      zmessageACK(Mwin,Bnofileselected);
      zd->zstat = 0;                                                       //  keep dialog active
   }

   return 1;
}


/**************************************************************************/

//  convert multiple RAW files to tiff, jpeg, or png using DCraw

namespace batch_dcraw
{
   char     location[400];
   char     **filelist;
   int      filecount;
   char     dcraw_command[100] = "";
   cchar    *filetype;
   int      bpc;
   float    resize;
   int      Fsharpen;
   int      amount, thresh;
};


void  m_batch_dcraw(GtkWidget *, cchar *menu)
{
   using namespace batch_dcraw;

   int  batch_dcraw_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   cchar       *title = ZTX("Batch Convert RAW Files (DCraw)");
   char        *tiffile, *infile, *outfile, *pp;
   int         zstat, ii, err, ftype;
   int         cc, ww2, hh2;
   PXM         *pxm1, *pxm2;

   F1_help_topic = "batch_raw";

   if (! Fdcraw) {                                                         //  v.14.07
      zmessageACK(Mwin,ZTX("DCraw not installed"));
      return;
   }

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       _____________________________________________________________________________
      | [x] [-] [_]       Batch Convert RAW Files (DCraw)                           |
      |                                                                             |
      | [Select Files]  N files selected                                            |
      | output location [_________________________________________________]         |
      | output file type  (o) JPEG  (o) PNG-8  (o) PNG-16  (o) TIFF-8  (o) TIFF-16  |
      | resize  (o) 1.0  (o) 3/4  (o) 2/3  (o) 1/2  (o) 1/3                         |
      | [x] Sharpen   amount [___]   threshold [___]                                |
      | --------------------------------------------------------------------------- |
      | white balance  (o) camera  (o) fixed  (o) calculated                        |
      | interpolation  (o) bilinear  (o) VNG  (o) PPG  (o) AHD                      |
      | color space    (o) raw  (o) sRGB  (o) Adobe  (o) wide  (o) Kodak  (o) XYZ   |
      | gamma curve    (o) default  (o) sRGB  (o) linear                            |
      | [ dcraw -T -6 -w -q 0 -o 1 -g 2.222 4.5       ]   [defaults]                |
      |                                                                             |
      |                                                         [proceed] [cancel]  |
      |_____________________________________________________________________________|

***/

   zd = zdialog_new(title,Mwin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hb1",Bnofileselected,"space=10");

   zdialog_add_widget(zd,"hbox","hbout","dialog");
   zdialog_add_widget(zd,"label","labout","hbout",ZTX("output location"),"space=5");
   zdialog_add_widget(zd,"entry","location","hbout",0,"space=5|expand");
   zdialog_add_widget(zd,"button","browse","hbout",Bbrowse,"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"label","labtype","hb2",ZTX("output file type"),"space=5");
   zdialog_add_widget(zd,"radio","JPEG","hb2","JPEG","space=5");
   zdialog_add_widget(zd,"radio","PNG-8","hb2","PNG-8","space=5");
   zdialog_add_widget(zd,"radio","PNG-16","hb2","PNG-16","space=5");
   zdialog_add_widget(zd,"radio","TIFF-8","hb2","TIFF-8","space=5");
   zdialog_add_widget(zd,"radio","TIFF-16","hb2","TIFF-16","space=5");

   zdialog_add_widget(zd,"hbox","hbsize","dialog");                                 //  v.14.06
   zdialog_add_widget(zd,"label","labsize","hbsize",ZTX("resize"),"space=5");
   zdialog_add_widget(zd,"label","space","hbsize",0,"space=5");
   zdialog_add_widget(zd,"radio","1.0","hbsize","1.0","space=5");
   zdialog_add_widget(zd,"radio","3/4","hbsize","3/4","space=5");
   zdialog_add_widget(zd,"radio","2/3","hbsize","2/3","space=5");
   zdialog_add_widget(zd,"radio","1/2","hbsize","1/2","space=5");
   zdialog_add_widget(zd,"radio","1/3","hbsize","1/3","space=5");

   zdialog_add_widget(zd,"hbox","hbsharp","dialog");                                //  v.14.06
   zdialog_add_widget(zd,"check","sharpen","hbsharp",ZTX("Sharpen"),"space=3");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labamount","hbsharp",ZTX("amount"),"space=3");
   zdialog_add_widget(zd,"spin","amount","hbsharp","0|400|1|100");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labthresh","hbsharp",ZTX("threshold"),"space=3");
   zdialog_add_widget(zd,"spin","thresh","hbsharp","0|100|1|20");

   zdialog_add_widget(zd,"hsep","hsep1","dialog",0,"space=5");

   zdialog_add_widget(zd,"hbox","hbvb","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hbvb",0,"space=5|homog");
   zdialog_add_widget(zd,"vbox","vb2","hbvb",0,"homog");

   zdialog_add_widget(zd,"label","labwbal","vb1",ZTX("white balance"));
   zdialog_add_widget(zd,"label","labinterp","vb1",ZTX("interpolation"));
   zdialog_add_widget(zd,"label","labcolor","vb1",ZTX("color space"));
   zdialog_add_widget(zd,"label","labgamma","vb1",ZTX("gamma curve"));

   zdialog_add_widget(zd,"hbox","hb3","vb2");
   zdialog_add_widget(zd,"radio","-w","hb3",ZTX("camera"),"space=3");
   zdialog_add_widget(zd,"radio","-W","hb3",ZTX("fixed"),"space=3");
   zdialog_add_widget(zd,"radio","-a","hb3",ZTX("calculated"),"space=3");

   zdialog_add_widget(zd,"hbox","hb4","vb2");
   zdialog_add_widget(zd,"radio","-q 0","hb4","bilinear","space=3");
   zdialog_add_widget(zd,"radio","-q 1","hb4","VNG","space=3");
   zdialog_add_widget(zd,"radio","-q 2","hb4","PPG","space=3");
   zdialog_add_widget(zd,"radio","-q 3","hb4","AHD","space=3");

   zdialog_add_widget(zd,"hbox","hb5","vb2");
   zdialog_add_widget(zd,"radio","-o 0","hb5","raw","space=3");
   zdialog_add_widget(zd,"radio","-o 1","hb5","sRGB","space=3");
   zdialog_add_widget(zd,"radio","-o 2","hb5","Adobe","space=3");
   zdialog_add_widget(zd,"radio","-o 3","hb5","Wide","space=3");
   zdialog_add_widget(zd,"radio","-o 4","hb5","Kodak","space=3");
   zdialog_add_widget(zd,"radio","-o 5","hb5","XYZ","space=3");

   zdialog_add_widget(zd,"hbox","hb6","vb2");
   zdialog_add_widget(zd,"radio","-g 2.222 4.5","hb6",ZTX("default"),"space=3");
   zdialog_add_widget(zd,"radio","-g 2.4 12.92","hb6","sRGB","space=3");
   zdialog_add_widget(zd,"radio","-g 1.0 1.0","hb6","linear","space=3");

   zdialog_add_widget(zd,"hbox","hb7","dialog",0,"space=3");
   zdialog_add_widget(zd,"entry","dcraw_command","hb7",0,"space=5|expand");
   zdialog_add_widget(zd,"button","defaults","hb7",ZTX("defaults"),"space=5");

   filelist = 0;                                                           //  set no files selected
   filecount = 0;
   *location = 0;

   zdialog_stuff(zd,"JPEG",1);                                             //  compensate GTK bug
   batch_dcraw_dialog_event(zd,"defaults");                                //  set defaults
   zdialog_restore_inputs(zd);                                             //  get prior inputs if any

   zdialog_run(zd,batch_dcraw_dialog_event,"save");                        //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   if (zstat != 1) goto cleanup;
   if (! filecount) goto cleanup;

   printz("dcraw command: %s \n",dcraw_command);

   m_viewmode(0,"F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Converting RAW files",500,200,Mwin);           //  status monitor popup window
   zmainloop();

   rawfile = 0;

   for (ii = 0; ii < filecount; ii++)                                      //  loop all RAW files
   {
      rawfile = filelist[ii];                                              //  filename.raw

      write_popup_text("write",rawfile);                                   //  write input raw file to log window
      zmainloop();

      ftype = image_file_type(rawfile);
      if (ftype != 3) {
         write_popup_text("write"," *** unknown RAW file type");
         continue;
      }

      err = shell_ack("%s \"%s\" ",dcraw_command,rawfile);                 //  convert filename.raw to filename.tiff
      if (err) {
         write_popup_text("write"," *** raw conversion failed");           //  conversion failed
         write_popup_text("write",wstrerror(err));
         zmainloop();
         continue;
      }

      tiffile = raw_to_tiff(rawfile);                                      //  converted file

      pxm1 = PXM_load(tiffile,0);                                          //  load tiff-16 output file
      if (! pxm1) {
         write_popup_text("write"," *** no dcraw .tiff output");
         zmainloop();
         zfree(tiffile);
         continue;
      }

      if (resize < 1.0)                                                    //  resize down if wanted
      {
         ww2 = resize * pxm1->ww;
         hh2 = resize * pxm1->hh;
         pxm2 = PXM_rescale(pxm1,ww2,hh2);
         PXM_free(pxm1);
         pxm1 = pxm2;

         if (! pxm1) {
            write_popup_text("write"," *** resize failed");
            zmainloop();
            zfree(tiffile);
            continue;
         }
      }

      if (Fsharpen)                                                        //  sharpen if wanted
         batch_sharp_func(pxm1,amount,thresh);

      outfile = zstrdup(tiffile);                                          //  have filename.tiff, 16 bits/color

      pp = strrchr(outfile,'.');                                           //  make *.jpg or *.png if wanted
      if (pp) strcpy(pp+1,filetype);

      if (strmatchN(filetype,"tif",3))                                     //  output .tif file, 8 bits/color
         err = PXM_TIFF_save(pxm1,outfile,bpc);                            //    or .tiff file, 16 bits/color

      else if (strmatch(filetype,"png"))                                   //  output .png file, 8/16 bits
         err = PXM_PNG_save(pxm1,outfile,bpc);

      else err = PXM_ANY_save(pxm1,outfile);                               //  output .jpg file, 8 bits

      if (err) {
         write_popup_text("write"," *** file type conversion failed");
         zmainloop();
         zfree(tiffile);
         zfree(outfile);
         continue;
      }

      if (! strmatch(outfile,tiffile))                                     //  if conversion was made,
         remove(tiffile);                                                  //    delete tiff-16 file

      exif_copy(rawfile,outfile,0,0,0);                                    //  copy metadata from RAW file

      if (*location && ! samedirk(location,outfile)) {
         infile = zstrdup(outfile);                                        //  copy to new location
         zfree(outfile);
         pp = strrchr(infile,'/');                                         //  /raw-location/filename.ext
         cc = strlen(pp);                                                  //               |
         outfile = zstrdup(location,cc+1);                                 //               pp
         strcat(outfile,pp);                                               //  /new-location/filename.ext
         err = shell_ack("cp -p \"%s\" \"%s\" ",infile,outfile);           //  copy to new location
         if (err) write_popup_text("write",wstrerror(err));
         remove(infile);                                                   //  remove tempfile
         zfree(infile);
      }

      f_open(outfile,0,0,0);                                               //  open converted file                v.15.03
      update_image_index(outfile);

      write_popup_text("write",outfile);                                   //  write output file to log window
      zfree(outfile);
      zfree(tiffile);
      zmainloop();
   }

   write_popup_text("write","COMPLETED");

   gallery(curr_dirk,"init");                                              //  update gallery file list           v.15.03
   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:                                                                   //  clean up and return

   if (filecount) {
      for (ii = 0; ii < filecount; ii++)                                   //  free memory
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
      filecount = 0;
   }

   rawfile = 0;
   Fblock = 0;
   return;
}


//  dialog event and completion callback function

int batch_dcraw_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batch_dcraw;

   int            ii, Fupdate = 0, err, cc;
   char           countmess[50], *ploc;
   static cchar   *wbal = 0, *intp = 0, *color = 0, *gamma = 0;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (! wbal) {                                                           //  get current options if not already
      wbal = "-w";
      zdialog_fetch(zd,"-W",ii);
      if (ii) wbal = "-W";
      zdialog_fetch(zd,"-a",ii);
      if (ii) wbal = "-a";
      Fupdate++;
   }

   if (! intp) {
      intp = "-q 0";
      zdialog_fetch(zd,"-q 1",ii);
      if (ii) intp = "-q 1";
      zdialog_fetch(zd,"-q 2",ii);
      if (ii) intp = "-q 2";
      zdialog_fetch(zd,"-q 3",ii);
      if (ii) intp = "-q 3";
      Fupdate++;
   }

   if (! color) {
      color = "-o 1";
      zdialog_fetch(zd,"-o 0",ii);
      if (ii) color = "-o 0";
      zdialog_fetch(zd,"-o 2",ii);
      if (ii) color = "-o 2";
      zdialog_fetch(zd,"-o 3",ii);
      if (ii) color = "-o 3";
      zdialog_fetch(zd,"-o 4",ii);
      if (ii) color = "-o 4";
      zdialog_fetch(zd,"-o 5",ii);
      if (ii) color = "-o 5";
      Fupdate++;
   }

   if (! gamma) {
      gamma = "-g 2.222 4.5";
      zdialog_fetch(zd,"-g 2.4 12.92",ii);
      if (ii) gamma = "-g 2.4 12.92";
      zdialog_fetch(zd,"-g 1.0 1.0",ii);
      if (ii) gamma = "-g 1.0 1.0";
      Fupdate++;
   }

   if (strmatch(event,"files"))                                            //  select images to convert
   {
      if (filecount) {                                                     //  free prior list
         for (ii = 0; ii < filecount; ii++)
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
         filecount = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      ii = 0;
      if (filelist)                                                        //  count files in list
         for (ii = 0; filelist[ii]; ii++);
      filecount = ii;

      snprintf(countmess,50,Bfileselected,ii);                             //  stuff count into dialog
      zdialog_stuff(zd,"fcount",countmess);
   }

   if (strmatch(event,"browse")) {
      ploc = zgetfile(ZTX("Select directory"),MWIN,"folder",curr_dirk);    //  new location browse
      if (! ploc) return 1;
      zdialog_stuff(zd,"location",ploc);
      zfree(ploc);
   }

   if (strmatch(event,"defaults"))                                         //  default all parameters
   {
      zdialog_stuff(zd,"JPEG",0);
      zdialog_stuff(zd,"PNG-8",0);
      zdialog_stuff(zd,"PNG-16",0);
      zdialog_stuff(zd,"TIFF-8",0);
      zdialog_stuff(zd,"TIFF-16",1);                                       //  output = TIFF-16

      zdialog_stuff(zd,"-w",1);                                            //  color balance = camera value
      zdialog_stuff(zd,"-W",0);
      zdialog_stuff(zd,"-a",0);
      wbal = "-w";

      zdialog_stuff(zd,"-q 0",1);                                          //  interpolation = bilinear
      zdialog_stuff(zd,"-q 1",0);
      zdialog_stuff(zd,"-q 2",0);
      zdialog_stuff(zd,"-q 3",0);
      intp = "-q 0";

      zdialog_stuff(zd,"-o 0",0);
      zdialog_stuff(zd,"-o 1",1);                                          //  color space = sRGB
      zdialog_stuff(zd,"-o 2",0);
      zdialog_stuff(zd,"-o 3",0);
      zdialog_stuff(zd,"-o 4",0);
      zdialog_stuff(zd,"-o 5",0);
      color = "-o 1";

      zdialog_stuff(zd,"-g 2.222 4.5",1);                                  //  gamma = default
      zdialog_stuff(zd,"-g 2.4 12.92",0);
      zdialog_stuff(zd,"-g 1.0 1.0",0);
      gamma = "-g 2.222 4.5";

      Fupdate++;
   }

   if (strstr("-w -W -a",event)) {                                         //  event is one of: -w -W -a
      zdialog_fetch(zd,event,ii);
      if (ii) {
         wbal = event;
         Fupdate++;
      }
   }

   if (strstr(event,"-q")) {                                               //  event is: -q N
      zdialog_fetch(zd,event,ii);
      if (ii) {
         intp = event;
         Fupdate++;
      }
   }

   if (strstr(event,"-o")) {                                               //  event is: -o N
      zdialog_fetch(zd,event,ii);
      if (ii) {
         color = event;
         Fupdate++;
      }
   }

   if (strstr(event,"-g")) {                                               //  event is: -g N1 N2
      zdialog_fetch(zd,event,ii);
      if (ii) {
         gamma = event;
         Fupdate++;
      }
   }

   if (Fupdate || ! *dcraw_command)                                        //  initialize or update dcraw command
   {
      snprintf(dcraw_command,100,"dcraw -T -6 %s %s %s %s",wbal,intp,color,gamma);
      zdialog_stuff(zd,"dcraw_command",dcraw_command);
   }

   if (zd->zstat != 1) return 0;                                           //  wait for [proceed]

   zd->zstat = 0;                                                          //  keep dialog active until inputs OK

   if (! filecount) {
      zmessageACK(Mwin,Bnofileselected);
      return 1;
   }

   zdialog_fetch(zd,"location",location,400);                              //  output location
   strTrim2(location);
   if (! blank_null(location)) {
      cc = strlen(location) - 1;
      if (location[cc] == '/') location[cc] = 0;                           //  remove trailing '/'
      err = check_create_dir(location);                                    //  create if needed
      if (err) return 1;
   }

   zdialog_fetch(zd,"JPEG",ii);                                            //  file type, color depth
   if (ii) { filetype = "jpg"; bpc = 8; }
   zdialog_fetch(zd,"PNG-8",ii);
   if (ii) { filetype = "png"; bpc = 8; }
   zdialog_fetch(zd,"PNG-16",ii);
   if (ii) { filetype = "png"; bpc = 16; }
   zdialog_fetch(zd,"TIFF-8",ii);
   if (ii) { filetype = "tif"; bpc = 8; }
   zdialog_fetch(zd,"TIFF-16",ii);
   if (ii) { filetype = "tiff"; bpc = 16; }

   resize = 1.0;
   zdialog_fetch(zd,"1.0",ii);                                             //  resize option                      v.14.06
   if (ii) resize = 1.0;
   zdialog_fetch(zd,"3/4",ii);
   if (ii) resize = 0.75;
   zdialog_fetch(zd,"2/3",ii);
   if (ii) resize = 0.666667;
   zdialog_fetch(zd,"1/2",ii);
   if (ii) resize = 0.50;
   zdialog_fetch(zd,"1/3",ii);
   if (ii) resize = 0.333333;

   zdialog_fetch(zd,"sharpen",Fsharpen);                                   //  sharpen option                     v.14.06
   zdialog_fetch(zd,"amount",amount);
   zdialog_fetch(zd,"thresh",thresh);

   zd->zstat = 1;                                                          //  dialog complete
   return 1;
}


/**************************************************************************/

//  convert multiple RAW files to tiff, jpeg, or png using Raw Therapee

namespace batch_rawtherapee
{
   char     location[400];
   char     **filelist;
   int      filecount;
   char     rawtherapee_command[100] = "";
   cchar    *filetype;
   int      bpc;
   float    resize;
   int      Fsharpen;
   int      amount, thresh;
};


void  m_batch_rawtherapee(GtkWidget *, cchar *menu)                        //  v.14.07
{
   using namespace batch_rawtherapee;

   int  batch_rawtherapee_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   cchar       *title = ZTX("Batch Convert RAW Files (Raw Therapee)");
   char        *tiffile, *infile, *outfile, *pp;
   int         zstat, ii, err, ftype;
   int         cc, ww2, hh2;
   PXM         *pxm1, *pxm2;

   F1_help_topic = "batch_raw";

   if (! Frawtherapee) {                                                   //  v.14.07
      zmessageACK(Mwin,ZTX("Raw Therapee not installed"));
      return;
   }

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       _____________________________________________________________________________
      | [x] [-] [_]       Batch Convert RAW Files (Raw Therapee)                    |
      |                                                                             |
      | [Select Files]  N files selected                                            |
      | output location [_________________________________________________]         |
      | output file type  (o) JPEG  (o) PNG-8  (o) PNG-16  (o) TIFF-8  (o) TIFF-16  |
      | resize  (o) 1.0  (o) 3/4  (o) 2/3  (o) 1/2  (o) 1/3                         |
      | [x] Sharpen   amount [___]   threshold [___]                                |
      |                                                         [proceed] [cancel]  |
      |_____________________________________________________________________________|

***/

   zd = zdialog_new(title,Mwin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hb1",Bnofileselected,"space=10");

   zdialog_add_widget(zd,"hbox","hbout","dialog");
   zdialog_add_widget(zd,"label","labout","hbout",ZTX("output location"),"space=5");
   zdialog_add_widget(zd,"entry","location","hbout",0,"space=5|expand");
   zdialog_add_widget(zd,"button","browse","hbout",Bbrowse,"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"label","labtype","hb2",ZTX("output file type"),"space=5");
   zdialog_add_widget(zd,"radio","JPEG","hb2","JPEG","space=5");
   zdialog_add_widget(zd,"radio","PNG-8","hb2","PNG-8","space=5");
   zdialog_add_widget(zd,"radio","PNG-16","hb2","PNG-16","space=5");
   zdialog_add_widget(zd,"radio","TIFF-8","hb2","TIFF-8","space=5");
   zdialog_add_widget(zd,"radio","TIFF-16","hb2","TIFF-16","space=5");

   zdialog_add_widget(zd,"hbox","hbsize","dialog");                                 //  v.14.06
   zdialog_add_widget(zd,"label","labsize","hbsize",ZTX("resize"),"space=5");
   zdialog_add_widget(zd,"label","space","hbsize",0,"space=5");
   zdialog_add_widget(zd,"radio","1.0","hbsize","1.0","space=5");
   zdialog_add_widget(zd,"radio","3/4","hbsize","3/4","space=5");
   zdialog_add_widget(zd,"radio","2/3","hbsize","2/3","space=5");
   zdialog_add_widget(zd,"radio","1/2","hbsize","1/2","space=5");
   zdialog_add_widget(zd,"radio","1/3","hbsize","1/3","space=5");

   zdialog_add_widget(zd,"hbox","hbsharp","dialog");                                //  v.14.06
   zdialog_add_widget(zd,"check","sharpen","hbsharp",ZTX("Sharpen"),"space=3");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labamount","hbsharp",ZTX("amount"),"space=3");
   zdialog_add_widget(zd,"spin","amount","hbsharp","0|400|1|100");
   zdialog_add_widget(zd,"label","space","hbsharp",0,"space=8");
   zdialog_add_widget(zd,"label","labthresh","hbsharp",ZTX("threshold"),"space=3");
   zdialog_add_widget(zd,"spin","thresh","hbsharp","0|100|1|20");

   filelist = 0;                                                           //  set no files selected
   filecount = 0;
   *location = 0;

   zdialog_stuff(zd,"JPEG",1);                                             //  compensate GTK bug
   zdialog_restore_inputs(zd);                                             //  get prior inputs if any

   zdialog_run(zd,batch_rawtherapee_dialog_event,"save");                  //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   if (zstat != 1) goto cleanup;
   if (! filecount) goto cleanup;

   m_viewmode(0,"F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Converting RAW files",500,200,Mwin);           //  status monitor popup window
   zmainloop();

   rawfile = 0;

   for (ii = 0; ii < filecount; ii++)                                      //  loop all RAW files
   {
      rawfile = filelist[ii];                                              //  filename.raw

      write_popup_text("write",rawfile);                                   //  write input raw file to log window
      zmainloop();

      ftype = image_file_type(rawfile);
      if (ftype != 3) {
         write_popup_text("write"," *** unknown RAW file type");
         continue;
      }

      snprintf(rawtherapee_command,100,"rawtherapee -t -Y -c ");

      err = shell_ack("%s \"%s\" ",rawtherapee_command,rawfile);           //  convert filename.raw to filename.tiff
      if (err) {
         write_popup_text("write"," *** raw conversion failed");           //  conversion failed
         write_popup_text("write",wstrerror(err));
         zmainloop();
         continue;
      }

      outfile = zstrdup(rawfile,5);                                        //  output is rawfilename.tif
      pp = strrchr(outfile,'.');
      strcpy(pp,".tif");
      tiffile = raw_to_tiff(rawfile);                                      //  rawfilename.tiff
      rename(outfile,tiffile);
      zfree(outfile);

      pxm1 = PXM_load(tiffile,0);                                          //  load tiff-16 output file
      if (! pxm1) {
         write_popup_text("write"," *** no Raw Therapee .tiff output");
         zmainloop();
         zfree(tiffile);
         continue;
      }

      if (resize < 1.0)                                                    //  resize down if wanted
      {
         ww2 = resize * pxm1->ww;
         hh2 = resize * pxm1->hh;
         pxm2 = PXM_rescale(pxm1,ww2,hh2);
         PXM_free(pxm1);
         pxm1 = pxm2;

         if (! pxm1) {
            write_popup_text("write"," *** resize failed");
            zmainloop();
            zfree(tiffile);
            continue;
         }
      }

      if (Fsharpen)                                                        //  sharpen if wanted
         batch_sharp_func(pxm1,amount,thresh);

      outfile = zstrdup(tiffile);                                          //  have filename.tiff, 16 bits/color

      pp = strrchr(outfile,'.');                                           //  make *.jpg or *.png if wanted
      if (pp) strcpy(pp+1,filetype);

      if (strmatchN(filetype,"tif",3))                                     //  output .tif file, 8 bits/color
         err = PXM_TIFF_save(pxm1,outfile,bpc);                            //    or .tiff file, 16 bits/color

      else if (strmatch(filetype,"png"))                                   //  output .png file, 8/16 bits
         err = PXM_PNG_save(pxm1,outfile,bpc);

      else err = PXM_ANY_save(pxm1,outfile);                               //  output .jpg file, 8 bits

      if (err) {
         write_popup_text("write"," *** file type conversion failed");
         zmainloop();
         zfree(tiffile);
         zfree(outfile);
         continue;
      }

      if (! strmatch(outfile,tiffile))                                     //  if conversion was made,
         remove(tiffile);                                                  //    delete tiff-16 file

      exif_copy(rawfile,outfile,0,0,0);                                    //  copy metadata from RAW file

      if (*location && ! samedirk(location,outfile)) {
         infile = zstrdup(outfile);                                        //  copy to new location
         zfree(outfile);
         pp = strrchr(infile,'/');                                         //  /raw-location/filename.ext
         cc = strlen(pp);                                                  //               |
         outfile = zstrdup(location,cc+1);                                 //               pp
         strcat(outfile,pp);                                               //  /new-location/filename.ext
         err = shell_ack("cp -p \"%s\" \"%s\" ",infile,outfile);           //  copy to new location
         if (err) write_popup_text("write",wstrerror(err));
         remove(infile);                                                   //  remove tempfile
         zfree(infile);
      }

      f_open(outfile,0,0,0);                                               //  open converted file                v.15.03
      update_image_index(outfile);

      write_popup_text("write",outfile);                                   //  write output file to log window
      zfree(outfile);
      zfree(tiffile);
      zmainloop();
   }

   write_popup_text("write","COMPLETED");

   gallery(curr_dirk,"init");                                              //  update gallery file list           v.15.03
   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:                                                                   //  clean up and return

   if (filecount) {
      for (ii = 0; ii < filecount; ii++)                                   //  free memory
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
      filecount = 0;
   }

   rawfile = 0;
   Fblock = 0;
   return;
}


//  dialog event and completion callback function

int batch_rawtherapee_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batch_rawtherapee;

   int            ii, err, cc;
   char           countmess[50], *ploc;

   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   
   if (strmatch(event,"files"))                                            //  select images to convert
   {
      if (filecount) {                                                     //  free prior list
         for (ii = 0; ii < filecount; ii++)
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
         filecount = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      ii = 0;
      if (filelist)                                                        //  count files in list
         for (ii = 0; filelist[ii]; ii++);
      filecount = ii;

      snprintf(countmess,50,Bfileselected,ii);                             //  stuff count into dialog
      zdialog_stuff(zd,"fcount",countmess);
   }

   if (strmatch(event,"browse")) {
      ploc = zgetfile(ZTX("Select directory"),MWIN,"folder",curr_dirk);    //  new location browse
      if (! ploc) return 1;
      zdialog_stuff(zd,"location",ploc);
      zfree(ploc);
   }

   if (zd->zstat != 1) return 0;                                           //  wait for [proceed]

   zd->zstat = 0;                                                          //  keep dialog active until inputs OK

   if (! filecount) {
      zmessageACK(Mwin,Bnofileselected);
      return 1;
   }

   zdialog_fetch(zd,"location",location,400);                              //  output location
   strTrim2(location);
   if (! blank_null(location)) {
      cc = strlen(location) - 1;
      if (location[cc] == '/') location[cc] = 0;                           //  remove trailing '/'
      err = check_create_dir(location);                                    //  create if needed
      if (err) return 1;
   }

   zdialog_fetch(zd,"JPEG",ii);                                            //  file type, color depth
   if (ii) { filetype = "jpg"; bpc = 8; }
   zdialog_fetch(zd,"PNG-8",ii);
   if (ii) { filetype = "png"; bpc = 8; }
   zdialog_fetch(zd,"PNG-16",ii);
   if (ii) { filetype = "png"; bpc = 16; }
   zdialog_fetch(zd,"TIFF-8",ii);
   if (ii) { filetype = "tif"; bpc = 8; }
   zdialog_fetch(zd,"TIFF-16",ii);
   if (ii) { filetype = "tiff"; bpc = 16; }

   resize = 1.0;
   zdialog_fetch(zd,"1.0",ii);                                             //  resize option                      v.14.06
   if (ii) resize = 1.0;
   zdialog_fetch(zd,"3/4",ii);
   if (ii) resize = 0.75;
   zdialog_fetch(zd,"2/3",ii);
   if (ii) resize = 0.666667;
   zdialog_fetch(zd,"1/2",ii);
   if (ii) resize = 0.50;
   zdialog_fetch(zd,"1/3",ii);
   if (ii) resize = 0.333333;

   zdialog_fetch(zd,"sharpen",Fsharpen);                                   //  sharpen option                     v.14.06
   zdialog_fetch(zd,"amount",amount);
   zdialog_fetch(zd,"thresh",thresh);

   zd->zstat = 1;                                                          //  dialog complete
   return 1;
}


/**************************************************************************/

//  find all duplicated files and create corresponding gallery of duplicates

void m_duplicates(GtkWidget *, const char *)                               //  v.14.09
{
   int  duplicates_dialog_event(zdialog *zd, cchar *event);

   PIXBUF      *pxb;
   GError      *gerror;
   FILE        *fid = 0;
   int         Nfiles = 0, Ndups = 0;                                      //  image file and duplicate counts
   char        **files = 0;                                                //  image file list
   char        **thumbs = 0;                                               //  64x64 thumbnail images
   int         *tcc = 0;                                                   //  thumbnail actual width x height
   uint8       *pix;
   zdialog     *zd;
   int         thumbsize, pixdiff, pixcount;
   char        text[100], gfile[100], *pp;
   int         zstat, ii, jj, kk, cc, err, ww, hh, ndup;
   int         m1, m2, sum, diff;

   F1_help_topic = "find_duplicates";

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

/***
       _________________________________________
      |         Find Duplicate Images           |
      |                                         |
      |  thumbnail size [ 64 |-+]               |
      |  pixel difference [ 2 |-+]              |
      |  pixel count [ 2 |-+]                   |
      |  Images: nnnnnn   Duplicates: nnn       |
      |  /topdirk/subdirk1/subdirk2/...         |
      |  imagefile.jpg                          |
      |                      [proceed] [cancel] |
      |_________________________________________|

***/

   zd = zdialog_new(ZTX("Find Duplicate Images"),Mwin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hbthumb","dialog");
   zdialog_add_widget(zd,"label","labthumb","hbthumb",ZTX("Thumbnail size"));
   zdialog_add_widget(zd,"spin","thumbsize","hbthumb","32|256|32|64","space=3");
   zdialog_add_widget(zd,"hbox","hbdiff","dialog");
   zdialog_add_widget(zd,"label","labdiff","hbdiff",ZTX("pixel difference"));
   zdialog_add_widget(zd,"spin","pixdiff","hbdiff","1|20|1|2","space=3");
   zdialog_add_widget(zd,"hbox","hbsum","dialog");
   zdialog_add_widget(zd,"label","labsum","hbsum",ZTX("pixel count"));
   zdialog_add_widget(zd,"spin","pixcount","hbsum","1|999|1|10","space=3");
   zdialog_add_widget(zd,"hbox","hbcount","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcount1","hbcount",ZTX("Images:"));
   zdialog_add_widget(zd,"label","labcount2","hbcount",ZTX("searching ..."));
   zdialog_add_widget(zd,"label","space","hbcount",0,"space=5");
   zdialog_add_widget(zd,"label","labdup1","hbcount",ZTX("Duplicates:"));
   zdialog_add_widget(zd,"label","labdup2","hbcount","0");
   zdialog_add_widget(zd,"hbox","hbdirk","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labdirk","hbdirk"," ","space=5");
   zdialog_add_widget(zd,"hbox","hbfile","dialog");
   zdialog_add_widget(zd,"label","labfile","hbfile"," ","space=5");

   zdialog_restore_inputs(zd);
   zdialog_resize(zd,300,0);
   zdialog_run(zd,duplicates_dialog_event);

   err = find_imagefiles(thumbdirk,files,Nfiles,1);                        //  get all thumbnail files
   if (err) {
      zmessageACK(Mwin,strerror(errno));
      goto cleanup;
   }

   if (Nfiles < 10) {
      zmessageACK(Mwin,ZTX("only %d image thumbnails found"),Nfiles);
      goto cleanup;
   }

   snprintf(text,100,"%d",Nfiles);                                         //  stuff thumbnail count in dialog
   zdialog_stuff(zd,"labcount2",text);

   zstat = zdialog_wait(zd);                                               //  wait for user inputs
   if (zstat != 1) goto cleanup;

   zdialog_fetch(zd,"thumbsize",thumbsize);                                //  thumbnail size to use
   zdialog_fetch(zd,"pixdiff",pixdiff);                                    //  pixel difference threshold
   zdialog_fetch(zd,"pixcount",pixcount);                                  //  pixel count threshold

   cc = Nfiles * sizeof(uint8 *);
   thumbs = (char **) zmalloc(cc);
   memset(thumbs,0,cc);

   cc = Nfiles * sizeof(int);
   tcc = (int *) zmalloc(cc);

   Ffuncbusy = 1;

   for (ii = 0; ii < Nfiles; ii++)                                         //  screen out thumbnails not          v.15.02
   {                                                                       //    within top image directories
      pp = thumb2imagefile(files[ii]);
      if (! pp) {
         zfree(files[ii]);
         files[ii] = 0;
         continue;
      }

      for (kk = 0; kk < Ntopdirks; kk++)
      {
         cc = strlen(topdirks[kk]);
         if (pp[cc] != '/') continue;
         if (strmatchN(pp,topdirks[kk],cc)) break;
      }

      zfree(pp);
      if (kk < Ntopdirks) continue;
      zfree(files[ii]);
      files[ii] = 0;
   }

   for (ii = 0; ii < Nfiles; ii++)                                         //  get thumbnails in memory
   {
      if (! files[ii]) continue;
      if (Fkillfunc) goto cleanup;

      kk = thumbsize;
      gerror = 0;
      pxb = gdk_pixbuf_new_from_file_at_size(files[ii],kk,kk,&gerror);
      if (! pxb) {
         printz("file: %s \n %s",files[ii],gerror->message);
         zfree(files[ii]);
         files[ii] = 0;
         continue;
      }

      ww = gdk_pixbuf_get_width(pxb);                                      //  actual thumbnail dimensions
      hh = gdk_pixbuf_get_height(pxb);                                     //  ( <= 64x64)
      pix = gdk_pixbuf_get_pixels(pxb);

      cc = ww * hh * 3;
      thumbs[ii] = (char *) zmalloc(cc);
      memcpy(thumbs[ii],pix,cc);
      tcc[ii] = cc;
      g_object_unref(pxb);
      zmainloop(10);                                                       //  keep GTK alive
   }

   for (ii = 0; ii < Nfiles; ii++)                                         //  replace thumbnail filespecs
   {                                                                       //    with corresp. image filespecs
      if (! files[ii]) continue;
      pp = thumb2imagefile(files[ii]);
      zfree(files[ii]);
      files[ii] = pp;
   }

   snprintf(gfile,100,"%s/search_results",tempdir);                        //  open file for gallery output
   fid = fopen(gfile,"w");
   if (! fid) goto filerror;

   Fbusy_goal = Nfiles;                                                    //  set up progress monitor
   Fbusy_done = 0;

   for (ii = 0; ii < Nfiles; ii++)                                         //  loop all thumbnails
   {
      if (Fkillfunc) goto cleanup;
      Fbusy_done++;

      if (! files[ii]) continue;
      pp = strrchr(files[ii],'/');
      if (! pp) continue;
      *pp = 0;
      zdialog_stuff(zd,"labdirk",files[ii]);                               //  update directory and file name
      zdialog_stuff(zd,"labfile",pp+1);                                    //    in dialog
      *pp = '/';
      zmainloop(10);                                                       //  keep GTK alive
      zsleep(0.001);

      for (jj = ndup = 0; jj < Nfiles; jj++)                               //  loop all thumbnails
      {
         if (jj == ii) continue;
         if (! files[jj]) continue;
         if (tcc[ii] != tcc[jj]) continue;                                 //  look for matching thumbnails

         for (kk = sum = 0; kk < tcc[ii]; kk++)
         {
            m1 = thumbs[ii][kk];
            m2 = thumbs[jj][kk];
            diff = abs(m1 - m2);
            if (diff < pixdiff) continue;
            if (++sum > pixcount) break;
         }

         if (kk < tcc[ii]) continue;                                       //  no match

         gerror = 0;                                                       //  thumbnails match
         pxb = gdk_pixbuf_new_from_file_at_size(files[jj],64,64,&gerror);  //  check image is readable
         if (pxb) g_object_unref(pxb);
         if (! pxb || gerror) {                                            //  no, ignore matching
            zfree(files[jj]);                                              //    "broken file" thumbnails
            files[jj] = 0;
            continue;
         }

         if (ndup == 0) fprintf(fid,"%s\n",files[ii]);                     //  first duplicate, output file name
         fprintf(fid,"%s\n",files[jj]);                                    //  output duplicate image file name
         zfree(files[jj]);
         files[jj] = 0;
         ndup++;

         Ndups++;
         snprintf(text,100,"%d",Ndups);                                    //  update duplicates found in dialog
         zdialog_stuff(zd,"labdup2",text);
      }
   }

   fclose(fid);
   fid = 0;

   free_resources();
   navi::gallerytype = 2;                                                  //  generate gallery of duplicate images
   gallery(gfile,"initF");
   m_viewmode(0,"G");
   gallery(0,"paint",0);                                                   //  position at top

cleanup:
   zdialog_free(zd);
   if (fid) fclose(fid);
   for (ii = 0; ii < Nfiles; ii++)
      if (files[ii]) zfree(files[ii]);
   if (thumbs)
      for (ii = 0; ii < Nfiles; ii++)
         if (thumbs[ii]) zfree(thumbs[ii]);
   if (files) zfree(files);
   if (thumbs) zfree(thumbs);
   if (tcc) zfree(tcc);
   fid = 0;
   Fbusy_goal = 0;
   Ffuncbusy = 0;
   Fkillfunc = 0;
   Fblock = 0;
   return;

filerror:
   zmessLogACK(Mwin,"file error: %s",strerror(errno));
   goto cleanup;
}


//  dialog event and completion function

int  duplicates_dialog_event(zdialog *zd, cchar *event)
{
   if (strmatch(event,"escape")) zd->zstat = 2;                            //  escape = cancel                    v.15.07
   if (! zd->zstat) return 1;                                              //  wait for user input
   if (zd->zstat != 1) Fkillfunc = 1;                                      //  [cancel] or [x]
   return 1;                                                               //  [proceed]
}


/**************************************************************************/

//  burn images to CD/DVD

void  m_burn(GtkWidget *, cchar *)
{
   int      ii, cc;
   char     **filelist, *imagefile, *bcommand;

   if (! Fbrasero) {                                                       //  v.14.07
      zmessageACK(Mwin,ZTX("Brasero not installed"));
      return;
   }

   if (checkpend("all")) return;                                           //  check nothing pending
   Fblock = 1;

   F1_help_topic = "burn_DVD";

   filelist = gallery_getfiles();                                          //  get list of files to burn
   if (! filelist) {
      Fblock = 0;
      return;
   }
   m_viewmode(0,"F");

   cc = 0;
   for (ii = 0; filelist[ii]; ii++)                                        //  get memory for brasero command line
      cc += strlen(filelist[ii]) + 4;

   bcommand = (char *) zmalloc(cc+20);
   strcpy(bcommand,"brasero ");
   cc = strlen(bcommand);

   for (ii = 0; filelist[ii]; ii++)                                        //  copy files to command line
   {
      imagefile = filelist[ii];
      strcpy(bcommand+cc," \"");
      cc += 2;
      strcpy(bcommand+cc,imagefile);
      cc += strlen(imagefile);
      strcpy(bcommand+cc,"\"");
      cc += 1;
      zfree(imagefile);
   }

   zfree(filelist);

   strcat(bcommand," &");                                                  //  brasero command in background
   shell_ack(bcommand);
   zfree(bcommand);

   Fblock = 0;
   return;
}



